Package: CFtime
Title: Using CF-Compliant Calendars with Climate Projection Data
Version: 1.7.3
Authors@R: 
    person("Patrick", "Van Laake", , "patrick@vanlaake.net", role = c("aut", "cre", "cph"))
Description: Support for all calendars as specified in the Climate and Forecast 
    (CF) Metadata Conventions for climate and forecasting data. The CF Metadata 
    Conventions is widely used for distributing files with climate observations 
    or projections, including the Coupled Model Intercomparison Project (CMIP) 
    data used by climate change scientists and the Intergovernmental Panel on
    Climate Change (IPCC). This package specifically allows the user to work 
    with any of the CF-compliant calendars (many of which are not compliant with 
    POSIXt). The CF time coordinate is formally defined in the CF Metadata 
    Conventions document available at <https://cf-convention.github.io/Data/cf-conventions/cf-conventions-1.13/cf-conventions.html#time-coordinate>.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: R6
Suggests: rmarkdown, testthat (>= 3.0.0), stringr
URL: https://r-cf.github.io/CFtime/, https://github.com/R-CF/CFtime
BugReports: https://github.com/R-CF/CFtime/issues
Config/testthat/edition: 3
Collate: 'api.R' 'CFCalendar.R' 'CFCalendar360.R' 'CFCalendar365.R'
        'CFCalendar366.R' 'CFCalendarJulian.R' 'CFCalendarNone.R'
        'CFCalendarProleptic.R' 'CFCalendarStandard.R'
        'CFCalendarTAI.R' 'CFCalendarUTC.R' 'CFClimatology.R'
        'CFtime-package.R' 'CFtime.R' 'helpers.R' 'zzz.R'
Config/Needs/website: rmarkdown
NeedsCompilation: no
Packaged: 2026-04-01 19:32:17 UTC; patrickvanlaake
Author: Patrick Van Laake [aut, cre, cph]
Maintainer: Patrick Van Laake <patrick@vanlaake.net>
Repository: CRAN
Date/Publication: 2026-04-02 09:10:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("+",CFClimatology)
S3method("+",CFTime)
S3method("==",CFTime)
S3method("[",CFTime)
S3method(as.character,CFTime)
S3method(cut,CFTime)
S3method(length,CFTime)
S3method(range,CFTime)
S3method(str,CFTime)
export("bounds<-")
export(CFClimatology)
export(CFTime)
export(CFfactor)
export(CFfactor_coverage)
export(CFfactor_units)
export(CFtime)
export(as_timestamp)
export(bounds)
export(calendar)
export(definition)
export(indexOf)
export(is_complete)
export(month_days)
export(offsets)
export(origin)
export(parse_timestamps)
export(resolution)
export(slice)
export(timezone)
export(unit)
import(R6)
importFrom(stats,na.omit)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# CFtime

<!-- badges: start -->

[![Lifecycle:
Stable](https://img.shields.io/badge/Lifecycle-Stable-green.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![CRAN
Status](https://www.r-pkg.org/badges/version/CFtime)](https://cran.r-project.org/package=CFtime)
[![CRAN
Downloads](https://cranlogs.r-pkg.org/badges/grand-total/CFtime)](https://cran.r-project.org/package=CFtime)
[![License: GPL
v3](https://img.shields.io/badge/License-MIT-blue.svg)](https://mit-license.org)
[![Commits since
release](https://img.shields.io/github/commits-since/R-CF/CFtime/latest.svg?color=green)](https://GitHub.com/R-CF/CFtime/commit/main/)
[![Last
commit](https://img.shields.io/github/last-commit/R-CF/CFtime)](https://github.com/R-CF/CFtime/commits/main)
[![R-CMD-check](https://github.com/R-CF/CFtime/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/R-CF/CFtime/actions/workflows/R-CMD-check.yaml)
[![codecov](https://codecov.io/gh/R-CF/CFtime/branch/main/graph/badge.svg)](https://app.codecov.io/gh/R-CF/CFtime)

<!-- badges: end -->

CFtime is an R package that supports working with [CF Metadata
Conventions](https://cf-convention.github.io/) time coordinates,
specifically geared to time-referencing data sets of climate projections
such as those produced under the [World Climate Research
Programme](https://www.wcrp-climate.org) and re-analysis data such as
ERA5 from the European Centre for Medium-range Weather Forecasts
(ECMWF).

The data sets include in their metadata an epoch, or origin, a point in
time from which other points in time are calculated. This epoch takes
the form of `days since 1949-12-01`, with each data collection (Coupled
Model Intercomparison Project (CMIP) generation, model, etc) having its
own epoch. The data itself has a temporal dimension if a coordinate
variable in the netCDF file has an attribute `units` with a string value
describing an epoch. The coordinate variable, say “time”, has data
values such as 43289, which are offsets from the epoch in units of the
epoch string (“days” in this case). To convert this offset to a date,
using a specific calendar, is what this package does. Given that the
calendars supported by the CF Metadata Conventions are not compatible
with `POSIXt`, this conversion is not trivial because the standard R
date-time operations do not give correct results. That it is important
to account for these differences is easily demonstrated:

``` r
library(CFtime)

# POSIXt calculations on a standard calendar
as.Date("1949-12-01") + 43289
#> [1] "2068-06-08"

# CFtime calculation on a "360_day" calendar
as_timestamp(CFtime("days since 1949-12-01", "360_day", 43289))
#> [1] "2070-02-30"
```

That’s a difference of nearly 21 months! (And yes, 30 February is a
valid date on a `360_day` calendar.)

All defined calendars of the CF Metadata Conventions are supported:

- `standard` (or the deprecated `gregorian`): This calendar is valid for
  the Common Era only; it starts at `0001-01-01T00:00:00`, i.e. 1
  January of year 1. Time periods prior to the introduction of the
  Gregorian calendar (`1582-10-15`) use the `julian` calendar that was
  in common use then. The 10-day gap between the Julian and Gregorian
  calendars is observed, so dates in the range 5 to 14 October 1582 are
  invalid.
- `proleptic_gregorian`: This calendar uses the Gregorian calendar for
  periods prior to the introduction of that calendar as well, and it
  extends to periods before the Common Era, e.g. year 0 and negative
  years.
- `tai`: International Atomic Time, a global standard for linear time
  based on multiple atomic clocks: it counts seconds since its start at
  `1958-01-01T00:00:00`. For presentation it uses the Gregorian
  calendar. Timestamps prior to its start are not allowed.
- `utc`: Coordinated Universal Time, the standard for civil timekeeping
  all over the world. It is based on International Atomic Time but it
  uses occasional leap seconds to remain synchronous with Earth’s
  rotation around the Sun; at the end of 2025 it is 37 seconds behind
  `tai`. It uses the Gregorian calendar with a start at
  `1972-01-01T00:00:00`; earlier timestamps are not allowed. Future
  timestamps are also not allowed because the insertion of leap seconds
  is unpredictable. Most computer clocks synchronize against UTC but
  calculations of periods do not consider leap seconds (but `CFtime`
  does).
- `julian`: The `julian` calendar has a leap year every four years,
  including centennial years. Otherwise it is the same as the `standard`
  calendar.
- `365_day` or `noleap`: This is a “model time” calendar in which no
  leap years occur. Year 0 exists, as well as years prior to that.
- `366_day` or `all_leap`: This is a “model time” calendar in which all
  years are leap years. Year 0 exists, as well as years prior to that.
- `360_day`: This is a “model time” calendar in which every year has 360
  days divided over 12 months of 30 days each. Year 0 exists, as well as
  years prior to that.
- `none`: Perpetual “calendar” for experiments that are simulated on a
  given instant during the year. All the elements in this calendar thus
  represent the same instant in time.

Use of custom calendars is currently not supported.

This package facilitates use of a suite of models of climate projections
that use different calendars in a consistent manner. This package is
particularly useful for working with climate projection data having a
daily or higher resolution, but it will work equally well on data with a
lower resolution.

Timestamps are generated using the [ISO8601
standard](https://en.wikipedia.org/wiki/ISO_8601).

Calendar-aware factors can be generated to support processing of data
using `tapply()` and similar functions. Merging of multiple data sets
and subsetting facilitate analysis while preserving computer resources.

## Working with CFtime

Check out the multiple articles that provide detailed instructions and
examples for use of this package.

## Installation

Get the latest stable version on CRAN:

``` r
install.packages("CFtime")
```

You can install the development version of CFtime from
[GitHub](https://github.com/) with:

``` r
# install.packages("devtools")
devtools::install_github("R-CF/CFtime")
```

## Coverage

This package has been tested with the following data sets:

- ERA5 (including multiple variables, levels, and mixed ERA5/ERA5T data)
- CMIP5
- CORDEX
- CMIP6
- ROMS

The package also operates on geographical and/or temporal subsets of
data sets so long as the subsetted data complies with the CF Metadata
Conventions. This includes subsetting in the [Climate Data
Store](https://cds.climate.copernicus.eu/#!/home). Subsetted data from
Climate4Impact is not automatically supported because the dimension
names are not compliant with the CF Metadata Conventions, use the
corresponding dimension names instead.

# cftime
Time-handling functionality from netcdf4-python

[![Build status](https://github.com/Unidata/cftime/workflows/cftime%20test/badge.svg)](https://github.com/Unidata/cftime/actions)
[![PyPI package](https://badge.fury.io/py/cftime.svg)](http://python.org/pypi/cftime)
[![Coverage Status](https://coveralls.io/repos/github/Unidata/cftime/badge.svg?branch=master)](https://coveralls.io/github/Unidata/cftime?branch=master)
[![Tag Status](https://img.shields.io/github/tag/UniData/cftime.svg)](https://github.com/Unidata/cftime/tags)
[![Release Status](https://img.shields.io/github/release/UniData/cftime.svg)](https://github.com/Unidata/cftime/releases)
[![Commits Status](https://img.shields.io/github/commits-since/UniData/cftime/latest.svg)](https://github.com/UniData/cftime/commits/master)
[![DOI](https://zenodo.org/badge/73107250.svg)](https://zenodo.org/badge/latestdoi/73107250)

## News
For details on the latest updates, see the [Changelog](https://github.com/Unidata/cftime/blob/master/Changelog).

10/20/2023:  Version 1.6.3 released.  Support for python 3.12, cython 3.0, strptime formats without separators.
 
9/18/2022:  Version 1.6.2 released.  strptime method added, fix for num2date failure on
empty integer array, date2num 'longdouble' keyword added. New wheel building workflow.

6/30/2022:  Version 1.6.1 released.  Fixes for numpy 1.23.0, updated CI/CD.

3/4/2022:  Version 1.6.0 released.  Big speed-ups for num2date, date2index bugfix for select != 'exact' when select='exact' works, fix for date2num with masked array inputs.

1/22/2022: Version 1.5.2 released (wheels for Apple M1 available on pypi for python 3.8,3.9 and 3.10). is_leap_year
function added (issue #259).

10/31/2021: Version 1.5.1.1 released (new binary wheels for python 3.10).

10/1/2021:  Version 1.5.1 released. Changed default behavior of ``proleptic_gregorian``
to has_year_zero=T (since it is allowed in ISO-8601 and CF does not specify the
year zero convention for this calendar). Raise warning message when trying
to create a calendar that is not supported by CF version 1.9 (no years < 1
allowed for 'standard'/'gregorian' or 'julian'  calendars).
Added support for "common_year" and "common_years" units for "noleap" 
and "365_day" calendars.
 
5/20/2021:  Version 1.5.0 released.  Includes support for astronomical year numbering
(including the year zero) for real-world calendars ('julian', 'gregorian'/'standard',
and 'proleptic_gregorian') using 'has_year_zero' `cftime.datetime` kwarg.
Added a 'change_calendar' `cftime.datetime` method to switch to another 
'real-world' calendar to enable comparison of instances with different calendars.
Some legacy classes and functions removed (`utime`, `JulianDayFromDate` and
`DateFromJulianDay`). The functionality of `JulianDayFromDate` and 
`DateFromJulianDay` is now available from the methods `cftime.datetime.toordinal`
and `cftime.datetime.fromordinal`.

2/2/2021:  Version 1.4.1 released. Restore use of calendar-specific subclasses
in `cftime.num2date`, `cftime.datetime.__add__`, and `cftime.datetime.__sub__`.
The use of this will be removed in a later release.
Add 'fromordinal' static method to create a cftime.datetime instance
from a julian day ordinal and calendar (inverse of 'toordinal').

2/1/2021:  Version 1.4.0 released.  License changed to MIT (GPL'ed code replaced).
Roundtrip accuracy improved for units other than microseconds. Added 
cftime.datetime.toordinal method, returns integer julian day number.

1/17/2021: Version 1.3.1 released.

11/16/2020:  Version 1.3.0 released. **API change**: The `cftime.datetime` constructor now creates 
 'calendar-aware' instances (default is `'standard'` calendar, if `calendar=''` or `None` the instance
 is not calendar aware and some methods, like `dayofwk`, `dayofyr`, `__add__` and `__sub__`, will not work)
 See discussion for issue [#198](https://github.com/Unidata/cftime/issues/198).
 The calendar specific sub-classes are now deprecated, but remain for now
 as stubs that just instantiate the base class and override `__repr__`.
 The default calendar in `cftime.date2num` has been changed from `'standard'` to `None`
 (the calendar associated with first input datetime object is used to define the calendar).

07/20/2020: Version 1.2.1 released.  Fixes a couple of regressions introduced in 1.2.0. See Changelog for details.

7/06/2020:  version 1.2.0 released. New microsecond accurate algorithm for date2num/num2date contributed by [spencerkclark](https://github.com/spencerkclark). Bugs fixed in masked array handling.

5/12/2020:  version 1.1.3 released.  Add isoformat method for compatibility with python datetime (issue #152).
 Make 'standard' default calendar for cftime.datetime so that dayofwk,dayofyr methods don't fail (issue #169).

4/20/2020:  version 1.1.2 released.  Code optimization, fix logic so `only_use_cftime_datetimes=False` works as 
 expected (issues [#158](https://github.com/Unidata/cftime/issues/158) and [#165](https://github.com/Unidata/cftime/issues/165)).

3/16/2020:  version 1.1.1 released.  Fix bug in microsecond formatting, ensure identical num2date results if input is an array of times, or a single scalar ([issue #143](https://github.com/Unidata/cftime/issues/143)).

2/12/2020:  version 1.1.0 released.  `cftime.datetime` instances are returned by default from `num2date`
(instead of returning python datetime instances where possible ([issue #136](https://github.com/Unidata/cftime/issues/136))).  `num2pydate`
convenience function added (always returns python datetime instances, [issue #134](https://github.com/Unidata/cftime/issues/134)). Fix for
fraction seconds in reference date string ([issue #140](https://github.com/Unidata/cftime/issues/140)). Added `daysinmonth` attribute 
([issue #137](https://github.com/Unidata/cftime/issues/137)).

10/25/2019:  version 1.0.4.2 released (fix for [issue #126](https://github.com/Unidata/cftime/issues/126)).

10/21/2019:  version 1.0.4 released.

12/05/2018:  version 1.0.3.4 released (just to fix a problem with the source 
tarball on pypi).

12/05/2018:  version 1.0.3.1 released.  Bugfix release (fixed issue with installation
when cython not installed, regression on 32-bit platforms, workaround for pandas 
compatibility).

12/01/2018:  version 1.0.3 released. Test coverage with coveralls.io, improved round-tripping accuracy for non-real world calendars (like `360_day`).

10/27/2018:  version 1.0.2 released. Improved accuracy (from approximately 1000 microseconds to 10 microseconds on x86
platforms). Refactored calendar calculations now allow for negative reference years. num2date function now more than an
order of magnitude faster. `months since` units now allowed, but only for `360_day` calendar.

08/15/2018:  version 1.0.1 released.

11/8/2016: `cftime` was split out of the [netcdf4-python](https://github.com/Unidata/netcdf4-python) package.

## Quick Start
* Clone GitHub repository (`git clone https://github.com/Unidata/cftime.git`), or get source tarball from [PyPI](https://pypi.python.org/pypi/cftime). Links to Windows and OS X precompiled binary packages are also available on [PyPI](https://pypi.python.org/pypi/cftime).

* Make sure [numpy](http://www.numpy.org/) and [Cython](http://cython.org/) are
  installed and you have [Python](https://www.python.org) 2.7 or newer.

* Run `python setup.py build`, then `python setup.py install` (with `sudo` if necessary).

* To run all the tests, execute `py.test`.

## Documentation
See the online [docs](http://unidata.github.io/cftime) for more details.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, include = FALSE---------------------------------------------------
library(CFtime)
library(ncdf4)

## -----------------------------------------------------------------------------
# Setting up
nc <- nc_open(list.files(path = system.file("extdata", package = "CFtime"), full.names = TRUE)[1])
cf <- CFtime(nc$dim$time$units, nc$dim$time$calendar, nc$dim$time$vals)

# Create monthly factors for a baseline epoch and early, mid and late 21st century epochs
baseline <- CFfactor(cf, epoch = 1991:2020)
future <- CFfactor(cf, epoch = list(early = 2021:2040, mid = 2041:2060, late = 2061:2080))
str(baseline)
str(future)

## -----------------------------------------------------------------------------
# Read the data from the NetCDF file.
# Keep degenerate dimensions so that we have a predictable data structure: 3-dimensional array.
# Converts units of kg m-2 s-1 to mm/day.
pr <- ncvar_get(nc, "pr", collapse_degen = FALSE) * 86400

# Assign dimnames(), optional.
dimnames(pr) <- list(nc$dim$lon$vals, nc$dim$lat$vals, CFtimestamp(cf))

# Get a global attribute from the file
experiment <- ncatt_get(nc, "")$experiment_id

nc_close(nc)

# Calculate the daily average precipitation per month for the baseline period
# and the three future epochs.
pr_base <- apply(pr, 1:2, tapply, baseline, mean)                         # an array
pr_future <- lapply(future, function(f) apply(pr, 1:2, tapply, f, mean))  # a list of arrays

# Calculate the precipitation anomalies for the future epochs against the baseline.
# Working with daily averages per month so we can simply subtract and then multiply by days 
# per month for each of the factor levels using the CF calendar.
ano <- mapply(function(pr, f) {(pr - pr_base) * CFfactor_units(cf, f)}, pr_future, future, SIMPLIFY = FALSE)

# Plot the results
plot(1:12, ano$early[,1,1], type = "o", col = "blue", ylim = c(-50, 40), xlim = c(1, 12), 
     main = paste0("Hamilton, New Zealand\nExperiment: ", experiment), 
     xlab = "month", ylab = "Precipitation anomaly (mm)")
lines(1:12, ano$mid[,1,1], type = "o", col = "green")
lines(1:12, ano$late[,1,1], type = "o", col = "red")

## -----------------------------------------------------------------------------
# Get the list of files that make up the ensemble members, here:
# GFDL ESM4 and MRI ESM2 models for experiment SSP2-4.5, precipitation, CMIP6 2015-01-01 to 2099-12-31
lf <- list.files(path = system.file("extdata", package = "CFtime"), full.names = TRUE)

# Loop over the files individually
# ano is here a list with each element holding the results for a single model
ano <- lapply(lf, function(fn) {
  nc <- nc_open(fn)
  cf <- CFtime(nc$dim$time$units, nc$dim$time$calendar, nc$dim$time$vals)
  pr <- ncvar_get(nc, "pr", collapse_degen = FALSE) * 86400
  nc_close(nc)

  baseline <- CFfactor(cf, epoch = 1991:2020)
  pr_base <- apply(pr, 1:2, tapply, baseline, mean)
  future <- CFfactor(cf, epoch = list(early = 2021:2040, mid = 2041:2060, late = 2061:2080))
  pr_future <- lapply(future, function(f) apply(pr, 1:2, tapply, f, mean))
  mapply(function(pr, f) {(pr - pr_base) * CFfactor_units(cf, f)}, pr_future, future, SIMPLIFY = FALSE)
})

# Epoch names
epochs <- c("early", "mid", "late")
dim(epochs) <- 3

# Build the ensemble for each epoch
# For each epoch, grab the data for each of the ensemble members, simplify to an array
# and take the mean per row (months, in this case)
ensemble <- apply(epochs, 1, function(e) {
  rowMeans(sapply(ano, function(a) a[[e]], simplify = T))})
colnames(ensemble) <- epochs
rownames(ensemble) <- rownames(ano[[1]][[1]])
ensemble

## ----eval = FALSE-------------------------------------------------------------
#  library(ncdf4)
#  library(abind)
#  
#  prepare_CORDEX <- function(fn, var) {
#    offsets <- vector("list", length(fn))
#    data <- vector("list", length(fn))
#    for (i in 1:length(fn)) {
#      nc <- nc_open(fn[i])
#      if (i == 1)
#        # Create an "empty" CFtime object, without elements
#        cf <- CFtime(nc$dim$time$units, nc$dim$time$calendar)
#  
#      # Make lists of all datum offsets and data arrays
#      offsets[[i]] <- as.vector(nc$dim$time$vals)
#      data[[i]] <- ncvar_get(nc, var,
#                             start = c(10, 10, 1), count = c(100, 100, -1), # spatial subsetting
#                             collapse_degen = FALSE)
#  
#      nc_close(nc)
#    }
#  
#    # Create a list for output with the CFtime instance assigned the offsets and
#    # the data bound in a single 3-dimensional array
#    list(CFtime = cf + unlist(offsets), data = abind(data, along = 3))
#  }

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, include = FALSE---------------------------------------------------
library(CFtime)
library(ncdf4)

## -----------------------------------------------------------------------------
# POSIXt calculations on a standard calendar - INCORRECT
as.Date("1949-12-01") + 43289

# CFtime calculation on a "360_day" calendar - CORRECT
# See below examples for details on the two functions
CFtimestamp(CFtime("days since 1949-12-01", "360_day", 43289))

## -----------------------------------------------------------------------------
# Create a CF time object from a definition string, a calendar and some offsets
cf <- CFtime("days since 1949-12-01", "360_day", 19830:90029)
cf

## -----------------------------------------------------------------------------
# Opening a data file that is included with the package and showing some attributes.
# Usually you would `list.files()` on a directory of your choice.
nc <- nc_open(list.files(path = system.file("extdata", package = "CFtime"), full.names = TRUE)[1])
attrs <- ncatt_get(nc, "")
attrs$title

# "Conventions" global attribute must have a string like "CF-1.*" for this package to work reliably
attrs$Conventions

# Create the CFtime instance from the metadata in the file.
cf <- CFtime(nc$dim$time$units, nc$dim$time$calendar, nc$dim$time$vals)
cf

## -----------------------------------------------------------------------------
library(RNetCDF)
nc <- open.nc(list.files(path = system.file("extdata", package = "CFtime"), full.names = TRUE)[1])
att.get.nc(nc, -1, "Conventions")
cf <- CFtime(att.get.nc(nc, "time", "units"), att.get.nc(nc, "time", "calendar"), var.get.nc(nc, "time"))
cf

## -----------------------------------------------------------------------------
dates <- CFtimestamp(cf, format = "date")
dates[1:10]

## -----------------------------------------------------------------------------
CFrange(cf)

## -----------------------------------------------------------------------------
# Create a dekad factor for the whole `cf` time series that was created above
f_k <- CFfactor(cf, "dekad")
str(f_k)

# Create monthly factors for a baseline epoch and early, mid and late 21st century epochs
baseline <- CFfactor(cf, epoch = 1991:2020)
future <- CFfactor(cf, epoch = list(early = 2021:2040, mid = 2041:2060, late = 2061:2080))
str(future)

## -----------------------------------------------------------------------------
# Is the time series complete?
CFcomplete(cf)

# How many time units fit in a factor level?
CFfactor_units(cf, baseline)

# What's the absolute and relative coverage of our time series
CFfactor_coverage(cf, baseline, "absolute")
CFfactor_coverage(cf, baseline, "relative")

## -----------------------------------------------------------------------------
# 4 years of data on a `365_day` calendar, keep 80% of values
n <- 365 * 4
cov <- 0.8
offsets <- sample(0:(n-1), n * cov)

cf <- CFtime("days since 2020-01-01", "365_day", offsets)
cf
# Note that there are about 1.25 days between observations

mon <- CFfactor(cf, "month")
CFfactor_coverage(cf, mon, "absolute")
CFfactor_coverage(cf, mon, "relative")

## -----------------------------------------------------------------------------
# Days in January and February
cf <- CFtime("days since 2023-01-01", "360_day", 0:59)
cf_days <- CFtimestamp(cf, "date")
as.Date(cf_days)

#' @title 360-day CF calendar
#'
#' @description This class represents a CF calendar of 360 days per year, evenly
#'   divided over 12 months of 30 days. This calendar is obviously not
#'   compatible with the standard POSIXt calendar.
#'
#'   This calendar supports dates before year 1 and includes the year 0.
#'
#' @aliases CFCalendar360
#' @docType class
CFCalendar360 <- R6::R6Class("CFCalendar360",
  inherit = CFCalendar,
  public = list(
    #' @description Create a new CF calendar.
    #' @param nm The name of the calendar. This must be "360_day". This argument
    #' is superfluous but maintained to be consistent with the initialization
    #' methods of the parent and sibling classes.
    #' @param definition The string that defines the units and the origin, as
    #' per the CF Metadata Conventions.
    #' @return A new instance of this class.
    initialize = function(nm, definition) {
      super$initialize(nm, definition)
    },

    #' @description Indicate which of the supplied dates are valid.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return Logical vector with the same length as argument `ymd` has rows
    #'   with `TRUE` for valid days and `FALSE` for invalid days, or `NA` where
    #'   the row in argument `ymd` has `NA` values.
    valid_days = function(ymd) {
      !is.na(ymd$year) & ymd$month >= 1L & ymd$month <= 12L & ymd$day >= 1L & ymd$day <= 30L
    },

    #' @description Determine the number of days in the month of the calendar.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return A vector indicating the number of days in each month for the
    #'   dates supplied as argument `ymd`. If no dates are supplied, the number
    #'   of days per month for the calendar as a vector of length 12.
    month_days = function(ymd = NULL) {
      if (is.null(ymd)) return(rep(30L, 12L))

      res <- rep(30L, nrow(ymd))
      res[which(is.na(ymd$year))] <- NA
      res
    },

    #' @description Indicate which years are leap years.
    #' @param yr Integer vector of years to test.
    #' @return Logical vector with the same length as argument `yr`. Since this
    #' calendar does not use leap days, all values will be `FALSE`, or `NA`
    #' where argument `yr` is `NA`.
    leap_year = function(yr) {
      res <- rep(FALSE, length(yr))
      res[which(is.na(yr))] <- NA
      res
    },

    #' @description Retrieve the day-of-year number for the supplied
    #'   `data.frame` of year, month and day, using this calendar.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return Integer vector as long as argument `ymd` has rows with the
    #'   day-of-year for each row.
    doy = function(ymd) {
      (ymd$month - 1L) * 30L + ymd$day
    },

    #' @description Calculate difference in days between a `data.frame` of time
    #'   parts and the origin.
    #'
    #' @param x `data.frame`. Dates to calculate the difference for.
    #'
    #' @return Integer vector of a length equal to the number of rows in
    #' argument `x` indicating the number of days between `x` and the `origin`,
    #' or `NA` for rows in `x` with `NA` values.
    date2offset = function(x) {
      (x$year - self$origin$year) * 360L + (x$month - self$origin$month) * 30L + x$day - self$origin$day
    },

    #' @description Calculate date parts from day differences from the origin.
    #'   This only deals with days as these are impacted by the calendar.
    #'   Hour-minute-second timestamp parts are handled in [CFCalendar].
    #'
    #' @param x Integer vector of days to add to the origin.
    #'
    #' @return A `data.frame` with columns 'year', 'month' and 'day' and as many
    #'   rows as the length of vector `x`.
    offset2date = function(x) {
      y <- self$origin$year + x %/% 360L
      m <- self$origin$month + (x %% 360L) %/% 30L
      d <- self$origin$day + x %% 30L
      over <- which(d > 30L)
      d[over] <- d[over] - 30L
      m[over] <- m[over] + 1L
      over <- which(m > 12L)
      m[over] <- m[over] - 12L
      y[over] <- y[over] + 1L
      data.frame(year = y, month = m, day = d, row.names = NULL)
    }
  )
)

#' @title CFTime class
#'
#' @description This class manages the "time" dimension of netCDF files that
#' follow the CF Metadata Conventions, and its productive use in R.
#'
#' The class has a field `cal` which holds a specific calendar from the
#' allowed types (all named calendars are supported). The calendar is
#' also implemented as a (hidden) class which converts netCDF file encodings to
#' timestamps as character strings, and vice-versa. Bounds information (the
#' period of time over which a timestamp is valid) is used when defined in the
#' netCDF file.
#'
#' Additionally, this class has functions to ease use of the netCDF "time"
#' information when processing data from netCDF files. Filtering and indexing of
#' time values is supported, as is the generation of factors.
#'
#' @export
#' @references
#'   https://cf-convention.github.io/Data/cf-conventions/cf-conventions-1.13/cf-conventions.html#time-coordinate
#' @docType class
CFTime <- R6::R6Class("CFTime",
  private = list(
    # The calendar of this instance, a descendant of the [CFCalendar] class.
    .cal = NULL,

    #A numeric vector of offsets from the origin of the calendar.
    .offsets = numeric(),

    # The average number of time units between offsets.
    .resolution = NA_real_,

    # The boundary values of the time instance.
    .bounds = NULL
  ),
  public = list(
    #' @description Create a new instance of this class.
    #' @param definition Character string of the units and origin of the
    #'   calendar.
    #' @param calendar Character string of the calendar to use. Must be one of
    #'   the values permitted by the CF Metadata Conventions. If `NULL`, the
    #'   "standard" calendar will be used.
    #' @param offsets Numeric or character vector, optional. When numeric, a
    #'   vector of offsets from the origin in the time series. When a character
    #'   vector of length 2 or more, timestamps in ISO8601 or UDUNITS format.
    initialize = function(definition, calendar = "standard", offsets = NULL) {
      if (is.null(calendar)) calendar <- "standard" # This may occur when "calendar" attribute is not defined in the NC file
      calendar <- tolower(calendar)
      private$.cal <- switch(calendar,
        "standard" = CFCalendarStandard$new(calendar, definition),
        "gregorian" = CFCalendarStandard$new(calendar, definition),
        "proleptic_gregorian" = CFCalendarProleptic$new(calendar, definition),
        "tai" = CFCalendarTAI$new(calendar, definition),
        "utc" = CFCalendarUTC$new(calendar, definition),
        "julian" = CFCalendarJulian$new(calendar, definition),
        "360_day" = CFCalendar360$new(calendar, definition),
        "365_day" = CFCalendar365$new(calendar, definition),
        "noleap" = CFCalendar365$new(calendar, definition),
        "366_day" = CFCalendar366$new(calendar, definition),
        "all_leap" = CFCalendar366$new(calendar, definition),
        "none" = CFCalendarNone$new(calendar, definition),
        stop("Invalid calendar specification", call. = FALSE)
      )

      if (is.null(offsets)) return()

      if (is.numeric(offsets)) {
        dim(offsets) <- NULL
        stopifnot(.validOffsets(offsets))

        if (length(offsets) > 1L) {
          private$.resolution <- (max(offsets) - min(offsets)) / (length(offsets) - 1L)
          if (any(diff(offsets) <= 0))
            warning("Offsets not monotonically increasing.", call. = FALSE)
        } else {
          private$.resolution <- NA_real_
        }
        private$.offsets <- as.numeric(offsets)
      } else if (is.character(offsets)) {
        time <- private$.cal$parse(offsets)
        if (anyNA(time$year)) stop("Argument `offsets` contains invalid timestamps", call. = FALSE) # nocov

        private$.offsets <- time$offset
        if (length(offsets) > 1L)
          private$.resolution <- (max(private$.offsets) - min(private$.offsets)) / (length(private$.offsets) - 1L)
        else
          private$.resolution <- NA_real_
        if (any(diff(private$.offsets) <= 0))
          warning("Offsets not monotonically increasing.", call. = FALSE)
      } else if (!is.null(offsets)) stop("Invalid offsets for CFTime object", call. = FALSE)
    },

    #' @description Print a summary of the `CFTime` object to the console.
    #' @param ... Ignored.
    #' @return Self, invisibly.
    print = function(...) {
      noff <- length(private$.offsets)
      if (noff == 0L) {
        el <- "  Elements: (no elements)\n"
        b  <- "  Bounds  : (not set)\n"
      } else {
        d <- self$range()
        if (inherits(private$.cal, "CFCalendarNone")) {
          el <- if (noff > 1L) {
            sprintf("  Elements: Series of %d elements at %s\n", noff, d[1L])
          } else
            paste("  Element : One element at", d[1L], "\n")
        } else {
          el <- if (noff > 1L) {
            sprintf("  Elements: [%s .. %s] (average of %f %s between %d elements)\n",
                   d[1L], d[2L], private$.resolution,
                   paste0(if (private$.cal$prefix_id) CFt$prefixes$name[private$.cal$prefix_id], CFt$units$name[private$.cal$unit]), noff)
          } else
            paste("  Element :", d[1L], "\n")
        }

        b <- if (is.null(private$.bounds)) "  Bounds  : not set\n"
             else "  Bounds  : set\n"
      }
      private$.cal$print()
      cat("\nTime series:\n",  el, b, sep = "")
      invisible(self)
    },

    #' @description This method returns the first and last timestamp of the time
    #'   series as a vector. Note that the offsets do not have to be sorted.
    #' @param format Value of "date" or "timestamp". Optionally, a
    #'   character string that specifies an alternate format.
    #' @param bounds Logical to indicate if the extremes from the bounds should
    #'   be used, if set. Defaults to `FALSE`.
    #' @return Vector of two character strings that represent the starting and
    #'   ending timestamps in the time series. If a `format` is supplied, that
    #'   format will be used. Otherwise, if all of the timestamps in the time
    #'   series have a time component of `00:00:00` the date of the timestamp is
    #'   returned, otherwise the full timestamp (without any time zone
    #'   information).
    range = function(format = "", bounds = FALSE) {
      if (length(private$.offsets) == 0L) return(c(NA_character_, NA_character_))
      if (!missing(format) && ((!is.character(format)) || length(format) != 1L))
        stop("`format` argument, when present, must be a character string with formatting specifiers", call. = FALSE) # nocov
      if (!is.logical(bounds) || length(bounds) != 1L)
        stop("`bounds` argument, when present, must be a single logical value", call. = FALSE) # nocov

      if (bounds) {
        bnds <- private$.bounds
        if (is.null(bnds)) time <- private$.cal$offsets2time(base::range(private$.offsets))
        else time <- private$.cal$offsets2time(c(bnds[1L, 1L], bnds[2L, length(private$.offsets)]))
      } else time <- private$.cal$offsets2time(base::range(private$.offsets))

      .format_format(time, private$.cal$timezone, format)
    },

    #' @description This method generates a vector of character strings or
    #'   `POSIXct`s that represent the date and time in a selectable combination
    #'   for each offset.
    #'
    #'   The character strings use the format `YYYY-MM-DDThh:mm:ss±hhmm`,
    #'   depending on the `format` specifier. The date in the string is not
    #'   necessarily compatible with `POSIXt` - in the `360_day` calendar
    #'   `2017-02-30` is valid and `2017-03-31` is not.
    #'
    #'   For the "proleptic_gregorian" calendar the output can also be generated
    #'   as a vector of `POSIXct` values by specifying `asPOSIX = TRUE`. The
    #'   same is possible for the "standard" and "gregorian" calendars but only
    #'   if all timestamps fall on or after 1582-10-15. If `asPOSIX = TRUE` is
    #'   specified while the calendar does not support it, an error will be
    #'   generated.
    #' @param format character. A character string with either of the values
    #'   "date" or "timestamp". If the argument is not specified, the format
    #'   used is "timestamp" if there is time information, "date" otherwise.
    #' @param asPOSIX logical. If `TRUE`, for "standard", "gregorian" and
    #'   "proleptic_gregorian" calendars the output is a vector of `POSIXct` -
    #'   for other calendars an error will be thrown. Default value is `FALSE`.
    #' @return A character vector where each element represents a moment in
    #'   time according to the `format` specifier.
    as_timestamp = function(format = NULL, asPOSIX = FALSE) {
      if (asPOSIX && !private$.cal$POSIX_compatible(private$.offsets))
        stop("Cannot make a POSIX timestamp with this calendar.", call. = FALSE)
      if (length(private$.offsets) == 0L) return()

      time <- private$.cal$offsets2time(private$.offsets)

      if (is.null(format))
        format <- ifelse(private$.cal$unit < 4L || .has_time(time), "timestamp", "date")
      else if (!(format %in% c("date", "timestamp")))
        stop("Format specifier not recognized", call. = FALSE) # nocov

      if (asPOSIX) {
        if (format == "date") ISOdate(time$year, time$month, time$day, 0L, tz = "UTC")
        else ISOdatetime(time$year, time$month, time$day, time$hour, time$minute, time$second, tz = "UTC")
      } else .format_format(time, private$.cal$timezone, format)
    },

    #' @description Format timestamps using a specific format string, using the
    #'   specifiers defined for the [base::strptime()] function, with
    #'   limitations. The only supported specifiers are `bBdeFhHImMpRSTYz%`.
    #'   Modifiers `E` and `O` are silently ignored. Other specifiers, including
    #'   their percent sign, are copied to the output as if they were adorning
    #'   text.
    #'
    #'   The formatting is largely oblivious to locale. The reason for this is
    #'   that certain dates in certain calendars are not POSIX-compliant and the
    #'   system functions necessary for locale information thus do not work
    #'   consistently. The main exception to this is the (abbreviated) names of
    #'   months (`bB`), which could be useful for pretty printing in the local
    #'   language. For separators and other locale-specific adornments, use
    #'   local knowledge instead of depending on system locale settings; e.g.
    #'   specify `%m/%d/%Y` instead of `%D`.
    #'
    #'   Week information, including weekday names, is not supported at all as a
    #'   "week" is not defined for non-standard CF calendars and not generally
    #'   useful for climate projection data. If you are working with observed
    #'   data and want to get pretty week formats, use the [as_timestamp()]
    #'   method to generate `POSIXct` timestamps (observed data generally uses a
    #'   "standard" calendar) and then use the [base::format()] function which
    #'   supports the full set of specifiers.
    #' @param format A character string with `strptime` format specifiers. If
    #'   omitted, the most economical format will be used: a full timestamp when
    #'   time information is available, a date otherwise.
    #' @param usetz Logical. Should the time zone offset be appended to the
    #'   output? This is always in numerical form, i.e. "-0800", from UTC.
    #'   Default is `FALSE`.
    #' @return A vector of character strings with a properly formatted
    #'   timestamp. Any format specifiers not recognized or supported will be
    #'   returned verbatim.
    format = function(format, usetz = FALSE) {
      if (length(private$.offsets) == 0L) return(character(0L))
      if (missing(format)) format <- ""
      else if (!is.character(format) || length(format) != 1L)
        stop("`format` argument must be a character string with formatting specifiers", call. = FALSE)

      ts <- private$.cal$offsets2time(private$.offsets)
      if ((usetz || grepl("%z$", format)) && .has_time(ts)) {
        tz <- private$.cal$timezone
        if (!grepl("%z$", format))
          format <- paste0(format, "%z")
      } else tz <- ""
      .format_format(ts, tz, format)
    },

    #' @description Find the index in the time series for each timestamp given
    #'   in argument `x`. Alternatively, when `x` is a numeric vector of index
    #'   values, return the valid indices of the same vector, with the side
    #'   effect being the attribute "CFTime" associated with the result.
    #'
    #'   Matching also returns index values for timestamps that fall between two
    #'   elements of the time series - this can lead to surprising results when
    #'   time series elements are positioned in the middle of an interval (as
    #'   the CF Metadata Conventions instruct us to "reasonably assume"): a time
    #'   series of days in January would be encoded in a netCDF file as
    #'   `c("2024-01-01 12:00:00", "2024-01-02 12:00:00", "2024-01-03 12:00:00",
    #'   ...)` so `x <- c("2024-01-01", "2024-01-02", "2024-01-03")` would
    #'   result in `(NA, 1, 2)` (or `(NA, 1.5, 2.5)` with `method = "linear"`)
    #'   because the date values in `x` are at midnight. This situation is
    #'   easily avoided by ensuring that this `CFTime` instance has bounds set
    #'   (use `bounds(y) <- TRUE` as a proximate solution if bounds are not
    #'   stored in the netCDF file).
    #'
    #'   If bounds are set, the indices are informed by those bounds. If the
    #'   bounds are not contiguous, returned values may be `NA` even if the
    #'   value of `x` falls between two valid timestamps.
    #'
    #'   Values of `x` that are not valid timestamps according to the calendar
    #'   of this `CFTime` instance will be returned as `NA`.
    #'
    #'   Argument `x` can also be a numeric vector of index values, in which
    #'   case the valid values in `x` are returned. If negative values are
    #'   passed, the positive counterparts will be excluded and then the
    #'   remainder returned. Positive and negative values may not be mixed.
    #'   Using a numeric vector has the side effect that the result has the
    #'   attribute "CFTime" describing the temporal dimension of the slice. If
    #'   index values outside of the range of `self` are provided, an error will
    #'   be thrown.
    #' @param x Vector of `character`, `POSIXt` or `Date` values to find indices
    #'   for, or a `numeric` vector.
    #' @param method Single value of "constant" or "linear". If `"constant"`,
    #'   return the index value for each match. If `"linear"`, return the index
    #'   value with any fractional value.
    #' @param rightmost.closed Whether or not to include the upper limit of
    #'   argument `x`. Default is `FALSE`. This argument is ignored when
    #'   argument `x` contains index values.
    #' @return A numeric vector giving indices into `self` for the valid values
    #'   of `x`. If there is at least 1 valid index, then attribute "CFTime"
    #'   contains an instance of `CFTime` that describes the dimension of
    #'   filtering the dataset associated with `self` with the result of this
    #'   method, excluding any `NA` values.
    indexOf = function(x, method = "constant", rightmost.closed = FALSE) {
      stopifnot(inherits(x, c("character", "POSIXt", "Date")) || is.numeric(x),
                method %in% c("constant", "linear"))

      if (is.numeric(x)) {
        if (!(all(x < 0, na.rm = TRUE) || all(x > 0, na.rm = TRUE)))
          stop("Cannot mix positive and negative index values", call. = FALSE)

        intv <- (1:length(private$.offsets))[x]
        xoff <- private$.offsets[x]
      } else {
        if (private$.cal$unit > 4L)
          stop("Parsing of timestamps on a 'month' or 'year' time unit is not supported.", call. = FALSE)

        xoff <- private$.cal$parse(as.character(x))$offset
        vals <- private$.offsets
        bnds <- self$get_bounds()
        if (!is.null(bnds)) {
          # Axis has bounds so get the closest coordinate first, allow for extremes
          intv <- .round(stats::approx(vals, 1L:length(vals), xoff, method = "linear", rule = 2)$y)
          # Test that xoff falls within the bounds of the coordinates
          valid <- if (rightmost.closed) (bnds[1L, intv] <= xoff) & (xoff <= bnds[2L, intv])
                   else (bnds[1L, intv] <= xoff) & (xoff < bnds[2L, intv])
          intv[!valid] <- NA
        } else {
          intv <- stats::approx(vals, 1L:length(vals), xoff, method = method)$y
          if (!rightmost.closed)
            intv[xoff == vals[length(vals)]] <- NA
          intv
        }
      }

      valid <- which(!is.na(intv))
      if (any(valid)) {
        t <- CFTime$new(private$.cal$definition, private$.cal$name, xoff[valid])
        bnds <- private$.bounds
        if (!is.null(bnds))
          t$set_bounds(bnds[, intv[valid], drop = FALSE])
        attr(intv, "CFTime") <- t
      }
      intv
    },

    #' @description Return boundary values.
    #' @param format A string specifying a format for output, optional.
    #' @return An array with `dims(2, length(offsets))` with values for the
    #'   boundaries. `NULL` if the boundaries have not been set.
    get_bounds = function(format) {
      bnds <- private$.bounds
      if (is.null(bnds) || missing(format)) return(bnds)

      ts <- private$.cal$offsets2time(as.vector(bnds))
      b <- .format_format(ts, private$.cal$timezone, format)
      dim(b) <- dim(bnds)
      b
    },

    #' @description Set or delete the boundary values of the `CFTime` instance.
    #' @param value The boundary values to set, in units of the offsets. A
    #'   matrix `(2, length(private$.offsets))`. If `NULL`, the boundaries are
    #'   deleted. If `TRUE`, make regular, consecutive boundaries.
    #' @return Self, invisibly.
    set_bounds = function(value) {
      if (is.null(value) || isFALSE(value)) private$.bounds <- NULL
      else if (isTRUE(value)) {
        len <- length(private$.offsets)
        b <- seq(from       = private$.offsets[1L] - private$.resolution * 0.5,
                 by         = private$.resolution,
                 length.out = len + 1L)
        private$.bounds <- rbind(b[1L:len], b[2L:(len+1L)])
      } else {
        off <- private$.offsets
        len <- length(off)

        warn <- NULL
        if (len == 0L)
          warn <- "Cannot set boundary values when there are no offsets"
        else if (is.matrix(value) && is.numeric(value)) {
          if (!all(dim(value) == c(2L, len)))
            warn <- "Replacement boundary values array has incorrect dimensions"
        } else warn <- "Replacement boundary values must be a numeric matrix or a single logical value"
        if (is.null(warn) && !(all(value[2L,] >= off) && all(off >= value[1L,])))
          warn <- "Values of the replacement boundary values must surround the offset values"

        if (is.null(warn))
          private$.bounds <- value
        else warning(warn, call. = FALSE)
      }
      invisible(self)
    },

    #' @description This method returns `TRUE` if the time series has uniformly
    #'   distributed time steps between the extreme values, `FALSE` otherwise.
    #'   First test without sorting; this should work for most data sets. If
    #'   not, only then offsets are sorted. For most data sets that will work
    #'   but for implied resolutions of month, season, year, etc based on a
    #'   "days" or finer calendar unit this will fail due to the fact that those
    #'   coarser units have a variable number of days per time step, in all
    #'   calendars except for `360_day`. For now, an approximate solution is
    #'   used that should work in all but the most non-conformal exotic
    #'   arrangements.
    #' @return `TRUE` if all time steps are equidistant, `FALSE` otherwise, or
    #'   `NA` if no offsets have been set.
    equidistant = function() {
      if (length(private$.offsets) == 0L) return(NA)
      out <- all(diff(private$.offsets) == private$.resolution)
      if (!out) {
        doff <- diff(sort(private$.offsets))
        out <- all(doff == private$.resolution)
        if (!out) {
          # Don't try to make sense of totally non-standard arrangements such as
          # calendar units "years" or "months" describing sub-daily time steps.
          # Also, 360_day calendar should be well-behaved so we don't want to get here.
          if (private$.cal$unit > 4L || inherits(private$.cal, "CFCalendar360")) return(FALSE)

          # Check if we have monthly or yearly data on a finer-scale calendar
          # This is all rather approximate but should be fine in most cases
          # This accommodates middle-of-the-time-period offsets as per the
          # CF Metadata Conventions
          # Please report problems at https://github.com/R-CF/CFtime/issues
          ddays <- range(doff) * CFt$units$per_day[private$.cal$unit]
          return((ddays[1] >= 28 && ddays[2] <= 31) ||      # months
                   (ddays[1] >= 8 && ddays[2] <= 11) ||     # dekads
                   (ddays[1] >= 90 && ddays[2] <= 92) ||    # seasons, quarters
                   (ddays[1] >= 365 && ddays[2] <= 366))    # years
        }
      }
      out
    },

    #' @description Given a vector of character timestamps, return a logical
    #'   vector of a length equal to the number of time steps in the time series
    #'   with values `TRUE` for those time steps that fall between the two
    #'   extreme values of the vector values, `FALSE` otherwise.
    #' @param extremes Character vector of timestamps that represent the
    #'   time period of interest. The extreme values are selected. Badly
    #'   formatted timestamps are silently dropped.
    #' @param closed Is the right side closed, i.e. included in the result?
    #'   Default is `FALSE`. A specification of `c("2022-01-01", "2023-01-01)`
    #'   will thus include all time steps that fall in the year 2022 when
    #'   `closed = FALSE` but include `2023-01-01` if that exact value is
    #'   present in the time series.
    #' @return A logical vector with a length equal to the number of time steps
    #'   in `self` with values `TRUE` for those time steps that fall between the
    #'   extreme values, `FALSE` otherwise.
    #'
    #'   An attribute 'CFTime' will have the same definition as `self` but with
    #'   offsets corresponding to the time steps falling between the two
    #'   extremes. If there are no values between the extremes, the attribute is
    #'   not set.
    slice = function(extremes, closed = FALSE) {
      if (!is.character(extremes) || length(extremes) < 1L)
        stop("Second argument must be a character vector of at least one timestamp.", call. = FALSE) # nocov

      off <- private$.offsets
      roff <- range(off)
      ext <- range(private$.cal$parse(extremes)$offset, na.rm = TRUE)
      if (all(is.na(ext)) || ext[1L] > roff[2L] || ext[2L] < roff[1L])
        out <- rep(FALSE, length(off))
      else {
        if (ext[1L] == ext[2L]) closed <- TRUE
        out <- if (closed) off >= ext[1L] & off <= ext[2L]
               else off >= ext[1L] & off < ext[2L]
        if (any(out)) {
          t <- CFTime$new(private$.cal$definition, private$.cal$name, off[out])
          bnds <- private$.bounds
          if (!is.null(bnds))
            t$set_bounds(bnds[, out, drop = FALSE])
          attr(out, "CFTime") <- t
        } else
          out <- rep(FALSE, length(off)) # nocov
      }
      out
    },

    #' @description Can the time series be converted to POSIXt?
    #' @return `TRUE` if the calendar support conversion to POSIXt, `FALSE`
    #' otherwise.
    POSIX_compatible = function() {
      private$.cal$POSIX_compatible(private$.offsets)
    },

    #' @description Create a factor for a `CFTime` instance.
    #'
    #'   When argument `breaks` is one of `"year", "season", "quarter", "month",
    #'   "dekad", "day"`, a factor is generated like by [CFfactor()]. When
    #'   `breaks` is a vector of character timestamps a factor is produced with
    #'   a level for every interval between timestamps. The last timestamp,
    #'   therefore, is only used to close the interval started by the
    #'   pen-ultimate timestamp - use a distant timestamp (e.g. `range(x)[2]`)
    #'   to ensure that all offsets to the end of the CFTime time series are
    #'   included, if so desired. The last timestamp will become the upper bound
    #'   in the `CFTime` instance that is returned as an attribute to this
    #'   function so a sensible value for the last timestamp is advisable.
    #'
    #'   This method works similar to [base::cut.POSIXt()] but there are some
    #'   differences in the arguments: for `breaks` the set of options is
    #'   different and no preceding integer is allowed, `labels` are always
    #'   assigned using values of `breaks`, and the interval is always
    #'   left-closed.
    #' @param breaks A character string of a factor period (see [CFfactor()] for
    #'   a description), or a character vector of timestamps that conform to the
    #'   calendar of `x`, with a length of at least 2. Timestamps must be given
    #'   in ISO8601 format, e.g. "2024-04-10 21:31:43".
    #' @return A factor with levels according to the `breaks` argument, with
    #'   attributes 'period', 'era' and 'CFTime'. When `breaks` is a factor
    #'   period, attribute 'period' has that value, otherwise it is '"day"'.
    #'   When `breaks` is a character vector of timestamps, attribute 'CFTime'
    #'   holds an instance of `CFTime` that has the same definition as `x`, but
    #'   with (ordered) offsets generated from the `breaks`. Attribute 'era'
    #'   is always -1.
    cut = function(breaks) {
      if (missing(breaks) || !is.character(breaks) || (len <- length(breaks)) < 1L)
        stop("Argument 'breaks' must be a character vector with at least 1 value", call. = FALSE) # nocov

      if(len == 1L) {
        breaks <- sub("s$", "", tolower(breaks))
        if (breaks %in% CFt$factor_periods)
          return(self$factor(breaks))
        else stop("Invalid specification of 'breaks'", call. = FALSE) # nocov
      }

      # breaks is a character vector of multiple timestamps
      if (private$.cal$unit > 4L) stop("Factorizing on a 'month' or 'year' time unit is not supported", call. = FALSE) # nocov
      time <- private$.cal$parse(breaks)
      if (anyNA(time$year))
        stop("Invalid specification of 'breaks'", call. = FALSE) # nocov
      sorted <- order(time$offset)
      ooff <- time$offset[sorted]
      intv <- findInterval(private$.offsets, ooff)
      intv[which(intv %in% c(0L, len))] <- NA
      f <- factor(intv, labels = breaks[sorted][1L:(len-1L)])

      # Attributes
      bnds <- rbind(ooff[1L:(len-1L)], ooff[2L:len])
      off  <- bnds[1L, ] + (bnds[2L, ] - bnds[1L, ]) * 0.5
      t <- CFTime$new(private$.cal$definition, private$.cal$name, off)
      t$set_bounds(bnds)
      attr(f, "period") <- "day"
      attr(f, "era")  <- -1L
      attr(f, "CFTime") <- t
      f
    },

    #' @description Generate a factor for the offsets, or a part thereof. This
    #'   is specifically interesting for creating factors from the date part of
    #'   the time series that aggregate the time series into longer time periods
    #'   (such as month) that can then be used to process daily CF data sets
    #'   using, for instance, `tapply()`.
    #'
    #'   The factor will respect the calendar that the time series is built on.
    #'
    #'   The factor will be generated in the order of the offsets. While typical
    #'   CF-compliant data sources use ordered time series there is, however, no
    #'   guarantee that the factor is ordered. For most processing with a factor
    #'   the ordering is of no concern.
    #'
    #'   If the `era` parameter is specified, either as a vector of years to
    #'   include in the factor, or as a list of such vectors, the factor will
    #'   only consider those values in the time series that fall within the list
    #'   of years, inclusive of boundary values. Other values in the factor will
    #'   be set to `NA`.
    #'
    #'   The following periods are supported by this method:
    #'
    #' \itemize{
    #'   \item `year`, the year of each offset is returned as "YYYY".
    #'   \item `season`, the meteorological season of each offset is returned as
    #'   "Sx", with x being 1-4, preceeded by "YYYY" if no `era` is
    #'   specified. Note that December dates are labeled as belonging to the
    #'   subsequent year, so the date "2020-12-01" yields "2021S1". This implies
    #'   that for standard CMIP files having one or more full years of data the
    #'   first season will have data for the first two months (January and
    #'   February), while the final season will have only a single month of data
    #'   (December).
    #'   \item `quarter`, the calendar quarter of each offset is returned as "Qx",
    #'   with x being 1-4, preceeded by "YYYY" if no `era` is specified.
    #'   \item `month`, the month of each offset is returned as "01" to
    #'   "12", preceeded by "YYYY-" if no `era` is specified. This is the default
    #'   period.
    #'   \item `dekad`, ten-day periods are returned as
    #'   "Dxx", where xx runs from "01" to "36", preceeded by "YYYY" if no `era`
    #'   is specified. Each month is subdivided in dekads as follows: 1- days 01 -
    #'   10; 2- days 11 - 20; 3- remainder of the month.
    #'   \item `day`, the month and day of each offset are returned as "MM-DD",
    #'   preceeded by "YYYY-" if no `era` is specified.
    #' }
    #'
    #'   It is not possible to create a factor for a period that is shorter than
    #'   the temporal resolution of the calendar. As an example, if the calendar
    #'   has a monthly unit, a dekad or day factor cannot be created.
    #'
    #'   Creating factors for other periods is not supported by this method.
    #'   Factors based on the timestamp information and not dependent on the
    #'   calendar can trivially be constructed from the output of the
    #'   [as_timestamp()] function.
    #'
    #'   Attribute 'CFTime' of the result contains a `CFTime` instance that is
    #'   valid for the result of applying the factor to a resource that this
    #'   instance is associated with. In other words, if `CFTime` instance 'At'
    #'   describes the temporal dimension of resource 'A' and a factor 'Af' is
    #'   generated from `Af <- At$factor()`, then `Bt <- attr(Af, "CFTime")`
    #'   describes the temporal dimension of the result of, say, `B <- apply(A,
    #'   1:2, tapply, Af, FUN)`. The 'CFTime' attribute contains a
    #'   [CFClimatology] instance for era factors.
    #' @param period character. A character string with one of the values
    #'   "year", "season", "quarter", "month" (the default), "dekad" or "day".
    #' @param era numeric or list, optional. Vector of years for which to
    #'   construct the factor, or a list whose elements are each a vector of
    #'   years. The extreme values of the supplied vector will be used. Note
    #'   that if a single year is specified that the result is valid, but not a
    #'   climatological statistic. If `era` is not specified, the factor will
    #'   use the entire time series for the factor.
    #' @return If `era` is a single vector or `NULL`, a factor with a length
    #'   equal to the number of offsets in this instance. If `era` is a list, a
    #'   list with the same number of elements and names as `era`, each
    #'   containing a factor. Elements in the factor will be set to `NA` for
    #'   time series values outside of the range of specified years.
    #'
    #'   The factor, or factors in the list, have attributes 'period', 'era' and
    #'   'CFTime'. Attribute 'period' holds the value of the `period` argument.
    #'   Attribute 'era' indicates the number of years that are included in the
    #'   era, or -1 if no `era` is provided. Attribute 'CFTime' holds an
    #'   instance of `CFTime` or `CFClimatology` that has the same definition as
    #'   this instance, but with offsets corresponding to the mid-point of the
    #'   factor levels.
    factor = function(period = "month", era = NULL) {
      if (length(private$.offsets) < 10L) stop("Cannot create a factor for very short time series", call. = FALSE) # nocov

      period <- tolower(period)
      if (!((length(period) == 1L) && (period %in% CFt$factor_periods)))
        stop("Period specifier must be a single value of a supported period", call. = FALSE) # nocov

      # No fine-grained period factors for coarse source data
      timestep <- CFt$units$seconds[private$.cal$unit] * private$.resolution;
      if ((period == "year") && (timestep > 86400 * 366) ||
          (period %in% c("season", "quarter")) && (timestep > 86400 * 90) || # Somewhat arbitrary
          (period == "month") && (timestep > 86400 * 31) ||
          (period == "dekad") && (timestep > 86400) ||       # Must be constructed from daily or finer data
          (period == "day") && (timestep > 86400))           # Must be no longer than a day
        stop("Cannot produce a short period factor from source data with long time interval", call. = FALSE) # nocov

      time <- private$.cal$offsets2time(private$.offsets)
      months <- c("01", "02", "03", "04", "05", "06", "07", "08", "09", "10", "11", "12")

      if (is.null(era)) {
        # Create the factor for the specified period as well as bounds dates for a
        # new CFtime instance for the factor. Lower bounds for the factor level is
        # easy, upper bound of last level takes effort.
        switch(period,
               "year"    = {
                 out <- as.factor(sprintf("%04d", time$year))
                 l  <- levels(out)
                 dt <- c(paste0(l, "-01-01"), sprintf("%04d-01-01", as.integer(l[nlevels(out)]) + 1L))
               },
               "season"  = {
                 if (!requireNamespace("stringr"))
                   stop("Must install package `stringr` to use this functionality.", call. = FALSE) # nocov

                 out <- as.factor(
                   ifelse(time$month == 12L, sprintf("%04dS1", time$year + 1L),
                          sprintf("%04dS%d", time$year, time$month %/% 3L + 1L)))
                 l  <- levels(out)
                 dt <- ifelse(substr(l, 6L, 6L) == "1", paste0(as.integer(substr(l, 1L, 4L)) - 1L, "-12-01"),
                              stringr::str_replace_all(l, c("S2" = "-03-01", "S3" = "-06-01", "S4" = "-09-01")))
                 ll <- l[nlevels(out)]
                 lp <- as.integer(substr(ll, 6L, 6L))
                 if (lp == 1L)
                   dt <- c(dt, sprintf("%04d-03-01", as.integer(substr(ll, 1L, 4L))))
                 else dt <- c(dt, sprintf("%s-%02d-01", substr(ll, 1L, 4L), lp * 3L))
               },
               "quarter" = {
                 if (!requireNamespace("stringr"))
                   stop("Must install package `stringr` to use this functionality.", call. = FALSE) # nocov

                 out <- as.factor(sprintf("%04dQ%d", time$year, (time$month - 1L) %/% 3L + 1L))
                 l  <- levels(out)
                 dt <- stringr::str_replace_all(l, c("Q1" = "-01-01", "Q2" = "-04-01", "Q3" = "-07-01", "Q4" = "-10-01"))
                 ll <- l[nlevels(out)]
                 lp <- as.integer(substr(ll, 6L, 6L))
                 if (lp == 4L)
                   dt <- c(dt, sprintf("%04d-01-01", as.integer(substr(ll, 1L, 4L)) + 1L))
                 else dt <- c(dt, sprintf("%s-%02d-01", substr(ll, 1L, 4L), lp * 3L + 1L))
               },
               "month" = {
                 out <- as.factor(sprintf("%04d-%s", time$year, months[time$month]))
                 l  <- levels(out)
                 dt <- paste0(l, "-01")
                 ll <- l[nlevels(out)]
                 lp <- as.integer(substr(ll, 6L, 7L))
                 if (lp == 12L)
                   dt <- c(dt, sprintf("%04d-01-01", as.integer(substr(ll, 1L, 4L)) + 1L))
                 else dt <- c(dt, sprintf("%s-%02d-01", substr(ll, 1L, 4L), lp + 1L))
               },
               "dekad" = {
                 out <- as.factor(sprintf("%04dD%02d", time$year, (time$month - 1L) * 3L + pmin.int((time$day - 1L) %/% 10L + 1L, 3L)))
                 l  <- levels(out)
                 dk <- as.integer(substr(l, 6L, 7L)) - 1L
                 dt <- sprintf("%s-%02d-%s", substr(l, 1L, 4L), dk %/% 3L + 1L, c("01", "11", "21")[dk %% 3L + 1L])
                 ll <- l[nlevels(out)]
                 lp <- as.integer(substr(ll, 6L, 7L))
                 yr <- as.integer(substr(ll, 1L, 4L))
                 if (lp == 36L)
                   dt <- c(dt, sprintf("%04d-01-01", yr + 1L))
                 else dt <- c(dt, sprintf("%04d-%02d-%s", yr, lp %/% 3L + 1L, c("01", "11", "21")[lp %% 3L + 1L]))
               },
               "day" = {
                 out <- as.factor(sprintf("%04d-%02d-%02d", time$year, time$month, time$day))
                 final <- levels(out)[nlevels(out)]
                 yr <- as.integer(substring(final, 1, 4))
                 mon <- as.integer(substring(final, 6, 7))
                 day <- as.integer(substring(final, 9, 10))
                 post <- private$.cal$add_day(data.frame(year = yr, month = mon, day = day))
                 post <- sprintf("%04d-%02d-%02d", post$year, post$month, post$day)
                 dt <- c(levels(out), post)
               }
        )

        # Convert bounds dates to an array of offsets, find mid-points, create new CFTime instance
        off  <- private$.cal$parse(dt)$offset
        off[is.na(off)] <- 0     # This can happen only when the time series starts at or close to the origin, for seasons
        noff <- length(off)
        bnds <- rbind(off[1L:(noff - 1L)], off[2L:noff])
        off  <- bnds[1L,] + (bnds[2L,] - bnds[1L,]) * 0.5
        new_time <- CFTime$new(private$.cal$definition, private$.cal$name, off)
        new_time$set_bounds(bnds)

        # Bind attributes to the factor
        attr(out, "era") <- -1L
        attr(out, "period") <- period
        attr(out, "CFTime") <- new_time
        return(out)
      }

      # Era factor
      if (is.numeric(era)) ep <- list(era)
      else if ((is.list(era) && all(unlist(lapply(era, is.numeric))))) ep <- era
      else stop("When specified, the `era` parameter must be a numeric vector or a list thereof", call. = FALSE)

      out <- lapply(ep, function(years) {
        years <- as.integer(years)
        yrs <- range(years)
        mid_yr <- as.integer(yrs[1L] + (yrs[2L] - yrs[1L]) * 0.5)
        switch(period,
          "year" = {
            f <- as.factor(ifelse(time$year %in% years, sprintf("%04d", time$year), NA_character_))
            n <- nlevels(f)
            if (n > 0L) {
              off <- private$.cal$parse(paste0(levels(f), "-07-01"))$offset
              last <- as.integer(levels(f)[n]) + 1L
              bnds <- c(rep(paste0(levels(f)[1L], "-01-01"), n), rep(paste0(last, "-01-01"), n))
            }
          },
          "season" = {
            f <- as.factor(ifelse((time$month == 12L) & ((time$year + 1L) %in% years), "S1",
                             ifelse((time$month < 12L) & (time$year %in% years), sprintf("S%d", time$month %/% 3L + 1L), NA_character_)))
            if (nlevels(f) > 0L) {
              s <- as.integer(substr(levels(f), 2, 2))
              mid_s <- c("-01-16T12", "-04-16", "-07-16T12", "-10-16T12")
              off <- private$.cal$parse(paste0(mid_yr, mid_s[s]))$offset
              bnds <- c(ifelse(s == 1L, sprintf("%d-12-01", yrs[1L] - 1L),
                                        sprintf("%d-%02d-01", yrs[1L], (s - 1L) * 3L)),
                        sprintf("%d-%02d-01", yrs[2L], s * 3L))
            }
          },
          "quarter" = {
            f <- as.factor(ifelse(time$year %in% years, sprintf("Q%d", (time$month - 1L) %/% 3L + 1L), NA_character_))
            if (nlevels(f) > 0L) {
              q <- as.integer(substr(levels(f), 2, 2))
              mid_q <- c("-02-15", "-05-16T12", "-08-16T12", "-11-16")
              off <- private$.cal$parse(paste0(mid_yr, mid_q[q]))$offset
              bnds <- c(sprintf("%d-%02d-01", yrs[1L], (q - 1L) * 3L + 1L),
                        ifelse(q == 4L, sprintf("%d-01-01", yrs[2L] + 1L),
                                        sprintf("%d-%02d-01", yrs[2L], q * 3L + 1L)))
            }
          },
          "month" = {
            f <- as.factor(ifelse(time$year %in% years, months[time$month], NA_character_))
            if (nlevels(f) > 0L) {
              mons <- as.numeric(levels(f))
              mid_mon <- c("-16T12", "-15", "-16T12", "-16", "-16T12", "-16", "-16T12", "-16T12", "-16", "-16T12", "-16", "-16T12")
              off <- private$.cal$parse(paste0(mid_yr, "-", levels(f), mid_mon[mons]))$offset
              bnds <- c(paste0(yrs[1L], "-", levels(f), "-01"),
                        ifelse(mons == 12L, sprintf("%d-01-01", yrs[2L] + 1L),
                                            sprintf("%d-%02d-01", yrs[2L], mons + 1L)))
            }
          },
          "dekad" = {
            f <- as.factor(ifelse(time$year %in% years, sprintf("D%02d", (time$month - 1L) * 3L + pmin.int((time$day - 1L) %/% 10L + 1L, 3L)), NA_character_))
            if (nlevels(f) > 0L) {
              dek <- as.integer(substr(levels(f), 2, 3))
              mon <- (dek - 1L) %/% 3L + 1L
              mid_dek <- c("06", "16", "26T12", "06", "16", "25",    "06", "16", "26T12", "06", "16", "26",
                           "06", "16", "26T12", "06", "16", "26",    "06", "16", "26T12", "06", "16", "26T12",
                           "06", "16", "26",    "06", "16", "26T12", "06", "16", "26",    "06", "16", "26T12")
              off <- private$.cal$parse(sprintf("%d-%02d-%s", mid_yr, mon, mid_dek[dek]))$offset
              day <- c("01", "11", "21")
              y2 <- rep(yrs[2L], length(dek))
              m2 <- mon
              ndx <- which(!(dek %% 3L))
              m2[ndx] <- m2[ndx] + 1L
              ndx <- which(dek == 36L)
              y2[ndx] <- yrs[2L] + 1L
              m2[ndx] <- 1L
              bnds <- c(sprintf("%d-%02d-%s", yrs[1L], mon, day[(dek - 1L) %% 3L + 1L]),
                        sprintf("%d-%02d-%s", y2, m2, day[dek %% 3L + 1L]))
            }
          },
          "day" = {
            f <- as.factor(ifelse(time$year %in% years, sprintf("%s-%02d", months[time$month], time$day), NA_character_))
            if (nlevels(f) > 0L) {
              mons <- as.integer(substr(levels(f), 1, 2))
              days <- as.integer(substr(levels(f), 4, 5))
              off <- private$.cal$parse(sprintf("%d-%s-%02dT12", mid_yr, months[mons], days))$offset
              # Add a day to the calendar for the final year
              y2off1 <- private$.cal$parse(paste0(yrs[2L], "-", levels(f)))$offset + CFt$units$per_day[private$.cal$unit]
              y2off1 <- private$.cal$offsets2time(y2off1)
              bnds <- c(sprintf("%d-%s-%02d", yrs[1L], months[mons], days),
                        sprintf("%d-%02d-%02d", y2off1$year, y2off1$month, y2off1$day))
            }
          }
        )
        attr(f, "era") <- length(years)
        attr(f, "period") <- period
        if (nlevels(f) > 0L) {
          bnds <- matrix(private$.cal$parse(bnds)$offset, nrow = 2L, byrow = TRUE)
          attr(f, "CFTime") <- CFClimatology$new(private$.cal$definition, private$.cal$name, off, bnds)
        }
        f
      })
      if (is.numeric(era)) out <- out[[1L]]
      else names(out) <- names(era)
      out
    },

    #' @description Given a factor as produced by `CFTime$factor()`, this method
    #'   will return a numeric vector with the number of time units in each
    #'   level of the factor.
    #'
    #'   The result of this method is useful to convert between absolute and
    #'   relative values. Climate change anomalies, for instance, are usually
    #'   computed by differencing average values between a future period and a
    #'   baseline period. Going from average values back to absolute values for
    #'   an aggregate period (which is typical for temperature and
    #'   precipitation, among other variables) is easily done with the result of
    #'   this method, without having to consider the specifics of the calendar
    #'   of the data set.
    #'
    #'   If the factor `f` is for an era (e.g. spanning multiple years and the
    #'   levels do not indicate the specific year), then the result will
    #'   indicate the number of time units of the period in a regular single
    #'   year. In other words, for an era of 2041-2060 and a monthly factor on a
    #'   standard calendar with a `days` unit, the result will be
    #'   `c(31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)`. Leap days are thus
    #'   only considered for the `366_day` and `all_leap` calendars.
    #'
    #'   Note that this function gives the number of time units in each level of
    #'   the factor - the actual number of data points in the time series per
    #'   factor level may be different. Use [CFfactor_coverage()] to determine
    #'   the actual number of data points or the coverage of data points
    #'   relative to the factor level.
    #' @param f A factor or a list of factors derived from the method
    #'   `CFTime$factor()`.
    #' @return If `f` is a factor, a numeric vector with a length equal to the
    #'   number of levels in the factor, indicating the number of time units in
    #'   each level of the factor. If `f` is a list of factors, a list with each
    #'   element a numeric vector as above.
    factor_units = function(f) {
      if (is.list(f)) factors <- f else factors <- list(f)
      if (!(all(unlist(lapply(factors, function(x) is.factor(x) && is.numeric(attr(x, "era")) &&
                              attr(x, "period") %in% CFt$factor_periods)))))
        stop("Argument `f` must be a factor generated by the method `CFTime$factor()`", call. = FALSE) # nocov

      out <- lapply(factors, function(fac) .factor_units(fac, private$.cal, CFt$units$per_day[private$.cal$unit]))
      if (is.factor(f)) out <- out[[1L]]
      out
    },

    #' @description Calculate the number of time elements, or the relative
    #' coverage, in each level of a factor generated by `CFTime$factor()`.
    #' @param f A factor or a list of factors derived from the method
    #'   `CFTime$factor()`.
    #' @param coverage "absolute" or "relative".
    #' @return If `f` is a factor, a numeric vector with a length equal to the
    #'   number of levels in the factor, indicating the number of units from the
    #'   time series contained in each level of the factor when
    #'   `coverage = "absolute"` or the proportion of units present relative to the
    #'   maximum number when `coverage = "relative"`. If `f` is a list of factors, a
    #'   list with each element a numeric vector as above.
    factor_coverage = function(f, coverage = "absolute") {
      if (is.list(f)) factors <- f else factors <- list(f)
      if (!(all(unlist(lapply(factors, function(x) is.factor(x) && is.numeric(attr(x, "era")) &&
                              attr(x, "period") %in% CFt$factor_periods)))))
        stop("Argument `f` must be a factor generated by the method `CFTime$factor()`", call. = FALSE) # nocov

      if (!(is.character(coverage) && coverage %in% c("absolute", "relative")))
        stop("Argument `coverage` must be a character string with a value of 'absolute' or 'relative'", call. = FALSE) # nocov

      if (coverage == "relative") {
        out <- lapply(factors, function(fac) {
          res <- tabulate(fac) / .factor_units(fac, private$.cal, CFt$units$per_day[private$.cal$unit])
          yrs <- attr(fac, "era")
          if (yrs > 0) res <- res / yrs
          return(res)
        })
      } else {
        out <- lapply(factors, tabulate)
      }

      if (is.factor(f)) out <- out[[1L]]
      out
    },

    #' @description Create a copy of the current instance. The copy is
    #'   completely separate from the current instance.
    #' @return A new `CFTime` instance with identical definition and set of
    #'   timestamps.
    copy = function() {
      new_time <- CFTime$new(private$.cal$definition, private$.cal$name, private$.offsets)
      if (!is.null(private$.bounds))
        new_time$set_bounds(private$.bounds)
      new_time
    },

    #' @description Get a new `CFTime` instance that is a subset of the current
    #' instance, including any boundary values.
    #' @param rng The numeric range of indices to subset this instance to.
    #' @return A new `CFTime` instance with identical definition and set of
    #'   timestamps according to the `rng` argument.
    subset = function(rng) {
      rng <- range(rng)
      if (!is.numeric(rng) || rng[1L] < 1L || rng[2L] > length(private$.offsets))
        stop("Range to subset is invalid.", call. = FALSE) # nocov

      new_time <- CFTime$new(private$.cal$definition, private$.cal$name, private$.offsets[rng[1L]:rng[2L]])
      if (!is.null(private$.bounds))
        new_time$set_bounds(private$.bounds[ , rng[1L]:rng[2L]])
      new_time
    },

    #' @description Retrieve the day-of-year number, either from the timestamps
    #'   in the current instance or from the supplied `data.frame` of year,
    #'   month and day. The `doy` is calendar-aware.
    #' @param ymd Optional `data.frame` with dates parsed into their parts in
    #'   columns `year`, `month` and `day`. Any other columns are disregarded.
    #' @return An integer vector of the same length as argument `ymd` or the
    #'   internal offsets if `ymd` is not specified with the day-of-year in the
    #'   calendar of this instance. Rows in argument `ymd` that do not hold a
    #'   valid date are returned as `NA`.
    doy = function(ymd) {
      if (missing(ymd) || is.null(ymd))
        ymd <- private$.cal$offsets2time(private$.offsets)
      else {
        valid <- private$.cal$valid_days(ymd)
        ymd$day[!valid] <- NA_integer_
      }
      if (!nrow(ymd))
        return(integer(0))

      private$.cal$doy(ymd)
    }
  ),
  active = list(
    #' @field calendar (read-only) The calendar of this `CFTime` instance, a
    #'   descendant of the [CFCalendar] class.
    calendar = function(value) {
      if (missing(value))
        private$.cal
    },

    #' @field cal (read-only) DEPRECATED. Use the `calendar` field instead.
    cal = function(value) {
      self$calendar
    },

    #' @field unit (read-only) The unit string of the calendar and time series.
    unit = function(value) {
      if (missing(value))
        CFt$units$name[private$.cal$unit]
    },

    #' @field offsets (read-only) A numeric vector of offsets from the origin of
    #'   the calendar.
    offsets = function(value) {
      if (missing(value))
        private$.offsets
    },

    #' @field resolution (read-only) The average number of time units between
    #'   offsets.
    resolution = function(value) {
      if (missing(value)) {
        private$.resolution
      }
    },

    #' @field length (read-only) Retrieve the number of offsets in the time
    #'   series.
    length = function(value) {
      if (missing(value))
        length(private$.offsets)
    },

    #' @field bounds Retrieve or set the bounds for the offsets. On setting, a
    #'   `matrix` with columns for `offsets` and low values in the first row,
    #'   high values in the second row. This may be simply `TRUE` to set regular
    #'   and consecutive bounds.
    bounds = function(value) {
      if (missing(value))
        private$.bounds
      else
        self$set_bounds(value)
    },

    #' @field friendlyClassName Character string with class name for display
    #'   purposes.
    friendlyClassName = function(value) {
      if (missing(value))
        "CFTime"
    }
  )
)

# ==============================================================================
# S3 functions

#' Compact display of a CFTime instance
#' @param object A `CFTime` instance.
#' @param ... Ignored.
#' @export
str.CFTime <- function(object, ...) {
  cat(object$friendlyClassName, " with origin [", object$calendar$definition,
      "] using calendar [", object$calendar$name,
      "] having ", length(object$offsets), " offset values", sep = "")
}

#' Subset a `CFTime` instance by position in the time series
#'
#' @param x A `CFTime` instance.
#' @param i A vector a positive integer values to indicate which values to
#'   extract from the time series by position. If negative values are passed,
#'   their positive counterparts will be excluded and then the remainder
#'   returned. Positive and negative values may not be mixed.
#' @param ... Ignored.
#' @return A numeric vector with those values of `i` (or the inverse, when
#'   negative) that are valid in `x`. If there is at least 1 valid result, then
#'   attribute "CFTime" of the returned value contains an instance of `CFTime`
#'   that describes the dimension of filtering the dataset associated with `x`
#'   with the result of this function, excluding any `NA` values.
#' @export
#' @references https://github.com/R-CF/CFtime/issues/20
"[.CFTime" = function(x, i = TRUE, ...) {
  x$indexOf(i)
}

test_that("CFClimatology works", {
  # Must have bounds
  expect_error(CFClimatology$new("days since 2025-01-01", "360_day", 15 + 30 * 0:11))

  # Bad bounds
  expect_warning(CFClimatology$new("days since 2025-01-01", "360_day", 15 + 30 * 0:11,
                                   matrix(c(80 * 0:11, 80 * 1:12), nrow = 2, byrow = T)))

  clim <- CFClimatology$new("days since 2025-01-01", "360_day", 15 + 30 * 0:11,
                            matrix(c(30 * 0:11, 30 * 1:12), nrow = 2, byrow = T))
  expect_true(inherits(clim, "CFClimatology"))
  expect_equal(clim$period, "month")
  expect_equal(clim$years, c(2025L, 2025L))
  x <- capture.output(clim$print())
  expect_match(x[6], "average of 30\\.000000 days between 12 elements\\)$")
  expect_equal(x[7], "  Period  : month")
  expect_equal(x[8], "  Year    : 2025")
})

test_that("Creating timestamps", {
  expect_error(as_timestamp("1-2-3"))          # No CFtime as first argument

  cf <- CFTime$new("hours since 2001-01-01", "365_day")
  expect_null(cf$as_timestamp()) # No offsets
  cf <- cf + 0L:2399L
  expect_error(cf$as_timestamp("d"))          # Wrong format specifier
  expect_error(cf$as_timestamp(asPOSIX = T))  # No POSIXt on a non-standard calendar
  expect_equal(length(cf$as_timestamp()), 2400L)

  cf <- CFTime$new("days since 2001-01-01", "standard", 0L:364L)
  expect_equal(length(cf$as_timestamp()), 365L)
  expect_equal(nchar(cf$as_timestamp()[1]), 10L) # date string
  expect_equal(length(cf$as_timestamp("date", TRUE)), 365L)
  expect_equal(length(cf$as_timestamp("timestamp", TRUE)), 365L)
  expect_equal(cf$as_timestamp(), as.character(cf))

  cf <- cf + 364.56764
  expect_equal(cf$range()[2], "2001-12-31T13:37:24.096")
})

test_that("Using format()", {
  cf <- CFTime$new("days since 2001-01-01 18:10:30-04", "365_day")
  expect_equal(length(cf$format()), 0L) # no offsets

  cf <- cf + 0:364

  expect_equal(cf$format()[1], "2001-01-01T18:10:30")      # format parameter missing
  expect_error(cf$format(123)) # format parameter must be character
  expect_error(cf$format(c("doesn't", "work", "either")))

  expect_equal(cf$format("Timestamp is: %%%F%%")[1], "Timestamp is: %2001-01-01%")
  expect_equal(cf$format("Timestamp is: %R")[1], "Timestamp is: 18:10")
  expect_equal(cf$format("%T%z")[1], "18:10:30-0400")
  #expect_equal(cf$format("%b")[c(1, 32, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335)], month.abb)  # en_EN only
  #expect_equal(cf$format("%B")[c(1, 32, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335)], month.name) #
  expect_equal(cf$format("%Od-%e-%I%p")[5], "05- 5-06PM")
})

test_that("Factor testing", {
  # No offsets
  cf <- CFTime$new("days since 2000-01-01", "365_day")
  expect_error(CFfactor())
  expect_error(CFfactor(cf))
  expect_error(CFfactor(cf, "zxcv"))
  expect_error(CFfactor(cf, c("day", "month")))
  expect_error(CFfactor(cf, "hour"))
  expect_error(CFfactor(cf, "month", "bad"))
  month_days <- c(31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
  leap_month_days <- c(31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
  dekad_days <- c(10, 10, 11, 10, 10, 8, 10, 10, 11, 10, 10, 10, 10, 10, 11, 10, 10, 10, 10, 10, 11, 10, 10, 11, 10, 10, 10, 10, 10, 11, 10, 10, 10, 10, 10, 11)
  leap_dekad_days <- c(10, 10, 11, 10, 10, 9, 10, 10, 11, 10, 10, 10, 10, 10, 11, 10, 10, 10, 10, 10, 11, 10, 10, 11, 10, 10, 10, 10, 10, 11, 10, 10, 10, 10, 10, 11)

  # Few offsets
  cf <- cf + 365L:370L
  expect_error(CFfactor(cf))

  cf <- cf + 371L:7664L # 20 years of offsets

  # Regular factors for all available periods
  np <- c(20, 81, 80, 240, 720, 7300)
  first <- c("2001", "2001S1", "2001Q1", "2001-01", "2001D01", "2001-01-01")
  last <-  c("2020", "2021S1", "2020Q4", "2020-12", "2020D36", "2020-12-31")
  for (p in 1:6) {
    f <- CFfactor(cf, CFt$factor_periods[p])
    expect_equal(as.character(f)[1L], first[p])
    expect_equal(as.character(f)[7300L], last[p])
    newcf <- attr(f, "CFTime")
    bnds <- bounds(newcf)
    expect_equal(definition(cf), definition(newcf))
    expect_true(is.matrix(bnds))
    expect_type(bnds, "double")
    expect_equal(dim(bnds), c(2, np[p]))
  }

  # Era factors for all available periods
  eras <- list(first = 2001L, double = 2002L:2003L, final3 = 2018L:2020L, outside = 2022L)
  lvls <- c(1L, 4L, 4L, 12L, 36L, 365L)
  expect_error(cf$factor("month", "bad"))
  for (p in 1:6) { # year, season, quarter, month, dekad, day
    f <- cf$factor(CFt$factor_periods[p], eras)
    expect_type(f, "list")
    expect_equal(length(f), 4L)
    expect_equal(length(f$first), 7300L)
    expect_equal(attr(f$first, "period"), CFt$factor_periods[p])
    expect_equal(attr(f$first, "era"), 1L)
    expect_null(attr(f$first, "zxcv"))
    if (p == 1L) {
      expect_equal(length(levels(f$first)), 1L)
      expect_equal(length(levels(f$double)), 2L)
      expect_equal(length(levels(f$final3)), 3L)
      expect_equal(length(levels(f$outside)), 0L)
    } else if (p %in% c(2L, 3L)) {
      expect_equal(length(levels(f$first)), 4L)
      expect_equal(length(levels(f$double)), 4L)
      expect_equal(length(levels(f$final3)), 4L)
      expect_equal(length(levels(f$outside)), 0L)
    } else {
      expect_equal(length(levels(f$first)), lvls[p])
      expect_equal(length(levels(f$double)), lvls[p])
      expect_equal(length(levels(f$final3)), lvls[p])
      expect_equal(length(levels(f$outside)), 0L)
    }
    lapply(1:3, function(n) {
      t <- attr(f[[n]], "CFTime")
      expect_true(inherits(t, "CFClimatology"))
      expect_equal(t$years, matrix(c(2001, 2001, 2002, 2003, 2018, 2020), nrow = 2)[,n])
      expect_equal(t$period, CFt$factor_periods[p])
    })
    expect_null(attr(f[[4]], "CFTime"))
  }

  # Single era value for all available periods
  for (p in 1:6) { # year, season, quarter, month, dekad, day
    f <- CFfactor(cf, CFt$factor_periods[p], 2002L)
    expect_s3_class(f, "factor")
    expect_equal(length(f), 7300L)
    expect_equal(length(levels(f)), lvls[p])
    expect_equal(attr(f, "period"), CFt$factor_periods[p])
    expect_equal(attr(f, "era"), 1L)
    expect_null(attr(f, "zxcv"))
  }

  # Units and coverage in factor levels
  expect_error(CFfactor_units("zxcv"))
  expect_error(CFfactor_units(cf, "zxcv"))
  expect_error((CFfactor_units(cf, list(12))))
  expect_error((CFfactor_units(cf, factor(letters))))
  expect_error((CFfactor_units(cf, list(factor(letters)))))

  f <- CFfactor(cf, "year")
  expect_true(all(CFfactor_units(cf, f) == 365L))
  expect_true(all(CFfactor_coverage(cf, f, "absolute") == 365L))
  expect_true(all(CFfactor_coverage(cf, f, "relative") == 1L))

  f <- CFfactor(cf, "season")
  expect_equal(sum(CFfactor_units(cf, f)), 7390L) # 20 yrs * 365 plus 90 bc DJF season present on both ends
  expect_true(all(CFfactor_units(cf, f) %in% 90L:92L))
  x <- CFfactor_coverage(cf, f, "absolute")
  expect_equal(x[1L], 59L)  # Jan + Feb of first year
  expect_equal(x[81L], 31L) # Dec of last year
  expect_true(all(x[2L:80L] %in% 90L:92L))
  x <- CFfactor_coverage(cf, f, "relative")
  expect_equal(x[1L] + x[81L], 1L)
  expect_true(all(x[2L:80L] == 1L))

  f <- CFfactor(cf, "quarter")
  expect_equal(sum(CFfactor_units(cf, f)), 7300L)
  expect_true(all(CFfactor_units(cf, f) %in% 90L:92L))
  x <- CFfactor_coverage(cf, f, "absolute")
  expect_equal(x[1L], 90L)
  expect_equal(x[80L], 92L)
  expect_true(all(x %in% 90L:92L))
  x <- CFfactor_coverage(cf, f, "relative")
  expect_true(all(x == 1L))

  f <- CFfactor(cf, "month")
  expect_equal(sum(CFfactor_units(cf, f)), 7300L)
  expect_true(all(CFfactor_coverage(cf, f, "absolute") == month_days))
  expect_true(all(CFfactor_coverage(cf, f, "relative") == 1L))

  f <- CFfactor(cf, "dekad")
  expect_equal(sum(CFfactor_units(cf, f)), 7300L)
  x <- CFfactor_coverage(cf, f, "absolute")
  expect_true(all(x == dekad_days))
  expect_true(all(CFfactor_coverage(cf, f, "relative") == 1L))

  f <- CFfactor(cf, "day")
  expect_equal(sum(CFfactor_units(cf, f)), 7300L)
  expect_true(all(CFfactor_coverage(cf, f, "absolute") == 1L))
  expect_true(all(CFfactor_coverage(cf, f, "relative") == 1L))

  # 360_day calendar
  cf360 <- CFtime("days since 2001-01-01", "360_day", 0:7199)
  f <- CFfactor(cf360, "month")
  expect_equal(sum(CFfactor_units(cf360, f)), 7200L)
  expect_true(all(CFfactor_coverage(cf360, f, "absolute") == 30L))
  expect_true(all(CFfactor_coverage(cf360, f, "relative") == 1L))

  # Units and coverage in factor levels with eras
  f <- CFfactor(cf, "year", eras)
  expect_true(all(unlist(CFfactor_units(cf, f)) == rep(365L, 6L)))
  expect_true(all(unlist(CFfactor_coverage(cf, f, "absolute")) == c(rep(365L, 6L), 0L)))
  expect_equal(sum(sapply(CFfactor_coverage(cf, f, "relative"), sum)), 3L)

  f <- CFfactor(cf, "season", eras)
  expect_true(all(sapply(CFfactor_units(cf, f), function(x) {all(x == c(90L, 92L, 92L, 91L))})))
  x <- CFfactor_coverage(cf, f, "absolute")
  expect_equal(x$first[1L], 59L)  # Jan + Feb of first year at beginning of time series
  expect_equal(x$double, c(180L, 184L, 184L, 182L)) # two full years
  expect_equal(x$final3, c(270L, 276L, 276L, 273L)) # three full years
  x <- unlist(CFfactor_coverage(cf, f, "relative"))
  expect_true(all(x[2L:12L] == 1L))

  f <- CFfactor(cf, "quarter", eras)
  expect_true(all(sapply(CFfactor_units(cf, f), function(x) {all(x == c(90L, 91L, 92L, 92L))})))
  x <- CFfactor_coverage(cf, f, "absolute")
  expect_equal(x$first[1L], 90L)
  expect_equal(x$double, c(180L, 182L, 184L, 184L)) # two full years
  expect_equal(x$final3, c(270L, 273L, 276L, 276L)) # three full years
  x <- unlist(CFfactor_coverage(cf, f, "relative"))
  expect_true(all(x[2L:12L] == 1L))

  f <- CFfactor(cf, "month", eras)
  expect_true(all(sapply(CFfactor_units(cf, f), function(x) {all(x == month_days)})))
  x <- CFfactor_coverage(cf, f, "absolute")
  expect_true(all(x$first == month_days))
  expect_true(all(x$double == month_days * 2L))
  expect_true(all(x$final3 == month_days * 3L))
  expect_true(all(unlist(CFfactor_coverage(cf, f, "relative")) == 1L))

  f <- CFfactor(cf, "dekad", eras)
  expect_true(all(sapply(CFfactor_units(cf, f), function(x) {all(x == dekad_days)})))
  x <- CFfactor_coverage(cf, f, "absolute")
  expect_true(all(x$first == dekad_days))
  expect_true(all(x$double == dekad_days * 2L))
  expect_true(all(x$final3 == dekad_days * 3L))
  expect_true(all(unlist(CFfactor_coverage(cf, f, "relative")) == 1L))

  f <- CFfactor(cf, "day", eras)
  expect_true(all(unlist(CFfactor_units(cf, f)) == 1L))
  x <- CFfactor_coverage(cf, f, "absolute")
  expect_true(all(x$first == 1L))
  expect_true(all(x$double == 2L))
  expect_true(all(x$final3 == 3L))
  expect_true(all(unlist(CFfactor_coverage(cf, f, "relative")) == 1L))

  # all_leap calendar
  cf366 <- CFtime("days since 2001-01-01", "all_leap", 0:7319)
  f <- CFfactor(cf366, "year", eras)
  expect_true(all(unlist(CFfactor_units(cf366, f)) == rep(366L, 6L)))
  expect_true(all(unlist(CFfactor_coverage(cf366, f, "absolute")) == c(rep(366L, 6L), 0L)))
  expect_equal(sum(sapply(CFfactor_coverage(cf366, f, "relative"), sum)), 3L)

  f <- CFfactor(cf366, "season", eras)
  expect_true(all(sapply(CFfactor_units(cf366, f), function(x) {all(x == c(91L, 92L, 92L, 91L))})))
  x <- CFfactor_coverage(cf366, f, "absolute")
  expect_equal(x$first[1L], 60L)  # Jan + Feb of first year at beginning of time series
  expect_equal(x$double, c(182L, 184L, 184L, 182L)) # two full years
  expect_equal(x$final3, c(273L, 276L, 276L, 273L)) # three full years
  x <- unlist(CFfactor_coverage(cf366, f, "relative"))
  #expect_equal(x[1], 60 / 90). # works in the console but not here
  expect_true(all(x[2L:12L] == 1L))

  f <- CFfactor(cf366, "month", eras)
  expect_true(all(sapply(CFfactor_units(cf366, f), function(x) {all(x == leap_month_days)})))
  x <- CFfactor_coverage(cf366, f, "absolute")
  expect_true(all(x$first == leap_month_days))
  expect_true(all(x$double == leap_month_days * 2L))
  expect_true(all(x$final3 == leap_month_days * 3L))
  expect_true(all(unlist(CFfactor_coverage(cf366, f, "relative")) == 1L))

  f <- CFfactor(cf366, "dekad", eras)
  expect_true(all(sapply(CFfactor_units(cf366, f), function(x) {all(x == leap_dekad_days)})))
  x <- CFfactor_coverage(cf366, f, "absolute")
  expect_true(all(x$first == leap_dekad_days))
  expect_true(all(x$double == leap_dekad_days * 2L))
  expect_true(all(x$final3 == leap_dekad_days * 3L))
  expect_true(all(unlist(CFfactor_coverage(cf366, f, "relative")) == 1L))

  f <- CFfactor(cf366, "day", eras)
  expect_true(all(unlist(CFfactor_units(cf366, f)) == 1L))
  x <- CFfactor_coverage(cf366, f, "absolute")
  expect_true(all(x$first == 1L))
  expect_true(all(x$double == 2L))
  expect_true(all(x$final3 == 3L))
  expect_true(all(unlist(CFfactor_coverage(cf366, f, "relative")) == 1L))

  # Factors on data not aligned by year
  t <- CFTime$new("days since 2020-07-01", "standard", 0:364)
  first <- c("2020", "2020S3", "2020Q3", "2020-07", "2020D19", "2020-07-01")
  last <-  c("2021", "2021S3", "2021Q2", "2021-06", "2021D18", "2021-06-30")
  for (p in 1:6) {
    f <- t$factor(CFt$factor_periods[p])
    expect_equal(as.character(f)[1L], first[p])
    expect_equal(as.character(f)[365L], last[p])
  }

  # Incomplete coverage
  n <- 365L * 20L
  cov <- 0.8
  offsets <- sample(0L:(n-1L), n * cov)
  expect_warning(cf <- CFtime("days since 2020-01-01", "365_day", offsets))
  f <- CFfactor(cf, "month")
  x <- CFfactor_coverage(cf, f, "absolute")
  expect_equal(sum(x), n * cov)
  x <- CFfactor_coverage(cf, f, "relative")
  expect_true((cov - 0.01) < mean(x) && mean(x) < (cov + 0.01))
})

test_that("cut() works", {
  cf <- CFtime("days since 2020-01-01", "360_day", 0:719)
  expect_error(cut("sfg"))
  expect_error(cut(cf))
  expect_error(cut(cf, breaks = 5))
  expect_error(cut(cf, ""))
  expect_error(cut(cf, "blah"))

  f <- cut(cf, "quarter")
  expect_equal(nlevels(f), 8)
  expect_equal(levels(f), c("2020Q1", "2020Q2", "2020Q3", "2020Q4", "2021Q1", "2021Q2", "2021Q3", "2021Q4"))

  f <- cut(cf, c("2021-01-01", "2020-04-03")) # out of order
  expect_s3_class(f, "factor")
  expect_equal(levels(f), "2020-04-03")

  f <- cut(cf, c("2020-01-01", "2020-06-17", "2021-01-01", "2021-04-12", "2401-01-01"))
  expect_equal(levels(f), c("2020-01-01", "2020-06-17", "2021-01-01", "2021-04-12"))
})

test_that("bounds works", {
  t <- CFTime$new("days since 2024-01-01", "standard")
  off <- seq(from = 0.5, by = 1, length.out = 10)
  bnds <- rbind(0:9, 1:10)

  expect_null(t$get_bounds())          # no offsets
  expect_warning(t$set_bounds(bnds))   #

  t <- t + off
  expect_null(t$bounds) # bounds not set
  t$set_bounds(bnds)
  expect_equal(t$bounds, bnds)
  expect_equal(t$get_bounds("%Y-%m-%d")[2,1:3], c("2024-01-02", "2024-01-03", "2024-01-04"))
  t$set_bounds(NULL)
  expect_null(t$bounds)

  expect_warning(bounds(t) <- matrix(1:12, nrow = 4))
  expect_warning(bounds(t) <- "plain wrong")
  expect_warning(bounds(t) <- bnds * 10)

  bounds(t) <- bnds
  expect_match(capture_output(t$print()), "Bounds  : set$")

  hr6 <- rbind(off - 0.25, off + 0.25)
  bounds(t) <- hr6
  expect_match(capture_output(t$print()), "Bounds  : set$")
  expect_equal(bounds(t), hr6)
  expect_equal(bounds(t, "%H")[,1], c("06", "18"))

  # Copying / subsetting with bounds
  expect_identical(t, t$copy())
  tsub <- t$subset(3:5)
  subbnds <- tsub$bounds
  expect_equal(subbnds, matrix(c(2.25, 2.75, 3.25, 3.75, 4.25, 4.75), nrow = 2))
})

test_that("indexOf() works", {
  off <- (23*360):(24*360-1) + 0.5 # year 2024 days
  t <- CFtime("days since 2001-01-01", "360_day", off)
  x <- c("1999-02-12", # pre-origin
         "2023-06-23", # pre-time series
         "2024-01-30",
         "2024-01-31", # non-existent date
         "2024-02-01",
         "2024-02-30",
         "2024-03-01",
         "2025-01-01") # post-time series

  expect_error(indexOf(x, t, method = 4))
  expect_error(indexOf(TRUE, t))
  expect_error(indexOf(x, CFtime("months since 2001-01-01", "standard", 0:23)))
  expect_error(indexOf(x, t, nomatch = "July")) # must be able to coerce to numeric

  expect_equal(indexOf(x, t)[1:8], c(NA, NA, 29, NA, 30, 59, 60, NA))
  expect_equal(indexOf(x, t, method = "linear")[1:8], c(NA, NA, 29.5, NA, 30.5, 59.5, 60.5, NA))

  n <- 1:3
  out <- indexOf(n, t)
  outt <- attr(out, "CFTime")
  expect_equal(as_timestamp(t)[1:3], as_timestamp(outt))
  n <- c(-1, -2, -3)
  out <- indexOf(n, t)
  outt <- attr(out, "CFTime")
  expect_equal(length(outt), 360 - 3)
  expect_equal(as_timestamp(t)[4:6], as_timestamp(outt)[1:3])
  expect_error(indexOf(c(-1, 1), t))

  # Attached CFtime must have valid timestamps in `x`
  out <- indexOf(x, t)
  outt <- attr(out, "CFTime")
  expect_equal(as_timestamp(outt), x[!is.na(out)])

  bounds(t) <- TRUE
  expect_equal(indexOf(x, t)[1:8], c(NA, NA, 30, NA, 31, 60, 61, NA))
  expect_equal(indexOf(x, t, method = "linear")[1:8], c(NA, NA, 30, NA, 31, 60, 61, NA))
})

#nocov start
# Create environment for global CFtime variables
CFt <- new.env(parent = emptyenv())

.onLoad <- function(libname, pkgname) {
  assign("CFunits", data.frame(unit = c("years", "year", "yr", "months", "month", "mon", "days", "day", "d", "hours", "hour", "hr", "h", "minutes", "minute", "min", "seconds", "second", "sec", "s"),
                               id   = c(6L, 6L, 6L, 5L, 5L, 5L, 4L, 4L, 4L, 3L, 3L, 3L, 3L, 2L, 2L, 2L, 1L, 1L, 1L, 1L)), envir = CFt)
  assign("units", data.frame(name     = c("seconds", "minutes", "hours", "days", "months", "years"),
                             seconds  = c(1L, 60L, 3600L, 86400L, 86400L * 30L, 86400L * 365L),
                             per_day  = c(86400, 1440, 24, 1, 1/30, 1/365)), envir = CFt)
  assign("prefixes", data.frame(name        = c("yocto", "zepto", "atto", "femto", "pico", "nano", "micro", "micro", "micro",
                                                "milli", "centi", "deci", "deca", "deka", "hecto", "kilo", "mega", "giga",
                                                "tera", "peta", "exa", "zetta", "yotta"),
                                abbrev      = c("y", "z", "a", "f", "p", "n", "u", "\u00B5", "\u03BC", "m", "c", "d", "da", "da",
                                                "h", "k", "M", "G", "T", "P", "E", "Z", "Y"),
                                multiplier  = c(1e-24, 1e-21, 1e-18, 1e-15, 1e-12, 1e-9, 1e-6, 1e-6, 1e-6, 1e-3, 1e-2, 1e-1,
                                                1e1, 1e1, 1e2, 1e3, 1e6, 1e9, 1e12, 1e15, 1e18, 1e21, 1e24)), envir = CFt)
  assign("factor_periods", c("year", "season", "quarter", "month", "dekad", "day"), envir = CFt)
  assign("eps", .Machine$double.eps^0.5, envir = CFt)
}
#nocov end

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, include = FALSE---------------------------------------------------
library(CFtime)
library(ncdfCF)

## -----------------------------------------------------------------------------
# POSIXt calculations on a standard calendar - INCORRECT
as.Date("1949-12-01") + 43289

# CFtime calculation on a "360_day" calendar - CORRECT
# See below examples for details on the two functions
as_timestamp(CFtime("days since 1949-12-01", "360_day", 43289))

## -----------------------------------------------------------------------------
# Create a CFTime object from a definition string, a calendar and some offsets
(t <- CFtime("days since 1949-12-01", "360_day", 19830:90029))

## -----------------------------------------------------------------------------
# Opening a data file that is included with the package.
# Usually you would `list.files()` on a directory of your choice.
fn <- list.files(path = system.file("extdata", package = "CFtime"), full.names = TRUE)[1]
(ds <- ncdfCF::open_ncdf(fn))

# "Conventions" global attribute must have a string like "CF-1.*" for this package to work reliably

# Look at the "time" axis
(time <- ds[["time"]])

# Get the CFTime instance from the "time" axis
(t <- time$time())

## ----eval = FALSE-------------------------------------------------------------
#  library(RNetCDF)
#  nc <- open.nc(fn)
#  att.get.nc(nc, -1, "Conventions")
#  t <- CFtime(att.get.nc(nc, "time", "units"),
#              att.get.nc(nc, "time", "calendar"),
#              var.get.nc(nc, "time"))
#  
#  library(ncdf4)
#  nc <- nc_open(fn)
#  nc_att_get(nc, 0, "Conventions")
#  t <- CFtime(nc$dim$time$units,
#              nc$dim$time$calendar,
#              nc$dim$time$vals)

## -----------------------------------------------------------------------------
dates <- t$as_timestamp(format = "date")
dates[1:10]

## -----------------------------------------------------------------------------
t$range()

## -----------------------------------------------------------------------------
# Create a dekad factor for the whole `t` time series that was created above
f_k <- t$factor("dekad")
str(f_k)

# Create monthly factors for a baseline era and early, mid and late 21st century eras
baseline <- t$factor(era = 1991:2020)
future <- t$factor(era = list(early = 2021:2040, mid = 2041:2060, late = 2061:2080))
str(future)

## -----------------------------------------------------------------------------
  (new_time <- attr(f_k, "CFTime"))

## -----------------------------------------------------------------------------
# Is the time series complete?
is_complete(t)

# How many time units fit in a factor level?
t$factor_units(baseline)

# What's the absolute and relative coverage of our time series
t$factor_coverage(baseline, "absolute")
t$factor_coverage(baseline, "relative")

## -----------------------------------------------------------------------------
# 4 years of data on a `365_day` calendar, keep 80% of values
n <- 365 * 4
cov <- 0.8
offsets <- sample(0:(n-1), n * cov)

(t <- CFtime("days since 2020-01-01", "365_day", offsets))
# Note that there are about 1.25 days between observations

mon <- t$factor("month")
t$factor_coverage(mon, "absolute")
t$factor_coverage(mon, "relative")

## -----------------------------------------------------------------------------
# 1970-01-01 is the origin of POSIXt
difftime(as.POSIXct("2024-01-01"), as.POSIXct("1970-01-01"), units = "sec")

# Leap seconds in UTC
.leap.seconds

## -----------------------------------------------------------------------------
# Days in January and February
t <- CFtime("days since 2023-01-01", "360_day", 0:59)
ts_days <- t$as_timestamp("date")
as.Date(ts_days)

#' @title CF calendar with no annual cycle
#'
#' @description This class represents a CF calendar with no annual cycle. All
#' datetimes in the calendar are the same. This is useful only for repeated
#' experiments simulating a fixed time in the year.
#'
#' @docType class
CFCalendarNone <- R6::R6Class("CFCalendarNone",
  inherit = CFCalendar,
  public = list(
    #' @description Create a new CF calendar.
    #' @param nm The name of the calendar. This must be "none". This argument
    #' is superfluous but maintained to be consistent with the initialization
    #' methods of the parent and sibling classes.
    #' @param definition The string that defines the units and the origin, as
    #' per the CF Metadata Conventions.
    #' @return A new instance of this class.
    initialize = function(nm, definition) {
      super$initialize(nm, definition)
    },

    #' @description Indicate which of the supplied dates are valid. In this
    #' calendar only one date is valid, namely the one that the calendar is
    #' initialized with.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return Logical vector with the same length as argument `ymd` has rows
    #'   with `TRUE` for valid days and `FALSE` for invalid days, or `NA` where
    #'   the row in argument `ymd` has `NA` values.
    valid_days = function(ymd) {
      if (!nrow(self$origin)) TRUE
      else
        ymd$year == self$origin$year &
        ymd$month == self$origin$month &
        ymd$day == self$origin$day
    },

    #' @description Determine the number of days in the month of the calendar.
    #'   This always returns a vector of `NA` values because this method has no
    #'   meaning for this class.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return A vector with `NA` values for the dates supplied as argument
    #'   `ymd`. If no dates are supplied, a vector of `NA` values of length 12.
    month_days = function(ymd = NULL) {
      if (is.null(ymd)) rep(NA, 12L) else rep(NA, nrow(ymd))
    },

    #' @description Indicate which years are leap years. This always returns a
    #'   vector of `NA` values because this method has no meaning for this
    #'   class.
    #' @param yr Integer vector of years to test.
    #' @return Vector with the same length as argument `yr` with `NA` values.
    leap_year = function(yr) {
      rep(NA, length(yr))
    },

    #' @description Retrieve the day-of-year number of the fixed data of this
    #'   calendar, using the Gregorian calendar.
    #' @param ymd `data.frame` with fixed date parsed into its parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return Integer with the day-of-year of the fixed date for this calendar.
    doy = function(ymd) {
      cum_month <- c(0L, 31L, 59L, 90L, 120L, 151L, 181L, 212L, 243L, 273L, 304L, 334L) # days in year before the month
      doy <- cum_month[ymd$month] + ymd$day
      if (((ymd$year %% 4L == 0L) & (ymd$year %% 100L > 0L)) | (ymd$year %% 400L == 0L))
        doy <- doy + 1L
      doy
    },

    #' @description Calculate difference in days between a `data.frame` of time
    #'   parts and the origin. The difference is always 0, given that
    #'   this is not a true calendar and there can be no calculations to
    #'   determine any difference.
    #' @param x `data.frame`. Dates to calculate the difference for.
    #' @return Integer vector of a length equal to the number of rows in
    #'   argument `x` indicating 0 for the days that equal the origin
    #'   of the calendar, or `NA` otherwise.
    date2offset = function(x) {
      res <- rep(0L, nrow(x))
      res[which(is.na(x$year))] <- NA_integer_
      res
    },

    #' @description Calculate date parts from day differences from the origin.
    #'   This always returns 0's for the year, month and day as other values are
    #'   not valid in this calendar. Hour-minute-second datetime parts are
    #'   handled in [CFCalendar].
    #' @param x Integer vector of days to add to the origin.
    #' @return A `data.frame` with columns 'year', 'month' and 'day' and as many
    #'   rows as the length of vector `x`. Rows with values of `x` other than 0
    #'   will return the values for the origin of the calendar nonetheless, in
    #'   accordance with the CF Metadata Conventions.
    offset2date = function(x) {
      len <- length(x)
      data.frame(year = rep(self$origin$year, len), month = rep(self$origin$month, len), day = rep(self$origin$day, len))
    }
  )
)

# Internal functions
#
# The functions in this source file are for internal use only.

# ==============================================================================
# Offsets and timestamp formatting

#' Validate offsets passed into a CFTime instance
#'
#' Tests the `offsets` values. Throws an error if the argument contains `NA` values.
#'
#' @param offsets The offsets to test
#'
#' @returns logical. `TRUE` if the offsets are valid, throws an error otherwise.
#' @noRd
.validOffsets <- function(offsets) {
  if (any(is.na(offsets))) stop("Offsets cannot contain `NA` values.", call. = FALSE)
  TRUE
}

#' Formatting of time strings from time elements
#'
#' This is an internal function that should not generally be used outside of
#' the CFtime package.
#'
#' @param t A `data.frame` representing timestamps.
#'
#' @returns A vector of character strings with a properly formatted time. If any
#' timestamp has a fractional second part, then all time strings will report
#' seconds at milli-second precision.
#' @noRd
.format_time <- function(t) {
  fsec <- t$second %% 1L
  if (any(fsec > 0L)) {
    paste0(sprintf("%02d:%02d:", t$hour, t$minute), ifelse(t$second < 10, "0", ""), sprintf("%.3f", t$second))
  } else {
    sprintf("%02d:%02d:%02d", t$hour, t$minute, t$second)
  }
}

#' Do the time elements have time-of-day information?
#'
#' If any time information > 0, then `TRUE` otherwise `FALSE`.
#'
#' This is an internal function that should not generally be used outside of
#' the CFtime package.
#'
#' @param t A `data.frame` representing timestamps.
#'
#' @returns `TRUE` if any timestamp has time-of-day information, `FALSE` otherwise.
#' @noRd
.has_time <- function(t) {
  any(t$hour > 0) || any(t$minute > 0) || any(t$second > 0)
}

#' Do formatting of timestamps with format specifiers
#'
#' @param ts `data.frame` of decomposed offsets.
#' @param tz Time zone character string.
#' @param format A character string with the format specifiers, or
#' "date" or "timestamp".
#' @returns Character vector of formatted timestamps.
#' @noRd
.format_format <- function(ts, tz, format) {
  if (format == "") format <- "timestamp"
  if (format == "timestamp" && sum(ts$hour, ts$minute, ts$second) == 0)
    format <- "date"

  if (format == "date") return(sprintf("%04d-%02d-%02d", ts$year, ts$month, ts$day))
  else if (format == "timestamp") return(sprintf("%04d-%02d-%02dT%s", ts$year, ts$month, ts$day, .format_time(ts)))

  # Expand any composite specifiers
  format <- stringr::str_replace_all(format, c("%F" = "%Y-%m-%d", "%R" = "%H:%M", "%T" = "%H:%M:%S"))

  # Splice in timestamp values for specifiers
  # nocov start
  if (grepl("%b|%h", format[1])) {
    mon <- strftime(ISOdatetime(2024, 1:12, 1, 0, 0, 0), "%b")
    format <- stringr::str_replace_all(format, "%b|%h", mon[ts$month])
  }
  if (grepl("%B", format[1])) {
    mon <- strftime(ISOdatetime(2024, 1:12, 1, 0, 0, 0), "%B")
    format <- stringr::str_replace_all(format, "%B", mon[ts$month])
  }
  # nocov end
  format <- stringr::str_replace_all(format, "%[O]?d", sprintf("%02d", ts$day))
  format <- stringr::str_replace_all(format, "%e", sprintf("%2d", ts$day))
  format <- stringr::str_replace_all(format, "%[O]?H", sprintf("%02d", ts$hour))
  format <- stringr::str_replace_all(format, "%[O]?I", sprintf("%02d", ts$hour %% 12))
  format <- stringr::str_replace_all(format, "%[O]?m", sprintf("%02d", ts$month))
  format <- stringr::str_replace_all(format, "%[O]?M", sprintf("%02d", ts$minute))
  format <- stringr::str_replace_all(format, "%p", ifelse(ts$hour < 12, "AM", "PM"))
  format <- stringr::str_replace_all(format, "%S", sprintf("%02d", as.integer(ts$second)))
  format <- stringr::str_replace_all(format, "%[E]?Y", sprintf("%04d", ts$year))
  format <- stringr::str_replace_all(format, "%z", tz)
  format <- stringr::str_replace_all(format, "%%", "%")
  format
}

# ==============================================================================
# Other internal functions

#' Calculate time units in factors
#'
#' @param f factor. Factor as generated by `CFfactor()`.
#' @param cal `CFCalendar` instance of the `CFTime` instance.
#' @param upd numeric. Number of units per day, from the `CFt` environment.
#' @returns A vector as long as the number of levels in the factor.
#' @noRd
.factor_units <- function(f, cal, upd) {
  period <- attr(f, "period")
  cal_class <- class(cal)[1L]

  res <- if (period == "day")
    rep(1L, nlevels(f))
  else if (cal_class == "CFCalendar360") {
    rep(c(360L, 90L, 90L, 30L, 10L, 1L)[which(CFt$factor_periods == period)], nlevels(f))
  } else {
    if (attr(f, "era") > 0L) {
      if (cal_class == "CFCalendar366") {
        switch(period,
               "year"    = rep(366L, nlevels(f)),
               "season"  = c(91L, 92L, 92L, 91L)[as.integer(substr(levels(f), 2, 2))],
               "quarter" = c(91L, 91L, 92L, 92L)[as.integer(levels(f))],
               "month"   = c(31L, 29L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[as.integer(levels(f))],
               "dekad"   = {
                 dk <- as.integer(substr(levels(f), 2L, 3L))
                 ifelse(dk %% 3L > 0L | dk %in% c(12L, 18L, 27L, 33L), 10L,
                        ifelse(dk %in% c(3L, 9L, 15L, 21L, 24L, 30L, 36L), 11L, 9L))
               }
        )
      } else {
        switch(period,
               "year"    = rep(365L, nlevels(f)),
               "season"  = c(90L, 92L, 92L, 91L)[as.integer(substr(levels(f), 2, 2))],
               "quarter" = c(90L, 91L, 92L, 92L)[as.integer(substr(levels(f), 2, 2))],
               "month"   = c(31L, 28L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[as.integer(levels(f))],
               "dekad"   = {
                 dk <- as.integer(substr(levels(f), 2L, 3L))
                 ifelse(dk %% 3L > 0L | dk %in% c(12L, 18L, 27L, 33L), 10L,
                        ifelse(dk %in% c(3L, 9L, 15L, 21L, 24L, 30L, 36L), 11L, 8L))
               }
        )
      }
    } else {  # not an era factor
      switch(period,
             "year"    = ifelse(cal$leap_year(as.integer(levels(f))), 366L, 365L),
             "season"  = {
               year <- as.integer(substr(levels(f), 1L, 4L))
               season <- as.integer(substr(levels(f), 6L, 6L))
               ifelse(cal$leap_year(year), c(91L, 92L, 92L, 91L)[season], c(90L, 92L, 92L, 91L)[season])
             },
             "quarter" = {
               year <- as.integer(substr(levels(f), 1L, 4L))
               qtr  <- as.integer(substr(levels(f), 6L, 6L))
               ifelse(cal$leap_year(year), c(91L, 91L, 92L, 92L)[qtr], c(90L, 91L, 92L, 92L)[qtr])
             },
             "month"   = {
               year  <- as.integer(substr(levels(f), 1L, 4L))
               month <- as.integer(substr(levels(f), 6L, 7L))
               ifelse(cal$leap_year(year), c(31L, 29L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[month],
                      c(31L, 28L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[month])
             },
             "dekad"   = {
               year  <- as.integer(substr(levels(f), 1L, 4L))
               dk <- as.integer(substr(levels(f), 6L, 7L))
               ifelse(dk %% 3L > 0L | dk %in% c(12L, 18L, 27L, 33L), 10L,
                      ifelse(dk %in% c(3L, 9L, 15L, 21L, 24L, 30L, 36L), 11L,
                             ifelse(cal$leap_year(year), 9L, 8L)))
             }
      )
    }
  }
  res * upd
}

# =====================
# Others

#' Round values `x` with .5 being rounded up.
#' Adapted from https://stackoverflow.com/a/12688836/3304426
#' @noRd
.round <- function(x) {
  posneg <- sign(x)
  z <- trunc(abs(x) + 0.5 + CFt$eps)
  z * posneg
}

#' @title International Atomic Time CF calendar
#'
#' @description This class represents a calendar based on the International
#'   Atomic Time. Validity is from 1958 onwards, with dates represented using
#'   the Gregorian calendar. Given that this "calendar" is based on a universal
#'   clock, the concepts of leap second, time zone and daylight savings time do
#'   not apply.
#'
#' @aliases CFCalendarTAI
#' @docType class
CFCalendarTAI <- R6::R6Class("CFCalendarTAI",
  inherit = CFCalendarProleptic,
  public = list(
    #' @description Indicate which of the supplied dates are valid.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. If present, the `tz` column is checked for
    #'   illegal time zone offsets. Any other columns are disregarded.
    #' @return Logical vector with the same length as argument `ymd` has rows
    #'   with `TRUE` for valid days and `FALSE` for invalid days, or `NA` where
    #'   the row in argument `ymd` has `NA` values.
    valid_days = function(ymd) {
      super$valid_days(ymd) & ymd$year >= 1958L
      # FIXME: TZ offsets
    }
  )
)

#' @title 366-day CF calendar
#'
#' @description This class represents a CF calendar of 366 days per year, having
#'   leap days in every year. This calendar is not compatible with the standard
#'   POSIXt calendar.
#'
#'   This calendar supports dates before year 1 and includes the year 0.
#'
#' @aliases CFCalendar366
#' @docType class
CFCalendar366 <- R6::R6Class("CFCalendar366",
  inherit = CFCalendar,
  private = list(
    # Rata Die, the number of days from the day before 0001-01-01 to
    # origin of this calendar. Used to convert offsets from the calendar origin
    # to the day before 0001-01-01 for arithmetic calculations.
    rd = 0L
  ),
  public = list(
    #' @description Create a new CF calendar of 366 days per year.
    #' @param nm The name of the calendar. This must be "366_day" or "all_leap".
    #' @param definition The string that defines the units and the origin, as
    #' per the CF Metadata Conventions.
    #' @return A new instance of this class.
    initialize = function(nm, definition) {
      super$initialize(nm, definition)
      private$rd <- self$date2offset(self$origin)
    },

    #' @description Indicate which of the supplied dates are valid.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return Logical vector with the same length as argument `ymd` has rows
    #'   with `TRUE` for valid days and `FALSE` for invalid days, or `NA` where
    #'   the row in argument `ymd` has `NA` values.
    valid_days = function(ymd) {
      !is.na(ymd$year) & ymd$month >= 1L & ymd$month <= 12L & ymd$day >= 1L &
      ymd$day <= c(31L, 29L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[ymd$month]
    },

    #' @description Determine the number of days in the month of the calendar.
    #' @param ymd `data.frame`, optional, with dates parsed into their parts.
    #' @return A vector indicating the number of days in each month for the
    #'   dates supplied as argument `ymd`. If no dates are supplied, the number
    #'   of days per month for the calendar as a vector of length 12.
    month_days = function(ymd = NULL) {
      if (is.null(ymd)) return(c(31L, 29L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L))

      res <- c(31L, 29L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[ymd$month]
      res[which(is.na(ymd$year))] <- NA
      res
    },

    #' @description Indicate which years are leap years.
    #' @param yr Integer vector of years to test.
    #' @return Logical vector with the same length as argument `yr`. Since in
    #'   this calendar all years have a leap day, all values will be `TRUE`, or
    #'   `NA` where argument `yr` is `NA`.
    leap_year = function(yr) {
      res <- rep(TRUE, length(yr))
      res[which(is.na(yr))] <- NA
      res
    },

    #' @description Retrieve the day-of-year number for the supplied
    #'   `data.frame` of year, month and day, using this calendar.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return Integer vector as long as argument `ymd` has rows with the
    #'   day-of-year for each row.
    doy = function(ymd) {
      cum_month <- c(0L, 31L, 60L, 91L, 121L, 152L, 182L, 213L, 244L, 274L, 305L, 335L)
      cum_month[ymd$month] + ymd$day
    },

    #' @description Calculate difference in days between a `data.frame` of time
    #'   parts and the origin.
    #'
    #' @param x `data.frame`. Dates to calculate the difference for.
    #'
    #' @return Integer vector of a length equal to the number of rows in
    #' argument `x` indicating the number of days between `x` and the `origin`,
    #' or `NA` for rows in `x` with `NA` values.
    date2offset = function(x) {
      corr <- ifelse(x$month <= 2L, 0L, -1L)
      366L * (x$year - 1L) + (367L * x$month - 362L) %/% 12L + corr + x$day - private$rd
    },

    #' @description Calculate date parts from day differences from the origin. This
    #'   only deals with days as these are impacted by the calendar.
    #'   Hour-minute-second timestamp parts are handled in [CFCalendar].
    #'
    #' @param x Integer vector of days to add to the origin.
    #' @return A `data.frame` with columns 'year', 'month' and 'day' and as many
    #'   rows as the length of vector `x`.
    offset2date = function(x) {
      d0 <- x - 1L + private$rd        # d0 is offset relative to year 0, 0-based
      yr <- d0 %/% 366L + 1L           # full years
      d1 <- d0 %% 366L                 # remaining days
      corr <- ifelse(d1 < 60L, 0L, 1L) # correct for days past February
      mon <- (12L * (d1 + corr) + 373L) %/% 367L
      day <- d1 - (367L * mon - 362L) %/% 12L + corr + 1L
      data.frame(year = yr, month = mon, day = day)
    }
  )
)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, include = FALSE---------------------------------------------------
library(CFtime)

## -----------------------------------------------------------------------------
# Setting up
library(ncdfCF)
fn <- list.files(path = system.file("extdata", package = "CFtime"), full.names = TRUE)[1]
(ds <- ncdfCF::open_ncdf(fn))

# The T axis, with name "time" has a CFTime instance
t <- ds[["time"]]$time()

# Create monthly factors for a baseline era and early, mid and late 21st century eras
baseline <- t$factor(era = 1991:2020)
future <- t$factor(era = list(early = 2021:2040, mid = 2041:2060, late = 2061:2080))
str(baseline)
str(future)

## -----------------------------------------------------------------------------
# Get the data for the "pr" data variable from the netCDF data set.
# The `CFData$array()` method ensures that data are in standard R orientation.
# Converts units of kg m-2 s-1 to mm/day.
pr <- ds[["pr"]]$data()$array() * 86400

# Get a global attribute from the file
experiment <- ds$attribute("experiment_id")

# Calculate the daily average precipitation per month for the baseline period
# and the three future eras.
pr_base <- apply(pr, 1:2, tapply, baseline, mean)                         # an array
pr_future <- lapply(future, function(f) apply(pr, 1:2, tapply, f, mean))  # a list of arrays

# Calculate the precipitation anomalies for the future eras against the baseline.
# Working with daily averages per month so we can simply subtract and then multiply by days 
# per month for each of the factor levels using the CF calendar.
ano <- mapply(function(pr, f) {(pr - pr_base) * t$factor_units(f)}, pr_future, future, SIMPLIFY = FALSE)

# Plot the results
plot(1:12, ano$early[,1,1], type = "o", col = "blue", ylim = c(-50, 40), xlim = c(1, 12), 
     main = paste0("Hamilton, New Zealand\nExperiment: ", experiment), 
     xlab = "month", ylab = "Precipitation anomaly (mm)")
lines(1:12, ano$mid[,1,1], type = "o", col = "green")
lines(1:12, ano$late[,1,1], type = "o", col = "red")

## -----------------------------------------------------------------------------
# Get the list of files that make up the ensemble members, here:
# GFDL ESM4 and MRI ESM2 models for experiment SSP2-4.5, precipitation, CMIP6 2015-01-01 to 2099-12-31
lf <- list.files(path = system.file("extdata", package = "CFtime"), full.names = TRUE)

# Loop over the files individually
# ano is here a list with each element holding the results for a single model
ano <- lapply(lf, function(fn) {
  ds <- ncdfCF::open_ncdf(fn)
  t <- ds[["time"]]$time()
  pr <- ds[["pr"]]$data()$array() * 86400

  baseline <- t$factor(era = 1991:2020)
  pr_base <- apply(pr, 1:2, tapply, baseline, mean)
  future <- t$factor(era = list(early = 2021:2040, mid = 2041:2060, late = 2061:2080))
  pr_future <- lapply(future, function(f) apply(pr, 1:2, tapply, f, mean))
  mapply(function(pr, f) {(pr - pr_base) * t$factor_units(f)}, pr_future, future, SIMPLIFY = FALSE)
})

# Era names
eras <- c("early", "mid", "late")
dim(eras) <- 3

# Build the ensemble for each era
# For each era, grab the data for each of the ensemble members, simplify to an array
# and take the mean per row (months, in this case)
ensemble <- apply(eras, 1, function(e) {
  rowMeans(sapply(ano, function(a) a[[e]], simplify = TRUE))})
colnames(ensemble) <- eras
rownames(ensemble) <- rownames(ano[[1]][[1]])
ensemble

## ----eval = FALSE-------------------------------------------------------------
#  library(ncdfCF)
#  library(abind)
#  
#  prepare_CORDEX <- function(fn, var, aoi) {
#    data <- vector("list", length(fn))
#    for (i in 1:length(fn)) {
#      ds <- ncdfCF::open_ncdf(fn[i])
#      if (i == 1) {
#        # Get a CFTime instance from the first file
#        t <- ds[["time"]]$time()
#      } else {
#        # Add offsets from the file and add to the CFTime instance
#        t <- t + ds[["time"]]$time()$offsets
#      }
#  
#      # Put the subsetted data array in the list
#      data[[i]] <- ds[[var]]$subset(aoi = aoi)$array()
#    }
#  
#    # Create a list for output with the CFTime instance and
#    # the data bound in a single 3-dimensional array
#    list(CFTime = t, data = abind(data, along = 3))
#  }

#' Create a CFTime object
#'
#' This function creates an instance of the [CFTime] class. The arguments to the
#' call are typically read from a CF-compliant data file with climatological
#' observations or climate projections. Specification of arguments can also be
#' made manually in a variety of combinations.
#'
#' A time series can also be constructed like a sequence. In this case argument
#' `offsets` should be `NULL` or missing and arguments `from` and `by` provided,
#' with either of arguments `to` or `length.out` indicating the end of the time
#' series. Arguments should be named to avoid ambiguity.
#'
#' @param definition A character string describing the time coordinate.
#' @param calendar A character string describing the calendar to use with the
#'   time dimension definition string. Default value is "standard".
#' @param offsets Numeric or character vector, optional. When numeric, a vector
#'   of offsets from the origin in the time series. When a character vector,
#'   timestamps in ISO8601 or UDUNITS format.
#' @param from,to Optional. Character timestamps in ISO8601 or UDUNITS format.
#'   When `from` is given, a sequence of timestamps is generated with `from` as
#'   the starting timestamp. Either argument `to` or `length.out` must be
#'   provided as well. Ignored when argument `offsets` is not `NULL`.
#' @param by Optional. A single character string representing a time interval,
#'   composed of a number and a time unit separated by a space, such as in "6
#'   hours". When argument `from` is supplied, argument `by` will be the
#'   separation between successive timestamps. Note that the time unit in the
#'   string does not have to be the same as the unit in the `definition`
#'   argument but it must be an allowable unit of time. Time interval units of
#'   "months" and "years" are strongly discouraged unless the same time unit is
#'   used in the `definition` argument - in all other cases there is a loss of
#'   precision due to the ambiguity in the time units.
#' @param length.out Optional. An numeric value that indicates the lengths of
#'   the time series to generate, rounded up if fractional. Ignored when
#'   argument `to` is provided.
#' @returns An instance of the `CFTime` class.
#' @export
#' @name CFtime-function
#' @examples
#' # Using numeric offset values - this is how a netCDF file works
#' CFtime("days since 1850-01-01", "julian", 0:364)
#'
#' # A time object with a single defined timestamp
#' CFtime("hours since 2023-01-01", "360_day", "2023-01-30T23:00")
#'
#' # A time series from a sequence with an end point
#' CFtime("days since 2023-01-01", from = "2020-01-01", to = "2023-12-31", by = "12 days")
#'
#' # A time series from a sequence with a specified length
#' CFtime("days since 2023-01-01", from = "2020-01-01T03:00:00", by = "6 hr", length.out = 31 * 4)
CFtime <- function(definition, calendar = "standard", offsets, from, to, by, length.out) {
  if (missing(offsets)) {
    if (missing(from))
      # Barebones CFTime without offsets
      CFTime$new(definition, calendar)
    else {
      if (missing(by))
        stop("When argument `from` is provided, argument `by` must be provided too.", call. = FALSE)
      else if (!is.character(from) || length(from) > 1L)
        stop("Argument `from` must be a single character timestamp value.", call. = FALSE)
      else
        t <- CFTime$new(definition, calendar, from)

      # Process argument `by`
      intv <- NA; intv_unit <- integer(0)
      if (is.character(by) && length(by) == 1L) {
        parts <- strsplit(by, " ", fixed = TRUE)[[1L]]
        if (length(parts) == 2L) {
          intv_unit <- CFt$CFunits[CFt$CFunits$unit == parts[2L], ]$id
          suppressWarnings(intv <- as.numeric(parts[1L]))
        }
      }
      if (is.na(intv) || !length(intv_unit))
        stop("Argument `by` is an invalid time interval.", call. = FALSE)

      doff <- intv * CFt$units$seconds[intv_unit] / CFt$units$seconds[t$calendar$unit]

      # End point of the time series
      from_off <- t$offsets # There is only a single offset at this point
      if (missing(to)) {
        if (missing(length.out))
          stop("One of arguments `to` or `length.out` must be supplied.", call. = FALSE)

        # Time series with a given length
        t + seq(from = from_off + doff, by = doff, length.out = length.out[1L] - 1L)
      } else {
        # Time series up to a finishing point
        to_off <- t$calendar$parse(to[1L])$offset
        if (is.na(to_off))
          stop("Argument `to` must be a single character timestamp value.")
        t + seq(from = from_off + doff, to = to_off, by = doff)
      }
    }
  } else
    CFTime$new(definition, calendar, offsets)
}

# ==============================================================================
# Functions to access CFTime properties and methods

#' @aliases properties
#' @title Properties of a CFTime object
#'
#' @description These functions return the properties of an instance of the
#'   [CFTime] class. The properties are all read-only, but offsets can be added
#'   using the `+` operator.
#'
#' @param t An instance of `CFTime`.
#'
#' @returns `calendar()` and `unit()` return a character string.
#'   `origin()` returns a data frame of timestamp elements with a single row
#'   of data. `timezone()` returns the calendar time zone as a character
#'   string. `offsets()` returns a vector of offsets or `NULL` if no offsets
#'   have been set.
#'
#' @examples
#' t <- CFtime("days since 1850-01-01", "julian", 0:364)
#' definition(t)
#' calendar(t)
#' unit(t)
#' timezone(t)
#' origin(t)
#' offsets(t)
#' resolution(t)

#' @describeIn properties The definition string of the `CFTime` instance.
#' @export
definition <- function(t) t$calendar$definition

#' @describeIn properties The calendar of the `CFTime` instance.
#' @export
calendar <- function(t) t$calendar$name

#' @describeIn properties The unit of the `CFTime` instance.
#' @export
unit <- function(t) t$unit

#' @describeIn properties The origin of the `CFTime` instance in timestamp elements.
#' @export
origin <- function(t) t$calendar$origin

#' @describeIn properties The time zone of the calendar of the `CFTime` instance as a character string.
#' @export
timezone <- function(t) t$calendar$timezone

#' @describeIn properties The offsets of the `CFTime` instance as a numeric vector.
#' @export
offsets <- function(t) t$offsets

#' @describeIn properties The average separation between the offsets in the `CFTime` instance.
#' @export
resolution <- function(t) t$resolution

#' Bounds of the time offsets
#'
#' CF-compliant netCDF files store time information as a single offset value for
#' each step along the dimension, typically centered on the valid interval of
#' the data (e.g. 12-noon for day data). Optionally, the lower and upper values
#' of the valid interval are stored in a so-called "bounds" variable, as an
#' array with two rows (lower and higher value) and a column for each offset.
#' With function `bounds()<-` those bounds can be set for a `CFTime` instance.
#' The bounds can be retrieved with the `bounds()` function.
#'
#' @param x A `CFTime` instance.
#' @param format Optional. A single string with format specifiers, see
#'   [CFtime::format()] for details.
#' @return If bounds have been set, an array of bounds values with dimensions
#'   `(2, length(offsets))`. The first row gives the lower bound, the second row
#'   the upper bound, with each column representing an offset of `x`. If the
#'   `format` argument is specified, the bounds values are returned as strings
#'   according to the format. `NULL` when no bounds have been set.
#' @aliases bounds
#' @export
#' @examples
#' t <- CFtime("days since 2024-01-01", "standard", seq(0.5, by = 1, length.out = 366))
#' as_timestamp(t)[1:3]
#' bounds(t) <- rbind(0:365, 1:366)
#' bounds(t)[, 1:3]
#' bounds(t, "%d-%b-%Y")[, 1:3]
bounds <- function(x, format) {
  x$get_bounds(format)
}

#' @rdname bounds
#' @param value A `matrix` (or `array`) with dimensions (2, length(offsets))
#'   giving the lower (first row) and higher (second row) bounds of each offset
#'   (this is the format that the CF Metadata Conventions uses for storage in
#'   netCDF files). Use `NULL` to unset any previously set bounds.
#' @export
`bounds<-` <- function(x, value) {
  x$set_bounds(value)
  x
}

#' The length of the offsets contained in the `CFTime` instance.
#'
#' @param x The `CFTime` instance whose length will be returned
#'
#' @return The number of offsets in the specified `CFTime` instance.
#' @export
#'
#' @examples
#' t <- CFtime("days since 1850-01-01", "julian", 0:364)
#' length(t)
length.CFTime <- function(x) base::length(x$offsets)

#' Return the timestamps contained in the `CFTime` instance
#' @param x The `CFTime` instance whose timestamps will be returned.
#' @param ... Ignored.
#' @return The timestamps in the specified `CFTime` instance.
#' @export
#' @examples
#' t <- CFtime("days since 1850-01-01", "julian", 0:364)
#' as.character(t)
as.character.CFTime <- function(x, ...) {
  x$as_timestamp()
}

#' Format the timestamps in the `CFTime` instance
#'
#'   The formatting is largely oblivious to locale. The reason for this is
#'   that certain dates in certain calendars are not POSIX-compliant and the
#'   system functions necessary for locale information thus do not work
#'   consistently. The main exception to this is the (abbreviated) names of
#'   months (`bB`), which could be useful for pretty printing in the local
#'   language. For separators and other locale-specific adornments, use
#'   local knowledge instead of depending on system locale settings; e.g.
#'   specify `%m/%d/%Y` instead of `%D`.
#'
#'   Week information, including weekday names, is not supported at all as a
#'   "week" is not defined for non-standard CF calendars and not generally
#'   useful for climate projection data. If you are working with observed
#'   data and want to get pretty week formats, use the [as_timestamp()]
#'   method to generate `POSIXct` timestamps (observed data generally uses a
#'   "standard" calendar) and then use the [base::format()] function which
#'   supports the full set of specifiers.
#' @param x The `CFTime` instance whose timestamps will be formatted.
#' @param format A character string. The default for the format methods is
#'   "%Y-%m-%dT%H:%M:%S" if any timestamp has a time component which is not
#'   midnight, and "%Y-%m-%d" otherwise. The only supported specifiers are
#'   `bBdeFhHImMpRSTYz%`. Modifiers `E` and `O` are silently ignored. Other
#'   specifiers, including their percent sign, are copied to the output as if
#'   they were adorning text.
#' @param tz Ignored.
#' @param usetz Logical. Should the time zone offset be appended to the output?
#'   This is always in numerical form, i.e. "-0800", from UTC.
#' @param ... Ignored.
format.CFTime <- function(x, format = "", tz = "", usetz = FALSE, ...) {
  x$format(format, usetz)
}

#' Create a factor for a `CFTime` instance
#'
#' Method for [base::cut()] applied to [CFTime] objects.
#'
#' When `breaks` is one of `"year", "season", "quarter", "month", "dekad",
#' "day"` a factor is generated like by [CFfactor()].
#'
#' When `breaks` is a vector of character timestamps a factor is produced with a
#' level for every interval between timestamps. The last timestamp, therefore,
#' is only used to close the interval started by the pen-ultimate timestamp -
#' use a distant timestamp (e.g. `range(x)[2]`) to ensure that all offsets to
#' the end of the `CFTime` time series are included, if so desired. The last
#' timestamp will become the upper bound in the `CFTime` instance that is
#' returned as an attribute to this function so a sensible value for the last
#' timestamp is advisable.
#'
#' This method works similar to [base::cut.POSIXt()] but there are some
#' differences in the arguments: for `breaks` the set of options is different
#' and no preceding integer is allowed, `labels` are always assigned using
#' values of `breaks`, and the interval is always left-closed.
#'
#' @param x An instance of `CFTime`.
#' @param breaks A character string of a factor period (see [CFfactor()] for a
#'   description), or a character vector of timestamps that conform to the
#'   calendar of `x`, with a length of at least 2. Timestamps must be given in
#'   ISO8601 format, e.g. "2024-04-10 21:31:43".
#' @param ... Ignored.
#' @returns A factor with levels according to the `breaks` argument, with
#'   attributes 'period', 'era' and 'CFTime'. When `breaks` is a factor
#'   period, attribute 'period' has that value, otherwise it is '"day"'. When
#'   `breaks` is a character vector of timestamps, attribute 'CFTime' holds an
#'   instance of `CFTime` that has the same definition as `x`, but with (ordered)
#'   offsets generated from the `breaks`. Attribute 'era' is always -1.
#' @aliases cut
#' @seealso [CFfactor()] produces a factor for several fixed periods, including
#'   for eras.
#' @export
#' @examples
#' x <- CFtime("days since 2021-01-01", "365_day", 0:729)
#' breaks <- c("2022-02-01", "2021-12-01", "2023-01-01")
#' cut(x, breaks)
cut.CFTime <- function (x, breaks, ...) {
  if (!inherits(x, "CFTime"))
    stop("Argument 'x' must be a CFTime instance", call. = FALSE) # nocov
  x$cut(breaks)
}

#' Find the index of timestamps in the time series
#'
#' Find the index in the time series for each timestamp given in argument `x`.
#' Values of `x` that are before the earliest value or after the latest value in
#' `y` will be returned as `NA`. Alternatively, when `x` is a numeric vector of
#' index values, return the valid indices of the same vector, with the side
#' effect being the attribute "CFTime" associated with the result.
#'
#' Timestamps can be provided as vectors of character strings, `POSIXct` or
#' `Date.`
#'
#' Matching also returns index values for timestamps that fall between two
#' elements of the time series - this can lead to surprising results when time
#' series elements are positioned in the middle of an interval (as the CF
#' Metadata Conventions instruct us to "reasonably assume"): a time series of
#' days in January would be encoded in a netCDF file as
#' `c("2024-01-01 12:00:00", "2024-01-02 12:00:00", "2024-01-03 12:00:00", ...)`
#' so `x <- c("2024-01-01", "2024-01-02", "2024-01-03")` would result in
#' `(NA, 1, 2)` (or `(NA, 1.5, 2.5)` with `method = "linear"`) because the date
#' values in `x` are at midnight. This situation is easily avoided by ensuring
#' that `y` has bounds set (use `bounds(y) <- TRUE` as a proximate solution if
#' bounds are not stored in the netCDF file). See the Examples.
#'
#' If bounds are set, the indices are informed by those bounds. If the
#' bounds are not contiguous, returned values may be `NA` even if the
#' value of `x` falls between two valid timestamps.
#'
#' Values of `x` that are not valid timestamps according to the calendar of `y`
#' will be returned as `NA`.
#'
#' `x` can also be a numeric vector of index values, in which case the valid
#' values in `x` are returned. If negative values are passed, the positive
#' counterparts will be excluded and then the remainder returned. Positive and
#' negative values may not be mixed. Using a numeric vector has the side effect
#' that the result has the attribute "CFTime" describing the temporal dimension
#' of the slice. If index values outside of the range of `y` (`1:length(y)`) are
#' provided, an error will be thrown.
#'
#' @param x Vector of `character`, `POSIXt` or `Date` values to find indices
#'   for, or a numeric vector.
#' @param y A [CFTime] instance.
#' @param method Single value of "constant" or "linear". If `"constant"` or when
#'   bounds are set on argument `y`, return the index value for each match. If
#'   `"linear"`, return the index value with any fractional value.
#' @param rightmost.closed Whether or not to include the upper limit of
#'   argument `x`. Default is `FALSE`. This argument is ignored when
#'   argument `x` contains index values.
#' @return A numeric vector giving indices into `y` for the values of
#'   `x`. If there is at least 1 valid index, then attribute "CFTime"
#'   contains an instance of `CFTime` that describes the dimension of
#'   filtering the dataset associated with `y` with the result of this
#'   method, excluding any `NA` values.
#' @export
#'
#' @examples
#' cf <- CFtime("days since 2020-01-01", "360_day", 1440:1799 + 0.5)
#' as_timestamp(cf)[1:3]
#' x <- c("2024-01-01", "2024-01-02", "2024-01-03")
#' indexOf(x, cf)
#' indexOf(x, cf, method = "linear")
#'
#' bounds(cf) <- TRUE
#' indexOf(x, cf)
#'
#' # Non-existent calendar day in a `360_day` calendar
#' x <- c("2024-03-30", "2024-03-31", "2024-04-01")
#' indexOf(x, cf)
#'
#' # Numeric x
#' indexOf(c(29, 30, 31), cf)
indexOf <- function(x, y, method = "constant", rightmost.closed = FALSE) {
  y$indexOf(x, method, rightmost.closed)
}

#' Extreme time series values
#'
#' Character representation of the extreme values in the time series.
#'
#' @param x An instance of the [CFTime] class.
#' @param format A character string with format specifiers, optional. If it is
#'   missing or an empty string, the most economical ISO8601 format is chosen:
#'   "date" when no time information is present in `x`, "timestamp" otherwise.
#'   Otherwise a suitable format specifier can be provided.
#' @param bounds Logical to indicate if the extremes from the bounds should be
#'   used, if set. Defaults to `FALSE`.
#' @param ... Ignored.
#' @param na.rm Ignored.
#' @return Vector of two character representations of the extremes of the time
#'   series.
#' @export
#' @examples
#' cf <- CFtime("days since 1850-01-01", "julian", 0:364)
#' range(cf)
#' range(cf, "%Y-%b-%e")
range.CFTime <- function(x, format = "", bounds = FALSE, ..., na.rm = FALSE) {
  x$range(format, bounds)
}

#' Indicates if the time series is complete
#'
#' This function indicates if the time series is complete, meaning that the time
#' steps are equally spaced and there are thus no gaps in the time series.
#'
#' This function gives exact results for time series where the nominal
#' *unit of separation* between observations in the time series is exact in
#' terms of the calendar unit. As an example, for a calendar unit of "days" where the
#' observations are spaced a fixed number of days apart the result is exact, but
#' if the same calendar unit is used for data that is on a monthly basis, the
#' *assessment* is approximate because the number of days per month is variable
#' and dependent on the calendar (the exception being the `360_day` calendar,
#' where the assessment is exact). The *result* is still correct in most cases
#' (including all CF-compliant data sets that the developers have seen) although
#' there may be esoteric constructions of CFTime and offsets that trip up this
#' implementation.
#'
#' @param x An instance of the [CFTime] class.
#' @returns logical. `TRUE` if the time series is complete, with no gaps;
#'   `FALSE` otherwise. If no offsets have been added to the `CFTime` instance,
#'   `NA` is returned.
#' @export
#' @examples
#' t <- CFtime("days since 1850-01-01", "julian", 0:364)
#' is_complete(t)
is_complete <- function(x) {
  if (!inherits(x, "CFTime")) stop("Argument must be an instance of `CFTime`", call. = FALSE)
  x$equidistant()
}

#' Which time steps fall within extreme values
#'
#' Given a vector of character timestamps, return a logical vector of a length
#' equal to the number of time steps in the time series with values `TRUE` for
#' those time steps that fall between the two extreme values of the vector
#' values, `FALSE` otherwise.
#'
#' If bounds were set these will be preserved.
#'
#' @param x The `CFTime` instance to operate on.
#' @param extremes Character vector of timestamps that represent the time period
#'   of interest. The extreme values are selected. Badly formatted timestamps
#'   are silently dropped.
#' @param rightmost.closed Is the right side closed, i.e. included in the
#'   result? Default is `FALSE`. A specification of `c("2022-01-01",
#'   "2023-01-01)` will thus include all time steps that fall in the year 2022
#'   when `closed = FALSE` but include `2023-01-01` if that exact value is
#'   present in the time series.
#' @returns A logical vector with a length equal to the number of time steps in
#'   `x` with values `TRUE` for those time steps that fall between the extreme
#'   values, `FALSE` otherwise.
#'
#'   An attribute 'CFTime' will have the same definition as `x` but with offsets
#'   corresponding to the time steps falling between the two extremes. If there
#'   are no values between the extremes, the attribute is `NULL`.
#' @export
#' @examples
#' t <- CFtime("hours since 2023-01-01 00:00:00", "standard", 0:23)
#' slice(t, c("2022-12-01", "2023-01-01 03:00"))
slice <- function(x, extremes, rightmost.closed = FALSE) {
  if (!inherits(x, "CFTime")) stop("First argument must be an instance of `CFTime`", call. = FALSE)
  x$slice(extremes, rightmost.closed)
}

#' Equivalence of CFTime objects
#'
#' This operator can be used to test if two [CFTime] objects represent the same
#' CF-convention time coordinates. Two `CFTime` objects are considered equivalent
#' if they have an equivalent calendar and the same offsets.
#'
#' @param e1,e2 Instances of the `CFTime` class.
#' @returns `TRUE` if the `CFTime` objects are equivalent, `FALSE` otherwise.
#' @export
#' @aliases CFtime-equivalent
#' @examples
#' e1 <- CFtime("days since 1850-01-01", "gregorian", 0:364)
#' e2 <- CFtime("days since 1850-01-01 00:00:00", "standard", 0:364)
#' e1 == e2
"==.CFTime" <- function(e1, e2)
  e1$calendar$is_equivalent(e2$calendar) &&
  length(e1$offsets) == length(e2$offsets) &&
  all(e1$offsets == e2$offsets)

#' Extend a CFTime or CFClimatology object
#'
#' A [CFTime] or [CFClimatology] instance can be extended with this operator,
#' using values from another `CFTime` or `CFClimatology` instance, respectively,
#' or a vector of numeric offsets or character timestamps. If the values come
#' from another `CFTime` or `CFClimatology` instance, the calendars of the two
#' instances must be compatible. If the calendars of the instances are not
#' compatible, an error is thrown.
#'
#' The resulting `CFTime` or `CFClimatology` instance will have the offsets of
#' the original instance, appended with offsets from argument `e2` in the order
#' that they are specified. If the new sequence of offsets is not monotonically
#' increasing a warning is generated (the COARDS metadata convention requires
#' offsets to be monotonically increasing).
#'
#' There is no reordering or removal of duplicates. This is because the time
#' series are usually associated with a data set and the correspondence between
#' the data in the files and the instance is thus preserved. When merging the
#' data sets described by this time series, the order must be identical to the
#' merging here.
#'
#' Note that when adding multiple vectors of offsets to a `CFTime` instance, it
#' is more efficient to first concatenate the vectors and then do a final
#' addition to the `CFTime` instance. So avoid `CFtime(definition, calendar, e1) + CFtime(definition, calendar, e2) + CFtime(definition, calendar, e3) + ...`
#' but rather do `CFtime(definition, calendar) + c(e1, e2, e3, ...)`. It is the
#' responsibility of the operator to ensure that the offsets of the different
#' data sets are in reference to the same calendar.
#'
#' Note also that `RNetCDF` and `ncdf4` packages both return the values of the
#' "time" dimension as a 1-dimensional array. You have to de-class the array to
#' a vector before adding offsets to an existing `CFtime` instance.
#'
#' If both operands to this operator are `CFTime` or `CFClimatology` instances
#' and both have bounds set, then their bounds will be preserved in the result.
#'
#' @param e1 Instance of the `CFTime` class.
#' @param e2 Instance of the `CFTime` class with a calendar compatible with that
#'   of argument `e1`. This argument may also be a numeric vector with offsets
#'   from the origin of argument `e1`, or a vector of `character` timestamps in
#'   ISO8601 or UDUNITS format.
#' @returns A `CFTime` object with the offsets of argument `e1` extended by the
#'   values from argument `e2`.
#' @export
#' @aliases CFtime-merge
#' @examples
#' e1 <- CFTime$new("days since 1850-01-01", "gregorian", 0:364)
#' e2 <- CFTime$new("days since 1850-01-01 00:00:00", "standard", 365:729)
#' e1 + e2
"+.CFTime" <- function(e1, e2) {
  diff <- NULL
  new_time <- if (inherits(e2, "CFTime")) {
    if (!e1$calendar$is_compatible(e2$calendar)) stop("Calendars not compatible", call. = FALSE) # nocov
    if (all(e1$calendar$origin[1:6] == e2$calendar$origin[1:6]))
      CFTime$new(e1$calendar$definition, e1$calendar$name, c(e1$offsets, e2$offsets))
    else {
      diff <- e1$calendar$parse(paste(e2$calendar$origin_date, e2$calendar$origin_time))$offset
      CFTime$new(e1$calendar$definition, e1$calendar$name, c(e1$offsets, e2$offsets + diff))
    }
  } else if (is.numeric(e2) && .validOffsets(e2)) {
    CFTime$new(e1$calendar$definition, e1$calendar$name, c(e1$offsets, e2))
  } else {
    time <- e1$calendar$parse(e2)
    if (anyNA(time$year)) stop("Argument `e2` contains invalid timestamps", call. = FALSE) # nocov
    CFTime$new(e1$calendar$definition, e1$calendar$name, c(e1$offsets, time$offset))
  }

  if (!is.null(be1 <- e1$bounds) && !is.null(be2 <- e2$bounds)) {
    if (!is.null(diff))
      be2 <- be2 + diff
    new_time$bounds <- cbind(be1, be2)
  }

  new_time
}

#' @title Extending a CFClimatology object
#' @rdname plus-.CFTime
#' @param c1,c2 Instances of the `CFClimatology` class, having compatible
#'   calendars.
#' @returns A `CFClimatology` object with the offsets of argument `c1` extended
#'   by the values from argument `c2`.
#' @export
"+.CFClimatology" <- function(c1, c2) {
  if (!inherits(c2, "CFClimatology"))
    stop("Can only merge a `CFClimatology` instance to another `CFClimatology`")
  if (!c1$calendar$is_compatible(c2$calendar)) stop("Calendars not compatible", call. = FALSE) # nocov

  if (all(c1$calendar$origin[1:6] == c2$calendar$origin[1:6]))
    CFClimatology$new(c1$calendar$definition, c1$calendar$name, c(c1$offsets, c2$offsets), cbind(c1$bounds, c2$bounds))
  else {
    diff <- c1$calendar$parse(paste(c2$calendar$origin_date, c2$calendar$origin_time))$offset
    CFClimatology$new(c1$calendar$definition, c1$calendar$name, c(c1$offsets, c2$offsets + diff), cbind(c1$bounds, c2$bounds + diff))
  }
}

# ==============================================================================
# Factors and coverage

#' Create a factor from the offsets in a `CFTime` instance
#'
#' With this function a factor can be generated for the time series, or a part
#' thereof, contained in the [CFTime] instance. This is specifically interesting
#' for creating factors from the date part of the time series that aggregate the
#' time series into longer time periods (such as month) that can then be used to
#' process daily CF data sets using, for instance, `tapply()`.
#'
#' The factor will respect the calendar that the time series is built on. For
#' `period`s longer than a day this will result in a factor where the calendar
#' is no longer relevant (because calendars impacts days, not dekads, months,
#' quarters, seasons or years).
#'
#' The factor will be generated in the order of the offsets of the `CFTime`
#' instance. While typical CF-compliant data sources use ordered time series
#' there is, however, no guarantee that the factor is ordered as multiple
#' `CFTime` objects may have been merged out of order. For most processing with
#' a factor the ordering is of no concern.
#'
#' If the `era` parameter is specified, either as a vector of years to include
#' in the factor, or as a list of such vectors, the factor will only consider
#' those values in the time series that fall within the list of years, inclusive
#' of boundary values. Other values in the factor will be set to `NA`. The years
#' need not be contiguous, within a single vector or among the list items, or in
#' order.
#'
#' The following periods are supported by this function:
#'
#' \itemize{
#'   \item `year`, the year of each offset is returned as "YYYY".
#'   \item `season`, the meteorological season of each offset is returned as
#'   "Sx", with x being 1-4, preceeded by "YYYY" if no `era` is
#'   specified. Note that December dates are labeled as belonging to the
#'   subsequent year, so the date "2020-12-01" yields "2021S1". This implies
#'   that for standard CMIP files having one or more full years of data the
#'   first season will have data for the first two months (January and
#'   February), while the final season will have only a single month of data
#'   (December).
#'   \item `quarter`, the calendar quarter of each offset is returned as "Qx",
#'   with x being 1-4, preceeded by "YYYY" if no `era` is specified.
#'   \item `month`, the month of each offset is returned as "01" to
#'   "12", preceeded by "YYYY-" if no `era` is specified. This is the default
#'   period.
#'   \item `dekad`, ten-day periods are returned as
#'   "Dxx", where xx runs from "01" to "36", preceeded by "YYYY" if no `era`
#'   is specified. Each month is subdivided in dekads as follows: 1- days 01 -
#'   10; 2- days 11 - 20; 3- remainder of the month.
#'   \item `day`, the month and day of each offset are returned as "MM-DD",
#'   preceeded by "YYYY-" if no `era` is specified.
#' }
#'
#' It is not possible to create a factor for a period that is shorter than the
#' temporal resolution of the source data set from which the `t` argument
#' derives. As an example, if the source data set has monthly data, a dekad or
#' day factor cannot be created.
#'
#' Creating factors for other periods is not supported by this function. Factors
#' based on the timestamp information and not dependent on the calendar can
#' trivially be constructed from the output of the [as_timestamp()] function.
#'
#' For non-era factors the attribute 'CFTime' of the result contains a `CFTime`
#' instance that is valid for the result of applying the factor to a data set
#' that the `t` argument is associated with. In other words, if `CFTime`
#' instance 'At' describes the temporal dimension of data set 'A' and a factor
#' 'Af' is generated like `Af <- CFfactor(At)`, then `Bt <- attr(Af, "CFTime")`
#' describes the temporal dimension of the result of, say,
#' `B <- apply(A, 1:2, tapply, Af, FUN)`. The 'CFTime' attribute is `NULL` for
#' era factors.
#'
#' @param t An instance of the `CFTime` class whose offsets will be used to
#'   construct the factor.
#' @param period character. A character string with one of the values "year",
#'   "season", "quarter", "month" (the default), "dekad" or "day".
#' @param era numeric or list, optional. Vector of years for which to
#'   construct the factor, or a list whose elements are each a vector of years.
#'   If `era` is not specified, the factor will use the entire time series for
#'   the factor.
#'
#' @returns If `era` is a single vector or not specified, a factor with a
#'   length equal to the number of offsets in `t`. If `era` is a list, a list
#'   with the same number of elements and names as `era`, each containing a
#'   factor. Elements in the factor will be set to `NA` for time series values
#'   outside of the range of specified years.
#'
#'   The factor, or factors in the list, have attributes 'period', 'era' and
#'   'CFTime'. Attribute 'period' holds the value of the `period` argument.
#'   Attribute 'era' indicates the number of years that are included in the
#'   era, or -1 if no `era` is provided. Attribute 'CFTime' holds an
#'   instance of `CFTime` that has the same definition as `t`, but with offsets
#'   corresponding to the mid-point of non-era factor levels; if the `era`
#'   argument is specified, attribute 'CFTime' is `NULL`.
#' @seealso [cut()] creates a non-era factor for arbitrary cut points.
#' @export
#'
#' @examples
#' t <- CFtime("days since 1949-12-01", "360_day", 19830:54029)
#'
#' # Create a dekad factor for the whole time series
#' f <- CFfactor(t, "dekad")
#'
#' # Create three monthly factors for early, mid and late 21st century eras
#' ep <- CFfactor(t, era = list(early = 2021:2040, mid = 2041:2060, late = 2061:2080))
CFfactor <- function(t, period = "month", era = NULL) {
  if (!(inherits(t, "CFTime"))) stop("First argument to CFfactor() must be an instance of the `CFTime` class", call. = FALSE) # nocov
  t$factor(period, era)
}

#' Number of base time units in each factor level
#'
#' Given a factor as returned by [CFfactor()] and the [CFTime] instance from
#' which the factor was derived, this function will return a numeric vector with
#' the number of time units in each level of the factor.
#'
#' The result of this function is useful to convert between absolute and
#' relative values. Climate change anomalies, for instance, are usually computed
#' by differencing average values between a future period and a baseline period.
#' Going from average values back to absolute values for an aggregate period
#' (which is typical for temperature and precipitation, among other variables)
#' is easily done with the result of this function, without having to consider
#' the specifics of the calendar of the data set.
#'
#' If the factor `f` is for an era (e.g. spanning multiple years and the
#' levels do not indicate the specific year), then the result will indicate the
#' number of time units of the period in a regular single year. In other words,
#' for an era of 2041-2060 and a monthly factor on a standard calendar with a
#' `days` unit, the result will be `c(31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)`.
#' Leap days are thus only considered for the `366_day` and `all_leap` calendars.
#'
#' Note that this function gives the number of time units in each level of the
#' factor - the actual number of data points in the `cf` instance per factor
#' level may be different. Use [CFfactor_coverage()] to determine the actual
#' number of data points or the coverage of data points relative to the factor
#' level.
#'
#' @param t An instance of `CFTime`.
#' @param f A factor or a list of factors derived from the
#'   parameter `t`. The factor or list thereof should generally be generated by
#'   the function [CFfactor()].
#'
#' @returns If `f` is a factor, a numeric vector with a length equal to the
#'   number of levels in the factor, indicating the number of time units in each
#'   level of the factor. If `f` is a list of factors, a list with each element
#'   a numeric vector as above.
#' @export
#'
#' @examples
#' t <- CFtime("days since 2001-01-01", "365_day", 0:364)
#' f <- CFfactor(t, "dekad")
#' CFfactor_units(t, f)
CFfactor_units <- function(t, f) {
  if (!inherits(t, "CFTime")) stop("First argument to `CFfactor_units()` must be an instance of the `CFTime` class", call. = FALSE)
  t$factor_units(f)
}

#' Coverage of time elements for each factor level
#'
#' This function calculates the number of time elements, or the relative
#' coverage, in each level of a factor generated by [CFfactor()].
#'
#' @param t An instance of [CFTime].
#' @param f factor or list. A factor or a list of factors derived from the
#'   parameter `t`. The factor or list thereof should generally be generated by
#'   the function [CFfactor()].
#' @param coverage "absolute" or "relative".
#' @returns If `f` is a factor, a numeric vector with a length equal to the
#'   number of levels in the factor, indicating the number of units from the
#'   time series in `t` contained in each level of the factor when
#'   `coverage = "absolute"` or the proportion of units present relative to the
#'   maximum number when `coverage = "relative"`. If `f` is a list of factors, a
#'   list with each element a numeric vector as above.
#' @export
#'
#' @examples
#' t <- CFtime("days since 2001-01-01", "365_day", 0:364)
#' f <- CFfactor(t, "dekad")
#' CFfactor_coverage(t, f, "absolute")
CFfactor_coverage <- function(t, f, coverage = "absolute") {
  if (!inherits(t, "CFTime")) stop("First argument to `CFfactor_coverage()` must be an instance of the `CFTime` class", call. = FALSE) # nocov
  t$factor_coverage(f, coverage)
}

# ==============================================================================
# Regular functions

#' Create a vector that represents CF timestamps
#'
#' This function generates a vector of character strings or `POSIXct`s that
#' represent the date and time in a selectable combination for each offset.
#'
#' The character strings use the format `YYYY-MM-DDThh:mm:ss±hhmm`, depending on
#' the `format` specifier. The date in the string is not necessarily compatible
#' with `POSIXt` - in the `360_day` calendar `2017-02-30` is valid and
#' `2017-03-31` is not.
#'
#' For the "proleptic_gregorian" calendar the output can also be generated as a
#' vector of `POSIXct` values by specifying `asPOSIX = TRUE`. The same is
#' possible for the "standard" and "gregorian" calendars but only if all
#' timestamps fall on or after 1582-10-15.
#'
#' @param t The `CFTime` instance that contains the offsets to use.
#' @param format character. A character string with either of the values "date"
#'   or "timestamp". If the argument is not specified, the format used is
#'   "timestamp" if there is time information, "date" otherwise.
#' @param asPOSIX logical. If `TRUE`, for "standard", "gregorian" and
#'   "proleptic_gregorian" calendars the output is a vector of `POSIXct` - for
#'   other calendars an error will be thrown. Default value is `FALSE`.
#' @seealso The [CFTime] `format()` method gives greater flexibility through
#'   the use of `strptime`-like format specifiers.
#' @returns A character vector where each element represents a moment in time
#'   according to the `format` specifier.
#' @export
#' @examples
#' t <- CFtime("hours since 2020-01-01", "standard", seq(0, 24, by = 0.25))
#' as_timestamp(t, "timestamp")
#'
#' t2 <- CFtime("days since 2002-01-21", "standard", 0:20)
#' tail(as_timestamp(t2, asPOSIX = TRUE))
#'
#' tail(as_timestamp(t2))
#'
#' tail(as_timestamp(t2 + 1.5))
as_timestamp <- function(t, format = NULL, asPOSIX = FALSE) {
  if (!(inherits(t, "CFTime")))
    stop("First argument to `as_timestamp()` must be an instance of the `CFTime` class", call. = FALSE)
  t$as_timestamp(format, asPOSIX)
}

#' Return the number of days in a month given a certain CF calendar
#'
#' Given a vector of dates as strings in ISO 8601 or UDUNITS format and a
#' [CFTime] object, this function will return a vector of the same length as the
#' dates, indicating the number of days in the month according to the calendar
#' specification. If no vector of days is supplied, the function will return an
#' integer vector of length 12 with the number of days for each month of the
#' calendar (disregarding the leap day for `standard` and `julian` calendars).
#'
#' @param t The `CFTime` instance to use.
#' @param x character. An optional vector of dates as strings with format
#'   `YYYY-MM-DD`. Any time part will be silently ingested.
#'
#' @returns A vector indicating the number of days in each month for the vector
#'   of dates supplied as argument `x`. Invalidly specified dates will result in
#'   an `NA` value. If no dates are supplied, the number of days per month for
#'   the calendar as a vector of length 12.
#'
#' @export
#' @seealso When working with factors generated by [CFfactor()], it is usually
#'   better to use [CFfactor_units()] as that will consider leap days for
#'   non-era factors. [CFfactor_units()] can also work with other time periods
#'   and calendar units, such as "hours per month", or "days per season".
#' @examples
#' dates <- c("2021-11-27", "2021-12-10", "2022-01-14", "2022-02-18")
#' t <- CFtime("days since 1850-01-01", "standard")
#' month_days(t, dates)
#'
#' t <- CFtime("days since 1850-01-01", "360_day")
#' month_days(t, dates)
#'
#' t <- CFtime("days since 1850-01-01", "all_leap")
#' month_days(t, dates)
#'
#' month_days(t)
month_days <- function(t, x = NULL) {
  stopifnot(inherits(t, "CFTime"))

  if (is.null(x))
    return(t$calendar$month_days())
  else {
    if (!(is.character(x))) stop("Argument `x` must be a character vector of dates in 'YYYY-MM-DD' format")

    ymd <- t$calendar$parse(x)
    if (anyNA(ymd$year)) warning("Some dates could not be parsed. Result contains `NA` values.", call. = FALSE)
    return(t$calendar$month_days(ymd))
  }
}

#' Parse series of timestamps in CF format to date-time elements
#'
#' This function will parse a vector of timestamps in ISO8601 or UDUNITS format
#' into a data frame with columns for the elements of the timestamp: year,
#' month, day, hour, minute, second, time zone. Those timestamps that could not
#' be parsed or which represent an invalid date in the indicated `CFtime`
#' instance will have `NA` values for the elements of the offending timestamp
#' (which will generate a warning).
#'
#' The supported formats are the *broken timestamp* format from the UDUNITS
#' library and ISO8601 *extended*, both with minor changes, as suggested by the
#' CF Metadata Conventions. In general, the format is `YYYY-MM-DD hh:mm:ss.sss
#' hh:mm`. The year can be from 1 to 4 digits and is interpreted literally, so
#' `79-10-24` is the day Mount Vesuvius erupted and destroyed Pompeii, not
#' `1979-10-24`. The year and month are mandatory, all other fields are
#' optional. There are defaults for all missing values, following the UDUNITS
#' and CF Metadata Conventions. Leading zeros can be omitted in the UDUNITS
#' format, but not in the ISO8601 format. The optional fractional part can have
#' as many digits as the precision calls for and will be applied to the smallest
#' specified time unit. In the result of this function, if the fraction is
#' associated with the minute or the hour, it is converted into a regular
#' `hh:mm:ss.sss` format, i.e. any fraction in the result is always associated
#' with the second, rounded down to milli-second accuracy. The separator between
#' the date and the time can be a single whitespace character or a `T`.
#'
#' The time zone is optional and should have at least the hour or `Z` if
#' present, the minute is optional. The time zone hour can have an optional
#' sign. In the UDUNITS format the separator between the time and the time zone
#' must be a single whitespace character, in ISO8601 there is no separation
#' between the time and the timezone. Time zone names are not supported (as
#' neither UDUNITS nor ISO8601 support them) and will cause parsing to fail when
#' supplied, with one exception: the designator "UTC" is silently dropped (i.e.
#' interpreted as "00:00").
#'
#' Currently only the extended formats (with separators between the elements)
#' are supported. The vector of timestamps may have any combination of ISO8601
#' and UDUNITS formats.
#'
#' @param t An instance of `CFTime` to use when parsing the date.
#' @param x Vector of character strings representing timestamps in
#'   ISO8601 extended or UDUNITS broken format.
#' @returns A `data.frame` with constituent elements of the parsed timestamps in
#'   numeric format. The columns are year, month, day, hour, minute, second
#'   (with an optional fraction), time zone (character string), and the
#'   corresponding offset value from the origin. Invalid input data will appear
#'   as `NA` - if this is the case, a warning message will be displayed - other
#'   missing information on input will use default values.
#' @importFrom stats na.omit
#' @export
#' @examples
#' t <- CFtime("days since 0001-01-01", "proleptic_gregorian")
#'
#' # This will have `NA`s on output and generate a warning
#' timestamps <- c("2012-01-01T12:21:34Z", "12-1-23", "today",
#'                 "2022-08-16T11:07:34.45-10", "2022-08-16 10.5+04")
#' parse_timestamps(t, timestamps)
parse_timestamps <- function(t, x) {
  stopifnot(is.character(x), inherits(t, "CFTime"))
  if (t$calendar$unit > 4) stop("Parsing of timestamps on a 'month' or 'year' time unit is not supported.", call. = FALSE)

  out <- t$calendar$parse(x)
  if (anyNA(out$year))
    warning("Some dates could not be parsed. Result contains `NA` values.") # nocov
  if (length(unique(na.omit(out$tz))) > 1)
    warning("Timestamps have multiple time zones. Some or all may be different from the calendar time zone.") # nocov
  else if (out$tz[1] != t$calendar$timezone)
    warning("Timestamps have time zone that is different from the calendar.") # nocov
  out
}

test_that("timestamp string parsing to offsets and deparsing of offsets to timestamps", {
  # This test tests: global CFt* constants, CFtime(), parse_timestamps(), as_timestamp():
  # decomposing offsets into timestamp elements, generating timestamp strings,
  # parsing timestamp strings back into timestamp elements, including negative
  # offsets.
  for (c in c("standard", "proleptic_gregorian", "julian", "tai", "360_day", "365_day", "366_day", "noleap", "all_leap")) {
    for (u in 1:4) {
      offsets <- -1000:1000
      def <- paste(CFt$units$name[u], "since 1992-08-20")
      t <- CFtime(def, c, offsets)
      time <- t$calendar$offsets2time(t$offsets)
      ts <- as_timestamp(t, "timestamp")
      cf2 <- CFtime(def, c)
      tp <- parse_timestamps(cf2, ts)
      expect_equal(tp, time)
    }
  }

  # Add no offsets, test return value
  t <- CFTime$new("days since 1992-08-20", "standard")
  res <- t$calendar$offsets2time()
  expect_equal(nrow(res), 0L)
  t <- CFTime$new("days since 1992-08-20", "utc")
  res <- t$calendar$offsets2time()
  expect_equal(nrow(res), 0L)
})

test_that("testing calendars with leap years", {
  # This test tests that for standard, proleptic_gregorian, julian and tai
  # calendars in leap years before/on/after the leap day function as needed.
  # Also testing year 2000 and 2100 offsets.
  for (c in c("standard", "proleptic_gregorian", "tai", "julian")) {
    for (d in c("1996-01-15", "1996-02-29", "1996-04-01")) {
      def <- paste("days since", d)
      t <- CFtime(def, c, c(1:2500, 36501:39000))
      time <- t$calendar$offsets2time(t$offsets)
      ts <- as_timestamp(t, "timestamp")
      cf2 <- CFtime(def, c)
      tp <- parse_timestamps(cf2, ts)
      expect_equal(tp, time)
    }
  }
})

# test_that("Testing milli-second timestamp string parsing to offsets and deparsing
#           of offsets to timestamps", {
#   # This test tests: global CFt* constants, CFtime(), parse_timestamps(), as_timestamp():
#   # decomposing offsets into milli-second timestamp elements, generating
#   # timestamp strings, parsing timestamp strings back into timestamp elements.
#   for (c in c("standard", "proleptic_gregorian", "julian", "tai", "360_day", "365_day", "366_day", "noleap", "all_leap")) {
#     for (u in 1:2) {
#       offsets <- runif(10000, max = 10000)
#       def <- paste(CFt$units$name[u], "since 1978-08-20 07:34:12.2")
#       expect_warning(t <- CFtime(def, c, offsets)) # not ordered
#       time <- t$calendar$offsets2time(t$offsets)
#       ts <- as_timestamp(t, "timestamp")
#       cf2 <- CFtime(def, c)
#       tp <- parse_timestamps(cf2, ts)
#       expect_equal(round(tp[c("second", "offset")], 2), round(time[c("second", "offset")], 2))
#     }
#   }
# })

test_that("Disallow parsing of timestamps on month and year calendar units", {
  for (c in c("standard", "proleptic_gregorian", "julian", "tai", "360_day", "365_day", "366_day", "noleap", "all_leap")) {
    for (u in 5:6) {
      def <- paste(CFt$units$name[u], "since 2020-01-01")
      t <- CFtime(def, c, 0:23)
      ts <- as_timestamp(t, "timestamp")
      expect_error(parse_timestamps(t, ts))
    }
  }
})

test_that("Gregorian/Julian calendar gap in the standard calendar", {
  t <- CFtime("days since 1582-10-01", "standard", 0:3)
  ts <- as_timestamp(t)
  expect_equal(ts[4], "1582-10-04")
  expect_false(t$calendar$POSIX_compatible(0:3))

  t <- t + 4:10
  ts <- as_timestamp(t)
  expect_equal(ts[5], "1582-10-15")
  expect_equal(ts[11], "1582-10-21")
  expect_equal(parse_timestamps(t, c("1582-09-30", ts))$offset, -1:10)

  t <- CFtime("days since 1582-10-20", "standard", -10:0)
  ts <- as_timestamp(t)
  expect_equal(ts[5], "1582-10-04")
  expect_equal(ts[6], "1582-10-15")

  expect_false(t$POSIX_compatible())            # use t$offsets
  expect_true(t$calendar$POSIX_compatible(0:100))    # use supplied offsets
  expect_true(t$calendar$POSIX_compatible(-5:100))
  expect_false(t$calendar$POSIX_compatible(-6:100))
})

test_that("Leap seconds in the utc calendar work fine", {
  expect_error(CFtime("seconds since 1792-01-01", "utc")) # No UTC in 1792
  t <- CFTime$new("seconds since 1972-06-30 23:59:57", "utc", 1:4)
  expect_equal(as_timestamp(t), c("1972-06-30T23:59:58", "1972-06-30T23:59:59",
                                  "1972-06-30T23:59:60", "1972-07-01T00:00:00"))

  expect_error(CFTime$new("seconds since 1973-06-30 23:59:60", "utc", -1:2)) # Bad leap second
  t <- CFTime$new("seconds since 1972-06-30 23:59:60", "utc", -1:2)
  expect_equal(as_timestamp(t), c("1972-06-30T23:59:59", "1972-06-30T23:59:60",
                                  "1972-07-01T00:00:00", "1972-07-01T00:00:01"))

  t <- CFTime$new("seconds since 2016-12-31 23:59:58", "utc", 1:4)
  expect_equal(as_timestamp(t), c("2016-12-31T23:59:59", "2016-12-31T23:59:60",
                                  "2017-01-01T00:00:00", "2017-01-01T00:00:01"))

  t <- CFTime$new("minutes since 2016-12-31 23:58:00", "utc", 1:4)
  expect_equal(as_timestamp(t), c("2016-12-31T23:59:00", "2016-12-31T23:59:60",
                                  "2017-01-01T00:00:59", "2017-01-01T00:01:59"))

  t <- CFTime$new("hours since 2016-12-31 22:00", "utc", 1:4)
  expect_equal(as_timestamp(t), c("2016-12-31T23:00:00", "2016-12-31T23:59:60",
                                  "2017-01-01T00:59:59", "2017-01-01T01:59:59"))

  t <- CFTime$new("days since 2016-12-30", "utc", 1:4)
  expect_equal(as_timestamp(t), c("2016-12-31T00:00:00", "2016-12-31T23:59:60",
                                  "2017-01-01T23:59:59", "2017-01-02T23:59:59"))
})

test_that("Fractional time parts and replace Z timezone with offset", {
  offsets <- c("1980-05-06Z", "1980-05-06 12.32Z", "1980-05-06 12:54.38Z", "1980-05-06 12:54:12.32Z")
  res <- matrix(c(0.00, 0.00, 0.80, 0.32,
                  0.0000, 0.2000, 0.3800, 0.2053,
                  0.0000, 0.3200, 0.9063, 0.9034,
                  0.0000, 0.5133, 0.5378, 0.5376), ncol = 4)
  for (c in c("standard", "proleptic_gregorian", "julian", "tai", "360_day", "365_day", "366_day", "noleap", "all_leap")) {
    for (u in 1:4) {
      def <- paste(CFt$units$name[u], "since 1978-08-20")
      t <- CFtime(def, c)
      p <- t$calendar$parse(offsets)
      expect_equal(round(p$offset %% 1, 4), res[,u])
      expect_equal(p$tz, rep("+0000", 4))
    }
  }

  offsets <- c("1980-05-06", "1980-05-06 12.32", "1980-05-06 12:54.38", "1980-05-06 12:54:12.32")
  res <- matrix(c(0.00, 0.00, 0.80, 0.32,
                  0.0333, 0.2333, 0.4133, 0.2387,
                  0.0006, 0.3206, 0.9069, 0.9040,
                  0.0000, 0.5134, 0.5378, 0.5377), ncol = 4)
  for (u in 1:4) {
    def <- paste(CFt$units$name[u], "since 1978-08-20")
    t <- CFtime(def, "utc")
    p <- t$calendar$parse(offsets)
    expect_equal(round(p$offset %% 1, 4), res[,u])
  }

})

test_that("test all variants of creating a CFtime object and useful functions", {
  # Call CFtime() without arguments
  expect_error(CFtime())

  # Call CFtime() with an invalid definition
  expect_error(CFtime("foo"))
  expect_error(CFtime("zxcv since 1991-01-01"))
  expect_error(CFtime("days since -991-01-01"))

  # Call CFtime() with a valid definition and an invalid calendar
  expect_error(CFtime("days since 1991-01-01", "foo"))
  expect_equal(calendar(CFtime("days since 1991-01-01", NULL, NULL)), "standard")
  expect_equal(definition(CFtime("days since 1991-01-01", NULL, NULL)), "days since 1991-01-01")
  expect_equal(resolution(CFtime("days since 1991-01-01", NULL, NULL)), NA_real_)

  # Call CFtime() with a valid definition and calendar but faulty offsets
  expect_error(CFtime("days since 1991-01-01", "standard", c(0:10, NA)))
  expect_error(CFtime("days since 1991-01-01", "utc", c(0:10, NA)))

  # CFtime() with only a definition
  t <- CFtime("d per 1991-01-01")

  expect_equal(origin(t)[1:3], data.frame(year = 1991, month = 1, day = 1))
  expect_equal(unit(t), "days")
  expect_equal(calendar(t), "standard")
  expect_equal(length(offsets(t)), 0L)

  # CFtime with only a definition and a calendar
  t <- CFtime("d per 1991-01-01", "julian")
  expect_match(capture_output(t$print()), "^CF calendar:")
  expect_match(capture_output(t$print()), "Elements: \\(no elements\\)\\n  Bounds  : \\(not set\\)$")

  expect_equal(origin(t)[1:3], data.frame(year = 1991, month = 1, day = 1))
  expect_equal(unit(t), "days")
  expect_equal(calendar(t), "julian")
  expect_equal(length(offsets(t)), 0L)

  # CFtime with a single offset
  t <- t + 15
  expect_equal(as_timestamp(t, "date"), "1991-01-16")
  expect_match(capture_output(t$print()), "Element : 1991-01-16 \\n  Bounds  : not set$")

  # Invalid offsets
  expect_error(CFtime("d per 1991-01-01", "julian", c(TRUE, FALSE, FALSE)))

  # Character offsets
  t <- CFtime("hours since 2023-01-01", "360_day", c("2023-01-01T00:00:00", "2023-04-30T23:00"))
  expect_equal(range(t, bounds = TRUE), c("2023-01-01T00:00:00", "2023-04-30T23:00:00")) # bounds have not been set
  expect_equal(length(as_timestamp(t, "timestamp")), 2)

  expect_warning(t <- CFtime("days since 2023-01-01", "366_day", c("2023-01-01", "2023-04-13", "2023-10-30", "2023-05-12")))
  expect_equal(range(t), c("2023-01-01", "2023-10-30"))

  # Merge two CFtime instances / extend offsets
  t1 <- CFtime("hours since 2001-01-01", "360_day", 0:99)
  t2 <- CFtime("hours since 2001-01-01", "julian", 100:199)
  expect_false(t1 == t2)
  expect_error(t1 + t2)

  t2 <- CFtime("hours since 2001-01-01", "360_day", 100:199)
  expect_false(t1 == t2)
  expect_equal(length(offsets(t1 + t2)), 200)

  expect_equal(length(offsets(t1 + 100:199)), 200)

  t1 <- CFtime("days since 2022-01-01", "365_day", 0:364)
  t2 <- CFtime("days since 2023-01-01", "365_day", 0:364)
  expect_match(capture_output(t1$print()), "between 365 elements\\)\\n  Bounds  : not set$")
  expect_true(length(offsets(t1 + t2)) == 730)
  expect_true(all(range(diff(offsets(t1 + t2))) == c(1, 1)))
  expect_warning(t3 <- t2 + t1)
  expect_true(length(offsets(t3)) == 730)
  expect_false((range(diff(offsets(t3))) == c(1, 1))[1])

  t2 <- t1 + c("2023-01-01", "2023-01-02")
  expect_match(capture_output(t1$print()), "between 365 elements\\)\\n  Bounds  : not set$")
  expect_equal(t2$resolution, 1)

  # Timezones
  expect_equal(timezone(t1), "+0000")
  expect_false(grepl("+0000", capture_output(t1$print()), fixed = TRUE))
  t1 <- CFtime("days since 2022-01-01 00:00:00+04", "365_day", 0:364)
  expect_true(grepl("+0400", capture_output(t1$print()), fixed = TRUE))

  # Time series completeness
  t <- CFtime("d per 1991-01-01", "julian")
  expect_error(is_complete("zxcv"))
  expect_true(is.na(is_complete(t)))
  expect_true(is_complete(t1))
  mid_months <- c("1950-01-16T12:00:00", "1950-02-15T00:00:00", "1950-03-16T12:00:00", "1950-04-16T00:00:00", "1950-05-16T12:00:00", "1950-06-16T00:00:00",
                  "1950-07-16T12:00:00", "1950-08-16T12:00:00", "1950-09-16T00:00:00", "1950-10-16T12:00:00", "1950-11-16T00:00:00", "1950-12-16T12:00:00")
  t <- CFtime("days since 1950-01-01", "standard", mid_months)
  expect_true(is_complete(t))
  ty <- CFtime("years since 2020-01-01", "standard", 0:19)
  expect_true(is_complete(ty))
  ty <- ty + 30:39
  expect_false(is_complete(ty))

  # Range
  t <- CFtime("days since 2001-01-01")
  expect_equal(range(t), c(NA_character_, NA_character_))
  t <- t + 0:1
  expect_error(range(t, 123))
  expect_error(range(t, c("asd %d", "%F")))
  expect_equal(range(t, "%Y-%B-%Od"), c("2001-January-01", "2001-January-02"))

  # Range on unsorted offsets
  random <- runif(100, min = 1, max = 99)
  expect_warning(t <- CFtime("days since 2001-01-01", offsets = c(0, random[1:50], 100, random[51:100])))
  expect_equal(range(t), c("2001-01-01", paste0(as.Date("2001-01-01") + 100)))

  # Subsetting
  t <- CFtime("hours since 2023-01-01 00:00:00", "standard", 0:239)
  expect_error(slice("zxcv"))
  expect_true(all(slice(t, c("2023-01-01", "2023-02-01"))))
  expect_true(length(which(slice(t, c("2023-01-01", "2023-01-01")))) == 1) # Identical extremes
  expect_true(length(which(slice(t, c("2023-01-01", "2023-05-01")))) == 240)
  expect_true(length(which(slice(t, c("2023-01-01 00:00", "2023-01-01 04:00")))) == 4)
  expect_true(length(which(slice(t, c("2023-01-01 04:00", "2023-01-01 00:00")))) == 4) # extremes in reverse order
  expect_true(slice(t, c("2022-01-01", "2023-01-02"))[1]) # early extreme before timeseries
  expect_true(all(!slice(t, c("2023-02-01", "2023-03-01")))) # both extremes outside time series
  expect_equal(sum(slice(t, c("2023-01-01 00:00", "2023-01-01 04:00", "2023-01-02 00:00"))), 24)
  t_slice <- slice(t, c("2023-01-01 00:00", "2023-01-01 04:00", "2023-01-02 00:00"), TRUE)
  expect_equal(sum(t_slice), 25)
  new_t <- attr(t_slice, "CFTime")
  expect_identical(new_t, t_sub <- t$subset(1:25))
  expect_identical(t_sub, t_sub$copy())

  t$bounds <- TRUE
  s <- t$slice(c("2023-01-01", "2023-05-01"))
  st <- attr(s, "CFTime")
  expect_equal(st$bounds, t$bounds)
})

test_that("Leap years on some calendars", {
  t <- CFTime$new("days since 2025-01-01", "360_day")
  expect_true(all(!t$calendar$leap_year(c(2000:2025))))
  t <- CFTime$new("days since 2025-01-01", "366_day")
  expect_true(all(t$calendar$leap_year(c(2000:2025))))
})

test_that("Calendar 'none'", {
  t <- CFTime$new("days since 2025-01-01", "none")
  expect_true(inherits(t$calendar, "CFCalendarNone"))
  t <- t + 0:4
  ymd <- data.frame(year = c(rep(2025, 5), NA), month = c(rep(1, 5), NA), day = c(1, 1, 1, 2, 3, NA))
  expect_equal(t$calendar$valid_days(ymd), c(T, T, T, F, F, NA))
  expect_equal(t$calendar$month_days(), rep(NA, 12))
  expect_equal(t$calendar$month_days(ymd), rep(NA, 6))
  expect_equal(t$calendar$leap_year(ymd$year), rep(NA, 6))
  expect_equal(t$calendar$date2offset(ymd), c(0, 0, 0, 0, 0, NA))
  expect_equal(t$calendar$offset2date(1:5), data.frame(year = rep(2025, 5), month = rep(1, 5), day = rep(1, 5)))
})

#' @title Julian CF calendar
#'
#' @description This class represents a Julian calendar of 365 days per year,
#'   with every fourth year being a leap year of 366 days. The months and the
#'   year align with the standard calendar. This calendar is not compatible with
#'   the standard POSIXt calendar.
#'
#'   This calendar starts on 1 January of year 1: 0001-01-01 00:00:00. Any dates
#'   before this will generate an error.
#'
#' @aliases CFCalendarJulian
#' @docType class
CFCalendarJulian <- R6::R6Class("CFCalendarJulian",
  inherit = CFCalendar,
  private = list(
    # Rata Die, the number of days from the day before 0001-01-01 to
    # origin of this calendar. Used to convert offsets from the calendar origin
    # to the day before 0001-01-01 for arithmetic calculations.
    rd = 0L
  ),
  public = list(
    #' @description Create a new CF calendar.
    #' @param nm The name of the calendar. This must be "julian". This argument
    #' is superfluous but maintained to be consistent with the initialization
    #' methods of the parent and sibling classes.
    #' @param definition The string that defines the units and the origin, as
    #' per the CF Metadata Conventions.
    #' @return A new instance of this class.
    initialize = function(nm, definition) {
      super$initialize(nm, definition)
      private$rd <- .julian_date2offset(self$origin, self$leap_year(self$origin$year))
    },

    #' @description Indicate which of the supplied dates are valid.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return Logical vector with the same length as argument `ymd` has rows
    #'   with `TRUE` for valid days and `FALSE` for invalid days, or `NA` where
    #'   the row in argument `ymd` has `NA` values.
    valid_days = function(ymd) {
      ymd$year >= 1L & ymd$month >= 1L & ymd$month <= 12L & ymd$day >= 1L &
      ifelse(self$leap_year(ymd$year),
        ymd$day <= c(31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)[ymd$month],
        ymd$day <= c(31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)[ymd$month])
    },

    #' @description Determine the number of days in the month of the calendar.
    #' @param ymd `data.frame`, optional, with dates parsed into their parts.
    #' @return A vector indicating the number of days in each month for the
    #'   dates supplied as argument `ymd`. If no dates are supplied, the number
    #'   of days per month for the calendar as a vector of length 12, for a
    #'   regular year without a leap day.
    month_days = function(ymd = NULL) {
      if (is.null(ymd)) return(c(31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31))

      ifelse(self$leap_year(ymd$year),
             c(31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)[ymd$month],
             c(31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)[ymd$month])
    },

    #' @description Indicate which years are leap years.
    #' @param yr Integer vector of years to test.
    #' @return Logical vector with the same length as argument `yr`. `NA` is
    #'   returned where elements in argument `yr` are `NA`.
    leap_year = function(yr) {
      yr %% 4L == 0L
    },

    #' @description Calculate difference in days between a `data.frame` of time
    #'   parts and the origin.
    #'
    #' @param x `data.frame`. Dates to calculate the difference for.
    #' @return Integer vector of a length equal to the number of rows in
    #'   argument `x` indicating the number of days between `x` and the origin
    #'   of the calendar, or `NA` for rows in `x` with `NA` values.
    date2offset = function(x) {
      .julian_date2offset(x, self$leap_year(x$year)) - private$rd
    },

    #' @description Calculate date parts from day differences from the origin. This
    #'   only deals with days as these are impacted by the calendar.
    #'   Hour-minute-second timestamp parts are handled in [CFCalendar].
    #'
    #' @param x Integer vector of days to add to the origin.
    #' @return A `data.frame` with columns 'year', 'month' and 'day' and as many
    #'   rows as the length of vector `x`.
    offset2date = function(x) {
      .julian_offset2date(x + private$rd)
    }
  )
)

# ==============================================================================
# The below functions use arithmetic offset calculation from date parts and
# vice-versa. These functions are R-ified from pseudo-functions in Reingold &
# Derschowitz, "Calendrical Calculations", 2018.

#' Dates to offset, from function `fixed-from-julian()`
#'
#' @param x `data.frame` with columns "year", "month" and "date"
#' @param leapyear Logical vector of the same length as `x` has rows indicating
#' for each row in `x` if this is a leap year.
#' @return Integer vector of offsets for the dates in `x`. The offsets are
#' relative to the day before 0001-01-01.
#' @noRd
.julian_date2offset <- function(x, leapyear) {
  year1 <- x$year - 1L
  corr <- ifelse(x$month <= 2L, 0L, as.integer(leapyear) - 2L)
  365L * year1 + year1 %/% 4L + (367L * x$month - 362L) %/% 12L + corr + x$day - 2L
}

#' Offsets to dates, from function `julian-from-fixed()` and support functions.
#'
#' @param x Integer vector of offsets. The offsets must be relative to the day
#' before 0001-01-01.
#' @return `data.frame` with date elements "year", "month" and "day".
#' @noRd
.julian_offset2date <- function(x) {
  yr <- (4 * (x + 1L) + 1464L) %/% 1461L
  leapyear <- yr %% 4L == 0L
  yr1 <- yr - 1L
  jan1 <- -2L + 365L * yr1 + yr1 %/% 4L + 1L
  prior_days <- x - jan1 + ifelse(x < jan1 + 59L + as.integer(leapyear), 0L, 2L - as.integer(leapyear))
  mon <- (12L * prior_days + 373L) %/% 367L
  day <- x - .julian_date2offset(data.frame(year = yr, month = mon, day = 1), leapyear) + 1L
  data.frame(year = yr, month = mon, day = day)
}

#' CFtime: working with CF Metadata Conventions "time" dimensions
#'
#' Support for all calendars as specified in the Climate and Forecast
#' (CF) Metadata Conventions for climate and forecasting data. The CF Metadata
#' Conventions is widely used for distributing files with climate observations
#' or projections, including the Coupled Model Intercomparison Project (CMIP)
#' data used by climate change scientists and the Intergovernmental Panel on
#' Climate Change (IPCC). This package specifically allows the user to work
#' with any of the CF-compliant calendars (many of which are not compliant with
#' POSIXt). The CF time coordinate is formally defined in the
#' [CF Metadata Conventions document](https://cf-convention.github.io/Data/cf-conventions/cf-conventions-1.13/cf-conventions.html#time-coordinate).
#'
#' This package also supports the creation of a "time" dimension, using class
#' `CFClimatology`, for climatological statistics as defined [here](https://cf-convention.github.io/Data/cf-conventions/cf-conventions-1.13/cf-conventions.html#climatological-statistics).
#'
#' The package can create a [CFTime] or [CFClimatology] instance from scratch or, more commonly, it
#' can use the dimension attributes and dimension variable values from a netCDF
#' resource. The package does not actually do any of the reading and the user is
#' free to use their netCDF package of preference. The recommended package to
#' use (with any netCDF resources) is [ncdfCF](https://cran.r-project.org/package=ncdfCF).
#' `ncdfCF` will automatically use this package to manage the "time" dimension
#' of any netCDF resource. As with this package, it reads and interprets the
#' attributes of the resource to apply the CF Metadata Conventions, supporting
#' axes, auxiliary coordinate variables, coordinate reference systems, etc.
#' Alternatively, for more basic netCDF reading and writing, the two main options
#' are [RNetCDF](https://cran.r-project.org/package=RNetCDF) and
#' [ncdf4](https://cran.r-project.org/package=ncdf4)).
#'
#' **Create, modify, inquire**
#' * [CFtime()]: Create a [CFTime] instance
#' * [`Properties`][properties] of the `CFTime` instance
#' * [parse_timestamps()]: Parse a vector of character timestamps into `CFTime` elements
#' * [`Compare`][CFtime-equivalent] two `CFTime` instances
#' * [`Merge`][CFtime-merge] two `CFTime` or `CFClimatology` instances, or append additional time steps to a `CFTime` instance
#' * [as_timestamp()] and [format()]: Generate a vector of character or `POSIXct` timestamps from a `CFTime` instance
#' * [range()]: Timestamps of the two endpoints in the time series
#' * `copy()`: Create a copy of a CFTime instance
#' * `subset()`: Create a new CFTime instance which is a subset by index positions of another CFTime instance
#' * [is_complete()]: Does the `CFTime` instance have a complete time series between endpoints?
#' * [month_days()]: How many days are there in a month using the calendar of the `CFTime` instance?
#' * `CFTime$doy()`: Retrieve the day-of-year of offsets in the `CFTime` instance or from supplied dates.
#'
#' **Factors and coverage**
#' * [CFfactor()] and [cut()]: Create factors for different time periods
#' * [CFfactor_units()]: How many units of time are there in each factor level?
#' * [CFfactor_coverage()]: How much data is available for each level of the factor?
#'
#' **Filtering and selection**
#' * [slice()]: Logical vector of time steps between two extreme points.
#' * [indexOf()]: Index values in the time series of given timestamps, possibly with
#' fractional part for interpolation.
#' @keywords internal
#' @aliases CFtime-package
"_PACKAGE"

## usethis namespace: start
## usethis namespace: end
NULL

#' @title Proleptic Gregorian CF calendar
#'
#' @description This class represents a standard CF calendar, but with the
#'   Gregorian calendar extended backwards to before the introduction of the
#'   Gregorian calendar. This calendar is compatible with the standard POSIXt
#'   calendar, but note that daylight savings time is not considered.
#'
#'   This calendar includes dates 1582-10-14 to 1582-10-05 (the gap between the
#'   Gregorian and Julian calendars, which is observed by the standard
#'   calendar), and extends to years before the year 1, including year 0.
#'
#' @aliases CFCalendarProleptic
#' @docType class
CFCalendarProleptic <- R6::R6Class("CFCalendarProleptic",
  inherit = CFCalendar,
  private = list(
    # Rata Die, the number of days from the day before 0001-01-01 to
    # origin of this calendar. Used to convert offsets from the calendar origin
    # to the day before 0001-01-01 for arithmetic calculations.
    rd = 0L
  ),
  public = list(
    #' @description Create a new CF calendar.
    #' @param nm The name of the calendar. This must be "proleptic_gregorian".
    #'   This argument is superfluous but maintained to be consistent with the
    #'   initialization methods of the parent and sibling classes.
    #' @param definition The string that defines the units and the origin, as
    #' per the CF Metadata Conventions.
    #' @return A new instance of this class.
    initialize = function(nm, definition) {
      super$initialize(nm, definition)
      private$rd <- .gregorian_date2offset(self$origin, self$leap_year(self$origin$year))
    },

    #' @description Indicate which of the supplied dates are valid.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return Logical vector with the same length as argument `ymd` has rows
    #'   with `TRUE` for valid days and `FALSE` for invalid days, or `NA` where
    #'   the row in argument `ymd` has `NA` values.
    valid_days = function(ymd) {
      !is.na(ymd$year) & ymd$month >= 1L & ymd$month <= 12L & ymd$day >= 1L &
      ifelse(self$leap_year(ymd$year),
        ymd$day <= c(31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)[ymd$month],
        ymd$day <= c(31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)[ymd$month])
    },

    #' @description Determine the number of days in the month of the calendar.
    #' @param ymd `data.frame`, optional, with dates parsed into their parts.
    #' @return Integer vector indicating the number of days in each month for
    #'   the dates supplied as argument `ymd`. If no dates are supplied, the
    #'   number of days per month for the calendar as a vector of length 12, for
    #'   a regular year without a leap day.
    month_days = function(ymd = NULL) {
      if (is.null(ymd)) return(c(31L, 28L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L))

      ifelse(self$leap_year(ymd$year),
             c(31L, 29L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[ymd$month],
             c(31L, 28L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[ymd$month])
    },

    #' @description Indicate which years are leap years.
    #' @param yr Integer vector of years to test.
    #' @return Logical vector with the same length as argument `yr`. `NA` is
    #'   returned where elements in argument `yr` are `NA`.
    leap_year = function(yr) {
      ((yr %% 4L == 0L) & (yr %% 100L > 0L)) | (yr %% 400L == 0L)
    },

    #' @description Indicate if the time series described using this calendar
    #'   can be safely converted to a standard date-time type (`POSIXct`,
    #'   `POSIXlt`, `Date`).
    #' @param offsets The offsets from the CFtime instance.
    #' @return `TRUE`.
    POSIX_compatible = function(offsets) {
      TRUE # nocov
    },

    #' @description Calculate difference in days between a `data.frame` of time
    #'   parts and the origin.
    #'
    #' @param x `data.frame`. Dates to calculate the difference for.
    #' @return Integer vector of a length equal to the number of rows in
    #' argument `x` indicating the number of days between `x` and the `origin`,
    #' or `NA` for rows in `x` with `NA` values.
    date2offset = function(x) {
      .gregorian_date2offset(x, self$leap_year(x$year)) - private$rd
    },

    #' @description Calculate date parts from day differences from the origin. This
    #'   only deals with days as these are impacted by the calendar.
    #'   Hour-minute-second timestamp parts are handled in [CFCalendar].
    #'
    #' @param x Integer vector of days to add to the origin.
    #' @return A `data.frame` with columns 'year', 'month' and 'day' and as many
    #'   rows as the length of vector `x`.
    offset2date = function(x) {
      .gregorian_offset2date(x + private$rd)
    }
  )
)

# The below functions use arithmetic offset calculation from date parts and
# vice-versa. These functions are R-ified from pseudo-functions in Reingold &
# Derschowitz, "Calendrical Calculations", 2018.

#' Dates to offset, from function `fixed-from-gregorian()`
#'
#' @param x `data.frame` with columns "year", "month" and "date"
#' @param leapyear Logical vector of the same length as `x` has rows indicating
#' for each row in `x` if this is a leap year.
#' @return Integer vector of offsets for the dates in `x`. The offsets are
#' relative to the day before 0001-01-01.
#' @noRd
.gregorian_date2offset <- function(x, leapyear) {
  year1 <- x$year - 1L
  corr <- ifelse(x$month <= 2L, 0L, as.integer(leapyear) - 2L)
  365L * year1 + year1 %/% 4L - year1 %/% 100L + year1 %/% 400L +
    (367L * x$month - 362L) %/% 12L + corr + x$day
}

#' Offsets to dates, from function `gregorian-from-fixed()` and support functions.
#'
#' @param x Integer vector of offsets. The offsets must be relative to the day
#' before 0001-01-01.
#' @return `data.frame` with date elements "year", "month" and "day".
#' @noRd
.gregorian_offset2date <- function(x) {
  d0 <- x - 1L
  n400 <- d0 %/% 146097L; d1 <- d0 %% 146097L
  n100 <- d1 %/% 36524L;  d2 <- d1 %% 36524L
  n4 <-   d2 %/% 1461L;   d3 <- d2 %% 1461L
  n1 <-   d3 %/% 365L
  yr <- 400L * n400 + 100L * n100 + 4L * n4 + n1
  yr <- ifelse(n100 == 4L | n1 == 4L, yr, yr + 1L)
  leapyear <- ((yr %% 4L == 0L) & (yr %% 100L > 0L)) | (yr %% 400L == 0L)
  yr1 <- yr - 1L
  jan1 <- 365L * yr1 + yr1 %/% 4L - yr1 %/% 100L + yr1 %/% 400L + 1L
  prior_days <- x - jan1 + ifelse(x < jan1 + 59L + as.integer(leapyear), 0L, 2L - as.integer(leapyear))
  mon <- (12L * prior_days + 373L) %/% 367L
  day <- x - .gregorian_date2offset(data.frame(year = yr, month = mon, day = 1), leapyear) + 1L
  data.frame(year = yr, month = mon, day = day)
}

#' @title Standard CF calendar
#'
#' @description This class represents a standard calendar of 365 or 366 days per
#'   year. This calendar is compatible with the standard POSIXt calendar for
#'   periods after the introduction of the Gregorian calendar, 1582-10-15
#'   00:00:00. The calendar starts at 0001-01-01 00:00:00, e.g. the start of the
#'   Common Era.
#'
#'   Note that this calendar, despite its name, is not the same as that used in
#'   ISO8601 or many computer systems for periods prior to the introduction of
#'   the Gregorian calendar. Use of the "proleptic_gregorian" calendar is
#'   recommended for periods before or straddling the introduction date, as that
#'   calendar is compatible with POSIXt on most OSes.
#'
#' @aliases CFCalendarStandard
#' @docType class
CFCalendarStandard <- R6::R6Class("CFCalendarStandard",
  inherit = CFCalendar,
  private = list(
    # Rata Die, the number of days from the day before 0001-01-01 to
    # origin of the Gregorian part of this calendar, if present. Used to
    # convert offsets from the Gregorian calendar origin to the day before
    # 0001-01-01 for arithmetic calculations.
    rd_greg = 0L,

    # Rata Die, the number of days from the day before 0001-01-01 to
    # origin of the Julian part of this calendar, if present. Used to convert
    # offsets from the Julian calendar origin to the day before 0001-01-01
    #for arithmetic calculations.
    rd_juli = 0L,

    # The integer offset for 1582-10-15 00:00:00, when the Gregorian
    # calendar started, or 1582-10-05, when the gap between Julian and
    # Gregorian calendars started. The former is set when the calendar origin
    # is more recent, the latter when the origin is prior to the gap.
    gap = -1L
  ),
  public = list(
    #' @description Create a new CF calendar. When called with the deprecated
    #' 'gregorian' it is automatically converted to the equivalent 'standard'.
    #' @param nm The name of the calendar. This must be "standard" or
    #'   "gregorian" (deprecated).
    #' @param definition The string that defines the units and the origin, as
    #' per the CF Metadata Conventions.
    #' @return A new instance of this class.
    initialize = function(nm, definition) {
      nm <- "standard"
      super$initialize(nm, definition)
      private$rd_greg <- .gregorian_date2offset(self$origin, self$leap_year(self$origin$year))
      private$rd_juli <- .julian_date2offset(self$origin, self$leap_year(self$origin$year))

      private$gap <- if (self$is_gregorian_date(self$origin))
        .gregorian_date2offset(data.frame(year = 1582, month = 10, day = 15), self$leap_year(self$origin$year)) - private$rd_greg
      else
        .julian_date2offset(data.frame(year = 1582, month = 10, day = 5), self$leap_year(self$origin$year)) - private$rd_juli
    },

    #' @description Indicate which of the supplied dates are valid.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return Logical vector with the same length as argument `ymd` has rows
    #'   with `TRUE` for valid days and `FALSE` for invalid days, or `NA` where
    #'   the row in argument `ymd` has `NA` values.
    valid_days = function(ymd) {
      ymd$year >= 1L & ymd$month >= 1L & ymd$month <= 12L & ymd$day >= 1L &
      ifelse(self$is_gregorian_date(ymd),
        # Gregorian calendar
        ifelse(self$leap_year(ymd$year),
               ymd$day <= c(31L, 29L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[ymd$month],
               ymd$day <= c(31L, 28L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[ymd$month]),
        # Julian calendar
        ifelse(ymd$year == 1582L & ymd$month == 10L & ymd$day > 4L,
               FALSE, # days 1582-10-05 - 1582-10-14 do not exist
               ifelse(ymd$year %% 4L == 0L,
                      ymd$day <= c(31L, 29L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[ymd$month],
                      ymd$day <= c(31L, 28L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[ymd$month])
        )
      )
    },

    #' @description Indicate which of the supplied dates are in the Gregorian
    #'   part of the calendar, e.g. 1582-10-15 or after.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return Logical vector with the same length as argument `ymd` has rows
    #'   with `TRUE` for days in the Gregorian part of the calendar and `FALSE`
    #'   otherwise, or `NA` where the row in argument `ymd` has `NA` values.
    is_gregorian_date = function(ymd) {
      ymd$year > 1582L | (ymd$year == 1582L & (ymd$month > 10L | (ymd$month == 10L & ymd$day >= 15L)))
    },

    #' @description Indicate if the time series described using this calendar
    #'   can be safely converted to a standard date-time type (`POSIXct`,
    #'   `POSIXlt`, `Date`). This is only the case if all offsets are for
    #'   timestamps fall on or after the start of the Gregorian calendar,
    #'   1582-10-15 00:00:00.
    #' @param offsets The offsets from the CFtime instance.
    #' @return `TRUE`.
    POSIX_compatible = function(offsets) {
      all(offsets >= private$gap)
    },

    #' @description Determine the number of days in the month of the calendar.
    #' @param ymd `data.frame`, optional, with dates parsed into their parts.
    #' @return A vector indicating the number of days in each month for the
    #'   dates supplied as argument `ymd`. If no dates are supplied, the number
    #'   of days per month for the calendar as a vector of length 12, for a
    #'   regular year without a leap day.
    month_days = function(ymd = NULL) {
      if (is.null(ymd)) return(c(31L, 28L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L))

      ifelse(self$leap_year(ymd$year),
        c(31L, 29L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[ymd$month],
        c(31L, 28L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[ymd$month])
    },

    #' @description Indicate which years are leap years.
    #' @param yr Integer vector of years to test.
    #' @return Logical vector with the same length as argument `yr`. `NA` is
    #'   returned where elements in argument `yr` are `NA`.
    leap_year = function(yr) {
      ifelse(yr <= 1582L,
             yr %% 4L == 0L,
             ((yr %% 4L == 0L) & (yr %% 100L > 0L)) | (yr %% 400L == 0L)
      )
    },

    #' @description Calculate difference in days between a `data.frame` of time
    #'   parts and the origin.
    #'
    #' @param x `data.frame`. Dates to calculate the difference for.
    #' @return Integer vector of a length equal to the number of rows in
    #'   argument `x` indicating the number of days between `x` and the origin
    #'   of the calendar, or `NA` for rows in `x` with `NA` values.
    date2offset = function(x) {
      leap <- self$leap_year(x$year)
      ifelse(self$is_gregorian_date(x),
             .gregorian_date2offset(x, leap),
             .julian_date2offset(x, leap)
      ) - if (private$gap > 0L) private$rd_juli else private$rd_greg
    },

    #' @description Calculate date parts from day differences from the origin. This
    #'   only deals with days as these are impacted by the calendar.
    #'   Hour-minute-second timestamp parts are handled in [CFCalendar].
    #'
    #' @param x Integer vector of days to add to the origin.
    #' @return A `data.frame` with columns 'year', 'month' and 'day' and as many
    #'   rows as the length of vector `x`.
    offset2date = function(x) {
      rd <- if (private$gap > 0L) private$rd_juli else private$rd_greg
      len <- length(x)
      gndx <- x >= private$gap & !is.na(x)
      if (any(gndx)) greg <- .gregorian_offset2date(x[gndx] + rd)
      else greg <- data.frame(year = integer(), month = integer(), day = integer())
      jndx <- x < private$gap & !is.na(x)
      if (any(jndx)) juli <- .julian_offset2date(x[jndx] + rd)
      else juli <- data.frame(year = integer(), month = integer(), day = integer())
      yr <- mon <- day <- rep(NA_integer_, len)
      yr[gndx]  <- greg$year;  yr[jndx]  <- juli$year
      mon[gndx] <- greg$month; mon[jndx] <- juli$month
      day[gndx] <- greg$day;   day[jndx] <- juli$day
      data.frame(year = yr, month = mon, day = day)
    }
  )
)

#' @name deprecated_functions
#' @title Deprecated functions
#'
#' @description
#' These functions are deprecated and should no longer be used in new code. The
#' below table gives the replacement function to use instead. The function
#' arguments of the replacement function are the same as those of the deprecated
#' function if no arguments are given in the table.
#'
#' | **Deprecated function** | **Replacement function** |
#' | ------------------- | -------------------- |
#' | CFcomplete()        | [is_complete()]      |
#' | CFmonth_days()      | [month_days()]       |
#' | CFparse()           | [parse_timestamps()] |
#' | CFrange()           | [range()]            |
#' | CFsubset()          | [slab()]             |
#' | CFtimestamp()       | [as_timestamp()]     |
#'
#' @param t,x,format,asPOSIX,extremes See replacement functions.
#'
#' @returns See replacement functions.

# nocov start

#' @rdname deprecated_functions
#' @export
CFtimestamp <- function(t, format = NULL, asPOSIX = FALSE) {
  warning("Function `CFtimestamp()` is deprecated. Use `as_timestamp()` instead.")
  as_timestamp(t, format, asPOSIX)
}

#' @rdname deprecated_functions
#' @export
CFmonth_days <- function(t, x = NULL) {
  warning("Function `CFmonth_days()` is deprecated. Use `month_days()` instead.")
  month_days(t, x)
}

#' @rdname deprecated_functions
#' @export
CFcomplete <- function(x) {
  warning("Function `CFcomplete()` is deprecated. Use `is_complete()` instead.")
  is_complete(x)
}

#' @rdname deprecated_functions
#' @export
CFsubset <- function(x, extremes) {
  warning("Function `CFsubset()` is deprecated. Use `slab()` instead.")
  slab(x, extremes)
}

#' @rdname deprecated_functions
#' @export
CFparse <- function(t, x) {
  warning("Function `CFparse()` is deprecated. Use `parse_timestamps()` instead.")
  parse_timestamps(t, x)
}

# nocov end

#' @title Coordinated Universal Time CF calendar
#'
#' @description This class represents a calendar based on the Coordinated
#'   Universal Time. Validity is from 1972 onwards, with dates represented using
#'   the Gregorian calendar, up to the present (so future timestamps are not
#'   allowed). Leap seconds are considered in all calculations. Also, time zone
#'   information is irrelevant and may not be given.
#'
#'   In general, the calendar should use a unit of time of a second. Minute,
#'   hour and day are allowed but discouraged. Month and year as time unit are
#'   not allowed as there is no practical way to maintain leap second accuracy.
#'
#' @aliases CFCalendarUTC
#' @docType class
CFCalendarUTC <- R6::R6Class("CFCalendarUTC",
  inherit = CFCalendarProleptic,
  private = list(
    # Leap days in various incarnations. More will be added in initialize().
    leapdays = data.frame(
      year      = c(1972L, 1972L, 1973L, 1974L, 1975L, 1976L, 1977L, 1978L, 1979L,
                    1981L, 1982L, 1983L, 1985L, 1987L, 1989L, 1990L, 1992L, 1993L,
                    1994L, 1995L, 1997L, 1998L, 2005L, 2008L, 2012L, 2015L, 2016L),
      month     = c(6L, 12L, 12L, 12L, 12L, 12L, 12L, 12L, 12L, 6L, 6L, 6L, 6L,
                    12L, 12L, 12L, 6L, 6L, 6L, 12L, 6L, 12L, 12L, 12L,  6L,  6L, 12L),
      day       = c(30L, 31L, 31L, 31L, 31L, 31L, 31L, 31L, 31L, 30L, 30L, 30L, 30L,
                    31L, 31L, 31L, 30L, 30L, 30L, 31L, 30L, 31L, 31L, 31L, 30L, 30L, 31L),
      compound  = c(19720630L, 19721231L, 19731231L, 19741231L, 19751231L, 19761231L,
                    19771231L, 19781231L, 19791231L, 19810630L, 19820630L, 19830630L,
                    19850630L, 19871231L, 19891231L, 19901231L, 19920630L, 19930630L,
                    19940630L, 19951231L, 19970630L, 19981231L, 20051231L, 20081231L,
                    20120630L, 20150630L, 20161231L),
      epochdays = c(182L, 366L, 731L, 1096L, 1461L, 1827L, 2192L, 2557L, 2922L,
                    3469L, 3834L, 4199L, 4930L, 5844L, 6575L, 6940L, 7487L, 7852L,
                    8217L, 8766L, 9313L, 9862L, 12419L, 13515L, 14792L, 15887L, 16437L)
    ),

    # Offset of the epoch from the origin, in calendar units. Set in initialize().
    epoch = 0,

    # Rata Die of the epoch.
    epoch_rd = 719893L,

    # Number of leap seconds applied to the origin of self. Set in initialize().
    origin_leapsecs = 0L,

    # Check if rows in a data.frame with seconds >= 60 are days when a leap
    # second was applied. Argument `cap` comes from method parse(). If not,
    # values in that row are set to `NA`.
    check_leap_seconds = function(cap) {
      # There will always be at least one row with 60 seconds or this method
      # would not be called
      leap_ndx <- which(cap$second >= 60)
      tst <- cap[leap_ndx, ]
      not2359 <- tst$hour != 23L | tst$minute != 59L
      tst <- tst$year * 10000L + tst$month * 100L + tst$day
      bad <- !(tst %in% private$leapdays$compound)
      cap[leap_ndx[bad | not2359], ] <- rep(NA, 7)
      cap
    }
  ),
  public = list(
    #' @description Create a new CF UTC calendar.
    #' @param nm The name of the calendar. This must be "utc".
    #' @param definition The string that defines the units and the origin, as
    #' per the CF Metadata Conventions.
    initialize = function(nm, definition) {
      super$initialize(nm, definition)
      if (self$unit > 4L)
        stop("Unit for an UTC calendar cannot be 'month' or 'year'.", call. = FALSE) # nocov

      # How many leap seconds have been applied to the origin?
      private$origin_leapsecs <- findInterval(self$origin$year * 10000L +
                                              self$origin$month * 100L +
                                              self$origin$day, private$leapdays$compound, left.open = TRUE)

      # Offset of the epoch from the origin
      private$epoch <- self$parse("1972-01-01")$offset

      # Add offsets in self calendar units of the leap seconds
      leapdates <- as.Date(paste(private$leapdays$year, private$leapdays$month, private$leapdays$day, sep = "-"))
      private$leapdays$seconds <-
        as.integer(c(unclass(difftime(leapdates, as.Date("1972-01-01"), units = "secs"))) +
                   CFt$units$per_day[1L] + (1:27))
    },

    #' @description Indicate which of the supplied dates are valid.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return Logical vector with the same length as argument `ymd` has rows
    #'   with `TRUE` for valid days and `FALSE` for invalid days, or `NA` where
    #'   the row in argument `ymd` has `NA` values.
    valid_days = function(ymd) {
      nw <- as.POSIXct(Sys.Date(), tz = "UTC")
      yr <- as.integer(substr(nw, 1, 4))
      mon <- as.integer(substr(nw, 6, 7))
      dt <- as.integer(substr(nw, 9, 10))
      super$valid_days(ymd) & ymd$year >= 1972L &
        (ymd$year < yr | (ymd$year == yr & (ymd$month < mon |
          (ymd$month == mon & ymd$day <= dt))))
    },

    #' @description Parsing a vector of date-time character strings into parts.
    #' This includes any leap seconds. Time zone indications are not allowed.
    #' @param d character. A character vector of date-times.
    #' @return A `data.frame` with columns year, month, day, hour, minute,
    #'   second, time zone, and offset. Invalid input data will appear as `NA`.
    #'   Note that the time zone is always "+0000" and is included to maintain
    #'   compatibility with results from other calendars.
    parse = function(d) {
      # Parsers
      # These parsers are specific to UTC, i.e. starting in 1972 and up to the
      # current year, and a possible leap second as value 60.

      # UDUNITS broken timestamp definition, with some changes
      # broken_timestamp	{broken_date}({space|T}+{broken_clock})? -- T not in definition but present in lexer code
      # broken_date		    {year}-{month}(-{day})?
      # year		        	[+-]?[0-9]{1,4}
      # month			        0?[1-9]|1[0-2]
      # day			          0?[1-9]|[1-2][0-9]|30|31
      # broken_clock		  {hour}:{minute}(:{second})?
      # hour			        [0-1]?[0-9]|2[0-3]      -- sign on hour not allowed
      # minute			      [0-5]?[0-9]
      # second		      	60|{minute}?
      # fractional part   (\.[0-9]*)?
      broken <- paste0(
        "^",                                    # anchor string at start
        "([+-]?[0-9]{1,4})",                    # year, with optional sign
        "-(0?[1-9]|1[012])",                    # month
        "(?:-(0?[1-9]|[12][0-9]|3[01]))?",      # day, optional
        "(?:[T ]",                              # if a time is following, separate with a single whitespace character or a "T"
        "([01]?[0-9]|2[0-3])",                  # hour
        ":([0-5]?[0-9])",                       # minute
        "(?::([0-6]?[0-9]))?",                  # second, optional
        "(?:\\.([0-9]*))?",                     # optional fractional part of the smallest specified unit
        ")?",                                   # close optional time capture group
        "(?:\\s",                               # if a time zone offset is following, separate with a single whitespace character
        "([+-])?([01]?[0-9]|2[0-3])",           # tz hour, with optional sign
        "(?::(00|15|30|45))?",                  # optional tz minute, only 4 possible values
        ")?",                                   # close optional timezone capture group
        "$"                                     # anchor string at end
      )

      iso8601 <- paste0(
        "^",
        "([0-9]{4})",
        "-(0[1-9]|1[012])",
        "-(0[1-9]|[12][0-9]|3[01])?",
        "(?:",
        "[T ]([01][0-9]|2[0-3])",
        "(?::([0-5][0-9]))?",
        "(?::([0-6][0-9]))?",
        "(?:[\\.,]([0-9]*))?",
        ")?",
        "(?:([Z+-])([01][0-9]|2[0-3])?(?::(00|15|30|45))?", ## FIXME: Z?, smaller number of captures
        ")?$"
      )

      # UDUNITS packed timestamp definition - NOT YET USED
      # packed_timestamp	{packed_date}({space|T}+{packed_clock})?  -- T and space only allowed in packed time follows
      # packed_date	    	{year}({month}{day}?)?           -- must be YYYYMMDD or else format is ambiguous, as per lexer code
      # packed_clock		  {hour}({minute}{second}?)?       -- must be HHMMSS to be unambiguous
      # packed <- stringi::stri_join(
      #   "^",                                    # anchor string at start
      #   "([+-]?[0-9]{4})",                      # year, with optional sign
      #   "(0[1-9]|1[012])?",                     # month, optional
      #   "(0[1-9]|[12][0-9]|3[01])?",            # day, optional
      #   "(?:[T,\\s]",                           # if a time is following, separate with a single whitespace character or a "T"
      #     "([01][0-9]|2[0-3])?",                # hour
      #     "([0-5][0-9])?",                      # minute, optional
      #     "([0-5]?[0-9](?:\\.[0-9]*)?)?",       # second, optional, with optional fractional part
      #   ")?",                                   # close optional time capture group
      #   "$"                                     # anchor string at end
      # )

      parse <- data.frame(year = integer(), month = integer(), day = integer(),
                          hour = integer(), minute = integer(), second = numeric(), frac = character(),
                          tz_sign = character(), tz_hour = character(), tz_min = character())

      cap <- utils::strcapture(iso8601, d, parse)
      missing <- which(is.na(cap$year))
      if (length(missing) > 0L)
        cap[missing,] <- utils::strcapture(broken, d[missing], parse)

      # Assign any fraction to the appropriate time part
      cap$frac[is.na(cap$frac)] <- "0"
      frac <- as.numeric(paste0("0.", cap$frac))
      if (sum(frac) > 0) {
        ndx <- which(!(is.na(cap$second)) & frac > 0)
        if (length(ndx) > 0L) cap$second[ndx] <- cap$second[ndx] + frac[ndx]
        ndx <- which(!(is.na(cap$minute)) & is.na(cap$second) & frac > 0)
        if (length(ndx) > 0L) cap$second[ndx] <- 60L * frac[ndx]
        ndx <- which(!(is.na(cap$hour)) & is.na(cap$minute) & frac > 0)
        if (length(ndx) > 0L) {
          secs <- 3600 * frac
          cap$minute[ndx] <- secs[ndx] %/% 60
          cap$second[ndx] <- secs[ndx] %% 60
        }
      }
      cap$frac <- NULL

      # Convert NA time parts to 0 - in CF default time is 00:00:00 when not specified
      cap$hour[is.na(cap$hour)] <- 0L
      cap$minute[is.na(cap$minute)] <- 0L
      cap$second[is.na(cap$second)] <- 0L

      # Set timezone to 00:00 to align cap data.frame with other calendars
      cap$tz_sign <- cap$tz_hour <- cap$tz_min <- NULL
      cap$tz <- "+0000"

      # Set optional date parts to 1 if not specified
      cap$month[is.na(cap$month)] <- 1L
      cap$day[is.na(cap$day)] <- 1L

      # Check date validity
      invalid <- !self$valid_days(cap)
      invalid[is.na(invalid)] <- TRUE
      if (sum(invalid) > 0L) cap[invalid,] <- rep(NA, 7)

      # Check that any supplied leap seconds coincide with official leap seconds
      if (any(cap$second >= 60, na.rm = TRUE))
        cap <- private$check_leap_seconds(cap)

      # Calculate offsets
      if (nrow(self$origin) == 0L) {        # if there's no origin yet, don't calculate offsets
        cap$offset <- rep(0, nrow(cap))     # this happens, f.i., when a CFCalendar is created
      } else {
        days <- self$date2offset(cap)
        chkdays <- cap$year * 10000L + cap$month * 100L + cap$day
        leapsecs <- findInterval(chkdays, private$leapdays$compound, left.open = TRUE) - private$origin_leapsecs
        cap$offset <- round((days * 86400 + (cap$hour - self$origin$hour[1]) * 3600 +
                             (cap$minute - self$origin$minute[1]) * 60 +
                             cap$second - self$origin$second + leapsecs) /
                            CFt$units$seconds[self$unit], 9)
      }
      cap
    },

    #' @description Decompose a vector of offsets, in units of the calendar, to
    #'   their timestamp values. This adds a specified amount of time to the
    #'   origin of a `CFTime` object.
    #' @param offsets Vector of numeric offsets to add to the origin of the
    #'   calendar.
    #' @return A `data.frame` with columns for the timestamp elements and as
    #'   many rows as there are offsets.
    offsets2time = function(offsets = NULL) {
      if(is.null(offsets) || (len <- length(offsets)) == 0L)
        return(data.frame(year = integer(), month = integer(), day = integer(),
                          hour = integer(), minute = integer(), second = numeric(),
                          tz = character(), offset = numeric()))

      # Base offsets on epoch, in seconds
      off <- (offsets - private$epoch) * CFt$units$seconds[self$unit]
      ndx <- findInterval(off, private$leapdays$seconds)
      remainder <- off - ifelse(ndx > 0L, private$leapdays$seconds[ndx], 0L)

      secs <- mins <- hrs <- days <- vector("numeric", len)
      frac <- rep(0, len)
      leap <- ndx < 27L & round(private$leapdays$seconds[ndx + 1L] - off, 5) <= 1L
      if (sum(leap) > 0L) {
        secs[leap] <- 60
        mins[leap] <- 59
        hrs[leap] <- 23
        days[leap] <- ifelse(ndx[leap] == 0L,
                             private$leapdays$epochdays[1L],
                             private$leapdays$epochdays[ndx + 1L] - private$leapdays$epochdays[ndx]) - 1L
      }
      not_leap <- !leap
      if (sum(not_leap) > 0L) {
        # Convert remainder to time parts, no leap seconds anymore
        days[not_leap] <- remainder[not_leap] %/% 86400L            # overflow days
        secs[not_leap] <- round(remainder[not_leap] %% 86400L, 3L)  # round down to milli-seconds to avoid errors
        #frac[not_leap] <- secs[not_leap] %% 1
        #secs[not_leap] <- secs[not_leap] %/% 1

        # Time elements for output
        hrs[not_leap] <- secs[not_leap] %/% 3600L
        mins[not_leap] <- (secs[not_leap] %% 3600L) %/% 60L
        secs[not_leap] <- secs[not_leap] %% 60L
      }

      # Now add days using the calendar
      epoch_days <- ifelse(ndx > 0L, private$leapdays$epochdays[ndx], 0L)
      out <- if (any(days != 0L))
        .gregorian_offset2date(as.integer(days) + private$epoch_rd + epoch_days)
      else
        data.frame(year = rep(self$origin$year, len),
                   month = rep(self$origin$month, len),
                   day = rep(self$origin$day, len))

      # Put it all back together again
      out$hour <- as.integer(hrs)
      out$minute <- as.integer(mins)
      out$second <- secs
      #if (sum(frac) > 0)
      #  out$second <- out$second + frac
      out$tz <- rep("+0000", len)

      out$offset <- offsets
      out
    }
  )
)

#' @title CFClimatology class
#'
#' @description This class represents a climatological time coordinate, both its
#' "calendar" and its "climatology" bounds values.
#'
#' The "calendar" portion is managed by the [CFTime] class, from which this
#' class is inherited. This class implements the specific behaviour of the
#' climatological bounds values and includes methods to query the structure of
#' the climatological data.
#'
#' @export
#' @references
#'   https://cf-convention.github.io/Data/cf-conventions/cf-conventions-1.13/cf-conventions.html#climatological-statistics
#' @docType class
CFClimatology <- R6::R6Class("CFClimatology",
  inherit = CFTime,
  private = list(
    # The base period over which the climatology is calculated. This must be
    # one of "year", "season", "quarter", "month", "dekad" or "day". Set in
    # `initialize()`.
    .period = "",

    # The years over which the climatology is calculated. Set in `initialize()`.
    .years = NA
  ),
  public = list(
    #' @description Create a new instance of this class.
    #' @param definition Character string of the units and origin of the
    #'   calendar.
    #' @param calendar Character string of the calendar to use. Must be one of
    #'   the values permitted by the CF Metadata Conventions. If `NULL`, the
    #'   "standard" calendar will be used.
    #' @param offsets Numeric or character vector. When numeric, a vector of
    #'   offsets from the origin in the time series. When a character vector of
    #'   length 2 or more, timestamps in ISO8601 or UDUNITS format.
    #' @param bounds The climatological bounds for the offsets. A `matrix` with
    #'   columns for `offsets` and low values in the first row, high values in
    #'   the second row. The bounds will oftentimes overlap or be discontinuous.
    initialize = function(definition, calendar = "standard", offsets, bounds) {
      if (!is.matrix(bounds) || !all(dim(bounds) == c(2, length(offsets))))
        stop("Argument 'bounds' must be a matrix with two rows and as many columns as the length of argument 'offset'", call. = FALSE)

      super$initialize(definition, calendar, offsets)
      self$set_bounds(bounds)

      # Determine period and years from the bounds of the first offset
      time <- private$.cal$offsets2time(bounds[1L:2L, 1L])
      private$.years <- time$year
      if (all(time$day == 1L)) {
        if (all(time$month == 1L)) {
          private$.period <- "year"
          private$.years <- c(time$year[1L], time$year[2L] - 1L)
        } else if (all(time$month %in% c(3L, 6L, 9L, 12L))) {
          private$.period <- "season"
          private$.years <- c(time$year[1L] + 1L, time$year[2L])
        } else if (all(time$month %in% c(1L, 4L, 7L, 10L)))
          private$.period <- "quarter"
        else
          private$.period <- "month"
      } else if (all(time$day %in% c(1L, 11L, 21L)))
        private$.period <- "dekad"
      else
        private$.period <- "day"
    },

    #' @description Print a summary of the `CFClimatology` object to the console.
    #' @param ... Ignored.
    #' @return `self` invisibly.
    print = function(...) {
      noff <- length(private$.offsets)
      d <- self$range()
      el <- if (noff > 1L) {
        sprintf("  Elements: [%s .. %s] (average of %f %s between %d elements)\n",
                d[1L], d[2L], private$.resolution, CFt$units$name[private$.cal$unit], noff)
      } else
        paste0("  Elements: ", d[1L], "\n")

      p <- paste0("  Period  : ", private$.period, "\n")
      if (private$.years[1L] == private$.years[2L]) y <- paste0("  Year    : ", private$.years[1L], "\n")
      else y <- paste0("  Years   : ", private$.years[1L], " - ", private$.years[2L], " (inclusive)\n")

      cal <- capture.output(private$.cal$print())
      cat(paste(cal, collapse = "\n"), "\nClimatological time series:\n",  el, p, y, sep = "")
      invisible(self)
    }
  ),
  active = list(
    #' @field period (read-only) Character string indicating the period during
    #'   the year over which the climatology was calculated.
    period = function(value) {
      if (missing(value))
        private$.period
    },

    #' @field years (read-only) Vector of two integer values indicating the
    #'   years over which the climatology was calculated.
    years = function(value) {
      if (missing(value))
        private$.years
    },

    #' @field friendlyClassName Character string with class name for display
    #'   purposes.
    friendlyClassName = function(value) {
      if (missing(value))
        "CFClimatology"
    }
  )
)

#' @title 365-day CF calendar
#'
#' @description This class represents a CF calendar of 365 days per year, having
#'   no leap days in any year. This calendar is not compatible with the standard
#'   POSIXt calendar.
#'
#'   This calendar supports dates before year 1 and includes the year 0.
#'
#' @aliases CFCalendar365
#' @docType class
CFCalendar365 <- R6::R6Class("CFCalendar365",
  inherit = CFCalendar,
  private = list(
    # Rata Die, the number of days from the day before 0001-01-01 to
    # origin of this calendar. Used to convert offsets from the calendar origin
    # to the day before 0001-01-01 for arithmetic calculations.
    rd = 0L
  ),
  public = list(
    #' @description Create a new CF calendar of 365 days per year.
    #' @param nm The name of the calendar. This must be "365_day" or "noleap".
    #' @param definition The string that defines the units and the origin, as
    #' per the CF Metadata Conventions.
    #' @return A new instance of this class.
    initialize = function(nm, definition) {
      super$initialize(nm, definition)
      private$rd <- self$date2offset(self$origin)
    },

    #' @description Indicate which of the supplied dates are valid.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return Logical vector with the same length as argument `ymd` has rows
    #'   with `TRUE` for valid days and `FALSE` for invalid days, or `NA` where
    #'   the row in argument `ymd` has `NA` values.
    valid_days = function(ymd) {
      ymd$year & ymd$month >= 1L & ymd$month <= 12L & ymd$day >= 1L &
      ymd$day <= c(31L, 28L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[ymd$month]
    },

    #' @description Determine the number of days in the month of the calendar.
    #' @param ymd `data.frame`, optional, with dates parsed into their parts.
    #' @return A vector indicating the number of days in each month for the
    #'   dates supplied as argument `ymd`. If no dates are supplied, the number
    #'   of days per month for the calendar as a vector of length 12.
    month_days = function(ymd = NULL) {
      if (is.null(ymd)) return(c(31L, 28L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L))

      res <- c(31L, 28L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)[ymd$month]
      res[which(is.na(ymd$year))] <- NA
      res
    },

    #' @description Indicate which years are leap years.
    #' @param yr Integer vector of years to test.
    #' @return Logical vector with the same length as argument `yr`. Since this
    #' calendar does not use leap days, all values will be `FALSE`, or `NA`
    #' where argument `yr` is `NA`.
    leap_year = function(yr) {
      res <- rep(FALSE, length(yr))
      res[which(is.na(yr))] <- NA
      res
    },

    #' @description Calculate difference in days between a `data.frame` of time
    #'   parts and the origin.
    #'
    #' @param x `data.frame`. Dates to calculate the difference for.
    #'
    #' @return Integer vector of a length equal to the number of rows in
    #' argument `x` indicating the number of days between `x` and the `origin`,
    #' or `NA` for rows in `x` with `NA` values.
    date2offset = function(x) {
      corr <- ifelse(x$month <= 2L, 0L, -2L)
      365L * (x$year - 1L) + (367L * x$month - 362L) %/% 12L + corr + x$day - private$rd
    },

    #' @description Calculate date parts from day differences from the origin. This
    #'   only deals with days as these are impacted by the calendar.
    #'   Hour-minute-second timestamp parts are handled in [CFCalendar].
    #'
    #' @param x Integer vector of days to add to the origin.
    #' @return A `data.frame` with columns 'year', 'month' and 'day' and as many
    #'   rows as the length of vector `x`.
    offset2date = function(x) {
      d0 <- x - 1L + private$rd        # d0 is offset relative to year 0, 0-based
      yr <- d0 %/% 365L + 1L           # full years
      d1 <- d0 %% 365L                 # remaining days
      corr <- ifelse(d1 < 59L, 0L, 2L) # correct for days past February
      mon <- (12L * (d1 + corr) + 373L) %/% 367L
      day <- d1 - (367L * mon - 362L) %/% 12L + corr + 1L
      data.frame(year = yr, month = mon, day = day)
    }
  )
)

#' @import R6
NULL

#' @title Basic CF calendar
#'
#' @description This class represents a basic CF calendar. It should not be
#'   instantiated directly; instead, use one of the descendant classes.
#'
#'   This internal class stores the information to represent date and time
#'   values using the CF conventions. An instance is created by the exported
#'   [CFTime] class, which also exposes the relevant properties of this class.
#'
#'   The following calendars are supported:
#'
#' \itemize{
#'   \item [`gregorian\standard`][CFCalendarStandard], the international standard calendar for civil use.
#'   \item [`proleptic_gregorian`][CFCalendarProleptic], the standard calendar but extending before 1582-10-15
#'   when the Gregorian calendar was adopted.
#'   \item [`tai`][CFCalendarTAI], International Atomic Time clock with dates expressed using the Gregorian calendar.
#'   \item [`utc`][CFCalendarUTC], Coordinated Universal Time clock with dates expressed using the Gregorian calendar.
#'   \item [`julian`][CFCalendarJulian], every fourth year is a leap year.
#'   \item [`noleap\365_day`][CFCalendar365], all years have 365 days.
#'   \item [`all_leap\366_day`][CFCalendar366], all years have 366 days.
#'   \item [`360_day`][CFCalendar360], all years have 360 days, divided over 12 months of 30 days.
#' }
#' @references
#'   https://cf-convention.github.io/Data/cf-conventions/cf-conventions-1.13/cf-conventions.html#calendar
#' @docType class
CFCalendar <- R6::R6Class("CFCalendar",
  private = list(
    .name = "",
    .definition = "",
    .unit = -1L,

    # The index into the CFt$prefixes data.frame, if a prefix is being used.
    # Default value of 0L indicates no prefix.
    .prefix = 0L,

    .origin = data.frame(),

    # This method parses the unit to see if it is known, including any prefix
    # for multiples or sub-multiples (section 3.1.3 of the conventions)
    parse_unit = function(x) {
      units <- CFt$CFunits$unit

      # Try exact match - case-insensitive
      u <- which(units == tolower(x))
      if (length(u))
        return(list(prefix = 0L, unit = CFt$CFunits$id[u]))

      # Try prefix + unit combinations - check longest unit names first
      usort <- units[order(nchar(units), decreasing = TRUE)]
      umatch <- which(endsWith(x, usort))
      if (length(umatch)) {
        for (m in 1:length(umatch)) {
          u <- usort[umatch[m]]
          pref <- sub(paste0(u, "$"), "", x)
          pmatch <- match(pref, CFt$prefixes$name)
          if (is.na(pmatch))
            pmatch <- match(pref, CFt$prefixes$abbrev)
          if (!is.na(pmatch))
            return(list(prefix = pmatch, unit = CFt$CFunits$id[match(u, units)]))
        }
      }
      return(list(prefix = 0L, unit = NA))
    }
  ),
  public = list(
    #' @description Create a new CF calendar.
    #' @param nm The name of the calendar. This must follow the CF Metadata
    #' Conventions.
    #' @param definition The string that defines the units and the origin, as
    #' per the CF Metadata Conventions.
    initialize = function(nm, definition) {
      stopifnot(length(definition) ==  1L, length(nm) == 1L)
      private$.name <- tolower(nm)
      private$.definition <- definition

      parts <- strsplit(definition, " ")[[1L]]
      if ((length(parts) < 3L) || !(tolower(parts[2L]) %in% c("since", "after", "from", "ref", "per")))
        stop("Definition string does not appear to be a CF-compliant time coordinate description", call. = FALSE)
      u <- private$parse_unit(parts[1L])
      if (is.na(u$unit)) stop("Unsupported unit: ", parts[1L], call. = FALSE)
      private$.unit <- u$unit
      private$.prefix <- u$prefix

      dt <- self$parse(paste(parts[3L:length(parts)], collapse = " "))
      if (is.na(dt$year[1L]))
        stop("Definition string does not appear to be a CF-compliant time coordinate description: invalid base date specification", call. = FALSE)
      private$.origin <- dt
    },

    #' @description Print information about the calendar to the console.
    #' @param ... Ignored.
    #' @return `self`, invisibly.
    print = function(...) {
      tz <- self$timezone
      if (tz == "+0000") tz <- ""
      cat("CF calendar:",
          "\n  Origin  : ", self$origin_date, "T", self$origin_time, tz,
          "\n  Units   : ", paste0(if (private$.prefix) CFt$prefixes$name[private$.prefix], CFt$units$name[private$.unit]),
          "\n  Type    : ", private$.name, "\n",
          sep = "")
      invisible(self)
    },

    #' @description Indicate which of the supplied dates are valid.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return `NULL`. A warning will be generated to the effect that a
    #' descendant class should be used for this method.
    valid_days = function(ymd) {
      warning("Use a descendant class from `CFCalendar` to call this method.", call. = FALSE) # nocov
      NULL                                                                                    # nocov
    },

    #' @description Add a day to the supplied dates.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return A `data.frame` like argument `ymd` but with a day added for every
    #'   row.
    add_day = function(ymd) {
      md <- self$month_days(ymd)
      ifelse(ymd$day < md,
             ymd$day <- ymd$day + 1L,
             {
               ymd$day <- 1L
               ifelse(ymd$month < 12L,
                      ymd$month <- ymd$month + 1L,
                      {
                        ymd$month <- 1L
                        ymd$year <- ymd$year + 1L
                      }
               )
             }
      )
      ymd
    },

    #' @description Indicate if the time series described using this calendar
    #'   can be safely converted to a standard date-time type (`POSIXct`,
    #'   `POSIXlt`, `Date`).
    #'
    #'   Only the 'standard' calendar and the 'proleptic_gregorian' calendar
    #'   when all dates in the time series are more recent than 1582-10-15
    #'   (inclusive) can be safely converted, so this method returns `FALSE` by
    #'   default to cover the majority of cases.
    #' @param offsets The offsets from the CFtime instance.
    #' @return `FALSE` by default.
    POSIX_compatible = function(offsets) {
      FALSE
    },

    #' @description This method tests if the `CFCalendar` instance in argument
    #'   `cal` is compatible with `self`, meaning that they are of the same
    #'   class and have the same unit and prefix. Calendars "standard", and
    #'   "gregorian" are compatible, as are the pairs of "365_day" and
    #'   "no_leap", and "366_day" and "all_leap".
    #' @param cal Instance of a descendant of the `CFCalendar` class.
    #' @return `TRUE` if the instance in argument `cal` is compatible with
    #'   `self`, `FALSE` otherwise.
    is_compatible = function(cal) {
      private$.unit == cal$unit && private$.prefix == cal$prefix_id &&
      class(self)[1L] == class(cal)[1L]
    },

    #' @description This method tests if the `CFCalendar` instance in argument
    #'   `cal` is equivalent to `self`, meaning that they are of the same class,
    #'   have the same unit and prefix, and equivalent origins. Calendars
    #'   "standard", and "gregorian" are equivalent, as are the pairs of
    #'   "365_day" and "no_leap", and "366_day" and "all_leap".
    #'
    #'   Note that the origins need not be identical, but their parsed values
    #'   have to be. "2000-01" is parsed the same as "2000-01-01 00:00:00", for
    #'   instance.
    #' @param cal Instance of a descendant of the `CFCalendar` class.
    #' @return `TRUE` if the instance in argument `cal` is equivalent to `self`,
    #'   `FALSE` otherwise.
    is_equivalent = function(cal) {
      all(private$.origin[1L,1L:6L] == cal$origin[1L,1L:6L]) &&  # Offset column is NA
      self$is_compatible(cal)
    },

    #' @description Parsing a vector of date-time character strings into parts.
    #' @param d character. A character vector of date-times.
    #' @return A `data.frame` with columns year, month, day, hour, minute,
    #'   second, time zone, and offset. Invalid input data will appear as `NA`.
    parse = function(d) {
      # Parsers

      # UDUNITS broken timestamp definition, with some changes
      # broken_timestamp	{broken_date}({space|T}+{broken_clock})? -- T not in definition but present in lexer code
      # broken_date		    {year}-{month}(-{day})?
      # year		        	[+-]?[0-9]{1,4}
      # month			        0?[1-9]|1[0-2]
      # day			          0?[1-9]|[1-2][0-9]|30|31
      # broken_clock		  {hour}:{minute}(:{second})?
      # hour			        [0-1]?[0-9]|2[0-3]      -- sign on hour not allowed, but see timezone
      # minute			      [0-5]?[0-9]
      # second		      	{minute}?               -- leap second not supported
      # fractional part   (\.[0-9]*)?
      # timezone          [+-]?{hour}(:{minute})? -- added, present in lexer code
      broken <- paste0(
        "^",                                    # anchor string at start
        "([+-]?[0-9]{1,4})",                    # year, with optional sign
        "-(0?[1-9]|1[012])",                    # month
        "(?:-(0?[1-9]|[12][0-9]|3[01]))?",      # day, optional
        "(?:[T ]",                              # if a time is following, separate with a single whitespace character or a "T"
        "([01]?[0-9]|2[0-3])",                  # hour
        ":([0-5]?[0-9])",                       # minute
        "(?::([0-5]?[0-9]))?",                  # second, optional
        "(?:\\.([0-9]*))?",                     # optional fractional part of the smallest specified unit
        ")?",                                   # close optional time capture group
        "(?:\\s",                               # if a time zone offset is following, separate with a single whitespace character
        "([+-])?([01]?[0-9]|2[0-3])",           # tz hour, with optional sign
        "(?::(00|15|30|45))?",                  # optional tz minute, only 4 possible values
        ")?",                                   # close optional timezone capture group
        "$"                                     # anchor string at end
      )

      iso8601 <- paste0(
        "^",
        "([0-9]{4})",
        "-(0[1-9]|1[012])",
        "-(0[1-9]|[12][0-9]|3[01])?",
        "(?:",
        "[T ]([01][0-9]|2[0-3])",
        "(?::([0-5][0-9]))?",
        "(?::([0-5][0-9]))?",
        "(?:[\\.,]([0-9]*))?",
        ")?",
        "(?:([Z+-])([01][0-9]|2[0-3])?(?::(00|15|30|45))?", ## FIXME: Z?, smaller number of captures
        ")?$"
      )

      # UDUNITS packed timestamp definition - NOT YET USED
      # packed_timestamp	{packed_date}({space|T}+{packed_clock})?  -- T and space only allowed in packed time follows
      # packed_date	    	{year}({month}{day}?)?           -- must be YYYYMMDD or else format is ambiguous, as per lexer code
      # packed_clock		  {hour}({minute}{second}?)?       -- must be HHMMSS to be unambiguous
      # timezone          [+-]?{hour}({minute})?           -- added, present in lexer code, must be HHMM
      # packed <- stringi::stri_join(
      #   "^",                                    # anchor string at start
      #   "([+-]?[0-9]{4})",                      # year, with optional sign
      #   "(0[1-9]|1[012])?",                     # month, optional
      #   "(0[1-9]|[12][0-9]|3[01])?",            # day, optional
      #   "(?:[T,\\s]",                           # if a time is following, separate with a single whitespace character or a "T"
      #     "([01][0-9]|2[0-3])?",                # hour
      #     "([0-5][0-9])?",                      # minute, optional
      #     "([0-5]?[0-9](?:\\.[0-9]*)?)?",       # second, optional, with optional fractional part
      #   ")?",                                   # close optional time capture group
      #   "(?:\\s",                               # if a time zone offset is following, separate with a single whitespace character
      #     "([+-]?[01][0-9]|2[0-3])?",           # hour, with optional sign
      #     "(00|15|30|45)?",                     # minute, only 4 possible values
      #   ")?",                                   # close optional timezone capture group
      #   "$"                                     # anchor string at end
      # )

      parse <- data.frame(year = integer(), month = integer(), day = integer(),
                          hour = integer(), minute = integer(), second = numeric(), frac = character(),
                          tz_sign = character(), tz_hour = character(), tz_min = character())

      # Drop "UTC", if given
      d <- trimws(gsub("UTC$", "", d))

      cap <- utils::strcapture(iso8601, d, parse)
      missing <- which(is.na(cap$year))
      if (length(missing) > 0L)
        cap[missing,] <- utils::strcapture(broken, d[missing], parse)

      # Assign any fraction to the appropriate time part
      cap$frac[is.na(cap$frac)] <- "0"
      frac <- as.numeric(paste0("0.", cap$frac))
      if (sum(frac) > 0) {
        ndx <- which(!(is.na(cap$second)) & frac > 0)
        if (length(ndx) > 0L) cap$second[ndx] <- cap$second[ndx] + frac[ndx]
        ndx <- which(!(is.na(cap$minute)) & is.na(cap$second) & frac > 0)
        if (length(ndx) > 0L) cap$second[ndx] <- 60L * frac[ndx]
        ndx <- which(!(is.na(cap$hour)) & is.na(cap$minute) & frac > 0)
        if (length(ndx) > 0L) {
          secs <- 3600 * frac
          cap$minute[ndx] <- secs[ndx] %/% 60
          cap$second[ndx] <- secs[ndx] %% 60
        }
      }
      cap$frac <- NULL

      # Convert NA time parts to 0 - in CF default time is 00:00:00 when not specified
      cap$hour[is.na(cap$hour)] <- 0L
      cap$minute[is.na(cap$minute)] <- 0L
      cap$second[is.na(cap$second)] <- 0L

      # Set timezone to default value where needed
      ndx <- which(cap$tz_sign == "Z")
      if (length(ndx) > 0L) {
        cap$tz_sign[ndx] <- "+"
        cap$tz_hour[ndx] <- "00"
        cap$tz_min[ndx] <- "00"
      }
      cap$tz <- paste0(ifelse(cap$tz_sign == "", "+", cap$tz_sign),
                       ifelse(cap$tz_hour == "", "00", cap$tz_hour),
                       ifelse(cap$tz_min == "", "00", cap$tz_min))
      cap$tz <- ifelse(cap$tz =="NANANA", "+0000", cap$tz)
      cap$tz_sign <- cap$tz_hour <- cap$tz_min <- NULL

      # Set optional date parts to 1 if not specified
      cap$month[is.na(cap$month)] <- 1L
      cap$day[is.na(cap$day)] <- 1L

      # Check date validity
      invalid <- !self$valid_days(cap)
      invalid[is.na(invalid)] <- TRUE
      if (sum(invalid) > 0L) cap[invalid,] <- rep(NA, 7)

      # Calculate offsets
      if (nrow(self$origin) == 0L) {        # if there's no origin yet, don't calculate offsets
        cap$offset <- rep(0, nrow(cap))     # this happens, f.i., when a CFCalendar is created
      } else {
        days <- self$date2offset(cap)
        pref <- if (private$.prefix) CFt$prefixes$multiplier[private$.prefix] else 1
        cap$offset <- round((days * 86400 + (cap$hour - private$.origin$hour[1]) * 3600 +
                             (cap$minute - private$.origin$minute[1]) * 60 +
                             cap$second - private$.origin$second) /
                             (CFt$units$seconds[private$.unit] * pref), 6)
      }
      cap
    },

    #' @description Retrieve the day-of-year number for the supplied
    #'   `data.frame` of year, month and day, using this calendar. This method
    #'   tests for leap years - calendars without leap years may override this
    #'   method for efficiency.
    #' @param ymd `data.frame` with dates parsed into their parts in columns
    #'   `year`, `month` and `day`. Any other columns are disregarded.
    #' @return Integer vector as long as argument `ymd` has rows with the
    #'   day-of-year for each row.
    doy = function(ymd) {
      cum_month <- c(0L, 31L, 59L, 90L, 120L, 151L, 181L, 212L, 243L, 273L, 304L, 334L) # days in year before the month
      doy <- cum_month[ymd$month] + ymd$day
      doy <- ifelse(self$leap_year(ymd$year) & ymd$month > 2L, doy + 1L, doy)
      doy
    },

    #' @description Decompose a vector of offsets, in units of the calendar, to
    #'   their timestamp values. This adds a specified amount of time to the
    #'   origin of a `CFTime` object.
    #'
    #'   This method may introduce inaccuracies where the calendar unit is
    #'   "months" or "years", due to the ambiguous definition of these units.
    #' @param offsets Vector of numeric offsets to add to the origin of the
    #'   calendar.
    #' @return A `data.frame` with columns for the timestamp elements and as
    #'   many rows as there are offsets.
    offsets2time = function(offsets = NULL) {
      if(is.null(offsets) || (len <- length(offsets)) == 0L)
        return(data.frame(year = integer(), month = integer(), day = integer(),
                          hour = integer(), minute = integer(), second = numeric(),
                          tz = character(), offset = numeric()))

      prefix <- if (private$.prefix) CFt$prefixes$multiplier[private$.prefix] else 1

      if (private$.unit <= 4L) { # Days, hours, minutes, seconds
        # First add time: convert to seconds first, then recompute time parts
        secs <- offsets * CFt$units$seconds[private$.unit] * prefix +
                private$.origin$hour * 3600 + private$.origin$minute * 60 + private$.origin$second
        days <- secs %/% 86400L            # overflow days
        secs <- round(secs %% 86400L, 3L)  # drop overflow days from time, round down to milli-seconds to avoid errors

        # Time elements for output
        hrs <- secs %/% 3600L
        mins <- (secs %% 3600L) %/% 60L
        secs <- secs %% 60L

        # Now add days using the calendar
        out <- if (any(days != 0L))
          self$offset2date(days)
        else
          data.frame(year = rep(private$.origin$year, len),
                     month = rep(private$.origin$month, len),
                     day = rep(private$.origin$day, len))

        # Put it all back together again
        out$hour <- hrs
        out$minute <- mins
        out$second <- secs
        out$tz <- rep(self$timezone, len)
      } else { # Months, years
        out <- private$.origin[rep(1L, len), ]
        if (private$.unit == 5L) { # Offsets are months
          months <- out$month + offsets * prefix - 1L
          out$month <- months %% 12L + 1L
          out$year <- out$year + months %/% 12L
        } else {               # Offsets are years
          out$year <- out$year + offsets * prefix
        }
      }
      out$offset <- offsets
      out
    }

  ),
  active = list(
    #' @field name (read-only) Name of the calendar, as per the CF Metadata
    #'   Conventions.
    name = function(value) {
      if (missing(value))
        private$.name
    },

    #' @field definition (read-only) The string that defines the units and the
    #'   origin, as per the CF Metadata Conventions.
    definition = function(value) {
      if (missing(value))
        private$.definition
    },

    #' @field unit (read-only) The numeric id of the unit of the calendar.
    unit = function(value) {
      if (missing(value))
        private$.unit
    },

    #' @field prefix_id (read-only) The index value of the prefix of the time
    #' unit. If the unit does not have a prefix, returns the value 0L.
    prefix_id = function(value) {
      if (missing(value))
        private$.prefix
    },

    #' @field origin (read-only) `data.frame` with fields for the origin of the
    #'   calendar.
    origin = function(value) {
      if (missing(value))
        private$.origin
    },

    #' @field origin_date (read-only) Character string with the date of the
    #'   calendar.
    origin_date = function(value) {
      if (missing(value)) {
        sprintf("%04d-%02d-%02d", self$origin$year, self$origin$month, self$origin$day)
      }
    },

    #' @field origin_time (read-only) Character string with the time of the
    #'   calendar.
    origin_time = function(value) {
      if (missing(value)) {
        .format_time(self$origin)
      }
    },

    #' @field timezone (read-only) Character string with the time zone of the
    #' origin of the calendar.
    timezone = function(value) {
      if (missing(value))
        self$origin$tz
    }
  )
)

#!/bin/bash
set -e # exit with nonzero exit code if anything fails
export GH_PAGES_DIR="$HOME/gh-pages"

# Decrypt and activate the deploy key
echo Setting up access...
openssl aes-256-cbc -K $encrypted_2f711532c6dd_key -iv $encrypted_2f711532c6dd_iv -in ${TRAVIS_BUILD_DIR}/ci/deploy_key.enc -out ${TRAVIS_BUILD_DIR}/ci/deploy_key -d
chmod 600 ${TRAVIS_BUILD_DIR}/ci/deploy_key
eval `ssh-agent -s`
ssh-add ${TRAVIS_BUILD_DIR}/ci/deploy_key

# Clone *this* git repo, but only the gh-pages branch.
echo Cloning gh-pages...
if [[ ! -d $GH_PAGES_DIR ]]; then
    git clone -q -b gh-pages --single-branch git@github.com:${TRAVIS_REPO_SLUG}.git $GH_PAGES_DIR
fi
cd $GH_PAGES_DIR

# inside this git repo we'll pretend to be a new user
git config user.name "Travis CI"
git config user.email "travis@nobody.org"

# The first and only commit to this new Git repo contains all the
# files present with the commit message "Deploy to GitHub Pages".
echo Updating docs...
rm -rf *
cp -R ${TRAVIS_BUILD_DIR}/docs/_build/html/* ./
touch .nojekyll

echo Staging...
git add -A .
git commit --amend --reset-author --no-edit

# Push up to gh-pages
echo Pushing...
git push --force origin gh-pages

---
title: "Working with CFtime"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Working with CFtime}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```


```{r setup, include = FALSE}
library(CFtime)
library(ncdf4)
```

## Climate change models and calendars

Around the world, many climate change models are being developed (100+) under
the umbrella of the [World Climate Research Programme](https://www.wcrp-climate.org) to assess the
rate of climate change. Published data is generally publicly available to download for
research and other (non-commercial) purposes through partner organizations in 
the Earth Systems Grid Federation.

The data are all formatted to comply with the [CF Metadata Conventions](http://cfconventions.org),
a set of standards to support standardization among research groups and published
data sets. These conventions greatly facilitate use and analysis of the climate
projections because standard processing work flows (should) work across
the various data sets.

On the flip side, the CF Metadata Conventions needs to cater to a wide range of
modeling requirements and that means that some of the areas covered by the standards are
more complex than might be assumed. One of those areas is the temporal dimension
of the data sets. The CF Metadata Conventions supports no less than nine different
calendar definitions, that, upon analysis, fall into five distinct calendars
(from the perspective of computation of climate projections):

 - `standard` or `gregorian`: The international civil calendar that is in common use in 
 many countries around the world, adopted by edict of Pope Gregory XIII in 1582
 and in effect from 15 October of that year. The `proleptic_gregorian` calendar
 is the same as the `gregorian` calendar, but with validity
 extended to periods prior to `1582-10-15`.
 - `julian`: Adopted in the year 45 BCE, every fourth year is a leap year. 
 Originally, the julian calendar did not have a monotonically increasing year 
 assigned to it and there are indeed several julian calendars in use around the
 world today with different years assigned to them. Common interpretation
 is currently that the year is the same as that of the standard calendar. The 
 julian calendar is currently 13 days behind the gregorian calendar.
 - `365_day` or `noleap`: No years have a leap day.
 - `366_day` or `all_leap`: All years have a leap day.
 - `360_day`: Every year has 12 months of 30 days each.

The three latter calendars are specific to the CF Metadata Conventions to
reduce computational complexities of working with dates. These three, and the
julian calendar, are not compliant with the standard `POSIXt` date/time
facilities in `R` and using standard date/time procedures would quickly lead to
problems. In the below code snippet, the date of `1949-12-01` is the *datum*
from which other dates are calculated. When adding 43,289 days to this *datum*
for a data set that uses the `360_day` calendar, that should yield a date some 120 
years after the *datum*:

```{r}
# POSIXt calculations on a standard calendar - INCORRECT
as.Date("1949-12-01") + 43289

# CFtime calculation on a "360_day" calendar - CORRECT
# See below examples for details on the two functions
CFtimestamp(CFtime("days since 1949-12-01", "360_day", 43289))
```

Using standard `POSIXt` calculations gives a result that is about 21 months off
from the correct date - obviously an undesirable situation. This example is far
from artificial: `1949-12-01` is the datum for all CORDEX data, covering the 
period 1951 - 2005 for historical experiments and the period 2006 - 2100 for RCP 
experiments (with some deviation between data sets), and several models used in 
the CORDEX set use the `360_day` calendar. The `365_day` or `noleap` calendar 
deviates by about 1 day every 4 years (disregarding centurial years), or about 
24 days in a century. The `366_day` or  `all_leap` calendar deviates by about 3 
days every 4 years, or about 76 days in a century.

The `CFtime` package deals with the complexity of the different calendars allowed
by the CF Metadata Conventions. It properly formats dates and times (even oddball
dates like `2070-02-30`) and it can generate calendar-aware factors for further 
processing of the data.

##### Time zones

The character of CF time series - a number of numerical offsets from a base date -
implies that there should only be a single time zone associated with the time
series. The time zone offset from UTC is stored in the datum and can be retrieved
with the `CFtimezone()` function. If a vector of character timestamps with time
zone information is parsed with the `CFparse()` function and the time zones are 
found to be different from the datum time zone, a warning message is generated
but the timestamp is interpreted as being in the datum time zone. No correction 
of timestamp to datum time zone is performed.

## Using CFtime to deal with calendars

Data sets that are compliant with the CF Metadata Conventions always include a 
*datum*, a specific point in time in reference to a specified *calendar*, from 
which other points in time are calculated by adding a specified *offset* of a
certain *unit*. This approach is encapsulated in the `CFtime` package by the S4 
class `CFtime`.

```{r}
# Create a CF time object from a definition string, a calendar and some offsets
cf <- CFtime("days since 1949-12-01", "360_day", 19830:90029)
cf
```

The `CFtime()` function takes a *datum* description (which is actually a unit -
"days" - in reference to a datum - "1949-12-01"), a calendar description, and a 
vector of *offsets* from that datum. Once a `CFtime` instance is created its 
datum and calendar cannot be changed anymore. Offsets may be added.

In practice, these parameters will be taken from the data set of interest. CF Metadata
Conventions require data sets to be in the NetCDF format, with all metadata
describing the data set included in a single file, including the mandatory 
"Conventions" global attribute which should have a string identifying the version 
of the CF Metadata Conventions that this file adheres to (among possible others).
Not surprisingly, all the pieces of interest are contained in the mandatory `time` 
dimension of the file. The process then becomes as follows, for a CMIP6 file of 
daily precipitation:

```{r}
# Opening a data file that is included with the package and showing some attributes.
# Usually you would `list.files()` on a directory of your choice.
nc <- nc_open(list.files(path = system.file("extdata", package = "CFtime"), full.names = TRUE)[1])
attrs <- ncatt_get(nc, "")
attrs$title

# "Conventions" global attribute must have a string like "CF-1.*" for this package to work reliably
attrs$Conventions

# Create the CFtime instance from the metadata in the file.
cf <- CFtime(nc$dim$time$units, nc$dim$time$calendar, nc$dim$time$vals)
cf
```

You can see from the global attribute "Conventions" that the file adheres to the
CF Metadata Conventions, among others. According to the CF conventions, `units` 
and `calendar` are required attributes of the `time` dimension in the NetCDF file, 
and `nc$dim$time$vals` are the offset values, or `dimnames()` in `R` terms, for 
the `time` dimension of the data.

The above example (and others in this vignette) use the `ncdf4` package. If you 
are using the `RNetCDF` package, checking for CF conventions and then creating a 
`CFtime` instance goes like this:

```{r}
library(RNetCDF)
nc <- open.nc(list.files(path = system.file("extdata", package = "CFtime"), full.names = TRUE)[1])
att.get.nc(nc, -1, "Conventions")
cf <- CFtime(att.get.nc(nc, "time", "units"), att.get.nc(nc, "time", "calendar"), var.get.nc(nc, "time"))
cf
```

The corresponding character representations of the time series can be easily generated:

```{r}
dates <- CFtimestamp(cf, format = "date")
dates[1:10]
```

...as well as the full range of the time series:

```{r}
CFrange(cf)
```

Note that in this latter case, if any of the timestamps in the time series have a time that is
other than `00:00:00` then the time of the extremes of the time series is also
displayed. This is a common occurrence because the CF Metadata Conventions
prescribe that the middle of the time period (month, day, etc) is recorded, which 
for months with 31 days would be something like `2005-01-15T12:00:00`.

## Supporting processing of climate projection data

When working with high resolution climate projection data, typically at a "day"
resolution, one of the processing steps would be to aggregate the data to some
lower resolution such as a dekad (10-day period), a month or a meteorological 
season, and then compute a derivative value such as the dekadal sum of precipitation,
monthly minimum/maximum daily temperature, or seasonal average daily short-wave
irradiance.

It is also possible to create factors for multiple "epochs" in one go. This greatly
reduces programming effort if you want to calculate anomalies over multiple future
periods. A complete example is provided in the vignette 
["Processing climate projection data"](Processing.html).

It is easy to generate the factors that you need once you have a `CFtime` instance 
prepared:

```{r}
# Create a dekad factor for the whole `cf` time series that was created above
f_k <- CFfactor(cf, "dekad")
str(f_k)

# Create monthly factors for a baseline epoch and early, mid and late 21st century epochs
baseline <- CFfactor(cf, epoch = 1991:2020)
future <- CFfactor(cf, epoch = list(early = 2021:2040, mid = 2041:2060, late = 2061:2080))
str(future)
```

For the "epoch" version, there are two interesting things to note here:

 - The epochs do not have to coincide with the boundaries of the time series. In the 
 example above, the time series starts in 2015, while the baseline epoch is from 1991.
 Obviously, the number of time steps from the time series that then fall within 
 this epoch will then be reduced.
 - The factor is always of the same length as the time series, with `NA` values
 where the time series values are not falling in the epoch. This ensures that the
 factor is compatible with the data set which the time series describes, such
 that functions like `tapply()` will not throw an error.
 
There are five periods defined for `CFfactor()`:

 - `year`, to summarize data to yearly timescales
 - `season`, the meteorological seasons. Note that the month of December will be
 added to the months of January and February of the following year, so the date 
 "2020-12-01" yields the factor value "2021-DJF".
 - `month`, monthly summaries, the default period.
 - `dekad`, 10-day period. Each month is subdivided in dekads as follows: (1) 
 days 01 - 10; (2) days 11 - 20; (3) remainder of the month.
 - `day`, to summarize sub-daily data.

##### Incomplete time series

You can test if your time series is complete with the function `CFcomplete()`.
A time series is considered complete if the time steps between the two extreme 
values are equally spaced. There is a "fuzzy" assessment of completeness for
time series with a datum unit of "days" or smaller where the time steps are
months or years apart - these have different lengths in days in different months 
or years (e.g. a leap year).

If your time series is incomplete, for instance because it has missing time steps, 
you should recognize that in your further processing. As an example, you might 
want to filter out months that have fewer than 90% of daily data from further 
processing or apply weights based on the actual coverage. 

```{r}
# Is the time series complete?
CFcomplete(cf)

# How many time units fit in a factor level?
CFfactor_units(cf, baseline)

# What's the absolute and relative coverage of our time series
CFfactor_coverage(cf, baseline, "absolute")
CFfactor_coverage(cf, baseline, "relative")
```

The time series is complete but coverage of the baseline epoch is only 
20%! Recall that the time series starts in 2015 while the baseline period in the
factor is for `1991:2020` so that's only 6 years of time series data out of 30 
years of the baseline factor.

An artificial example of missing data:

```{r}
# 4 years of data on a `365_day` calendar, keep 80% of values
n <- 365 * 4
cov <- 0.8
offsets <- sample(0:(n-1), n * cov)

cf <- CFtime("days since 2020-01-01", "365_day", offsets)
cf
# Note that there are about 1.25 days between observations

mon <- CFfactor(cf, "month")
CFfactor_coverage(cf, mon, "absolute")
CFfactor_coverage(cf, mon, "relative")
```

Keep in mind, though, that there are data sets where the time unit is lower than
the intended resolution of the data. Since the CF conventions recommend that the
coarsest time unit is "day", many files with monthly data sets have a definition
like `days since 2016-01-01` with offset values for the middle of the month like
`15, 44, 74, 104, ...`. Even in these scenarios you can verify that your data 
set is complete with the function `CFcomplete()`.

## CFtime and POSIXt

The CF Metadata Conventions support nine different calendars. Of these, only the
`standard`, `gregorian` and `proleptic_gregorian` calendars are fully compatible
with POSIXt. The other calendars have varying degrees of discrepancies:

 - `julian`: Every fouth year is a leap year. Dates like `2100-02-29` and
 `2200-02-29` are valid.
 - `365_day` or `noleap`: No leap year exists. `2020-02-29` does not occur.
 - `366_day` or `all_leap`: All years are leap years.
 - `360_day`: All months have 30 days in every year. This means that 31 January,
 March, May, July, August, October and December never occur, while 29 and 30
 February occur in every year.
 
Converting time series using these incompatible calendars to `Date`s is likely
to produce problems. This is most pronounced for the `360_day` calendar:

```{r}
# Days in January and February
cf <- CFtime("days since 2023-01-01", "360_day", 0:59)
cf_days <- CFtimestamp(cf, "date")
as.Date(cf_days)
```

31 January is missing from the vector of `Date`s because the `360_day` calendar 
does not include it and 29 and 30 February are `NA`s because POSIXt rejects them.
This will produce problems later on when processing your data.

The general advice is therefore: **do not convert CFtime objects to Date objects**
unless you are sure that the `CFtime` object uses a POSIXt-compatible calendar.

##### So how do I compare climate projection data with different calendars?

One reason to convert the "time" dimension from different climate projection data
sets is to be able to compare the data from different models and produce a
multi-model ensemble. The correct procedure to do this is to first calculate 
**for each data set individually** the property of interest (e.g. average daily 
rainfall per month anomaly for some future period relative to a baseline period), 
which will typically involve aggregation to a lower resolution (such as from 
daily data to monthly averages), and only then combine the aggregate data from 
multiple data sets to compute statistically interesting properties (such as 
average among models and standard deviation, etc).

Once data is aggregated from daily or higher-resolution values, the different 
calendars no longer matter (although if you do need to convert averaged data to
absolute data you should use `CFfactor_units()` to make sure that you use the
correct scaling factor).

Otherwise, there really shouldn't be any reason to convert the time series in the
data files to `Date`s. Climate projection data is virtually never compared on a 
day-to-day basis between different models and neither does complex date arithmetic 
make much sense (such as adding intervals) - `CFtime` can support basic arithmetic
by manipulation the offsets of the `CFtime` object. The character representations
that are produced are perfectly fine to use for `dimnames()` on an array or as
`rownames()` in a `data.frame` and these also support basic logical operations
such as `"2023-02-30" < "2023-03-01"`. So ask yourself, do you really need `Date`s
when working with unprocessed climate projection data? (If so, [open an issue on 
GitHub](https://github.com/pvanlaake/CFtime/issues)).

A complete example of creating a multi-model ensemble is provided in the vignette 
["Processing climate projection data"](Processing.html).

## Final observations

 - This package is intended to facilitate processing of climate projection
 data. It is not a full implementation of the CF Metadata Conventions "time"
 component.
 - In parsing and deparsing of offsets and timestamps, data is rounded to
 3 digits of precision of the unit of the datum. When using a description of time
 that is very different from the datum unit, this may lead to some loss of 
 precision due to rounding errors. For instance, if milli-second precision is 
 required, use a unit of "seconds". The authors have no knowledge of published
 climate projection data that requires milli-second precision so for the
 intended use of the package this issue is marginal.

---
title: "Processing climate projection data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Processing climate projection data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, include = FALSE}
library(CFtime)
```

Climate projection data sets are produced in a variety of formats but all conform 
to the CF Metadata Conventions. NetCDF data files, in particular, are highly
structured and relatively easy to process. That said, it is very important to
maintain a proper processing workflow such that the small changes in the climate
projections are maintained and revealed through analysis. In this document, the
basic workflow with varying calendars is described.

> In this vignette we are using the `ncdfCF` package as that provides the easiest
interface to work with netCDF files. Package `CFtime` is integrated into `ncdfCF`
which makes working with time dimensions in netCDF seamless.  
> Packages `RNetCDF` and `ncdf4` can work with `CFtime` as well but then the
"intelligence" built into `ncdfCF` is not available, such as automatically
identifying axes and data orientation. Other packages like `terra` and `stars`
are not recommended because they do not provide access to the specifics of the
time dimension of the data and do not consider any calendars other than
"proleptic_gregorian".

## Processing climate projection data

Individual files containing climate projections contain global, regional or local
data, typically on a rectangular latitude-longitude grid, for a single parameter
such as "near-surface temperature", and for a number of time steps. An analysis
workflow then consists of a number of steps:

 - Download the appropriate data files for your desired combination of model, experiment, realization,
 geography, time range, parameter, ... (called a "data suite" henceforth). If your 
 analysis involves multiple parameters (such as temperature and precipitation to 
 estimate crop productivity), repeat the process for all parameters. If you want 
 to make a multi-model ensemble to reduce model bias, repeat again for all desired model, 
 experiment and realization combinations ("ensemble member"). You end up having 
 one or more data suites to work with.
 - Take all files in a data suite and extract the data. Process the data in the 
 data suite. Since the data are (mostly) 3-dimensional arrays, this will involve 
 binding the arrays on the right dimension and then do something like 
 `apply(data, 1:2, tapply, f, fun)` (following the CF Metadata Conventions,
 dimensions 1 and 2 are "longitude" and "latitude", respectively; the third 
 dimension is "time"; none of this is standardized though and deviations are out
 there; package `ncdfCF` can help with the `CFData$array()` method). Repeat for 
 the data suite for each ensemble member.
 - Combine the above results as your workflow requires. Frequently this involves 
 computing "anomalies": ratio the data for one or more future periods to a 
 baseline period. Repeat for each ensemble member.
 - Construct the multi-model ensemble from the individual ensemble members.
 
Apart from the first step of obtaining the data, the steps lend themselves well
to automation. The catch, however, is in the factor `f` to use with `tapply()`.
The different models (in your ensemble) use different calendars, meaning that
different factors are required. The `CFtime` package can help out.

The `CFTime$factor()` method produces a factor that respects the calendar of the
data files. The method comes in two operating modes:

 - Plain vanilla mode produces a factor for a time period across the entire
 time series. The factor level includes the year. This would be useful to calculate 
 mean temperature for every month in every year, for instance.
 - When one or more "eras" (periods of interest) are provided, the factor level 
 no longer includes the year and can be used to calculate, for instance, the mean
 temperature per period of interest in the era (e.g. average March temperature
 in the era 2041-2060).
 
```{r}
# Setting up
library(ncdfCF)
fn <- list.files(path = system.file("extdata", package = "CFtime"), full.names = TRUE)[1]
(ds <- ncdfCF::open_ncdf(fn))

# The T axis, with name "time" has a CFTime instance
t <- ds[["time"]]$time()

# Create monthly factors for a baseline era and early, mid and late 21st century eras
baseline <- t$factor(era = 1991:2020)
future <- t$factor(era = list(early = 2021:2040, mid = 2041:2060, late = 2061:2080))
str(baseline)
str(future)
```

Building on the examples above of opening a file, creating a `CFTime` instance
and a suitable factor for one data suite, here daily rainfall, the actual
processing of the data into precipitation anomalies for 3 periods relative to a
baseline period could look like this:

```{r}
# Get the data for the "pr" data variable from the netCDF data set.
# The `CFData$array()` method ensures that data are in standard R orientation.
# Converts units of kg m-2 s-1 to mm/day.
pr <- ds[["pr"]]$data()$array() * 86400

# Get a global attribute from the file
experiment <- ds$attribute("experiment_id")

# Calculate the daily average precipitation per month for the baseline period
# and the three future eras.
pr_base <- apply(pr, 1:2, tapply, baseline, mean)                         # an array
pr_future <- lapply(future, function(f) apply(pr, 1:2, tapply, f, mean))  # a list of arrays

# Calculate the precipitation anomalies for the future eras against the baseline.
# Working with daily averages per month so we can simply subtract and then multiply by days 
# per month for each of the factor levels using the CF calendar.
ano <- mapply(function(pr, f) {(pr - pr_base) * t$factor_units(f)}, pr_future, future, SIMPLIFY = FALSE)

# Plot the results
plot(1:12, ano$early[,1,1], type = "o", col = "blue", ylim = c(-50, 40), xlim = c(1, 12), 
     main = paste0("Hamilton, New Zealand\nExperiment: ", experiment), 
     xlab = "month", ylab = "Precipitation anomaly (mm)")
lines(1:12, ano$mid[,1,1], type = "o", col = "green")
lines(1:12, ano$late[,1,1], type = "o", col = "red")
```

Looks like Hadley will be needing rubber boots in spring and autumn back home!

The interesting feature, working from opening the netCDF file down to plotting, is that
the specifics of the CF calendar that the data suite uses do not have to be considered
anywhere in the processing workflow: the `CFtime` package provides the functionality.
Data suites using another CF calendar are processed exactly the same.

## Combining data from different models with different calendars

Different climate projection data sets can use different calendars. It is
absolutely essential to respect the calendar of the different data sets because
the underlying solar and atmospheric physics are based on those calendars as
well.

In a typical situation, a researcher would construct a multi-model ensemble to
remove or reduce the bias in any given model. The data sets composing the
ensemble might well use different calendars. The correct way of constructing an
ensemble is to perform the desired analysis on every ensemble member individually
and to combine them only in the final step and to then perform any ensemble operations
such as computing confidence intervals. The design of the `CFtime` package makes
it easy to do this, through its heavy use of lists. Building on the previous
example, let's make a multi-model ensemble of 2 models (not much of an ensemble
but such are the limitations of including data with packages - the example easily
extends to a larger set of ensemble members).

```{r}
# Get the list of files that make up the ensemble members, here:
# GFDL ESM4 and MRI ESM2 models for experiment SSP2-4.5, precipitation, CMIP6 2015-01-01 to 2099-12-31
lf <- list.files(path = system.file("extdata", package = "CFtime"), full.names = TRUE)

# Loop over the files individually
# ano is here a list with each element holding the results for a single model
ano <- lapply(lf, function(fn) {
  ds <- ncdfCF::open_ncdf(fn)
  t <- ds[["time"]]$time()
  pr <- ds[["pr"]]$data()$array() * 86400

  baseline <- t$factor(era = 1991:2020)
  pr_base <- apply(pr, 1:2, tapply, baseline, mean)
  future <- t$factor(era = list(early = 2021:2040, mid = 2041:2060, late = 2061:2080))
  pr_future <- lapply(future, function(f) apply(pr, 1:2, tapply, f, mean))
  mapply(function(pr, f) {(pr - pr_base) * t$factor_units(f)}, pr_future, future, SIMPLIFY = FALSE)
})

# Era names
eras <- c("early", "mid", "late")
dim(eras) <- 3

# Build the ensemble for each era
# For each era, grab the data for each of the ensemble members, simplify to an array
# and take the mean per row (months, in this case)
ensemble <- apply(eras, 1, function(e) {
  rowMeans(sapply(ano, function(a) a[[e]], simplify = TRUE))})
colnames(ensemble) <- eras
rownames(ensemble) <- rownames(ano[[1]][[1]])
ensemble
```

Here we simply compute the average of the monthly precipitation anomaly over the
ensemble members. In a more typical scenario, you would use the values from the 
individual models and to apply a more suitable analysis, such as calculating 
the confidence interval or model agreement.

One significant advantage of this processing workflow is that it is easily 
parallelized: the bulk of the work goes into computing the anomalies, `ano`, and
this is [embarrassingly parallel](https://en.wikipedia.org/wiki/Embarrassingly_parallel)
because they read their own data and produce independent outputs. Use 
[package future](https://cran.r-project.org/package=future) or something similar 
to easily make the code run on all available CPU cores.

## Working with multiple files in a single data suite

Due to the large size of typical climate projection data files, it is
common to have a data suite that is contained in multiple files. A case in point
is the CORDEX data set which breaks up the experiment period of 2006 - 2100 into
19 files of 5 years each, with each file covering a single parameter (temperature,
precipitation, etc) over an entire domain (such as Europe, South Asia, Central 
America and the Caribbean, etc). The CFtime package can streamline processing of
such multi-file data suites as well.

Assuming that you have your CORDEX files in a directory on disk, organized by domain
and other properties such as the variable, GCM/RCM combination, experiment, etc, 
the process of preparing the files for processing could be encoded in a function 
as below. The argument `fn` is a list of file names to process, and `var` is the
variable contained in the files. (There are no checks on argument sanity here, 
which should really be included. This function only makes sense for a single
[domain, GCM/RCM, experiment, variable] combination. Also be aware of data size,
CORDEX files are huge and stitching all domain data together will easily exhaust
available memory and it may thus lead to very large swap files and very poor 
performance - use the `ncdfCF::CFVariable$subset()` method to read spatial or
temporal chunks of data to avoid such problems.)

```{r eval = FALSE}
library(ncdfCF)
library(abind)

prepare_CORDEX <- function(fn, var, aoi) {
  data <- vector("list", length(fn))
  for (i in 1:length(fn)) {
    ds <- ncdfCF::open_ncdf(fn[i])
    if (i == 1) {
      # Get a CFTime instance from the first file
      t <- ds[["time"]]$time()
    } else {
      # Add offsets from the file and add to the CFTime instance
      t <- t + ds[["time"]]$time()$offsets
    }
    
    # Put the subsetted data array in the list
    data[[i]] <- ds[[var]]$subset(aoi = aoi)$array()
  }
  
  # Create a list for output with the CFTime instance and
  # the data bound in a single 3-dimensional array
  list(CFTime = t, data = abind(data, along = 3))
}
```

Calling this function like `prepare_CORDEX(list.files(path = "~/CC/CORDEX/CAM",
pattern = "^pr.*\\.nc$", full.names = TRUE), "pr", ncdfCF::aoi(0, 20, 30, 50))` 
will yield a list of netCDF files with precipitation data, with the resulting 
`CFTime` instance describing the full temporal extent covered by the data files, 
as well as the data bound on the temporal dimension, ready for further processing.

When working like this it is imperative that the offsets and the data arrays are
added to their final structures *in exactly the same order*. It is not necessary
that the offsets (and the data) themselves are in order, but the correspondence
between offsets and data needs to be maintained. (`list.files()` produces a list
in alphabetical order by default, which for most climate projection files
produces offsets in chronological order.)

## Acknowledgements

The results presented contain modified data from Copernicus Climate Change 
Service information, 2023-2025. Neither the European Commission nor ECMWF is responsible 
for any use that may be made of the Copernicus information or data it contains.

We acknowledge the World Climate Research Programme, which, through its Working 
Group on Coupled Modelling, coordinated and promoted CMIP6. We thank the climate 
modeling groups for producing and making available their model output, the Earth 
System Grid Federation (ESGF) for archiving the data and providing access, and the 
multiple funding agencies who support CMIP6 and ESGF.

The two data sets used as examples in this vignette carry the following license
statements:

 - **pr_day_GFDL-ESM4_ssp245_r1i1p1f1_gr1_20150101-20991231_v20180701.nc:** CMIP6 
model data produced by NOAA-GFDL is licensed under a Creative Commons Attribution-ShareAlike 
4.0 International License (https://creativecommons.org/licenses/). Consult
https://pcmdi.llnl.gov/CMIP6/TermsOfUse/ for terms of use governing CMIP6 output, 
including citation requirements and proper acknowledgment. Further information 
about this data, including some limitations, can be found via the further_info_url 
(recorded as a global attribute in this file). The data producers and data providers 
make no warranty, either express or implied, including, but not limited to, warranties 
of merchantability and fitness for a particular purpose. All liabilities arising 
from the supply of the information (including any liability arising in negligence) 
are excluded to the fullest extent permitted by law.
 - **pr_day_MRI-ESM2-0_ssp245_r1i1p1f1_gn_20150101-20991231_v20190603.nc:** CMIP6 
model data produced by MRI is licensed under a Creative Commons Attribution-ShareAlike 
4.0 International License (https://creativecommons.org/licenses/). Consult
https://pcmdi.llnl.gov/CMIP6/TermsOfUse/ for terms of use governing CMIP6 output, 
including citation requirements and proper acknowledgment. Further information about 
this data, including some limitations, can be found via the further_info_url (recorded
as a global attribute in this file). The data producers and data providers make no 
warranty, either express or implied, including, but not limited to, warranties of 
merchantability and fitness for a particular purpose. All liabilities arising from
the supply of the information (including any liability arising in negligence) are 
excluded to the fullest extent permitted by law.

---
title: "Working with CFtime"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Working with CFtime}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```


```{r setup, include = FALSE}
library(CFtime)
library(ncdfCF)
```

## Climate change models and calendars

Around the world, many climate change models are being developed (100+) under
the umbrella of the [World Climate Research Programme](https://www.wcrp-climate.org) to assess the
rate of climate change. Published data is generally publicly available to download for
research and other (non-commercial) purposes through partner organizations in 
the Earth Systems Grid Federation.

The data are all formatted to comply with the [CF Metadata Conventions](http://cfconventions.org),
a set of standards to support standardization among research groups and published
data sets. These conventions greatly facilitate use and analysis of the climate
projections because standard processing work flows (should) work across
the various data sets.

On the flip side, the CF Metadata Conventions needs to cater to a wide range of
modeling requirements and that means that some of the areas covered by the standards are
more complex than might be assumed. One of those areas is the temporal dimension
of the data sets. The CF Metadata Conventions supports no less than 11 different
calendar definitions, that, upon analysis, fall into 8 distinct calendars
(from the perspective of computation of climate projections):

 - `standard` (or `gregorian`): The Gregorian calendar that is in common use in 
 many countries around the world, adopted by edict of Pope Gregory XIII in 1582
 and in effect from 15 October of that year. The earliest valid time in this
 calendar is 0001-01-01 00:00:00 (1 January of year 1) as year 0 does not exist
 and the CF Metadata Conventions require the year to be positive, but noting
 that a Julian calendar is used in periods before the Gregorian calendar was
 introduced.
 - `proleptic_gregorian`: This is the Gregorian calendar with validity 
 extended to periods prior to `1582-10-15`, including a year 0 and negative years.
 This calendar is being used in most OSes and is what is being used by R.
 - `tai`: International Atomic Time, a global standard for linear time: it 
counts seconds since its start at 1958-01-01 00:00:00. For presentation it uses
the Gregorian calendar. Timestamps prior to its start are not allowed.
- `utc`: Coordinated Universal Time, the standard for civil timekeeping all
over the world. It is based on International Atomic Time but it uses occasional
leap seconds to remain synchronous with Earth's rotation around the Sun; at the
end of 2024 it is 37 seconds behind `tai`. It uses the Gregorian calendar with a
start at 1972-01-01 00:00:00; earlier timestamps are not allowed. Future timestamps
are also not allowed because the insertion of leap seconds is unpredictable.
Most computer clocks use UTC.
- `julian`: Adopted in the year 45 BCE, every fourth year is a leap year. 
 Originally, the Julian calendar did not have a monotonically increasing year 
 assigned to it and there are indeed several Julian calendars in use around the
 world today with different years assigned to them. Common interpretation
 is currently that the year is the same as that of the Gregorian calendar. The 
 Julian calendar is currently 13 days behind the Gregorian calendar. As with the 
 standard calendar, the earliest valid time in this calendar is 0001-01-01 00:00:00.
 - `365_day` or `noleap`: "Model time" in which no years have a leap day. 
 Negative years are allowed and year 0 exists.
 - `366_day` or `all_leap`: "Model time" in which all years have a leap day. 
 Negative years are allowed and year 0 exists.
 - `360_day`: "Model time" in which every year has 12 months of 30 days each. 
 Negative years are allowed and year 0 exists.

The three latter calendars of model time are specific to the CF Metadata 
Conventions to reduce computational complexities of working with dates. None of 
the 8 calendars are compliant with the standard `POSIXt` date/time facilities in 
`R` and using standard date/time functions would quickly lead to problems. See 
the section on "CFtime and POSIXt", below, for a detailed description of the 
discrepancies between the CF calendars and POSIXt.

In the below code snippet, the date of `1949-12-01` is the *origin* from which 
other dates are calculated. When adding 43,289 days to this origin for a data 
set that uses the `360_day` calendar, that should yield a date some 120 years 
after the origin:

```{r}
# POSIXt calculations on a standard calendar - INCORRECT
as.Date("1949-12-01") + 43289

# CFtime calculation on a "360_day" calendar - CORRECT
# See below examples for details on the two functions
as_timestamp(CFtime("days since 1949-12-01", "360_day", 43289))
```

Using standard `POSIXt` calculations gives a result that is about 21 months off
from the correct date - obviously an undesirable situation. This example is far
from artificial: `1949-12-01` is the origin for all CORDEX data, covering the 
period 1950 - 2005 for historical experiments and the period 2006 - 2100 for RCP 
experiments (with some deviation between data sets), and several models used in 
the CORDEX set use the `360_day` calendar. The `365_day` or `noleap` calendar 
deviates by about 1 day every 4 years (disregarding centurial years), or about 
24 days in a century. The `366_day` or  `all_leap` calendar deviates by about 3 
days every 4 years, or about 76 days in a century.

The `CFtime` package deals with the complexity of the different calendars allowed
by the CF Metadata Conventions. It properly formats dates and times (even oddball
dates like `2070-02-30`) and it can generate calendar-aware factors for further 
processing of the data.

##### Time zones

The character of CF time series - a number of numerical offsets from a base date -
implies that there should only be a single time zone associated with the time
series, and then only for the `standard` and `proleptic_gregorian` calendars.
For the other calendars a time zone can be set but it will have no effect.
Daylight savings time information is never considered by `CFtime` so the user
should take care to avoid entering times with DST.

The time zone offset from UTC is stored in the `CFTime` instance and can be 
retrieved with the `timezone()` function. If a vector of character timestamps with time
zone information is parsed with the `parse_timestamps()` function and the time zones are 
found to be different from the `CFTime` time zone, a warning message is generated
but the timestamp is interpreted as being in the `CFTime` time zone. No correction 
of timestamp to `CFTime` time zone is performed.

The concept of time zones does not apply to the `utc` and `tai` calendars as
they represent universal time, i.e. the indicated time is valid all over the
globe. Timestamps passed to these calendars should not have a time zone
indicated, but if there are, anything other than a 0 offset will generate an
error.

## Using CFtime to deal with calendars

Data sets that are compliant with the CF Metadata Conventions always include a 
*origin*, a specific point in time in reference to a specified *calendar*, from 
which other points in time are calculated by adding a specified *offset* of a
certain *unit*. This approach is encapsulated in the `CFtime` package by the R6 
class `CFTime`.

```{r}
# Create a CFTime object from a definition string, a calendar and some offsets
(t <- CFtime("days since 1949-12-01", "360_day", 19830:90029))
```

The `CFtime()` function takes a description (which is actually a unit -
"days" - in reference to an origin - "1949-12-01"), a calendar description, and a 
vector of *offsets* from that origin. Once a `CFTime` instance is created its 
origin and calendar cannot be changed anymore. Offsets may be added.

In practice, these parameters will be taken from the data set of interest. CF Metadata
Conventions require data sets to be in the netCDF format, with all metadata
describing the data set included in a single file, including the mandatory 
"Conventions" global attribute which should have a string identifying the version 
of the CF Metadata Conventions that this file adheres to (among possible others).
Not surprisingly, all the pieces of interest are contained in the "time"
dimension of the file. The process then becomes as follows, for a CMIP6 file of 
daily precipitation:

```{r}
# Opening a data file that is included with the package.
# Usually you would `list.files()` on a directory of your choice.
fn <- list.files(path = system.file("extdata", package = "CFtime"), full.names = TRUE)[1]
(ds <- ncdfCF::open_ncdf(fn))

# "Conventions" global attribute must have a string like "CF-1.*" for this package to work reliably

# Look at the "time" axis
(time <- ds[["time"]])

# Get the CFTime instance from the "time" axis
(t <- time$time())
```

You can see from the global attribute "Conventions" that the file adheres to the
CF Metadata Conventions, among others. According to the CF conventions, `units` 
and `calendar` are required attributes of the "time" dimension in the netCDF file.

The above example (and others in this vignette) use the `ncdfCF` package. If you 
are using the `RNetCDF` or `ncdf4` package, checking for CF conventions and then 
creating a `CFTime` instance goes like this:

```{r, eval = FALSE}
library(RNetCDF)
nc <- open.nc(fn)
att.get.nc(nc, -1, "Conventions")
t <- CFtime(att.get.nc(nc, "time", "units"), 
            att.get.nc(nc, "time", "calendar"), 
            var.get.nc(nc, "time"))

library(ncdf4)
nc <- nc_open(fn)
nc_att_get(nc, 0, "Conventions")
t <- CFtime(nc$dim$time$units, 
            nc$dim$time$calendar, 
            nc$dim$time$vals)
```

The character representations of the time series can be easily generated:

```{r}
dates <- t$as_timestamp(format = "date")
dates[1:10]
```

...as well as the range of the time series:

```{r}
t$range()
```

Note that in this latter case, if any of the timestamps in the time series have a time that is
other than `00:00:00` then the time of the extremes of the time series is also
displayed. This is a common occurrence because the CF Metadata Conventions
prescribe that the middle of the time period (month, day, etc) is recorded, which 
for months with 31 days would be something like `2005-01-15T12:00:00`.

## Supporting processing of climate projection data

When working with high resolution climate projection data, typically at a "day"
resolution, one of the processing steps would be to aggregate the data to some
lower resolution such as a dekad (10-day period), a month or a meteorological 
season, and then compute a derivative value such as the dekadal sum of precipitation,
monthly minimum/maximum daily temperature, or seasonal average daily short-wave
irradiance.

It is also possible to create factors for multiple "eras" in one go. This greatly
reduces programming effort if you want to calculate anomalies over multiple future
periods. A complete example is provided in the vignette 
["Processing climate projection data"](Processing.html).

It is easy to generate the factors that you need once you have a `CFTime` instance 
prepared:

```{r}
# Create a dekad factor for the whole `t` time series that was created above
f_k <- t$factor("dekad")
str(f_k)

# Create monthly factors for a baseline era and early, mid and late 21st century eras
baseline <- t$factor(era = 1991:2020)
future <- t$factor(era = list(early = 2021:2040, mid = 2041:2060, late = 2061:2080))
str(future)
```

For the "era" version, there are two interesting things to note here:

 - The eras do not have to coincide with the boundaries of the time series. In the 
 example above, the time series starts in 2015, while the baseline era is from 1991.
 Obviously, the number of time steps from the time series that then fall within 
 this era will then be reduced.
 - The factor is always of the same length as the time series, with `NA` values
 where the time series values are not falling in the era. This ensures that the
 factor is compatible with the data set which the time series describes, such
 that functions like `tapply()` will not throw an error.
 
There are six periods defined for factoring:

 - `year`, to summarize data to yearly timescales
 - `season`, the meteorological seasons. Note that the month of December will be
 added to the months of January and February of the following year, so the date 
 "2020-12-01" yields the factor value "2021S1".
 - `quarter`, the standard quarters of the year.
 - `month`, monthly summaries, the default period.
 - `dekad`, 10-day period. Each month is subdivided in dekads as follows: (1) 
 days 01 - 10; (2) days 11 - 20; (3) remainder of the month.
 - `day`, to summarize sub-daily data.

##### New "time" dimension

A `CFTime` instance describes the "time" dimension of an associated data set. When
you process that dimension of the data set using `CFTime$factor()` or another method
to filter or otherwise subset the "time" dimension, the resulting data set will
have a different "time" dimension. To associate a proper `CFTime` instance with
your processing result, the methods in this package return that `CFTime` instance
as an attribute:

```{r}
  (new_time <- attr(f_k, "CFTime"))
```

In the vignette ["Processing climate projection data"](Processing.html) is a
fully worked out example of this.

##### Incomplete time series

You can test if your time series is complete with the function `is_complete()`.
A time series is considered complete if the time steps between the two extreme 
values are equally spaced. There is a "fuzzy" assessment of completeness for
time series with a datum unit of "days" or smaller where the time steps are
months or years apart - these have different lengths in days in different months 
or years (e.g. a leap year).

If your time series is incomplete, for instance because it has missing time steps, 
you should recognize that in your further processing. As an example, you might 
want to filter out months that have fewer than 90% of daily data from further 
processing or apply weights based on the actual coverage. 

```{r}
# Is the time series complete?
is_complete(t)

# How many time units fit in a factor level?
t$factor_units(baseline)

# What's the absolute and relative coverage of our time series
t$factor_coverage(baseline, "absolute")
t$factor_coverage(baseline, "relative")
```

The time series is complete but coverage of the baseline era is only 
20%! Recall that the time series starts in 2015 while the baseline period in the
factor is for `1991:2020` so that's only 6 years of time series data out of 30 
years of the baseline factor.

An artificial example of missing data:

```{r}
# 4 years of data on a `365_day` calendar, keep 80% of values
n <- 365 * 4
cov <- 0.8
offsets <- sample(0:(n-1), n * cov)

(t <- CFtime("days since 2020-01-01", "365_day", offsets))
# Note that there are about 1.25 days between observations

mon <- t$factor("month")
t$factor_coverage(mon, "absolute")
t$factor_coverage(mon, "relative")
```

Keep in mind, though, that there are data sets where the time unit is lower than
the intended resolution of the data. Since the CF conventions recommend that the
coarsest time unit is "day", many files with monthly data sets have a definition
like `days since 2016-01-01` with offset values for the middle of the month like
`15, 44, 74, 104, ...`. Even in these scenarios you can verify that your data 
set is complete with the function `CFcomplete()`.

## CFtime and POSIXt

The CF Metadata Conventions supports 11 different calendars. None of these are 
fully compatible with POSIXt, the basis for timekeeping on virtually all 
computers. The reason for this is that POSIXt does not consider leap seconds
(just like the `tai` calendar) but computer clocks are periodically synchronized
using Network Time Protocol servers that report UTC time. The problem is easily
demonstrated:

```{r}
# 1970-01-01 is the origin of POSIXt
difftime(as.POSIXct("2024-01-01"), as.POSIXct("1970-01-01"), units = "sec")

# Leap seconds in UTC
.leap.seconds
```

`difftime()` is off by 27 seconds, the number of leap seconds in UTC since their
introduction in 1972. Your computer may have the correct time based on UTC, but
calculations over periods that include leap seconds are always off by a number 
of seconds.

##### Duh!

If 27 seconds is of no concern to you or your application - perhaps your data 
has a daily resolution - then you can safely forget about the leap seconds in
several of the calendars, in particular `standard` (for periods after 1582-10-15),
`proleptic_gregorian` and `tai`. The `utc` calendar does account for leap
seconds so consider if you should use that - this is the only calendar that 
considers leap seconds in calculation. These calendars support the generation of
timestamps in POSIXct with the `as_timestamp()` function but note the potential
for a discrepancy due to the presence of leap seconds.

##### When seconds count

If second accuracy is of concern, then you should carefully consider the time
keeping in the source of your data and use a matching calendar. The `utc`
calendar is a sensible option if your equipment synchronizes time with an NTP 
server or a computer that does so. Even then you should ensure that time is
accurate after a new leap second is introduced.

##### Bigger problems

The other calendars have discrepancies with POSIXt that are much larger, namely
one or more days. These calendars do not support POSIXct timestamps and an error
will be thrown if you try. If you really want the timestamps in POSIXct then you
can generate the timestamps as character strings using this package, and then
convert to a `POSIXct` or `Date` using the available R tools. Converting time 
series using these incompatible calendars to `POSIXct` or `Date` is likely
to produce problems. This is most pronounced for the `360_day` calendar:

```{r}
# Days in January and February
t <- CFtime("days since 2023-01-01", "360_day", 0:59)
ts_days <- t$as_timestamp("date")
as.Date(ts_days)
```

31 January is missing from the vector of `Date`s because the `360_day` calendar 
does not include it and 29 and 30 February are `NA`s because POSIXt rejects them.
This will produce problems later on when processing your data.

The general advice is therefore: **do not convert CFTime objects to Date objects**
unless you are sure that the `CFTime` object uses a POSIXt-compatible calendar.

The degree of incompatibility for the various calendars is as follows:

 - `standard`: Only valid for periods after 1582-10-15. The preceeding period 
 uses the Julian calendar.
 - `julian`: Every fouth year is a leap year. Dates like `2100-02-29` and
 `2200-02-29` are valid.
 - `365_day` or `noleap`: No leap year exists. `2020-02-29` does not occur.
 - `366_day` or `all_leap`: All years are leap years.
 - `360_day`: All months have 30 days in every year. This means that 31 January,
 March, May, July, August, October and December never occur, while 29 and 30
 February occur in every year.

##### So how do I compare climate projection data with different calendars?

One reason to convert the time dimension from different climate projection data
sets is to be able to compare the data from different models and produce a
multi-model ensemble. The correct procedure to do this is to first calculate 
**for each data set individually** the property of interest (e.g. average daily 
rainfall per month anomaly for some future period relative to a baseline period), 
which will typically involve aggregation to a lower resolution (such as from 
daily data to monthly averages), and only then combine the aggregate data from 
multiple data sets to compute statistically interesting properties (such as 
average among models and standard deviation, etc).

Once data is aggregated from daily or higher-resolution values to a lower
temporal resolution - such as a "month" - the different calendars no longer 
matter (although if you do need to convert averaged data (e.g. average daily
precipitation in a month) to absolute data (e.g. precipitation per month) you 
should use `CFfactor_units()` to make sure that you use the correct scaling 
factor).

Otherwise, there really shouldn't be any reason to convert the time series in the
data files to `Date`s. Climate projection data is virtually never compared on a 
day-to-day basis between different models and neither does complex date arithmetic 
make much sense (such as adding intervals) - `CFtime` can support basic arithmetic
by manipulation the offsets of the `CFTime` object. The character representations
that are produced are perfectly fine to use for `dimnames()` on an array or as
`rownames()` in a `data.frame` and these also support basic logical operations
such as `"2023-02-30" < "2023-03-01"`. So ask yourself, do you really need `Date`s
when working with unprocessed climate projection data? (If so, [open an issue on 
GitHub](https://github.com/pvanlaake/CFtime/issues)).

A complete example of creating a multi-model ensemble is provided in the vignette 
["Processing climate projection data"](Processing.html).

## Final observations

 - This package is intended to facilitate processing of climate projection
 data. It is not a full implementation of the CF Metadata Conventions "time"
 component.
 - In parsing and deparsing of offsets and timestamps, data is rounded to
 3 digits of precision of the unit of the calendar. When using a description of 
 time that is very different from the calendar unit, this may lead to some loss 
 of precision due to rounding errors. For instance, if milli-second precision is 
 required, use a unit of "seconds". The authors have no knowledge of published
 climate projection data that requires milli-second precision so for the
 intended use of the package this issue is marginal.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{bounds}
\alias{bounds}
\alias{bounds<-}
\title{Bounds of the time offsets}
\usage{
bounds(x, format)

bounds(x) <- value
}
\arguments{
\item{x}{A \code{CFTime} instance.}

\item{format}{Optional. A single string with format specifiers, see
\code{\link[=format]{format()}} for details.}

\item{value}{A \code{matrix} (or \code{array}) with dimensions (2, length(offsets))
giving the lower (first row) and higher (second row) bounds of each offset
(this is the format that the CF Metadata Conventions uses for storage in
netCDF files). Use \code{NULL} to unset any previously set bounds.}
}
\value{
If bounds have been set, an array of bounds values with dimensions
\verb{(2, length(offsets))}. The first row gives the lower bound, the second row
the upper bound, with each column representing an offset of \code{x}. If the
\code{format} argument is specified, the bounds values are returned as strings
according to the format. \code{NULL} when no bounds have been set.
}
\description{
CF-compliant netCDF files store time information as a single offset value for
each step along the dimension, typically centered on the valid interval of
the data (e.g. 12-noon for day data). Optionally, the lower and upper values
of the valid interval are stored in a so-called "bounds" variable, as an
array with two rows (lower and higher value) and a column for each offset.
With function \verb{bounds()<-} those bounds can be set for a \code{CFTime} instance.
The bounds can be retrieved with the \code{bounds()} function.
}
\examples{
t <- CFtime("days since 2024-01-01", "standard", seq(0.5, by = 1, length.out = 366))
as_timestamp(t)[1:3]
bounds(t) <- rbind(0:365, 1:366)
bounds(t)[, 1:3]
bounds(t, "\%d-\%b-\%Y")[, 1:3]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{==.CFTime}
\alias{==.CFTime}
\alias{CFtime-equivalent}
\title{Equivalence of CFTime objects}
\usage{
\method{==}{CFTime}(e1, e2)
}
\arguments{
\item{e1, e2}{Instances of the \code{CFTime} class.}
}
\value{
\code{TRUE} if the \code{CFTime} objects are equivalent, \code{FALSE} otherwise.
}
\description{
This operator can be used to test if two \link{CFTime} objects represent the same
CF-convention time coordinates. Two \code{CFTime} objects are considered equivalent
if they have an equivalent calendar and the same offsets.
}
\examples{
e1 <- CFtime("days since 1850-01-01", "gregorian", 0:364)
e2 <- CFtime("days since 1850-01-01 00:00:00", "standard", 0:364)
e1 == e2
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{range.CFTime}
\alias{range.CFTime}
\title{Extreme time series values}
\usage{
\method{range}{CFTime}(x, format = "", bounds = FALSE, ..., na.rm = FALSE)
}
\arguments{
\item{x}{An instance of the \link{CFTime} class.}

\item{format}{A character string with format specifiers, optional. If it is
missing or an empty string, the most economical ISO8601 format is chosen:
"date" when no time information is present in \code{x}, "timestamp" otherwise.
Otherwise a suitable format specifier can be provided.}

\item{bounds}{Logical to indicate if the extremes from the bounds should be
used, if set. Defaults to \code{FALSE}.}

\item{...}{Ignored.}

\item{na.rm}{Ignored.}
}
\value{
Vector of two character representations of the extremes of the time
series.
}
\description{
Character representation of the extreme values in the time series.
}
\examples{
cf <- CFtime("days since 1850-01-01", "julian", 0:364)
range(cf)
range(cf, "\%Y-\%b-\%e")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{month_days}
\alias{month_days}
\title{Return the number of days in a month given a certain CF calendar}
\usage{
month_days(t, x = NULL)
}
\arguments{
\item{t}{The \code{CFTime} instance to use.}

\item{x}{character. An optional vector of dates as strings with format
\code{YYYY-MM-DD}. Any time part will be silently ingested.}
}
\value{
A vector indicating the number of days in each month for the vector
of dates supplied as argument \code{x}. Invalidly specified dates will result in
an \code{NA} value. If no dates are supplied, the number of days per month for
the calendar as a vector of length 12.
}
\description{
Given a vector of dates as strings in ISO 8601 or UDUNITS format and a
\link{CFTime} object, this function will return a vector of the same length as the
dates, indicating the number of days in the month according to the calendar
specification. If no vector of days is supplied, the function will return an
integer vector of length 12 with the number of days for each month of the
calendar (disregarding the leap day for \code{standard} and \code{julian} calendars).
}
\examples{
dates <- c("2021-11-27", "2021-12-10", "2022-01-14", "2022-02-18")
t <- CFtime("days since 1850-01-01", "standard")
month_days(t, dates)

t <- CFtime("days since 1850-01-01", "360_day")
month_days(t, dates)

t <- CFtime("days since 1850-01-01", "all_leap")
month_days(t, dates)

month_days(t)
}
\seealso{
When working with factors generated by \code{\link[=CFfactor]{CFfactor()}}, it is usually
better to use \code{\link[=CFfactor_units]{CFfactor_units()}} as that will consider leap days for
non-era factors. \code{\link[=CFfactor_units]{CFfactor_units()}} can also work with other time periods
and calendar units, such as "hours per month", or "days per season".
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{definition}
\alias{definition}
\alias{properties}
\alias{calendar}
\alias{unit}
\alias{origin}
\alias{timezone}
\alias{offsets}
\alias{resolution}
\title{Properties of a CFTime object}
\usage{
definition(t)

calendar(t)

unit(t)

origin(t)

timezone(t)

offsets(t)

resolution(t)
}
\arguments{
\item{t}{An instance of \code{CFTime}.}
}
\value{
\code{calendar()} and \code{unit()} return a character string.
\code{origin()} returns a data frame of timestamp elements with a single row
of data. \code{timezone()} returns the calendar time zone as a character
string. \code{offsets()} returns a vector of offsets or \code{NULL} if no offsets
have been set.
}
\description{
These functions return the properties of an instance of the
\link{CFTime} class. The properties are all read-only, but offsets can be added
using the \code{+} operator.
}
\section{Functions}{
\itemize{
\item \code{definition()}: The definition string of the \code{CFTime} instance.

\item \code{calendar()}: The calendar of the \code{CFTime} instance.

\item \code{unit()}: The unit of the \code{CFTime} instance.

\item \code{origin()}: The origin of the \code{CFTime} instance in timestamp elements.

\item \code{timezone()}: The time zone of the calendar of the \code{CFTime} instance as a character string.

\item \code{offsets()}: The offsets of the \code{CFTime} instance as a numeric vector.

\item \code{resolution()}: The average separation between the offsets in the \code{CFTime} instance.

}}
\examples{
t <- CFtime("days since 1850-01-01", "julian", 0:364)
definition(t)
calendar(t)
unit(t)
timezone(t)
origin(t)
offsets(t)
resolution(t)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CFCalendar360.R
\docType{class}
\name{CFCalendar360}
\alias{CFCalendar360}
\title{360-day CF calendar}
\description{
This class represents a CF calendar of 360 days per year, evenly
divided over 12 months of 30 days. This calendar is obviously not
compatible with the standard POSIXt calendar.

This calendar supports dates before year 1 and includes the year 0.
}
\section{Super class}{
\code{\link[CFtime:CFCalendar]{CFtime::CFCalendar}} -> \code{CFCalendar360}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-CFCalendar360-new}{\code{CFCalendar360$new()}}
\item \href{#method-CFCalendar360-valid_days}{\code{CFCalendar360$valid_days()}}
\item \href{#method-CFCalendar360-month_days}{\code{CFCalendar360$month_days()}}
\item \href{#method-CFCalendar360-leap_year}{\code{CFCalendar360$leap_year()}}
\item \href{#method-CFCalendar360-doy}{\code{CFCalendar360$doy()}}
\item \href{#method-CFCalendar360-date2offset}{\code{CFCalendar360$date2offset()}}
\item \href{#method-CFCalendar360-offset2date}{\code{CFCalendar360$offset2date()}}
\item \href{#method-CFCalendar360-clone}{\code{CFCalendar360$clone()}}
}
}
\if{html}{\out{
<details><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="POSIX_compatible"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-POSIX_compatible'><code>CFtime::CFCalendar$POSIX_compatible()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="add_day"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-add_day'><code>CFtime::CFCalendar$add_day()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="is_compatible"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-is_compatible'><code>CFtime::CFCalendar$is_compatible()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="is_equivalent"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-is_equivalent'><code>CFtime::CFCalendar$is_equivalent()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="offsets2time"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-offsets2time'><code>CFtime::CFCalendar$offsets2time()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="parse"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-parse'><code>CFtime::CFCalendar$parse()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="print"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-print'><code>CFtime::CFCalendar$print()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar360-new"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar360-new}{}}}
\subsection{Method \code{new()}}{
Create a new CF calendar.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar360$new(nm, definition)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{nm}}{The name of the calendar. This must be "360_day". This argument
is superfluous but maintained to be consistent with the initialization
methods of the parent and sibling classes.}

\item{\code{definition}}{The string that defines the units and the origin, as
per the CF Metadata Conventions.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new instance of this class.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar360-valid_days"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar360-valid_days}{}}}
\subsection{Method \code{valid_days()}}{
Indicate which of the supplied dates are valid.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar360$valid_days(ymd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame} with dates parsed into their parts in columns
\code{year}, \code{month} and \code{day}. Any other columns are disregarded.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Logical vector with the same length as argument \code{ymd} has rows
with \code{TRUE} for valid days and \code{FALSE} for invalid days, or \code{NA} where
the row in argument \code{ymd} has \code{NA} values.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar360-month_days"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar360-month_days}{}}}
\subsection{Method \code{month_days()}}{
Determine the number of days in the month of the calendar.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar360$month_days(ymd = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame} with dates parsed into their parts in columns
\code{year}, \code{month} and \code{day}. Any other columns are disregarded.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A vector indicating the number of days in each month for the
dates supplied as argument \code{ymd}. If no dates are supplied, the number
of days per month for the calendar as a vector of length 12.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar360-leap_year"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar360-leap_year}{}}}
\subsection{Method \code{leap_year()}}{
Indicate which years are leap years.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar360$leap_year(yr)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{yr}}{Integer vector of years to test.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Logical vector with the same length as argument \code{yr}. Since this
calendar does not use leap days, all values will be \code{FALSE}, or \code{NA}
where argument \code{yr} is \code{NA}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar360-doy"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar360-doy}{}}}
\subsection{Method \code{doy()}}{
Retrieve the day-of-year number for the supplied
\code{data.frame} of year, month and day, using this calendar.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar360$doy(ymd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame} with dates parsed into their parts in columns
\code{year}, \code{month} and \code{day}. Any other columns are disregarded.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Integer vector as long as argument \code{ymd} has rows with the
day-of-year for each row.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar360-date2offset"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar360-date2offset}{}}}
\subsection{Method \code{date2offset()}}{
Calculate difference in days between a \code{data.frame} of time
parts and the origin.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar360$date2offset(x)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{x}}{\code{data.frame}. Dates to calculate the difference for.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Integer vector of a length equal to the number of rows in
argument \code{x} indicating the number of days between \code{x} and the \code{origin},
or \code{NA} for rows in \code{x} with \code{NA} values.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar360-offset2date"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar360-offset2date}{}}}
\subsection{Method \code{offset2date()}}{
Calculate date parts from day differences from the origin.
This only deals with days as these are impacted by the calendar.
Hour-minute-second timestamp parts are handled in \link{CFCalendar}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar360$offset2date(x)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{x}}{Integer vector of days to add to the origin.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{data.frame} with columns 'year', 'month' and 'day' and as many
rows as the length of vector \code{x}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar360-clone"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar360-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar360$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{slice}
\alias{slice}
\title{Which time steps fall within extreme values}
\usage{
slice(x, extremes, rightmost.closed = FALSE)
}
\arguments{
\item{x}{The \code{CFTime} instance to operate on.}

\item{extremes}{Character vector of timestamps that represent the time period
of interest. The extreme values are selected. Badly formatted timestamps
are silently dropped.}

\item{rightmost.closed}{Is the right side closed, i.e. included in the
result? Default is \code{FALSE}. A specification of \verb{c("2022-01-01", "2023-01-01)} will thus include all time steps that fall in the year 2022
when \code{closed = FALSE} but include \code{2023-01-01} if that exact value is
present in the time series.}
}
\value{
A logical vector with a length equal to the number of time steps in
\code{x} with values \code{TRUE} for those time steps that fall between the extreme
values, \code{FALSE} otherwise.

An attribute 'CFTime' will have the same definition as \code{x} but with offsets
corresponding to the time steps falling between the two extremes. If there
are no values between the extremes, the attribute is \code{NULL}.
}
\description{
Given a vector of character timestamps, return a logical vector of a length
equal to the number of time steps in the time series with values \code{TRUE} for
those time steps that fall between the two extreme values of the vector
values, \code{FALSE} otherwise.
}
\details{
If bounds were set these will be preserved.
}
\examples{
t <- CFtime("hours since 2023-01-01 00:00:00", "standard", 0:23)
slice(t, c("2022-12-01", "2023-01-01 03:00"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{slab}
\alias{slab}
\title{Which time steps fall within two extreme values}
\usage{
slab(x, extremes, rightmost.closed = FALSE)
}
\arguments{
\item{x, extremes, rightmost.closed}{See \code{slice()}.}
}
\value{
See \code{slice()}.
}
\description{
Avoid using this function, use \code{\link[=slice]{slice()}} instead. This function will be
deprecated in the near future.
}
\examples{
t <- CFtime("hours since 2023-01-01 00:00:00", "standard", 0:23)
slab(t, c("2022-12-01", "2023-01-01 03:00"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{cut.CFTime}
\alias{cut.CFTime}
\alias{cut}
\title{Create a factor for a \code{CFTime} instance}
\usage{
\method{cut}{CFTime}(x, breaks, ...)
}
\arguments{
\item{x}{An instance of \code{CFTime}.}

\item{breaks}{A character string of a factor period (see \code{\link[=CFfactor]{CFfactor()}} for a
description), or a character vector of timestamps that conform to the
calendar of \code{x}, with a length of at least 2. Timestamps must be given in
ISO8601 format, e.g. "2024-04-10 21:31:43".}

\item{...}{Ignored.}
}
\value{
A factor with levels according to the \code{breaks} argument, with
attributes 'period', 'era' and 'CFTime'. When \code{breaks} is a factor
period, attribute 'period' has that value, otherwise it is '"day"'. When
\code{breaks} is a character vector of timestamps, attribute 'CFTime' holds an
instance of \code{CFTime} that has the same definition as \code{x}, but with (ordered)
offsets generated from the \code{breaks}. Attribute 'era' is always -1.
}
\description{
Method for \code{\link[base:cut]{base::cut()}} applied to \link{CFTime} objects.
}
\details{
When \code{breaks} is one of \verb{"year", "season", "quarter", "month", "dekad", "day"} a factor is generated like by \code{\link[=CFfactor]{CFfactor()}}.

When \code{breaks} is a vector of character timestamps a factor is produced with a
level for every interval between timestamps. The last timestamp, therefore,
is only used to close the interval started by the pen-ultimate timestamp -
use a distant timestamp (e.g. \code{range(x)[2]}) to ensure that all offsets to
the end of the \code{CFTime} time series are included, if so desired. The last
timestamp will become the upper bound in the \code{CFTime} instance that is
returned as an attribute to this function so a sensible value for the last
timestamp is advisable.

This method works similar to \code{\link[base:cut.POSIXt]{base::cut.POSIXt()}} but there are some
differences in the arguments: for \code{breaks} the set of options is different
and no preceding integer is allowed, \code{labels} are always assigned using
values of \code{breaks}, and the interval is always left-closed.
}
\examples{
x <- CFtime("days since 2021-01-01", "365_day", 0:729)
breaks <- c("2022-02-01", "2021-12-01", "2023-01-01")
cut(x, breaks)
}
\seealso{
\code{\link[=CFfactor]{CFfactor()}} produces a factor for several fixed periods, including
for eras.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CFCalendarNone.R
\docType{class}
\name{CFCalendarNone}
\alias{CFCalendarNone}
\title{CF calendar with no annual cycle}
\description{
This class represents a CF calendar with no annual cycle. All
datetimes in the calendar are the same. This is useful only for repeated
experiments simulating a fixed time in the year.
}
\section{Super class}{
\code{\link[CFtime:CFCalendar]{CFtime::CFCalendar}} -> \code{CFCalendarNone}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-CFCalendarNone-new}{\code{CFCalendarNone$new()}}
\item \href{#method-CFCalendarNone-valid_days}{\code{CFCalendarNone$valid_days()}}
\item \href{#method-CFCalendarNone-month_days}{\code{CFCalendarNone$month_days()}}
\item \href{#method-CFCalendarNone-leap_year}{\code{CFCalendarNone$leap_year()}}
\item \href{#method-CFCalendarNone-doy}{\code{CFCalendarNone$doy()}}
\item \href{#method-CFCalendarNone-date2offset}{\code{CFCalendarNone$date2offset()}}
\item \href{#method-CFCalendarNone-offset2date}{\code{CFCalendarNone$offset2date()}}
\item \href{#method-CFCalendarNone-clone}{\code{CFCalendarNone$clone()}}
}
}
\if{html}{\out{
<details><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="POSIX_compatible"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-POSIX_compatible'><code>CFtime::CFCalendar$POSIX_compatible()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="add_day"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-add_day'><code>CFtime::CFCalendar$add_day()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="is_compatible"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-is_compatible'><code>CFtime::CFCalendar$is_compatible()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="is_equivalent"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-is_equivalent'><code>CFtime::CFCalendar$is_equivalent()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="offsets2time"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-offsets2time'><code>CFtime::CFCalendar$offsets2time()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="parse"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-parse'><code>CFtime::CFCalendar$parse()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="print"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-print'><code>CFtime::CFCalendar$print()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarNone-new"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarNone-new}{}}}
\subsection{Method \code{new()}}{
Create a new CF calendar.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarNone$new(nm, definition)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{nm}}{The name of the calendar. This must be "none". This argument
is superfluous but maintained to be consistent with the initialization
methods of the parent and sibling classes.}

\item{\code{definition}}{The string that defines the units and the origin, as
per the CF Metadata Conventions.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new instance of this class.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarNone-valid_days"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarNone-valid_days}{}}}
\subsection{Method \code{valid_days()}}{
Indicate which of the supplied dates are valid. In this
calendar only one date is valid, namely the one that the calendar is
initialized with.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarNone$valid_days(ymd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame} with dates parsed into their parts in columns
\code{year}, \code{month} and \code{day}. Any other columns are disregarded.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Logical vector with the same length as argument \code{ymd} has rows
with \code{TRUE} for valid days and \code{FALSE} for invalid days, or \code{NA} where
the row in argument \code{ymd} has \code{NA} values.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarNone-month_days"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarNone-month_days}{}}}
\subsection{Method \code{month_days()}}{
Determine the number of days in the month of the calendar.
This always returns a vector of \code{NA} values because this method has no
meaning for this class.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarNone$month_days(ymd = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame} with dates parsed into their parts in columns
\code{year}, \code{month} and \code{day}. Any other columns are disregarded.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A vector with \code{NA} values for the dates supplied as argument
\code{ymd}. If no dates are supplied, a vector of \code{NA} values of length 12.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarNone-leap_year"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarNone-leap_year}{}}}
\subsection{Method \code{leap_year()}}{
Indicate which years are leap years. This always returns a
vector of \code{NA} values because this method has no meaning for this
class.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarNone$leap_year(yr)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{yr}}{Integer vector of years to test.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Vector with the same length as argument \code{yr} with \code{NA} values.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarNone-doy"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarNone-doy}{}}}
\subsection{Method \code{doy()}}{
Retrieve the day-of-year number of the fixed data of this
calendar, using the Gregorian calendar.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarNone$doy(ymd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame} with fixed date parsed into its parts in columns
\code{year}, \code{month} and \code{day}. Any other columns are disregarded.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Integer with the day-of-year of the fixed date for this calendar.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarNone-date2offset"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarNone-date2offset}{}}}
\subsection{Method \code{date2offset()}}{
Calculate difference in days between a \code{data.frame} of time
parts and the origin. The difference is always 0, given that
this is not a true calendar and there can be no calculations to
determine any difference.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarNone$date2offset(x)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{x}}{\code{data.frame}. Dates to calculate the difference for.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Integer vector of a length equal to the number of rows in
argument \code{x} indicating 0 for the days that equal the origin
of the calendar, or \code{NA} otherwise.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarNone-offset2date"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarNone-offset2date}{}}}
\subsection{Method \code{offset2date()}}{
Calculate date parts from day differences from the origin.
This always returns 0's for the year, month and day as other values are
not valid in this calendar. Hour-minute-second datetime parts are
handled in \link{CFCalendar}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarNone$offset2date(x)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{x}}{Integer vector of days to add to the origin.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{data.frame} with columns 'year', 'month' and 'day' and as many
rows as the length of vector \code{x}. Rows with values of \code{x} other than 0
will return the values for the origin of the calendar nonetheless, in
accordance with the CF Metadata Conventions.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarNone-clone"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarNone-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarNone$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{+.CFTime}
\alias{+.CFTime}
\alias{CFtime-merge}
\alias{+.CFClimatology}
\title{Extend a CFTime or CFClimatology object}
\usage{
\method{+}{CFTime}(e1, e2)

\method{+}{CFClimatology}(c1, c2)
}
\arguments{
\item{e1}{Instance of the \code{CFTime} class.}

\item{e2}{Instance of the \code{CFTime} class with a calendar compatible with that
of argument \code{e1}. This argument may also be a numeric vector with offsets
from the origin of argument \code{e1}, or a vector of \code{character} timestamps in
ISO8601 or UDUNITS format.}

\item{c1, c2}{Instances of the \code{CFClimatology} class, having compatible
calendars.}
}
\value{
A \code{CFTime} object with the offsets of argument \code{e1} extended by the
values from argument \code{e2}.

A \code{CFClimatology} object with the offsets of argument \code{c1} extended
by the values from argument \code{c2}.
}
\description{
A \link{CFTime} or \link{CFClimatology} instance can be extended with this operator,
using values from another \code{CFTime} or \code{CFClimatology} instance, respectively,
or a vector of numeric offsets or character timestamps. If the values come
from another \code{CFTime} or \code{CFClimatology} instance, the calendars of the two
instances must be compatible. If the calendars of the instances are not
compatible, an error is thrown.
}
\details{
The resulting \code{CFTime} or \code{CFClimatology} instance will have the offsets of
the original instance, appended with offsets from argument \code{e2} in the order
that they are specified. If the new sequence of offsets is not monotonically
increasing a warning is generated (the COARDS metadata convention requires
offsets to be monotonically increasing).

There is no reordering or removal of duplicates. This is because the time
series are usually associated with a data set and the correspondence between
the data in the files and the instance is thus preserved. When merging the
data sets described by this time series, the order must be identical to the
merging here.

Note that when adding multiple vectors of offsets to a \code{CFTime} instance, it
is more efficient to first concatenate the vectors and then do a final
addition to the \code{CFTime} instance. So avoid \code{CFtime(definition, calendar, e1) + CFtime(definition, calendar, e2) + CFtime(definition, calendar, e3) + ...}
but rather do \code{CFtime(definition, calendar) + c(e1, e2, e3, ...)}. It is the
responsibility of the operator to ensure that the offsets of the different
data sets are in reference to the same calendar.

Note also that \code{RNetCDF} and \code{ncdf4} packages both return the values of the
"time" dimension as a 1-dimensional array. You have to de-class the array to
a vector before adding offsets to an existing \code{CFtime} instance.

If both operands to this operator are \code{CFTime} or \code{CFClimatology} instances
and both have bounds set, then their bounds will be preserved in the result.
}
\examples{
e1 <- CFTime$new("days since 1850-01-01", "gregorian", 0:364)
e2 <- CFTime$new("days since 1850-01-01 00:00:00", "standard", 365:729)
e1 + e2
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CFtime.R
\name{str.CFTime}
\alias{str.CFTime}
\title{Compact display of a CFTime instance}
\usage{
\method{str}{CFTime}(object, ...)
}
\arguments{
\item{object}{A \code{CFTime} instance.}

\item{...}{Ignored.}
}
\description{
Compact display of a CFTime instance
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{CFtime-function}
\alias{CFtime-function}
\alias{CFtime}
\title{Create a CFTime object}
\usage{
CFtime(definition, calendar = "standard", offsets, from, to, by, length.out)
}
\arguments{
\item{definition}{A character string describing the time coordinate.}

\item{calendar}{A character string describing the calendar to use with the
time dimension definition string. Default value is "standard".}

\item{offsets}{Numeric or character vector, optional. When numeric, a vector
of offsets from the origin in the time series. When a character vector,
timestamps in ISO8601 or UDUNITS format.}

\item{from, to}{Optional. Character timestamps in ISO8601 or UDUNITS format.
When \code{from} is given, a sequence of timestamps is generated with \code{from} as
the starting timestamp. Either argument \code{to} or \code{length.out} must be
provided as well. Ignored when argument \code{offsets} is not \code{NULL}.}

\item{by}{Optional. A single character string representing a time interval,
composed of a number and a time unit separated by a space, such as in "6
hours". When argument \code{from} is supplied, argument \code{by} will be the
separation between successive timestamps. Note that the time unit in the
string does not have to be the same as the unit in the \code{definition}
argument but it must be an allowable unit of time. Time interval units of
"months" and "years" are strongly discouraged unless the same time unit is
used in the \code{definition} argument - in all other cases there is a loss of
precision due to the ambiguity in the time units.}

\item{length.out}{Optional. An numeric value that indicates the lengths of
the time series to generate, rounded up if fractional. Ignored when
argument \code{to} is provided.}
}
\value{
An instance of the \code{CFTime} class.
}
\description{
This function creates an instance of the \link{CFTime} class. The arguments to the
call are typically read from a CF-compliant data file with climatological
observations or climate projections. Specification of arguments can also be
made manually in a variety of combinations.
}
\details{
A time series can also be constructed like a sequence. In this case argument
\code{offsets} should be \code{NULL} or missing and arguments \code{from} and \code{by} provided,
with either of arguments \code{to} or \code{length.out} indicating the end of the time
series. Arguments should be named to avoid ambiguity.
}
\examples{
# Using numeric offset values - this is how a netCDF file works
CFtime("days since 1850-01-01", "julian", 0:364)

# A time object with a single defined timestamp
CFtime("hours since 2023-01-01", "360_day", "2023-01-30T23:00")

# A time series from a sequence with an end point
CFtime("days since 2023-01-01", from = "2020-01-01", to = "2023-12-31", by = "12 days")

# A time series from a sequence with a specified length
CFtime("days since 2023-01-01", from = "2020-01-01T03:00:00", by = "6 hr", length.out = 31 * 4)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{CFfactor_coverage}
\alias{CFfactor_coverage}
\title{Coverage of time elements for each factor level}
\usage{
CFfactor_coverage(t, f, coverage = "absolute")
}
\arguments{
\item{t}{An instance of \link{CFTime}.}

\item{f}{factor or list. A factor or a list of factors derived from the
parameter \code{t}. The factor or list thereof should generally be generated by
the function \code{\link[=CFfactor]{CFfactor()}}.}

\item{coverage}{"absolute" or "relative".}
}
\value{
If \code{f} is a factor, a numeric vector with a length equal to the
number of levels in the factor, indicating the number of units from the
time series in \code{t} contained in each level of the factor when
\code{coverage = "absolute"} or the proportion of units present relative to the
maximum number when \code{coverage = "relative"}. If \code{f} is a list of factors, a
list with each element a numeric vector as above.
}
\description{
This function calculates the number of time elements, or the relative
coverage, in each level of a factor generated by \code{\link[=CFfactor]{CFfactor()}}.
}
\examples{
t <- CFtime("days since 2001-01-01", "365_day", 0:364)
f <- CFfactor(t, "dekad")
CFfactor_coverage(t, f, "absolute")
}

% Generated by roxygen2: do not edit by hand % Please edit documentation in R/CFClimatology.R \docType{class} \name{CFClimatology} \alias{CFClimatology} \title{CFClimatology class} \description{ This class represents a climatological time coordinate, both its "calendar" and its "climatology" bounds values. The "calendar" portion is managed by the \link{CFTime} class, from which this class is inherited. This class implements the specific behaviour of the climatological bounds values and includes methods to query the structure of the climatological data. } \references{ https://cf-convention.github.io/Data/cf-conventions/cf-conventions-1.13/cf-conventions.html#climatological-statistics } \section{Super class}{ \code{\link[CFtime:CFTime]{CFtime::CFTime}} -> \code{CFClimatology} } \section{Active bindings}{ \if{html}{\out{
}} \describe{ \item{\code{period}}{(read-only) Character string indicating the period during the year over which the climatology was calculated.} \item{\code{years}}{(read-only) Vector of two integer values indicating the years over which the climatology was calculated.} \item{\code{friendlyClassName}}{Character string with class name for display purposes.} } \if{html}{\out{
}} } \section{Methods}{ \subsection{Public methods}{ \itemize{ \item \href{#method-CFClimatology-new}{\code{CFClimatology$new()}} \item \href{#method-CFClimatology-print}{\code{CFClimatology$print()}} \item \href{#method-CFClimatology-clone}{\code{CFClimatology$clone()}} } } \if{html}{\out{
Inherited methods

- `CFtime::CFTime$POSIX_compatible()`
- `CFtime::CFTime$as_timestamp()`
- `CFtime::CFTime$copy()`
- `CFtime::CFTime$cut()`
- `CFtime::CFTime$doy()`
- `CFtime::CFTime$equidistant()`
- `CFtime::CFTime$factor()`
- `CFtime::CFTime$factor_coverage()`
- `CFtime::CFTime$factor_units()`
- `CFtime::CFTime$format()`
- `CFtime::CFTime$get_bounds()`
- `CFtime::CFTime$indexOf()`
- `CFtime::CFTime$range()`
- `CFtime::CFTime$set_bounds()`
- `CFtime::CFTime$slice()`
- `CFtime::CFTime$subset()`

}} \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-CFClimatology-new}{}}} \subsection{Method \code{new()}}{ Create a new instance of this class. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{CFClimatology$new(definition, calendar = "standard", offsets, bounds)}\if{html}{\out{
}} } \subsection{Arguments}{ \if{html}{\out{
}} \describe{ \item{\code{definition}}{Character string of the units and origin of the calendar.} \item{\code{calendar}}{Character string of the calendar to use. Must be one of the values permitted by the CF Metadata Conventions. If \code{NULL}, the "standard" calendar will be used.} \item{\code{offsets}}{Numeric or character vector. When numeric, a vector of offsets from the origin in the time series. When a character vector of length 2 or more, timestamps in ISO8601 or UDUNITS format.} \item{\code{bounds}}{The climatological bounds for the offsets. A \code{matrix} with columns for \code{offsets} and low values in the first row, high values in the second row. The bounds will oftentimes overlap or be discontinuous.} } \if{html}{\out{
}} } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-CFClimatology-print}{}}} \subsection{Method \code{print()}}{ Print a summary of the \code{CFClimatology} object to the console. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{CFClimatology$print(...)}\if{html}{\out{
}} } \subsection{Arguments}{ \if{html}{\out{
}} \describe{ \item{\code{...}}{Ignored.} } \if{html}{\out{
}} } \subsection{Returns}{ \code{self} invisibly. } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-CFClimatology-clone}{}}} \subsection{Method \code{clone()}}{ The objects of this class are cloneable with this method. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{CFClimatology$clone(deep = FALSE)}\if{html}{\out{
}} } \subsection{Arguments}{ \if{html}{\out{
}} \describe{ \item{\code{deep}}{Whether to make a deep clone.} } \if{html}{\out{
}} } } }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CFtime-package.R
\docType{package}
\name{CFtime-package}
\alias{CFtime-package}
\title{CFtime: working with CF Metadata Conventions "time" dimensions}
\description{
Support for all calendars as specified in the Climate and Forecast
(CF) Metadata Conventions for climate and forecasting data. The CF Metadata
Conventions is widely used for distributing files with climate observations
or projections, including the Coupled Model Intercomparison Project (CMIP)
data used by climate change scientists and the Intergovernmental Panel on
Climate Change (IPCC). This package specifically allows the user to work
with any of the CF-compliant calendars (many of which are not compliant with
POSIXt). The CF time coordinate is formally defined in the
\href{https://cf-convention.github.io/Data/cf-conventions/cf-conventions-1.13/cf-conventions.html#time-coordinate}{CF Metadata Conventions document}.
}
\details{
This package also supports the creation of a "time" dimension, using class
\code{CFClimatology}, for climatological statistics as defined \href{https://cf-convention.github.io/Data/cf-conventions/cf-conventions-1.13/cf-conventions.html#climatological-statistics}{here}.

The package can create a \link{CFTime} or \link{CFClimatology} instance from scratch or, more commonly, it
can use the dimension attributes and dimension variable values from a netCDF
resource. The package does not actually do any of the reading and the user is
free to use their netCDF package of preference. The recommended package to
use (with any netCDF resources) is \href{https://cran.r-project.org/package=ncdfCF}{ncdfCF}.
\code{ncdfCF} will automatically use this package to manage the "time" dimension
of any netCDF resource. As with this package, it reads and interprets the
attributes of the resource to apply the CF Metadata Conventions, supporting
axes, auxiliary coordinate variables, coordinate reference systems, etc.
Alternatively, for more basic netCDF reading and writing, the two main options
are \href{https://cran.r-project.org/package=RNetCDF}{RNetCDF} and
\href{https://cran.r-project.org/package=ncdf4}{ncdf4}).

\strong{Create, modify, inquire}
\itemize{
\item \code{\link[=CFtime]{CFtime()}}: Create a \link{CFTime} instance
\item \code{\link[=properties]{Properties}} of the \code{CFTime} instance
\item \code{\link[=parse_timestamps]{parse_timestamps()}}: Parse a vector of character timestamps into \code{CFTime} elements
\item \code{\link[=CFtime-equivalent]{Compare}} two \code{CFTime} instances
\item \code{\link[=CFtime-merge]{Merge}} two \code{CFTime} or \code{CFClimatology} instances, or append additional time steps to a \code{CFTime} instance
\item \code{\link[=as_timestamp]{as_timestamp()}} and \code{\link[=format]{format()}}: Generate a vector of character or \code{POSIXct} timestamps from a \code{CFTime} instance
\item \code{\link[=range]{range()}}: Timestamps of the two endpoints in the time series
\item \code{copy()}: Create a copy of a CFTime instance
\item \code{subset()}: Create a new CFTime instance which is a subset by index positions of another CFTime instance
\item \code{\link[=is_complete]{is_complete()}}: Does the \code{CFTime} instance have a complete time series between endpoints?
\item \code{\link[=month_days]{month_days()}}: How many days are there in a month using the calendar of the \code{CFTime} instance?
\item \code{CFTime$doy()}: Retrieve the day-of-year of offsets in the \code{CFTime} instance or from supplied dates.
}

\strong{Factors and coverage}
\itemize{
\item \code{\link[=CFfactor]{CFfactor()}} and \code{\link[=cut]{cut()}}: Create factors for different time periods
\item \code{\link[=CFfactor_units]{CFfactor_units()}}: How many units of time are there in each factor level?
\item \code{\link[=CFfactor_coverage]{CFfactor_coverage()}}: How much data is available for each level of the factor?
}

\strong{Filtering and selection}
\itemize{
\item \code{\link[=slice]{slice()}}: Logical vector of time steps between two extreme points.
\item \code{\link[=indexOf]{indexOf()}}: Index values in the time series of given timestamps, possibly with
fractional part for interpolation.
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://r-cf.github.io/CFtime/}
  \item \url{https://github.com/R-CF/CFtime}
  \item Report bugs at \url{https://github.com/R-CF/CFtime/issues}
}

}
\author{
\strong{Maintainer}: Patrick Van Laake \email{patrick@vanlaake.net} [copyright holder]

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CFCalendar.R
\docType{class}
\name{CFCalendar}
\alias{CFCalendar}
\title{Basic CF calendar}
\description{
This class represents a basic CF calendar. It should not be
instantiated directly; instead, use one of the descendant classes.

This internal class stores the information to represent date and time
values using the CF conventions. An instance is created by the exported
\link{CFTime} class, which also exposes the relevant properties of this class.

The following calendars are supported:

\itemize{
\item \code{\link[=CFCalendarStandard]{gregorian\\standard}}, the international standard calendar for civil use.
\item \code{\link[=CFCalendarProleptic]{proleptic_gregorian}}, the standard calendar but extending before 1582-10-15
when the Gregorian calendar was adopted.
\item \code{\link[=CFCalendarTAI]{tai}}, International Atomic Time clock with dates expressed using the Gregorian calendar.
\item \code{\link[=CFCalendarUTC]{utc}}, Coordinated Universal Time clock with dates expressed using the Gregorian calendar.
\item \code{\link[=CFCalendarJulian]{julian}}, every fourth year is a leap year.
\item \code{\link[=CFCalendar365]{noleap\\365_day}}, all years have 365 days.
\item \code{\link[=CFCalendar366]{all_leap\\366_day}}, all years have 366 days.
\item \code{\link[=CFCalendar360]{360_day}}, all years have 360 days, divided over 12 months of 30 days.
}
}
\references{
https://cf-convention.github.io/Data/cf-conventions/cf-conventions-1.13/cf-conventions.html#calendar
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{name}}{(read-only) Name of the calendar, as per the CF Metadata
Conventions.}

\item{\code{definition}}{(read-only) The string that defines the units and the
origin, as per the CF Metadata Conventions.}

\item{\code{unit}}{(read-only) The numeric id of the unit of the calendar.}

\item{\code{prefix_id}}{(read-only) The index value of the prefix of the time
unit. If the unit does not have a prefix, returns the value 0L.}

\item{\code{origin}}{(read-only) \code{data.frame} with fields for the origin of the
calendar.}

\item{\code{origin_date}}{(read-only) Character string with the date of the
calendar.}

\item{\code{origin_time}}{(read-only) Character string with the time of the
calendar.}

\item{\code{timezone}}{(read-only) Character string with the time zone of the
origin of the calendar.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-CFCalendar-new}{\code{CFCalendar$new()}}
\item \href{#method-CFCalendar-print}{\code{CFCalendar$print()}}
\item \href{#method-CFCalendar-valid_days}{\code{CFCalendar$valid_days()}}
\item \href{#method-CFCalendar-add_day}{\code{CFCalendar$add_day()}}
\item \href{#method-CFCalendar-POSIX_compatible}{\code{CFCalendar$POSIX_compatible()}}
\item \href{#method-CFCalendar-is_compatible}{\code{CFCalendar$is_compatible()}}
\item \href{#method-CFCalendar-is_equivalent}{\code{CFCalendar$is_equivalent()}}
\item \href{#method-CFCalendar-parse}{\code{CFCalendar$parse()}}
\item \href{#method-CFCalendar-doy}{\code{CFCalendar$doy()}}
\item \href{#method-CFCalendar-offsets2time}{\code{CFCalendar$offsets2time()}}
\item \href{#method-CFCalendar-clone}{\code{CFCalendar$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar-new"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar-new}{}}}
\subsection{Method \code{new()}}{
Create a new CF calendar.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar$new(nm, definition)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{nm}}{The name of the calendar. This must follow the CF Metadata
Conventions.}

\item{\code{definition}}{The string that defines the units and the origin, as
per the CF Metadata Conventions.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar-print"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar-print}{}}}
\subsection{Method \code{print()}}{
Print information about the calendar to the console.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Ignored.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{self}, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar-valid_days"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar-valid_days}{}}}
\subsection{Method \code{valid_days()}}{
Indicate which of the supplied dates are valid.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar$valid_days(ymd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame} with dates parsed into their parts in columns
\code{year}, \code{month} and \code{day}. Any other columns are disregarded.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL}. A warning will be generated to the effect that a
descendant class should be used for this method.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar-add_day"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar-add_day}{}}}
\subsection{Method \code{add_day()}}{
Add a day to the supplied dates.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar$add_day(ymd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame} with dates parsed into their parts in columns
\code{year}, \code{month} and \code{day}. Any other columns are disregarded.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{data.frame} like argument \code{ymd} but with a day added for every
row.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar-POSIX_compatible"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar-POSIX_compatible}{}}}
\subsection{Method \code{POSIX_compatible()}}{
Indicate if the time series described using this calendar
can be safely converted to a standard date-time type (\code{POSIXct},
\code{POSIXlt}, \code{Date}).

Only the 'standard' calendar and the 'proleptic_gregorian' calendar
when all dates in the time series are more recent than 1582-10-15
(inclusive) can be safely converted, so this method returns \code{FALSE} by
default to cover the majority of cases.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar$POSIX_compatible(offsets)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{offsets}}{The offsets from the CFtime instance.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{FALSE} by default.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar-is_compatible"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar-is_compatible}{}}}
\subsection{Method \code{is_compatible()}}{
This method tests if the \code{CFCalendar} instance in argument
\code{cal} is compatible with \code{self}, meaning that they are of the same
class and have the same unit and prefix. Calendars "standard", and
"gregorian" are compatible, as are the pairs of "365_day" and
"no_leap", and "366_day" and "all_leap".
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar$is_compatible(cal)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{cal}}{Instance of a descendant of the \code{CFCalendar} class.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{TRUE} if the instance in argument \code{cal} is compatible with
\code{self}, \code{FALSE} otherwise.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar-is_equivalent"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar-is_equivalent}{}}}
\subsection{Method \code{is_equivalent()}}{
This method tests if the \code{CFCalendar} instance in argument
\code{cal} is equivalent to \code{self}, meaning that they are of the same class,
have the same unit and prefix, and equivalent origins. Calendars
"standard", and "gregorian" are equivalent, as are the pairs of
"365_day" and "no_leap", and "366_day" and "all_leap".

Note that the origins need not be identical, but their parsed values
have to be. "2000-01" is parsed the same as "2000-01-01 00:00:00", for
instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar$is_equivalent(cal)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{cal}}{Instance of a descendant of the \code{CFCalendar} class.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{TRUE} if the instance in argument \code{cal} is equivalent to \code{self},
\code{FALSE} otherwise.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar-parse"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar-parse}{}}}
\subsection{Method \code{parse()}}{
Parsing a vector of date-time character strings into parts.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar$parse(d)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{d}}{character. A character vector of date-times.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{data.frame} with columns year, month, day, hour, minute,
second, time zone, and offset. Invalid input data will appear as \code{NA}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar-doy"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar-doy}{}}}
\subsection{Method \code{doy()}}{
Retrieve the day-of-year number for the supplied
\code{data.frame} of year, month and day, using this calendar. This method
tests for leap years - calendars without leap years may override this
method for efficiency.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar$doy(ymd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame} with dates parsed into their parts in columns
\code{year}, \code{month} and \code{day}. Any other columns are disregarded.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Integer vector as long as argument \code{ymd} has rows with the
day-of-year for each row.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar-offsets2time"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar-offsets2time}{}}}
\subsection{Method \code{offsets2time()}}{
Decompose a vector of offsets, in units of the calendar, to
their timestamp values. This adds a specified amount of time to the
origin of a \code{CFTime} object.

This method may introduce inaccuracies where the calendar unit is
"months" or "years", due to the ambiguous definition of these units.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar$offsets2time(offsets = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{offsets}}{Vector of numeric offsets to add to the origin of the
calendar.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{data.frame} with columns for the timestamp elements and as
many rows as there are offsets.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendar-clone"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendar-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendar$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{is_complete}
\alias{is_complete}
\title{Indicates if the time series is complete}
\usage{
is_complete(x)
}
\arguments{
\item{x}{An instance of the \link{CFTime} class.}
}
\value{
logical. \code{TRUE} if the time series is complete, with no gaps;
\code{FALSE} otherwise. If no offsets have been added to the \code{CFTime} instance,
\code{NA} is returned.
}
\description{
This function indicates if the time series is complete, meaning that the time
steps are equally spaced and there are thus no gaps in the time series.
}
\details{
This function gives exact results for time series where the nominal
\emph{unit of separation} between observations in the time series is exact in
terms of the calendar unit. As an example, for a calendar unit of "days" where the
observations are spaced a fixed number of days apart the result is exact, but
if the same calendar unit is used for data that is on a monthly basis, the
\emph{assessment} is approximate because the number of days per month is variable
and dependent on the calendar (the exception being the \verb{360_day} calendar,
where the assessment is exact). The \emph{result} is still correct in most cases
(including all CF-compliant data sets that the developers have seen) although
there may be esoteric constructions of CFTime and offsets that trip up this
implementation.
}
\examples{
t <- CFtime("days since 1850-01-01", "julian", 0:364)
is_complete(t)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CFCalendarJulian.R
\docType{class}
\name{CFCalendarJulian}
\alias{CFCalendarJulian}
\title{Julian CF calendar}
\description{
This class represents a Julian calendar of 365 days per year,
with every fourth year being a leap year of 366 days. The months and the
year align with the standard calendar. This calendar is not compatible with
the standard POSIXt calendar.

This calendar starts on 1 January of year 1: 0001-01-01 00:00:00. Any dates
before this will generate an error.
}
\section{Super class}{
\code{\link[CFtime:CFCalendar]{CFtime::CFCalendar}} -> \code{CFCalendarJulian}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-CFCalendarJulian-new}{\code{CFCalendarJulian$new()}}
\item \href{#method-CFCalendarJulian-valid_days}{\code{CFCalendarJulian$valid_days()}}
\item \href{#method-CFCalendarJulian-month_days}{\code{CFCalendarJulian$month_days()}}
\item \href{#method-CFCalendarJulian-leap_year}{\code{CFCalendarJulian$leap_year()}}
\item \href{#method-CFCalendarJulian-date2offset}{\code{CFCalendarJulian$date2offset()}}
\item \href{#method-CFCalendarJulian-offset2date}{\code{CFCalendarJulian$offset2date()}}
\item \href{#method-CFCalendarJulian-clone}{\code{CFCalendarJulian$clone()}}
}
}
\if{html}{\out{
<details><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="POSIX_compatible"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-POSIX_compatible'><code>CFtime::CFCalendar$POSIX_compatible()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="add_day"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-add_day'><code>CFtime::CFCalendar$add_day()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="doy"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-doy'><code>CFtime::CFCalendar$doy()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="is_compatible"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-is_compatible'><code>CFtime::CFCalendar$is_compatible()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="is_equivalent"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-is_equivalent'><code>CFtime::CFCalendar$is_equivalent()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="offsets2time"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-offsets2time'><code>CFtime::CFCalendar$offsets2time()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="parse"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-parse'><code>CFtime::CFCalendar$parse()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="print"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-print'><code>CFtime::CFCalendar$print()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarJulian-new"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarJulian-new}{}}}
\subsection{Method \code{new()}}{
Create a new CF calendar.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarJulian$new(nm, definition)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{nm}}{The name of the calendar. This must be "julian". This argument
is superfluous but maintained to be consistent with the initialization
methods of the parent and sibling classes.}

\item{\code{definition}}{The string that defines the units and the origin, as
per the CF Metadata Conventions.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new instance of this class.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarJulian-valid_days"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarJulian-valid_days}{}}}
\subsection{Method \code{valid_days()}}{
Indicate which of the supplied dates are valid.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarJulian$valid_days(ymd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame} with dates parsed into their parts in columns
\code{year}, \code{month} and \code{day}. Any other columns are disregarded.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Logical vector with the same length as argument \code{ymd} has rows
with \code{TRUE} for valid days and \code{FALSE} for invalid days, or \code{NA} where
the row in argument \code{ymd} has \code{NA} values.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarJulian-month_days"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarJulian-month_days}{}}}
\subsection{Method \code{month_days()}}{
Determine the number of days in the month of the calendar.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarJulian$month_days(ymd = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame}, optional, with dates parsed into their parts.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A vector indicating the number of days in each month for the
dates supplied as argument \code{ymd}. If no dates are supplied, the number
of days per month for the calendar as a vector of length 12, for a
regular year without a leap day.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarJulian-leap_year"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarJulian-leap_year}{}}}
\subsection{Method \code{leap_year()}}{
Indicate which years are leap years.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarJulian$leap_year(yr)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{yr}}{Integer vector of years to test.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Logical vector with the same length as argument \code{yr}. \code{NA} is
returned where elements in argument \code{yr} are \code{NA}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarJulian-date2offset"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarJulian-date2offset}{}}}
\subsection{Method \code{date2offset()}}{
Calculate difference in days between a \code{data.frame} of time
parts and the origin.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarJulian$date2offset(x)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{x}}{\code{data.frame}. Dates to calculate the difference for.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Integer vector of a length equal to the number of rows in
argument \code{x} indicating the number of days between \code{x} and the origin
of the calendar, or \code{NA} for rows in \code{x} with \code{NA} values.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarJulian-offset2date"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarJulian-offset2date}{}}}
\subsection{Method \code{offset2date()}}{
Calculate date parts from day differences from the origin. This
only deals with days as these are impacted by the calendar.
Hour-minute-second timestamp parts are handled in \link{CFCalendar}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarJulian$offset2date(x)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{x}}{Integer vector of days to add to the origin.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{data.frame} with columns 'year', 'month' and 'day' and as many
rows as the length of vector \code{x}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarJulian-clone"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarJulian-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarJulian$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{parse_timestamps}
\alias{parse_timestamps}
\title{Parse series of timestamps in CF format to date-time elements}
\usage{
parse_timestamps(t, x)
}
\arguments{
\item{t}{An instance of \code{CFTime} to use when parsing the date.}

\item{x}{Vector of character strings representing timestamps in
ISO8601 extended or UDUNITS broken format.}
}
\value{
A \code{data.frame} with constituent elements of the parsed timestamps in
numeric format. The columns are year, month, day, hour, minute, second
(with an optional fraction), time zone (character string), and the
corresponding offset value from the origin. Invalid input data will appear
as \code{NA} - if this is the case, a warning message will be displayed - other
missing information on input will use default values.
}
\description{
This function will parse a vector of timestamps in ISO8601 or UDUNITS format
into a data frame with columns for the elements of the timestamp: year,
month, day, hour, minute, second, time zone. Those timestamps that could not
be parsed or which represent an invalid date in the indicated \code{CFtime}
instance will have \code{NA} values for the elements of the offending timestamp
(which will generate a warning).
}
\details{
The supported formats are the \emph{broken timestamp} format from the UDUNITS
library and ISO8601 \emph{extended}, both with minor changes, as suggested by the
CF Metadata Conventions. In general, the format is \verb{YYYY-MM-DD hh:mm:ss.sss hh:mm}. The year can be from 1 to 4 digits and is interpreted literally, so
\code{79-10-24} is the day Mount Vesuvius erupted and destroyed Pompeii, not
\code{1979-10-24}. The year and month are mandatory, all other fields are
optional. There are defaults for all missing values, following the UDUNITS
and CF Metadata Conventions. Leading zeros can be omitted in the UDUNITS
format, but not in the ISO8601 format. The optional fractional part can have
as many digits as the precision calls for and will be applied to the smallest
specified time unit. In the result of this function, if the fraction is
associated with the minute or the hour, it is converted into a regular
\code{hh:mm:ss.sss} format, i.e. any fraction in the result is always associated
with the second, rounded down to milli-second accuracy. The separator between
the date and the time can be a single whitespace character or a \code{T}.

The time zone is optional and should have at least the hour or \code{Z} if
present, the minute is optional. The time zone hour can have an optional
sign. In the UDUNITS format the separator between the time and the time zone
must be a single whitespace character, in ISO8601 there is no separation
between the time and the timezone. Time zone names are not supported (as
neither UDUNITS nor ISO8601 support them) and will cause parsing to fail when
supplied, with one exception: the designator "UTC" is silently dropped (i.e.
interpreted as "00:00").

Currently only the extended formats (with separators between the elements)
are supported. The vector of timestamps may have any combination of ISO8601
and UDUNITS formats.
}
\examples{
t <- CFtime("days since 0001-01-01", "proleptic_gregorian")

# This will have `NA`s on output and generate a warning
timestamps <- c("2012-01-01T12:21:34Z", "12-1-23", "today",
                "2022-08-16T11:07:34.45-10", "2022-08-16 10.5+04")
parse_timestamps(t, timestamps)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CFtime.R
\name{[.CFTime}
\alias{[.CFTime}
\title{Subset a \code{CFTime} instance by position in the time series}
\usage{
\method{[}{CFTime}(x, i = TRUE, ...)
}
\arguments{
\item{x}{A \code{CFTime} instance.}

\item{i}{A vector a positive integer values to indicate which values to
extract from the time series by position. If negative values are passed,
their positive counterparts will be excluded and then the remainder
returned. Positive and negative values may not be mixed.}

\item{...}{Ignored.}
}
\value{
A numeric vector with those values of \code{i} (or the inverse, when
negative) that are valid in \code{x}. If there is at least 1 valid result, then
attribute "CFTime" of the returned value contains an instance of \code{CFTime}
that describes the dimension of filtering the dataset associated with \code{x}
with the result of this function, excluding any \code{NA} values.
}
\description{
Subset a \code{CFTime} instance by position in the time series
}
\references{
https://github.com/R-CF/CFtime/issues/20
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{format.CFTime}
\alias{format.CFTime}
\title{Format the timestamps in the \code{CFTime} instance}
\usage{
\method{format}{CFTime}(x, format = "", tz = "", usetz = FALSE, ...)
}
\arguments{
\item{x}{The \code{CFTime} instance whose timestamps will be formatted.}

\item{format}{A character string. The default for the format methods is
"\%Y-\%m-\%dT\%H:\%M:\%S" if any timestamp has a time component which is not
midnight, and "\%Y-\%m-\%d" otherwise. The only supported specifiers are
\verb{bBdeFhHImMpRSTYz\%}. Modifiers \code{E} and \code{O} are silently ignored. Other
specifiers, including their percent sign, are copied to the output as if
they were adorning text.}

\item{tz}{Ignored.}

\item{usetz}{Logical. Should the time zone offset be appended to the output?
This is always in numerical form, i.e. "-0800", from UTC.}

\item{...}{Ignored.}
}
\description{
The formatting is largely oblivious to locale. The reason for this is
that certain dates in certain calendars are not POSIX-compliant and the
system functions necessary for locale information thus do not work
consistently. The main exception to this is the (abbreviated) names of
months (\code{bB}), which could be useful for pretty printing in the local
language. For separators and other locale-specific adornments, use
local knowledge instead of depending on system locale settings; e.g.
specify \verb{\%m/\%d/\%Y} instead of \verb{\%D}.
}
\details{
Week information, including weekday names, is not supported at all as a
"week" is not defined for non-standard CF calendars and not generally
useful for climate projection data. If you are working with observed
data and want to get pretty week formats, use the \code{\link[=as_timestamp]{as_timestamp()}}
method to generate \code{POSIXct} timestamps (observed data generally uses a
"standard" calendar) and then use the \code{\link[base:format]{base::format()}} function which
supports the full set of specifiers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{CFfactor}
\alias{CFfactor}
\title{Create a factor from the offsets in a \code{CFTime} instance}
\usage{
CFfactor(t, period = "month", era = NULL)
}
\arguments{
\item{t}{An instance of the \code{CFTime} class whose offsets will be used to
construct the factor.}

\item{period}{character. A character string with one of the values "year",
"season", "quarter", "month" (the default), "dekad" or "day".}

\item{era}{numeric or list, optional. Vector of years for which to
construct the factor, or a list whose elements are each a vector of years.
If \code{era} is not specified, the factor will use the entire time series for
the factor.}
}
\value{
If \code{era} is a single vector or not specified, a factor with a
length equal to the number of offsets in \code{t}. If \code{era} is a list, a list
with the same number of elements and names as \code{era}, each containing a
factor. Elements in the factor will be set to \code{NA} for time series values
outside of the range of specified years.

The factor, or factors in the list, have attributes 'period', 'era' and
'CFTime'. Attribute 'period' holds the value of the \code{period} argument.
Attribute 'era' indicates the number of years that are included in the
era, or -1 if no \code{era} is provided. Attribute 'CFTime' holds an
instance of \code{CFTime} that has the same definition as \code{t}, but with offsets
corresponding to the mid-point of non-era factor levels; if the \code{era}
argument is specified, attribute 'CFTime' is \code{NULL}.
}
\description{
With this function a factor can be generated for the time series, or a part
thereof, contained in the \link{CFTime} instance. This is specifically interesting
for creating factors from the date part of the time series that aggregate the
time series into longer time periods (such as month) that can then be used to
process daily CF data sets using, for instance, \code{tapply()}.
}
\details{
The factor will respect the calendar that the time series is built on. For
\code{period}s longer than a day this will result in a factor where the calendar
is no longer relevant (because calendars impacts days, not dekads, months,
quarters, seasons or years).

The factor will be generated in the order of the offsets of the \code{CFTime}
instance. While typical CF-compliant data sources use ordered time series
there is, however, no guarantee that the factor is ordered as multiple
\code{CFTime} objects may have been merged out of order. For most processing with
a factor the ordering is of no concern.

If the \code{era} parameter is specified, either as a vector of years to include
in the factor, or as a list of such vectors, the factor will only consider
those values in the time series that fall within the list of years, inclusive
of boundary values. Other values in the factor will be set to \code{NA}. The years
need not be contiguous, within a single vector or among the list items, or in
order.

The following periods are supported by this function:

\itemize{
\item \code{year}, the year of each offset is returned as "YYYY".
\item \code{season}, the meteorological season of each offset is returned as
"Sx", with x being 1-4, preceeded by "YYYY" if no \code{era} is
specified. Note that December dates are labeled as belonging to the
subsequent year, so the date "2020-12-01" yields "2021S1". This implies
that for standard CMIP files having one or more full years of data the
first season will have data for the first two months (January and
February), while the final season will have only a single month of data
(December).
\item \code{quarter}, the calendar quarter of each offset is returned as "Qx",
with x being 1-4, preceeded by "YYYY" if no \code{era} is specified.
\item \code{month}, the month of each offset is returned as "01" to
"12", preceeded by "YYYY-" if no \code{era} is specified. This is the default
period.
\item \code{dekad}, ten-day periods are returned as
"Dxx", where xx runs from "01" to "36", preceeded by "YYYY" if no \code{era}
is specified. Each month is subdivided in dekads as follows: 1- days 01 -
10; 2- days 11 - 20; 3- remainder of the month.
\item \code{day}, the month and day of each offset are returned as "MM-DD",
preceeded by "YYYY-" if no \code{era} is specified.
}

It is not possible to create a factor for a period that is shorter than the
temporal resolution of the source data set from which the \code{t} argument
derives. As an example, if the source data set has monthly data, a dekad or
day factor cannot be created.

Creating factors for other periods is not supported by this function. Factors
based on the timestamp information and not dependent on the calendar can
trivially be constructed from the output of the \code{\link[=as_timestamp]{as_timestamp()}} function.

For non-era factors the attribute 'CFTime' of the result contains a \code{CFTime}
instance that is valid for the result of applying the factor to a data set
that the \code{t} argument is associated with. In other words, if \code{CFTime}
instance 'At' describes the temporal dimension of data set 'A' and a factor
'Af' is generated like \code{Af <- CFfactor(At)}, then \code{Bt <- attr(Af, "CFTime")}
describes the temporal dimension of the result of, say,
\code{B <- apply(A, 1:2, tapply, Af, FUN)}. The 'CFTime' attribute is \code{NULL} for
era factors.
}
\examples{
t <- CFtime("days since 1949-12-01", "360_day", 19830:54029)

# Create a dekad factor for the whole time series
f <- CFfactor(t, "dekad")

# Create three monthly factors for early, mid and late 21st century eras
ep <- CFfactor(t, era = list(early = 2021:2040, mid = 2041:2060, late = 2061:2080))
}
\seealso{
\code{\link[=cut]{cut()}} creates a non-era factor for arbitrary cut points.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CFCalendarStandard.R
\docType{class}
\name{CFCalendarStandard}
\alias{CFCalendarStandard}
\title{Standard CF calendar}
\description{
This class represents a standard calendar of 365 or 366 days per
year. This calendar is compatible with the standard POSIXt calendar for
periods after the introduction of the Gregorian calendar, 1582-10-15
00:00:00. The calendar starts at 0001-01-01 00:00:00, e.g. the start of the
Common Era.

Note that this calendar, despite its name, is not the same as that used in
ISO8601 or many computer systems for periods prior to the introduction of
the Gregorian calendar. Use of the "proleptic_gregorian" calendar is
recommended for periods before or straddling the introduction date, as that
calendar is compatible with POSIXt on most OSes.
}
\section{Super class}{
\code{\link[CFtime:CFCalendar]{CFtime::CFCalendar}} -> \code{CFCalendarStandard}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-CFCalendarStandard-new}{\code{CFCalendarStandard$new()}}
\item \href{#method-CFCalendarStandard-valid_days}{\code{CFCalendarStandard$valid_days()}}
\item \href{#method-CFCalendarStandard-is_gregorian_date}{\code{CFCalendarStandard$is_gregorian_date()}}
\item \href{#method-CFCalendarStandard-POSIX_compatible}{\code{CFCalendarStandard$POSIX_compatible()}}
\item \href{#method-CFCalendarStandard-month_days}{\code{CFCalendarStandard$month_days()}}
\item \href{#method-CFCalendarStandard-leap_year}{\code{CFCalendarStandard$leap_year()}}
\item \href{#method-CFCalendarStandard-date2offset}{\code{CFCalendarStandard$date2offset()}}
\item \href{#method-CFCalendarStandard-offset2date}{\code{CFCalendarStandard$offset2date()}}
\item \href{#method-CFCalendarStandard-clone}{\code{CFCalendarStandard$clone()}}
}
}
\if{html}{\out{
<details><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="add_day"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-add_day'><code>CFtime::CFCalendar$add_day()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="doy"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-doy'><code>CFtime::CFCalendar$doy()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="is_compatible"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-is_compatible'><code>CFtime::CFCalendar$is_compatible()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="is_equivalent"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-is_equivalent'><code>CFtime::CFCalendar$is_equivalent()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="offsets2time"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-offsets2time'><code>CFtime::CFCalendar$offsets2time()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="parse"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-parse'><code>CFtime::CFCalendar$parse()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="print"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-print'><code>CFtime::CFCalendar$print()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarStandard-new"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarStandard-new}{}}}
\subsection{Method \code{new()}}{
Create a new CF calendar. When called with the deprecated
'gregorian' it is automatically converted to the equivalent 'standard'.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarStandard$new(nm, definition)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{nm}}{The name of the calendar. This must be "standard" or
"gregorian" (deprecated).}

\item{\code{definition}}{The string that defines the units and the origin, as
per the CF Metadata Conventions.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new instance of this class.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarStandard-valid_days"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarStandard-valid_days}{}}}
\subsection{Method \code{valid_days()}}{
Indicate which of the supplied dates are valid.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarStandard$valid_days(ymd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame} with dates parsed into their parts in columns
\code{year}, \code{month} and \code{day}. Any other columns are disregarded.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Logical vector with the same length as argument \code{ymd} has rows
with \code{TRUE} for valid days and \code{FALSE} for invalid days, or \code{NA} where
the row in argument \code{ymd} has \code{NA} values.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarStandard-is_gregorian_date"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarStandard-is_gregorian_date}{}}}
\subsection{Method \code{is_gregorian_date()}}{
Indicate which of the supplied dates are in the Gregorian
part of the calendar, e.g. 1582-10-15 or after.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarStandard$is_gregorian_date(ymd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame} with dates parsed into their parts in columns
\code{year}, \code{month} and \code{day}. Any other columns are disregarded.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Logical vector with the same length as argument \code{ymd} has rows
with \code{TRUE} for days in the Gregorian part of the calendar and \code{FALSE}
otherwise, or \code{NA} where the row in argument \code{ymd} has \code{NA} values.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarStandard-POSIX_compatible"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarStandard-POSIX_compatible}{}}}
\subsection{Method \code{POSIX_compatible()}}{
Indicate if the time series described using this calendar
can be safely converted to a standard date-time type (\code{POSIXct},
\code{POSIXlt}, \code{Date}). This is only the case if all offsets are for
timestamps fall on or after the start of the Gregorian calendar,
1582-10-15 00:00:00.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarStandard$POSIX_compatible(offsets)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{offsets}}{The offsets from the CFtime instance.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{TRUE}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarStandard-month_days"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarStandard-month_days}{}}}
\subsection{Method \code{month_days()}}{
Determine the number of days in the month of the calendar.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarStandard$month_days(ymd = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame}, optional, with dates parsed into their parts.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A vector indicating the number of days in each month for the
dates supplied as argument \code{ymd}. If no dates are supplied, the number
of days per month for the calendar as a vector of length 12, for a
regular year without a leap day.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarStandard-leap_year"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarStandard-leap_year}{}}}
\subsection{Method \code{leap_year()}}{
Indicate which years are leap years.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarStandard$leap_year(yr)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{yr}}{Integer vector of years to test.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Logical vector with the same length as argument \code{yr}. \code{NA} is
returned where elements in argument \code{yr} are \code{NA}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarStandard-date2offset"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarStandard-date2offset}{}}}
\subsection{Method \code{date2offset()}}{
Calculate difference in days between a \code{data.frame} of time
parts and the origin.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarStandard$date2offset(x)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{x}}{\code{data.frame}. Dates to calculate the difference for.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Integer vector of a length equal to the number of rows in
argument \code{x} indicating the number of days between \code{x} and the origin
of the calendar, or \code{NA} for rows in \code{x} with \code{NA} values.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarStandard-offset2date"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarStandard-offset2date}{}}}
\subsection{Method \code{offset2date()}}{
Calculate date parts from day differences from the origin. This
only deals with days as these are impacted by the calendar.
Hour-minute-second timestamp parts are handled in \link{CFCalendar}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarStandard$offset2date(x)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{x}}{Integer vector of days to add to the origin.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{data.frame} with columns 'year', 'month' and 'day' and as many
rows as the length of vector \code{x}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarStandard-clone"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarStandard-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarStandard$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{as_timestamp}
\alias{as_timestamp}
\title{Create a vector that represents CF timestamps}
\usage{
as_timestamp(t, format = NULL, asPOSIX = FALSE)
}
\arguments{
\item{t}{The \code{CFTime} instance that contains the offsets to use.}

\item{format}{character. A character string with either of the values "date"
or "timestamp". If the argument is not specified, the format used is
"timestamp" if there is time information, "date" otherwise.}

\item{asPOSIX}{logical. If \code{TRUE}, for "standard", "gregorian" and
"proleptic_gregorian" calendars the output is a vector of \code{POSIXct} - for
other calendars an error will be thrown. Default value is \code{FALSE}.}
}
\value{
A character vector where each element represents a moment in time
according to the \code{format} specifier.
}
\description{
This function generates a vector of character strings or \code{POSIXct}s that
represent the date and time in a selectable combination for each offset.
}
\details{
The character strings use the format \verb{YYYY-MM-DDThh:mm:ss±hhmm}, depending on
the \code{format} specifier. The date in the string is not necessarily compatible
with \code{POSIXt} - in the \verb{360_day} calendar \code{2017-02-30} is valid and
\code{2017-03-31} is not.

For the "proleptic_gregorian" calendar the output can also be generated as a
vector of \code{POSIXct} values by specifying \code{asPOSIX = TRUE}. The same is
possible for the "standard" and "gregorian" calendars but only if all
timestamps fall on or after 1582-10-15.
}
\examples{
t <- CFtime("hours since 2020-01-01", "standard", seq(0, 24, by = 0.25))
as_timestamp(t, "timestamp")

t2 <- CFtime("days since 2002-01-21", "standard", 0:20)
tail(as_timestamp(t2, asPOSIX = TRUE))

tail(as_timestamp(t2))

tail(as_timestamp(t2 + 1.5))
}
\seealso{
The \link{CFTime} \code{format()} method gives greater flexibility through
the use of \code{strptime}-like format specifiers.
}

% Generated by roxygen2: do not edit by hand % Please edit documentation in R/CFCalendarUTC.R \docType{class} \name{CFCalendarUTC} \alias{CFCalendarUTC} \title{Coordinated Universal Time CF calendar} \description{ This class represents a calendar based on the Coordinated Universal Time. Validity is from 1972 onwards, with dates represented using the Gregorian calendar, up to the present (so future timestamps are not allowed). Leap seconds are considered in all calculations. Also, time zone information is irrelevant and may not be given. In general, the calendar should use a unit of time of a second. Minute, hour and day are allowed but discouraged. Month and year as time unit are not allowed as there is no practical way to maintain leap second accuracy. } \section{Super classes}{ \code{\link[CFtime:CFCalendar]{CFtime::CFCalendar}} -> \code{\link[CFtime:CFCalendarProleptic]{CFtime::CFCalendarProleptic}} -> \code{CFCalendarUTC} } \section{Methods}{ \subsection{Public methods}{ \itemize{ \item \href{#method-CFCalendarUTC-new}{\code{CFCalendarUTC$new()}} \item \href{#method-CFCalendarUTC-valid_days}{\code{CFCalendarUTC$valid_days()}} \item \href{#method-CFCalendarUTC-parse}{\code{CFCalendarUTC$parse()}} \item \href{#method-CFCalendarUTC-offsets2time}{\code{CFCalendarUTC$offsets2time()}} \item \href{#method-CFCalendarUTC-clone}{\code{CFCalendarUTC$clone()}} } } \if{html}{\out{
Inherited methods

- `CFtime::CFCalendar$add_day()`
- `CFtime::CFCalendar$doy()`
- `CFtime::CFCalendar$is_compatible()`
- `CFtime::CFCalendar$is_equivalent()`
- `CFtime::CFCalendar$print()`
- `CFtime::CFCalendarProleptic$POSIX_compatible()`
- `CFtime::CFCalendarProleptic$date2offset()`
- `CFtime::CFCalendarProleptic$leap_year()`
- `CFtime::CFCalendarProleptic$month_days()`
- `CFtime::CFCalendarProleptic$offset2date()`

}} \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-CFCalendarUTC-new}{}}} \subsection{Method \code{new()}}{ Create a new CF UTC calendar. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{CFCalendarUTC$new(nm, definition)}\if{html}{\out{
}} } \subsection{Arguments}{ \if{html}{\out{
}} \describe{ \item{\code{nm}}{The name of the calendar. This must be "utc".} \item{\code{definition}}{The string that defines the units and the origin, as per the CF Metadata Conventions.} } \if{html}{\out{
}} } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-CFCalendarUTC-valid_days}{}}} \subsection{Method \code{valid_days()}}{ Indicate which of the supplied dates are valid. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{CFCalendarUTC$valid_days(ymd)}\if{html}{\out{
}} } \subsection{Arguments}{ \if{html}{\out{
}} \describe{ \item{\code{ymd}}{\code{data.frame} with dates parsed into their parts in columns \code{year}, \code{month} and \code{day}. Any other columns are disregarded.} } \if{html}{\out{
}} } \subsection{Returns}{ Logical vector with the same length as argument \code{ymd} has rows with \code{TRUE} for valid days and \code{FALSE} for invalid days, or \code{NA} where the row in argument \code{ymd} has \code{NA} values. } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-CFCalendarUTC-parse}{}}} \subsection{Method \code{parse()}}{ Parsing a vector of date-time character strings into parts. This includes any leap seconds. Time zone indications are not allowed. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{CFCalendarUTC$parse(d)}\if{html}{\out{
}} } \subsection{Arguments}{ \if{html}{\out{
}} \describe{ \item{\code{d}}{character. A character vector of date-times.} } \if{html}{\out{
}} } \subsection{Returns}{ A \code{data.frame} with columns year, month, day, hour, minute, second, time zone, and offset. Invalid input data will appear as \code{NA}. Note that the time zone is always "+0000" and is included to maintain compatibility with results from other calendars. } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-CFCalendarUTC-offsets2time}{}}} \subsection{Method \code{offsets2time()}}{ Decompose a vector of offsets, in units of the calendar, to their timestamp values. This adds a specified amount of time to the origin of a \code{CFTime} object. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{CFCalendarUTC$offsets2time(offsets = NULL)}\if{html}{\out{
}} } \subsection{Arguments}{ \if{html}{\out{
}} \describe{ \item{\code{offsets}}{Vector of numeric offsets to add to the origin of the calendar.} } \if{html}{\out{
}} } \subsection{Returns}{ A \code{data.frame} with columns for the timestamp elements and as many rows as there are offsets. } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-CFCalendarUTC-clone}{}}} \subsection{Method \code{clone()}}{ The objects of this class are cloneable with this method. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{CFCalendarUTC$clone(deep = FALSE)}\if{html}{\out{
}} } \subsection{Arguments}{ \if{html}{\out{
}} \describe{ \item{\code{deep}}{Whether to make a deep clone.} } \if{html}{\out{
}} } } }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CFCalendarProleptic.R
\docType{class}
\name{CFCalendarProleptic}
\alias{CFCalendarProleptic}
\title{Proleptic Gregorian CF calendar}
\description{
This class represents a standard CF calendar, but with the
Gregorian calendar extended backwards to before the introduction of the
Gregorian calendar. This calendar is compatible with the standard POSIXt
calendar, but note that daylight savings time is not considered.

This calendar includes dates 1582-10-14 to 1582-10-05 (the gap between the
Gregorian and Julian calendars, which is observed by the standard
calendar), and extends to years before the year 1, including year 0.
}
\section{Super class}{
\code{\link[CFtime:CFCalendar]{CFtime::CFCalendar}} -> \code{CFCalendarProleptic}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-CFCalendarProleptic-new}{\code{CFCalendarProleptic$new()}}
\item \href{#method-CFCalendarProleptic-valid_days}{\code{CFCalendarProleptic$valid_days()}}
\item \href{#method-CFCalendarProleptic-month_days}{\code{CFCalendarProleptic$month_days()}}
\item \href{#method-CFCalendarProleptic-leap_year}{\code{CFCalendarProleptic$leap_year()}}
\item \href{#method-CFCalendarProleptic-POSIX_compatible}{\code{CFCalendarProleptic$POSIX_compatible()}}
\item \href{#method-CFCalendarProleptic-date2offset}{\code{CFCalendarProleptic$date2offset()}}
\item \href{#method-CFCalendarProleptic-offset2date}{\code{CFCalendarProleptic$offset2date()}}
\item \href{#method-CFCalendarProleptic-clone}{\code{CFCalendarProleptic$clone()}}
}
}
\if{html}{\out{
<details><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="add_day"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-add_day'><code>CFtime::CFCalendar$add_day()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="doy"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-doy'><code>CFtime::CFCalendar$doy()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="is_compatible"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-is_compatible'><code>CFtime::CFCalendar$is_compatible()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="is_equivalent"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-is_equivalent'><code>CFtime::CFCalendar$is_equivalent()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="offsets2time"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-offsets2time'><code>CFtime::CFCalendar$offsets2time()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="parse"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-parse'><code>CFtime::CFCalendar$parse()</code></a></span></li>
<li><span class="pkg-link" data-pkg="CFtime" data-topic="CFCalendar" data-id="print"><a href='../../CFtime/html/CFCalendar.html#method-CFCalendar-print'><code>CFtime::CFCalendar$print()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarProleptic-new"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarProleptic-new}{}}}
\subsection{Method \code{new()}}{
Create a new CF calendar.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarProleptic$new(nm, definition)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{nm}}{The name of the calendar. This must be "proleptic_gregorian".
This argument is superfluous but maintained to be consistent with the
initialization methods of the parent and sibling classes.}

\item{\code{definition}}{The string that defines the units and the origin, as
per the CF Metadata Conventions.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new instance of this class.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarProleptic-valid_days"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarProleptic-valid_days}{}}}
\subsection{Method \code{valid_days()}}{
Indicate which of the supplied dates are valid.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarProleptic$valid_days(ymd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame} with dates parsed into their parts in columns
\code{year}, \code{month} and \code{day}. Any other columns are disregarded.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Logical vector with the same length as argument \code{ymd} has rows
with \code{TRUE} for valid days and \code{FALSE} for invalid days, or \code{NA} where
the row in argument \code{ymd} has \code{NA} values.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarProleptic-month_days"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarProleptic-month_days}{}}}
\subsection{Method \code{month_days()}}{
Determine the number of days in the month of the calendar.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarProleptic$month_days(ymd = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{\code{data.frame}, optional, with dates parsed into their parts.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Integer vector indicating the number of days in each month for
the dates supplied as argument \code{ymd}. If no dates are supplied, the
number of days per month for the calendar as a vector of length 12, for
a regular year without a leap day.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarProleptic-leap_year"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarProleptic-leap_year}{}}}
\subsection{Method \code{leap_year()}}{
Indicate which years are leap years.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarProleptic$leap_year(yr)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{yr}}{Integer vector of years to test.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Logical vector with the same length as argument \code{yr}. \code{NA} is
returned where elements in argument \code{yr} are \code{NA}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarProleptic-POSIX_compatible"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarProleptic-POSIX_compatible}{}}}
\subsection{Method \code{POSIX_compatible()}}{
Indicate if the time series described using this calendar
can be safely converted to a standard date-time type (\code{POSIXct},
\code{POSIXlt}, \code{Date}).
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarProleptic$POSIX_compatible(offsets)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{offsets}}{The offsets from the CFtime instance.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{TRUE}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarProleptic-date2offset"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarProleptic-date2offset}{}}}
\subsection{Method \code{date2offset()}}{
Calculate difference in days between a \code{data.frame} of time
parts and the origin.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarProleptic$date2offset(x)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{x}}{\code{data.frame}. Dates to calculate the difference for.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Integer vector of a length equal to the number of rows in
argument \code{x} indicating the number of days between \code{x} and the \code{origin},
or \code{NA} for rows in \code{x} with \code{NA} values.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarProleptic-offset2date"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarProleptic-offset2date}{}}}
\subsection{Method \code{offset2date()}}{
Calculate date parts from day differences from the origin. This
only deals with days as these are impacted by the calendar.
Hour-minute-second timestamp parts are handled in \link{CFCalendar}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarProleptic$offset2date(x)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{x}}{Integer vector of days to add to the origin.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{data.frame} with columns 'year', 'month' and 'day' and as many
rows as the length of vector \code{x}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFCalendarProleptic-clone"></a>}}
\if{latex}{\out{\hypertarget{method-CFCalendarProleptic-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFCalendarProleptic$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand % Please edit documentation in R/CFCalendarTAI.R \docType{class} \name{CFCalendarTAI} \alias{CFCalendarTAI} \title{International Atomic Time CF calendar} \description{ This class represents a calendar based on the International Atomic Time. Validity is from 1958 onwards, with dates represented using the Gregorian calendar. Given that this "calendar" is based on a universal clock, the concepts of leap second, time zone and daylight savings time do not apply. } \section{Super classes}{ \code{\link[CFtime:CFCalendar]{CFtime::CFCalendar}} -> \code{\link[CFtime:CFCalendarProleptic]{CFtime::CFCalendarProleptic}} -> \code{CFCalendarTAI} } \section{Methods}{ \subsection{Public methods}{ \itemize{ \item \href{#method-CFCalendarTAI-valid_days}{\code{CFCalendarTAI$valid_days()}} \item \href{#method-CFCalendarTAI-clone}{\code{CFCalendarTAI$clone()}} } } \if{html}{\out{
Inherited methods

- `CFtime::CFCalendar$add_day()`
- `CFtime::CFCalendar$doy()`
- `CFtime::CFCalendar$is_compatible()`
- `CFtime::CFCalendar$is_equivalent()`
- `CFtime::CFCalendar$offsets2time()`
- `CFtime::CFCalendar$parse()`
- `CFtime::CFCalendar$print()`
- `CFtime::CFCalendarProleptic$POSIX_compatible()`
- `CFtime::CFCalendarProleptic$date2offset()`
- `CFtime::CFCalendarProleptic$initialize()`
- `CFtime::CFCalendarProleptic$leap_year()`
- `CFtime::CFCalendarProleptic$month_days()`
- `CFtime::CFCalendarProleptic$offset2date()`

}} \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-CFCalendarTAI-valid_days}{}}} \subsection{Method \code{valid_days()}}{ Indicate which of the supplied dates are valid. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{CFCalendarTAI$valid_days(ymd)}\if{html}{\out{
}} } \subsection{Arguments}{ \if{html}{\out{
}} \describe{ \item{\code{ymd}}{\code{data.frame} with dates parsed into their parts in columns \code{year}, \code{month} and \code{day}. If present, the \code{tz} column is checked for illegal time zone offsets. Any other columns are disregarded.} } \if{html}{\out{
}} } \subsection{Returns}{ Logical vector with the same length as argument \code{ymd} has rows with \code{TRUE} for valid days and \code{FALSE} for invalid days, or \code{NA} where the row in argument \code{ymd} has \code{NA} values. } } \if{html}{\out{

}} \if{html}{\out{}} \if{latex}{\out{\hypertarget{method-CFCalendarTAI-clone}{}}} \subsection{Method \code{clone()}}{ The objects of this class are cloneable with this method. \subsection{Usage}{ \if{html}{\out{
}}\preformatted{CFCalendarTAI$clone(deep = FALSE)}\if{html}{\out{
}} } \subsection{Arguments}{ \if{html}{\out{
}} \describe{ \item{\code{deep}}{Whether to make a deep clone.} } \if{html}{\out{
}} } } }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{as.character.CFTime}
\alias{as.character.CFTime}
\title{Return the timestamps contained in the \code{CFTime} instance}
\usage{
\method{as.character}{CFTime}(x, ...)
}
\arguments{
\item{x}{The \code{CFTime} instance whose timestamps will be returned.}

\item{...}{Ignored.}
}
\value{
The timestamps in the specified \code{CFTime} instance.
}
\description{
Return the timestamps contained in the \code{CFTime} instance
}
\examples{
t <- CFtime("days since 1850-01-01", "julian", 0:364)
as.character(t)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{indexOf}
\alias{indexOf}
\title{Find the index of timestamps in the time series}
\usage{
indexOf(x, y, method = "constant", rightmost.closed = FALSE)
}
\arguments{
\item{x}{Vector of \code{character}, \code{POSIXt} or \code{Date} values to find indices
for, or a numeric vector.}

\item{y}{A \link{CFTime} instance.}

\item{method}{Single value of "constant" or "linear". If \code{"constant"} or when
bounds are set on argument \code{y}, return the index value for each match. If
\code{"linear"}, return the index value with any fractional value.}

\item{rightmost.closed}{Whether or not to include the upper limit of
argument \code{x}. Default is \code{FALSE}. This argument is ignored when
argument \code{x} contains index values.}
}
\value{
A numeric vector giving indices into \code{y} for the values of
\code{x}. If there is at least 1 valid index, then attribute "CFTime"
contains an instance of \code{CFTime} that describes the dimension of
filtering the dataset associated with \code{y} with the result of this
method, excluding any \code{NA} values.
}
\description{
Find the index in the time series for each timestamp given in argument \code{x}.
Values of \code{x} that are before the earliest value or after the latest value in
\code{y} will be returned as \code{NA}. Alternatively, when \code{x} is a numeric vector of
index values, return the valid indices of the same vector, with the side
effect being the attribute "CFTime" associated with the result.
}
\details{
Timestamps can be provided as vectors of character strings, \code{POSIXct} or
\code{Date.}

Matching also returns index values for timestamps that fall between two
elements of the time series - this can lead to surprising results when time
series elements are positioned in the middle of an interval (as the CF
Metadata Conventions instruct us to "reasonably assume"): a time series of
days in January would be encoded in a netCDF file as
\code{c("2024-01-01 12:00:00", "2024-01-02 12:00:00", "2024-01-03 12:00:00", ...)}
so \code{x <- c("2024-01-01", "2024-01-02", "2024-01-03")} would result in
\verb{(NA, 1, 2)} (or \verb{(NA, 1.5, 2.5)} with \code{method = "linear"}) because the date
values in \code{x} are at midnight. This situation is easily avoided by ensuring
that \code{y} has bounds set (use \code{bounds(y) <- TRUE} as a proximate solution if
bounds are not stored in the netCDF file). See the Examples.

If bounds are set, the indices are informed by those bounds. If the
bounds are not contiguous, returned values may be \code{NA} even if the
value of \code{x} falls between two valid timestamps.

Values of \code{x} that are not valid timestamps according to the calendar of \code{y}
will be returned as \code{NA}.

\code{x} can also be a numeric vector of index values, in which case the valid
values in \code{x} are returned. If negative values are passed, the positive
counterparts will be excluded and then the remainder returned. Positive and
negative values may not be mixed. Using a numeric vector has the side effect
that the result has the attribute "CFTime" describing the temporal dimension
of the slice. If index values outside of the range of \code{y} (\code{1:length(y)}) are
provided, an error will be thrown.
}
\examples{
cf <- CFtime("days since 2020-01-01", "360_day", 1440:1799 + 0.5)
as_timestamp(cf)[1:3]
x <- c("2024-01-01", "2024-01-02", "2024-01-03")
indexOf(x, cf)
indexOf(x, cf, method = "linear")

bounds(cf) <- TRUE
indexOf(x, cf)

# Non-existent calendar day in a `360_day` calendar
x <- c("2024-03-30", "2024-03-31", "2024-04-01")
indexOf(x, cf)

# Numeric x
indexOf(c(29, 30, 31), cf)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{length.CFTime}
\alias{length.CFTime}
\title{The length of the offsets contained in the \code{CFTime} instance.}
\usage{
\method{length}{CFTime}(x)
}
\arguments{
\item{x}{The \code{CFTime} instance whose length will be returned}
}
\value{
The number of offsets in the specified \code{CFTime} instance.
}
\description{
The length of the offsets contained in the \code{CFTime} instance.
}
\examples{
t <- CFtime("days since 1850-01-01", "julian", 0:364)
length(t)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CFtime.R
\docType{class}
\name{CFTime}
\alias{CFTime}
\title{CFTime class}
\description{
This class manages the "time" dimension of netCDF files that
follow the CF Metadata Conventions, and its productive use in R.

The class has a field \code{cal} which holds a specific calendar from the
allowed types (all named calendars are supported). The calendar is
also implemented as a (hidden) class which converts netCDF file encodings to
timestamps as character strings, and vice-versa. Bounds information (the
period of time over which a timestamp is valid) is used when defined in the
netCDF file.

Additionally, this class has functions to ease use of the netCDF "time"
information when processing data from netCDF files. Filtering and indexing of
time values is supported, as is the generation of factors.
}
\references{
https://cf-convention.github.io/Data/cf-conventions/cf-conventions-1.13/cf-conventions.html#time-coordinate
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{calendar}}{(read-only) The calendar of this \code{CFTime} instance, a
descendant of the \link{CFCalendar} class.}

\item{\code{cal}}{(read-only) DEPRECATED. Use the \code{calendar} field instead.}

\item{\code{unit}}{(read-only) The unit string of the calendar and time series.}

\item{\code{offsets}}{(read-only) A numeric vector of offsets from the origin of
the calendar.}

\item{\code{resolution}}{(read-only) The average number of time units between
offsets.}

\item{\code{length}}{(read-only) Retrieve the number of offsets in the time
series.}

\item{\code{bounds}}{Retrieve or set the bounds for the offsets. On setting, a
\code{matrix} with columns for \code{offsets} and low values in the first row,
high values in the second row. This may be simply \code{TRUE} to set regular
and consecutive bounds.}

\item{\code{friendlyClassName}}{Character string with class name for display
purposes.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-CFTime-new}{\code{CFTime$new()}}
\item \href{#method-CFTime-print}{\code{CFTime$print()}}
\item \href{#method-CFTime-range}{\code{CFTime$range()}}
\item \href{#method-CFTime-as_timestamp}{\code{CFTime$as_timestamp()}}
\item \href{#method-CFTime-format}{\code{CFTime$format()}}
\item \href{#method-CFTime-indexOf}{\code{CFTime$indexOf()}}
\item \href{#method-CFTime-get_bounds}{\code{CFTime$get_bounds()}}
\item \href{#method-CFTime-set_bounds}{\code{CFTime$set_bounds()}}
\item \href{#method-CFTime-equidistant}{\code{CFTime$equidistant()}}
\item \href{#method-CFTime-slice}{\code{CFTime$slice()}}
\item \href{#method-CFTime-POSIX_compatible}{\code{CFTime$POSIX_compatible()}}
\item \href{#method-CFTime-cut}{\code{CFTime$cut()}}
\item \href{#method-CFTime-factor}{\code{CFTime$factor()}}
\item \href{#method-CFTime-factor_units}{\code{CFTime$factor_units()}}
\item \href{#method-CFTime-factor_coverage}{\code{CFTime$factor_coverage()}}
\item \href{#method-CFTime-copy}{\code{CFTime$copy()}}
\item \href{#method-CFTime-subset}{\code{CFTime$subset()}}
\item \href{#method-CFTime-doy}{\code{CFTime$doy()}}
\item \href{#method-CFTime-clone}{\code{CFTime$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-new"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-new}{}}}
\subsection{Method \code{new()}}{
Create a new instance of this class.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$new(definition, calendar = "standard", offsets = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{definition}}{Character string of the units and origin of the
calendar.}

\item{\code{calendar}}{Character string of the calendar to use. Must be one of
the values permitted by the CF Metadata Conventions. If \code{NULL}, the
"standard" calendar will be used.}

\item{\code{offsets}}{Numeric or character vector, optional. When numeric, a
vector of offsets from the origin in the time series. When a character
vector of length 2 or more, timestamps in ISO8601 or UDUNITS format.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-print"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-print}{}}}
\subsection{Method \code{print()}}{
Print a summary of the \code{CFTime} object to the console.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Ignored.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Self, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-range"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-range}{}}}
\subsection{Method \code{range()}}{
This method returns the first and last timestamp of the time
series as a vector. Note that the offsets do not have to be sorted.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$range(format = "", bounds = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{format}}{Value of "date" or "timestamp". Optionally, a
character string that specifies an alternate format.}

\item{\code{bounds}}{Logical to indicate if the extremes from the bounds should
be used, if set. Defaults to \code{FALSE}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Vector of two character strings that represent the starting and
ending timestamps in the time series. If a \code{format} is supplied, that
format will be used. Otherwise, if all of the timestamps in the time
series have a time component of \code{00:00:00} the date of the timestamp is
returned, otherwise the full timestamp (without any time zone
information).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-as_timestamp"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-as_timestamp}{}}}
\subsection{Method \code{as_timestamp()}}{
This method generates a vector of character strings or
\code{POSIXct}s that represent the date and time in a selectable combination
for each offset.

The character strings use the format \verb{YYYY-MM-DDThh:mm:ss±hhmm},
depending on the \code{format} specifier. The date in the string is not
necessarily compatible with \code{POSIXt} - in the \verb{360_day} calendar
\code{2017-02-30} is valid and \code{2017-03-31} is not.

For the "proleptic_gregorian" calendar the output can also be generated
as a vector of \code{POSIXct} values by specifying \code{asPOSIX = TRUE}. The
same is possible for the "standard" and "gregorian" calendars but only
if all timestamps fall on or after 1582-10-15. If \code{asPOSIX = TRUE} is
specified while the calendar does not support it, an error will be
generated.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$as_timestamp(format = NULL, asPOSIX = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{format}}{character. A character string with either of the values
"date" or "timestamp". If the argument is not specified, the format
used is "timestamp" if there is time information, "date" otherwise.}

\item{\code{asPOSIX}}{logical. If \code{TRUE}, for "standard", "gregorian" and
"proleptic_gregorian" calendars the output is a vector of \code{POSIXct} -
for other calendars an error will be thrown. Default value is \code{FALSE}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A character vector where each element represents a moment in
time according to the \code{format} specifier.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-format"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-format}{}}}
\subsection{Method \code{format()}}{
Format timestamps using a specific format string, using the
specifiers defined for the \code{\link[base:strptime]{base::strptime()}} function, with
limitations. The only supported specifiers are \verb{bBdeFhHImMpRSTYz\%}.
Modifiers \code{E} and \code{O} are silently ignored. Other specifiers, including
their percent sign, are copied to the output as if they were adorning
text.

The formatting is largely oblivious to locale. The reason for this is
that certain dates in certain calendars are not POSIX-compliant and the
system functions necessary for locale information thus do not work
consistently. The main exception to this is the (abbreviated) names of
months (\code{bB}), which could be useful for pretty printing in the local
language. For separators and other locale-specific adornments, use
local knowledge instead of depending on system locale settings; e.g.
specify \verb{\%m/\%d/\%Y} instead of \verb{\%D}.

Week information, including weekday names, is not supported at all as a
"week" is not defined for non-standard CF calendars and not generally
useful for climate projection data. If you are working with observed
data and want to get pretty week formats, use the \code{\link[=as_timestamp]{as_timestamp()}}
method to generate \code{POSIXct} timestamps (observed data generally uses a
"standard" calendar) and then use the \code{\link[base:format]{base::format()}} function which
supports the full set of specifiers.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$format(format, usetz = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{format}}{A character string with \code{strptime} format specifiers. If
omitted, the most economical format will be used: a full timestamp when
time information is available, a date otherwise.}

\item{\code{usetz}}{Logical. Should the time zone offset be appended to the
output? This is always in numerical form, i.e. "-0800", from UTC.
Default is \code{FALSE}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A vector of character strings with a properly formatted
timestamp. Any format specifiers not recognized or supported will be
returned verbatim.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-indexOf"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-indexOf}{}}}
\subsection{Method \code{indexOf()}}{
Find the index in the time series for each timestamp given
in argument \code{x}. Alternatively, when \code{x} is a numeric vector of index
values, return the valid indices of the same vector, with the side
effect being the attribute "CFTime" associated with the result.

Matching also returns index values for timestamps that fall between two
elements of the time series - this can lead to surprising results when
time series elements are positioned in the middle of an interval (as
the CF Metadata Conventions instruct us to "reasonably assume"): a time
series of days in January would be encoded in a netCDF file as
\code{c("2024-01-01 12:00:00", "2024-01-02 12:00:00", "2024-01-03 12:00:00", ...)} so \code{x <- c("2024-01-01", "2024-01-02", "2024-01-03")} would
result in \verb{(NA, 1, 2)} (or \verb{(NA, 1.5, 2.5)} with \code{method = "linear"})
because the date values in \code{x} are at midnight. This situation is
easily avoided by ensuring that this \code{CFTime} instance has bounds set
(use \code{bounds(y) <- TRUE} as a proximate solution if bounds are not
stored in the netCDF file).

If bounds are set, the indices are informed by those bounds. If the
bounds are not contiguous, returned values may be \code{NA} even if the
value of \code{x} falls between two valid timestamps.

Values of \code{x} that are not valid timestamps according to the calendar
of this \code{CFTime} instance will be returned as \code{NA}.

Argument \code{x} can also be a numeric vector of index values, in which
case the valid values in \code{x} are returned. If negative values are
passed, the positive counterparts will be excluded and then the
remainder returned. Positive and negative values may not be mixed.
Using a numeric vector has the side effect that the result has the
attribute "CFTime" describing the temporal dimension of the slice. If
index values outside of the range of \code{self} are provided, an error will
be thrown.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$indexOf(x, method = "constant", rightmost.closed = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{x}}{Vector of \code{character}, \code{POSIXt} or \code{Date} values to find indices
for, or a \code{numeric} vector.}

\item{\code{method}}{Single value of "constant" or "linear". If \code{"constant"},
return the index value for each match. If \code{"linear"}, return the index
value with any fractional value.}

\item{\code{rightmost.closed}}{Whether or not to include the upper limit of
argument \code{x}. Default is \code{FALSE}. This argument is ignored when
argument \code{x} contains index values.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A numeric vector giving indices into \code{self} for the valid values
of \code{x}. If there is at least 1 valid index, then attribute "CFTime"
contains an instance of \code{CFTime} that describes the dimension of
filtering the dataset associated with \code{self} with the result of this
method, excluding any \code{NA} values.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-get_bounds"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-get_bounds}{}}}
\subsection{Method \code{get_bounds()}}{
Return boundary values.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$get_bounds(format)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{format}}{A string specifying a format for output, optional.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
An array with \code{dims(2, length(offsets))} with values for the
boundaries. \code{NULL} if the boundaries have not been set.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-set_bounds"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-set_bounds}{}}}
\subsection{Method \code{set_bounds()}}{
Set or delete the boundary values of the \code{CFTime} instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$set_bounds(value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{value}}{The boundary values to set, in units of the offsets. A
matrix \verb{(2, length(private$.offsets))}. If \code{NULL}, the boundaries are
deleted. If \code{TRUE}, make regular, consecutive boundaries.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Self, invisibly.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-equidistant"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-equidistant}{}}}
\subsection{Method \code{equidistant()}}{
This method returns \code{TRUE} if the time series has uniformly
distributed time steps between the extreme values, \code{FALSE} otherwise.
First test without sorting; this should work for most data sets. If
not, only then offsets are sorted. For most data sets that will work
but for implied resolutions of month, season, year, etc based on a
"days" or finer calendar unit this will fail due to the fact that those
coarser units have a variable number of days per time step, in all
calendars except for \verb{360_day}. For now, an approximate solution is
used that should work in all but the most non-conformal exotic
arrangements.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$equidistant()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{TRUE} if all time steps are equidistant, \code{FALSE} otherwise, or
\code{NA} if no offsets have been set.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-slice"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-slice}{}}}
\subsection{Method \code{slice()}}{
Given a vector of character timestamps, return a logical
vector of a length equal to the number of time steps in the time series
with values \code{TRUE} for those time steps that fall between the two
extreme values of the vector values, \code{FALSE} otherwise.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$slice(extremes, closed = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{extremes}}{Character vector of timestamps that represent the
time period of interest. The extreme values are selected. Badly
formatted timestamps are silently dropped.}

\item{\code{closed}}{Is the right side closed, i.e. included in the result?
Default is \code{FALSE}. A specification of \verb{c("2022-01-01", "2023-01-01)}
will thus include all time steps that fall in the year 2022 when
\code{closed = FALSE} but include \code{2023-01-01} if that exact value is
present in the time series.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A logical vector with a length equal to the number of time steps
in \code{self} with values \code{TRUE} for those time steps that fall between the
extreme values, \code{FALSE} otherwise.

An attribute 'CFTime' will have the same definition as \code{self} but with
offsets corresponding to the time steps falling between the two
extremes. If there are no values between the extremes, the attribute is
not set.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-POSIX_compatible"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-POSIX_compatible}{}}}
\subsection{Method \code{POSIX_compatible()}}{
Can the time series be converted to POSIXt?
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$POSIX_compatible()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{TRUE} if the calendar support conversion to POSIXt, \code{FALSE}
otherwise.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-cut"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-cut}{}}}
\subsection{Method \code{cut()}}{
Create a factor for a \code{CFTime} instance.

When argument \code{breaks} is one of \verb{"year", "season", "quarter", "month", "dekad", "day"}, a factor is generated like by \code{\link[=CFfactor]{CFfactor()}}. When
\code{breaks} is a vector of character timestamps a factor is produced with
a level for every interval between timestamps. The last timestamp,
therefore, is only used to close the interval started by the
pen-ultimate timestamp - use a distant timestamp (e.g. \code{range(x)[2]})
to ensure that all offsets to the end of the CFTime time series are
included, if so desired. The last timestamp will become the upper bound
in the \code{CFTime} instance that is returned as an attribute to this
function so a sensible value for the last timestamp is advisable.

This method works similar to \code{\link[base:cut.POSIXt]{base::cut.POSIXt()}} but there are some
differences in the arguments: for \code{breaks} the set of options is
different and no preceding integer is allowed, \code{labels} are always
assigned using values of \code{breaks}, and the interval is always
left-closed.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$cut(breaks)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{breaks}}{A character string of a factor period (see \code{\link[=CFfactor]{CFfactor()}} for
a description), or a character vector of timestamps that conform to the
calendar of \code{x}, with a length of at least 2. Timestamps must be given
in ISO8601 format, e.g. "2024-04-10 21:31:43".}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A factor with levels according to the \code{breaks} argument, with
attributes 'period', 'era' and 'CFTime'. When \code{breaks} is a factor
period, attribute 'period' has that value, otherwise it is '"day"'.
When \code{breaks} is a character vector of timestamps, attribute 'CFTime'
holds an instance of \code{CFTime} that has the same definition as \code{x}, but
with (ordered) offsets generated from the \code{breaks}. Attribute 'era'
is always -1.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-factor"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-factor}{}}}
\subsection{Method \code{factor()}}{
Generate a factor for the offsets, or a part thereof. This
is specifically interesting for creating factors from the date part of
the time series that aggregate the time series into longer time periods
(such as month) that can then be used to process daily CF data sets
using, for instance, \code{tapply()}.

The factor will respect the calendar that the time series is built on.

The factor will be generated in the order of the offsets. While typical
CF-compliant data sources use ordered time series there is, however, no
guarantee that the factor is ordered. For most processing with a factor
the ordering is of no concern.

If the \code{era} parameter is specified, either as a vector of years to
include in the factor, or as a list of such vectors, the factor will
only consider those values in the time series that fall within the list
of years, inclusive of boundary values. Other values in the factor will
be set to \code{NA}.

The following periods are supported by this method:

\itemize{
\item \code{year}, the year of each offset is returned as "YYYY".
\item \code{season}, the meteorological season of each offset is returned as
"Sx", with x being 1-4, preceeded by "YYYY" if no \code{era} is
specified. Note that December dates are labeled as belonging to the
subsequent year, so the date "2020-12-01" yields "2021S1". This implies
that for standard CMIP files having one or more full years of data the
first season will have data for the first two months (January and
February), while the final season will have only a single month of data
(December).
\item \code{quarter}, the calendar quarter of each offset is returned as "Qx",
with x being 1-4, preceeded by "YYYY" if no \code{era} is specified.
\item \code{month}, the month of each offset is returned as "01" to
"12", preceeded by "YYYY-" if no \code{era} is specified. This is the default
period.
\item \code{dekad}, ten-day periods are returned as
"Dxx", where xx runs from "01" to "36", preceeded by "YYYY" if no \code{era}
is specified. Each month is subdivided in dekads as follows: 1- days 01 -
10; 2- days 11 - 20; 3- remainder of the month.
\item \code{day}, the month and day of each offset are returned as "MM-DD",
preceeded by "YYYY-" if no \code{era} is specified.
}

It is not possible to create a factor for a period that is shorter than
the temporal resolution of the calendar. As an example, if the calendar
has a monthly unit, a dekad or day factor cannot be created.

Creating factors for other periods is not supported by this method.
Factors based on the timestamp information and not dependent on the
calendar can trivially be constructed from the output of the
\code{\link[=as_timestamp]{as_timestamp()}} function.

Attribute 'CFTime' of the result contains a \code{CFTime} instance that is
valid for the result of applying the factor to a resource that this
instance is associated with. In other words, if \code{CFTime} instance 'At'
describes the temporal dimension of resource 'A' and a factor 'Af' is
generated from \code{Af <- At$factor()}, then \code{Bt <- attr(Af, "CFTime")}
describes the temporal dimension of the result of, say, \code{B <- apply(A, 1:2, tapply, Af, FUN)}. The 'CFTime' attribute contains a
\link{CFClimatology} instance for era factors.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$factor(period = "month", era = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{period}}{character. A character string with one of the values
"year", "season", "quarter", "month" (the default), "dekad" or "day".}

\item{\code{era}}{numeric or list, optional. Vector of years for which to
construct the factor, or a list whose elements are each a vector of
years. The extreme values of the supplied vector will be used. Note
that if a single year is specified that the result is valid, but not a
climatological statistic. If \code{era} is not specified, the factor will
use the entire time series for the factor.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
If \code{era} is a single vector or \code{NULL}, a factor with a length
equal to the number of offsets in this instance. If \code{era} is a list, a
list with the same number of elements and names as \code{era}, each
containing a factor. Elements in the factor will be set to \code{NA} for
time series values outside of the range of specified years.

The factor, or factors in the list, have attributes 'period', 'era' and
'CFTime'. Attribute 'period' holds the value of the \code{period} argument.
Attribute 'era' indicates the number of years that are included in the
era, or -1 if no \code{era} is provided. Attribute 'CFTime' holds an
instance of \code{CFTime} or \code{CFClimatology} that has the same definition as
this instance, but with offsets corresponding to the mid-point of the
factor levels.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-factor_units"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-factor_units}{}}}
\subsection{Method \code{factor_units()}}{
Given a factor as produced by \code{CFTime$factor()}, this method
will return a numeric vector with the number of time units in each
level of the factor.

The result of this method is useful to convert between absolute and
relative values. Climate change anomalies, for instance, are usually
computed by differencing average values between a future period and a
baseline period. Going from average values back to absolute values for
an aggregate period (which is typical for temperature and
precipitation, among other variables) is easily done with the result of
this method, without having to consider the specifics of the calendar
of the data set.

If the factor \code{f} is for an era (e.g. spanning multiple years and the
levels do not indicate the specific year), then the result will
indicate the number of time units of the period in a regular single
year. In other words, for an era of 2041-2060 and a monthly factor on a
standard calendar with a \code{days} unit, the result will be
\code{c(31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)}. Leap days are thus
only considered for the \verb{366_day} and \code{all_leap} calendars.

Note that this function gives the number of time units in each level of
the factor - the actual number of data points in the time series per
factor level may be different. Use \code{\link[=CFfactor_coverage]{CFfactor_coverage()}} to determine
the actual number of data points or the coverage of data points
relative to the factor level.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$factor_units(f)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{f}}{A factor or a list of factors derived from the method
\code{CFTime$factor()}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
If \code{f} is a factor, a numeric vector with a length equal to the
number of levels in the factor, indicating the number of time units in
each level of the factor. If \code{f} is a list of factors, a list with each
element a numeric vector as above.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-factor_coverage"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-factor_coverage}{}}}
\subsection{Method \code{factor_coverage()}}{
Calculate the number of time elements, or the relative
coverage, in each level of a factor generated by \code{CFTime$factor()}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$factor_coverage(f, coverage = "absolute")}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{f}}{A factor or a list of factors derived from the method
\code{CFTime$factor()}.}

\item{\code{coverage}}{"absolute" or "relative".}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
If \code{f} is a factor, a numeric vector with a length equal to the
number of levels in the factor, indicating the number of units from the
time series contained in each level of the factor when
\code{coverage = "absolute"} or the proportion of units present relative to the
maximum number when \code{coverage = "relative"}. If \code{f} is a list of factors, a
list with each element a numeric vector as above.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-copy"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-copy}{}}}
\subsection{Method \code{copy()}}{
Create a copy of the current instance. The copy is
completely separate from the current instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$copy()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A new \code{CFTime} instance with identical definition and set of
timestamps.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-subset"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-subset}{}}}
\subsection{Method \code{subset()}}{
Get a new \code{CFTime} instance that is a subset of the current
instance, including any boundary values.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$subset(rng)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{rng}}{The numeric range of indices to subset this instance to.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{CFTime} instance with identical definition and set of
timestamps according to the \code{rng} argument.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-doy"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-doy}{}}}
\subsection{Method \code{doy()}}{
Retrieve the day-of-year number, either from the timestamps
in the current instance or from the supplied \code{data.frame} of year,
month and day. The \code{doy} is calendar-aware.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$doy(ymd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ymd}}{Optional \code{data.frame} with dates parsed into their parts in
columns \code{year}, \code{month} and \code{day}. Any other columns are disregarded.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
An integer vector of the same length as argument \code{ymd} or the
internal offsets if \code{ymd} is not specified with the day-of-year in the
calendar of this instance. Rows in argument \code{ymd} that do not hold a
valid date are returned as \code{NA}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFTime-clone"></a>}}
\if{latex}{\out{\hypertarget{method-CFTime-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CFTime$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{deprecated_functions}
\alias{deprecated_functions}
\alias{CFtimestamp}
\alias{CFmonth_days}
\alias{CFcomplete}
\alias{CFsubset}
\alias{CFparse}
\title{Deprecated functions}
\usage{
CFtimestamp(t, format = NULL, asPOSIX = FALSE)

CFmonth_days(t, x = NULL)

CFcomplete(x)

CFsubset(x, extremes)

CFparse(t, x)
}
\arguments{
\item{t, x, format, asPOSIX, extremes}{See replacement functions.}
}
\value{
See replacement functions.
}
\description{
These functions are deprecated and should no longer be used in new code. The
below table gives the replacement function to use instead. The function
arguments of the replacement function are the same as those of the deprecated
function if no arguments are given in the table.\tabular{ll}{
   \strong{Deprecated function} \tab \strong{Replacement function} \cr
   CFcomplete() \tab \code{\link[=is_complete]{is_complete()}} \cr
   CFmonth_days() \tab \code{\link[=month_days]{month_days()}} \cr
   CFparse() \tab \code{\link[=parse_timestamps]{parse_timestamps()}} \cr
   CFrange() \tab \code{\link[=range]{range()}} \cr
   CFsubset() \tab \code{\link[=slab]{slab()}} \cr
   CFtimestamp() \tab \code{\link[=as_timestamp]{as_timestamp()}} \cr
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{CFfactor_units}
\alias{CFfactor_units}
\title{Number of base time units in each factor level}
\usage{
CFfactor_units(t, f)
}
\arguments{
\item{t}{An instance of \code{CFTime}.}

\item{f}{A factor or a list of factors derived from the
parameter \code{t}. The factor or list thereof should generally be generated by
the function \code{\link[=CFfactor]{CFfactor()}}.}
}
\value{
If \code{f} is a factor, a numeric vector with a length equal to the
number of levels in the factor, indicating the number of time units in each
level of the factor. If \code{f} is a list of factors, a list with each element
a numeric vector as above.
}
\description{
Given a factor as returned by \code{\link[=CFfactor]{CFfactor()}} and the \link{CFTime} instance from
which the factor was derived, this function will return a numeric vector with
the number of time units in each level of the factor.
}
\details{
The result of this function is useful to convert between absolute and
relative values. Climate change anomalies, for instance, are usually computed
by differencing average values between a future period and a baseline period.
Going from average values back to absolute values for an aggregate period
(which is typical for temperature and precipitation, among other variables)
is easily done with the result of this function, without having to consider
the specifics of the calendar of the data set.

If the factor \code{f} is for an era (e.g. spanning multiple years and the
levels do not indicate the specific year), then the result will indicate the
number of time units of the period in a regular single year. In other words,
for an era of 2041-2060 and a monthly factor on a standard calendar with a
\code{days} unit, the result will be \code{c(31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)}.
Leap days are thus only considered for the \verb{366_day} and \code{all_leap} calendars.

Note that this function gives the number of time units in each level of the
factor - the actual number of data points in the \code{cf} instance per factor
level may be different. Use \code{\link[=CFfactor_coverage]{CFfactor_coverage()}} to determine the actual
number of data points or the coverage of data points relative to the factor
level.
}
\examples{
t <- CFtime("days since 2001-01-01", "365_day", 0:364)
f <- CFfactor(t, "dekad")
CFfactor_units(t, f)
}
